@@ -54,12 +54,16 @@ defmodule Application do
   You can also change the application environment dynamically by using functions
   such as `put_env/3` and `delete_env/2`.
 
-  > Note: The config files `config/config.exs` and `config/runtime.exs`
+  > #### Environment in libraries {: .tip}
+  >
+  > The config files `config/config.exs` and `config/runtime.exs`
   > are rarely used by libraries. Libraries typically define their environment
-  > in the `def application` function of their `mix.exs`. Configuration files
+  > in the `application/0` function of their `mix.exs`. Configuration files
   > are rather used by applications to configure their libraries.
 
-  > Note: Each application is responsible for its own environment. Do not
+  > #### Reading the environment of other applications {: .warning}
+  >
+  > Each application is responsible for its own environment. Do not
   > use the functions in this module for directly accessing or modifying
   > the environment of other applications. Whenever you change the application
   > environment, Elixir's build tool will only recompile the files that
@@ -161,6 +165,13 @@ defmodule Application do
         end
       end
 
+  > #### `use Application` {: .info}
+  >
+  > When you `use Application`, the `Application` module will
+  > set `@behaviour Application` and define an overridable
+  > definition for the `c:stop/1` function, which is required
+  > by Erlang/OTP.
+
   The `c:start/2` callback has to spawn and link a supervisor and return `{:ok,
   pid}` or `{:ok, pid, state}`, where `pid` is the PID of the supervisor, and
   `state` is an optional application state. `args` is the second element of the
@@ -425,6 +436,28 @@ defmodule Application do
   @type value :: term
   @type state :: term
   @type start_type :: :normal | {:takeover, node} | {:failover, node}
+
+  @typedoc """
+  Specifies the type of the application:
+
+    * `:permanent` - if `app` terminates, all other applications and the entire
+      node are also terminated.
+
+    * `:transient` - if `app` terminates with `:normal` reason, it is reported
+      but no other applications are terminated. If a transient application
+      terminates abnormally, all other applications and the entire node are
+      also terminated.
+
+    * `:temporary` - if `app` terminates, it is reported but no other
+      applications are terminated (the default).
+
+  Note that it is always possible to stop an application explicitly by calling
+  `stop/1`. Regardless of the type of the application, no other applications will
+  be affected.
+
+  Note also that the `:transient` type is of little practical use, since when a
+  supervision tree terminates, the reason is set to `:shutdown`, not `:normal`.
+  """
   @type restart_type :: :permanent | :transient | :temporary
 
   @doc """
@@ -632,10 +665,14 @@ defmodule Application do
   If the configuration parameter does not exist, the function returns the
   `default` value.
 
-  > **Important:** you must use this function to read only your own application
+  > #### Warning {: .warning}
+  >
+  > You must use this function to read only your own application
   > environment. Do not read the environment of other applications.
 
-  > **Important:** if you are writing a library to be used by other developers,
+  > #### Application environment in libraries {: .info}
+  >
+  > If you are writing a library to be used by other developers,
   > it is generally recommended to avoid the application environment, as the
   > application environment is effectively a global storage. For more information,
   > read our [library guidelines](library-guidelines.md).
@@ -689,10 +726,14 @@ defmodule Application do
 
   If the configuration parameter does not exist, the function returns `:error`.
 
-  > **Important:** you must use this function to read only your own application
+  > #### Warning {: .warning}
+  >
+  > You must use this function to read only your own application
   > environment. Do not read the environment of other applications.
 
-  > **Important:** if you are writing a library to be used by other developers,
+  > #### Application environment in info
+  >
+  > If you are writing a library to be used by other developers,
   > it is generally recommended to avoid the application environment, as the
   > application environment is effectively a global storage. For more information,
   > read our [library guidelines](library-guidelines.md).
@@ -712,10 +753,14 @@ defmodule Application do
 
   If the configuration parameter does not exist, raises `ArgumentError`.
 
-  > **Important:** you must use this function to read only your own application
+  > #### Warning {: .warning}
+  >
+  > You must use this function to read only your own application
   > environment. Do not read the environment of other applications.
 
-  > **Important:** if you are writing a library to be used by other developers,
+  > #### Application environment in info
+  >
+  > If you are writing a library to be used by other developers,
   > it is generally recommended to avoid the application environment, as the
   > application environment is effectively a global storage. For more information,
   > read our [library guidelines](library-guidelines.md).
@@ -807,17 +852,13 @@ defmodule Application do
   end
 
   @doc """
-  Ensures the given `app` is started.
+  Ensures the given `app` is started with `t:restart_type/0`.
 
   Same as `start/2` but returns `:ok` if the application was already
-  started. This is useful in scripts and in test setup, where test
-  applications need to be explicitly started:
-
-      :ok = Application.ensure_started(:my_test_dep)
-
+  started.
   """
-  @spec ensure_started(app, restart_type) :: :ok | {:error, term}
-  def ensure_started(app, type \\ :temporary) when is_atom(app) do
+  @spec ensure_started(app, restart_type()) :: :ok | {:error, term}
+  def ensure_started(app, type \\ :temporary) when is_atom(app) and is_atom(type) do
     :application.ensure_started(app, type)
   end
 
@@ -838,19 +879,54 @@ defmodule Application do
   end
 
   @doc """
-  Ensures the given `app` and its applications are started.
+  Ensures the given `app` or `apps` and their child applications are started.
+
+  The second argument is either the `t:restart_type/1` (for consistency with
+  `start/2`) or a keyword list.
+
+  ## Options
+
+    * `:type` - if the application should be started in `:permanent`,
+      `:temporary`, or `:transient`. See `t:restart_type/1` for more information.
+
+    * `:mode` - (since v1.15.0) if the applications should be started serially
+      or concurrently. This option requires Erlang/OTP 26+.
 
-  Same as `start/2` but also starts the applications listed under
-  `:applications` in the `.app` file in case they were not previously
-  started.
   """
-  @spec ensure_all_started(app, restart_type) :: {:ok, [app]} | {:error, {app, term}}
-  def ensure_all_started(app, type \\ :temporary) when is_atom(app) do
-    :application.ensure_all_started(app, type)
+  @spec ensure_all_started(app | [app], type: restart_type(), mode: :serial | :concurrent) ::
+          {:ok, [app]} | {:error, term}
+  @spec ensure_all_started(app | [app], restart_type()) ::
+          {:ok, [app]} | {:error, term}
+  def ensure_all_started(app_or_apps, type_or_opts \\ [])
+
+  def ensure_all_started(app, type) when is_atom(type) do
+    ensure_all_started(app, type: type)
+  end
+
+  def ensure_all_started(app, opts) when is_atom(app) do
+    ensure_all_started([app], opts)
+  end
+
+  @compile {:no_warn_undefined, {:application, :ensure_all_started, 3}}
+
+  def ensure_all_started(apps, opts) when is_list(apps) and is_list(opts) do
+    opts = Keyword.validate!(opts, type: :temporary, mode: :serial)
+
+    if function_exported?(:application, :ensure_all_started, 3) do
+      :application.ensure_all_started(apps, opts[:type], opts[:mode])
+    else
+      # TODO: Remove this clause when we require Erlang/OTP 26+
+      Enum.reduce_while(apps, {:ok, []}, fn app, {:ok, acc} ->
+        case :application.ensure_all_started(app, opts[:type]) do
+          {:ok, apps} -> {:cont, {:ok, apps ++ acc}}
+          {:error, e} -> {:halt, {:error, e}}
+        end
+      end)
+    end
   end
 
   @doc """
-  Starts the given `app`.
+  Starts the given `app` with `t:restart_type/0`.
 
   If the `app` is not loaded, the application will first be loaded using `load/1`.
   Any included application, defined in the `:included_applications` key of the
@@ -862,29 +938,9 @@ defmodule Application do
 
   In case you want to automatically load **and start** all of `app`'s dependencies,
   see `ensure_all_started/2`.
-
-  The `type` argument specifies the type of the application:
-
-    * `:permanent` - if `app` terminates, all other applications and the entire
-      node are also terminated.
-
-    * `:transient` - if `app` terminates with `:normal` reason, it is reported
-      but no other applications are terminated. If a transient application
-      terminates abnormally, all other applications and the entire node are
-      also terminated.
-
-    * `:temporary` - if `app` terminates, it is reported but no other
-      applications are terminated (the default).
-
-  Note that it is always possible to stop an application explicitly by calling
-  `stop/1`. Regardless of the type of the application, no other applications will
-  be affected.
-
-  Note also that the `:transient` type is of little practical use, since when a
-  supervision tree terminates, the reason is set to `:shutdown`, not `:normal`.
   """
-  @spec start(app, restart_type) :: :ok | {:error, term}
-  def start(app, type \\ :temporary) when is_atom(app) do
+  @spec start(app, restart_type()) :: :ok | {:error, term}
+  def start(app, type \\ :temporary) when is_atom(app) and is_atom(type) do
     :application.start(app, type)
   end
 
@@ -1075,7 +1131,7 @@ defmodule Application do
     "bad application start specs: #{inspect(spec)}"
   end
 
-  defp do_format_error({'no such file or directory', file}) do
+  defp do_format_error({~c"no such file or directory", file}) do
     "could not find application file: #{file}"
   end
 
