@@ -4,7 +4,7 @@ defmodule Date do
 
   The Date struct contains the fields year, month, day and calendar.
   New dates can be built with the `new/3` function or using the
-  `~D` (see `Kernel.sigil_D/2`) sigil:
+  `~D` (see `sigil_D/2`) sigil:
 
       iex> ~D[2000-01-01]
       ~D[2000-01-01]
@@ -31,7 +31,12 @@ defmodule Date do
 
   Comparisons in Elixir using `==/2`, `>/2`, `</2` and similar are structural
   and based on the `Date` struct fields. For proper comparison between
-  dates, use the `compare/2` function.
+  dates, use the `compare/2` function. The existence of the `compare/2`
+  function in this module also allows using `Enum.min/2` and `Enum.max/2`
+  functions to get the minimum and maximum date of an `Enum`. For example:
+
+      iex>  Enum.min([~D[2017-03-31], ~D[2017-04-01]], Date)
+      ~D[2017-03-31]
 
   ## Using epochs
 
@@ -72,19 +77,23 @@ defmodule Date do
   ## Examples
 
       iex> Date.range(~D[1999-01-01], ~D[2000-01-01])
-      #DateRange<~D[1999-01-01], ~D[2000-01-01]>
+      Date.range(~D[1999-01-01], ~D[2000-01-01])
 
   A range of dates implements the `Enumerable` protocol, which means
   functions in the `Enum` module can be used to work with
   ranges:
 
       iex> range = Date.range(~D[2001-01-01], ~D[2002-01-01])
+      iex> range
+      Date.range(~D[2001-01-01], ~D[2002-01-01])
       iex> Enum.count(range)
       366
-      iex> Enum.member?(range, ~D[2001-02-01])
+      iex> ~D[2001-02-01] in range
       true
-      iex> Enum.reduce(range, 0, fn _date, acc -> acc - 1 end)
-      -366
+      iex> Enum.take(range, 3)
+      [~D[2001-01-01], ~D[2001-01-02], ~D[2001-01-03]]
+      iex> for d <- Date.range(~D[2023-03-01], ~D[2023-04-01]), Date.day_of_week(d) == 7, do: d
+      [~D[2023-03-05], ~D[2023-03-12], ~D[2023-03-19], ~D[2023-03-26]]
 
   """
   @doc since: "1.5.0"
@@ -92,19 +101,61 @@ defmodule Date do
   def range(%{calendar: calendar} = first, %{calendar: calendar} = last) do
     {first_days, _} = to_iso_days(first)
     {last_days, _} = to_iso_days(last)
-
-    %Date.Range{
-      first: %Date{calendar: calendar, year: first.year, month: first.month, day: first.day},
-      last: %Date{calendar: calendar, year: last.year, month: last.month, day: last.day},
-      first_in_iso_days: first_days,
-      last_in_iso_days: last_days
-    }
+    # TODO: Deprecate inferring a range with a step of -1 on Elixir v1.16
+    step = if first_days <= last_days, do: 1, else: -1
+    range(first, first_days, last, last_days, calendar, step)
   end
 
   def range(%{calendar: _, year: _, month: _, day: _}, %{calendar: _, year: _, month: _, day: _}) do
     raise ArgumentError, "both dates must have matching calendars"
   end
 
+  @doc """
+  Returns a range of dates with a step.
+
+  ## Examples
+
+      iex> range = Date.range(~D[2001-01-01], ~D[2002-01-01], 2)
+      iex> range
+      Date.range(~D[2001-01-01], ~D[2002-01-01], 2)
+      iex> Enum.count(range)
+      183
+      iex> ~D[2001-01-03] in range
+      true
+      iex> Enum.take(range, 3)
+      [~D[2001-01-01], ~D[2001-01-03], ~D[2001-01-05]]
+
+  """
+  @doc since: "1.12.0"
+  @spec range(Calendar.date(), Calendar.date(), step :: pos_integer | neg_integer) ::
+          Date.Range.t()
+  def range(%{calendar: calendar} = first, %{calendar: calendar} = last, step)
+      when is_integer(step) and step != 0 do
+    {first_days, _} = to_iso_days(first)
+    {last_days, _} = to_iso_days(last)
+    range(first, first_days, last, last_days, calendar, step)
+  end
+
+  def range(
+        %{calendar: _, year: _, month: _, day: _} = first,
+        %{calendar: _, year: _, month: _, day: _} = last,
+        step
+      ) do
+    raise ArgumentError,
+          "both dates must have matching calendar and the step must be a " <>
+            "non-zero integer, got: #{inspect(first)}, #{inspect(last)}, #{step}"
+  end
+
+  defp range(first, first_days, last, last_days, calendar, step) do
+    %Date.Range{
+      first: %Date{calendar: calendar, year: first.year, month: first.month, day: first.day},
+      last: %Date{calendar: calendar, year: last.year, month: last.month, day: last.day},
+      first_in_iso_days: first_days,
+      last_in_iso_days: last_days,
+      step: step
+    }
+  end
+
   @doc """
   Returns the current date in UTC.
 
@@ -273,7 +324,7 @@ defmodule Date do
 
   @doc """
   Parses the extended "Dates" format described by
-  [ISO 8601:2004](https://en.wikipedia.org/wiki/ISO_8601).
+  [ISO 8601:2019](https://en.wikipedia.org/wiki/ISO_8601).
 
   The year parsed by this function is limited to four digits.
 
@@ -298,7 +349,7 @@ defmodule Date do
 
   @doc """
   Parses the extended "Dates" format described by
-  [ISO 8601:2004](https://en.wikipedia.org/wiki/ISO_8601).
+  [ISO 8601:2019](https://en.wikipedia.org/wiki/ISO_8601).
 
   Raises if the format is invalid.
 
@@ -323,7 +374,7 @@ defmodule Date do
 
   @doc """
   Converts the given `date` to
-  [ISO 8601:2004](https://en.wikipedia.org/wiki/ISO_8601).
+  [ISO 8601:2019](https://en.wikipedia.org/wiki/ISO_8601).
 
   By default, `Date.to_iso8601/2` returns dates formatted in the "extended"
   format, for human readability. It also supports the "basic" format through passing the `:basic` option.
@@ -500,8 +551,8 @@ defmodule Date do
     end
   end
 
-  def compare(date1, date2) do
-    if Calendar.compatible_calendars?(date1.calendar, date2.calendar) do
+  def compare(%{calendar: calendar1} = date1, %{calendar: calendar2} = date2) do
+    if Calendar.compatible_calendars?(calendar1, calendar2) do
       case {to_iso_days(date1), to_iso_days(date2)} do
         {first, second} when first > second -> :gt
         {first, second} when first < second -> :lt
@@ -517,6 +568,44 @@ defmodule Date do
     end
   end
 
+  @doc """
+  Returns true if the first date is strictly earlier than the second.
+
+  ## Examples
+
+      iex> Date.before?(~D[2021-01-01], ~D[2022-02-02])
+      true
+      iex> Date.before?(~D[2021-01-01], ~D[2021-01-01])
+      false
+      iex> Date.before?(~D[2022-02-02], ~D[2021-01-01])
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec before?(Calendar.date(), Calendar.date()) :: boolean()
+  def before?(date1, date2) do
+    compare(date1, date2) == :lt
+  end
+
+  @doc """
+  Returns true if the first date is strictly later than the second.
+
+  ## Examples
+
+      iex> Date.after?(~D[2022-02-02], ~D[2021-01-01])
+      true
+      iex> Date.after?(~D[2021-01-01], ~D[2021-01-01])
+      false
+      iex> Date.after?(~D[2021-01-01], ~D[2022-02-02])
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec after?(Calendar.date(), Calendar.date()) :: boolean()
+  def after?(date1, date2) do
+    compare(date1, date2) == :gt
+  end
+
   @doc """
   Converts the given `date` from its calendar to the given `calendar`.
 
@@ -659,11 +748,12 @@ defmodule Date do
     end
   end
 
-  defp to_iso_days(%{calendar: Calendar.ISO, year: year, month: month, day: day}) do
+  @doc false
+  def to_iso_days(%{calendar: Calendar.ISO, year: year, month: month, day: day}) do
     {Calendar.ISO.date_to_iso_days(year, month, day), {0, 86_400_000_000}}
   end
 
-  defp to_iso_days(%{calendar: calendar, year: year, month: month, day: day}) do
+  def to_iso_days(%{calendar: calendar, year: year, month: month, day: day}) do
     calendar.naive_datetime_to_iso_days(year, month, day, 0, 0, 0, {0, 0})
   end
 
@@ -787,8 +877,6 @@ defmodule Date do
       ~D[2020-07-05]
       iex> Date.end_of_week(~D[2020-07-06], :sunday)
       ~D[2020-07-11]
-      iex> Date.end_of_week(~D[2020-07-06], :sunday)
-      ~D[2020-07-11]
       iex> Date.end_of_week(~D[2020-07-06], :saturday)
       ~D[2020-07-10]
       iex> Date.end_of_week(~N[2020-07-11 01:23:45])
@@ -900,8 +988,14 @@ defmodule Date do
   @spec year_of_era(Calendar.date()) :: {Calendar.year(), non_neg_integer()}
   def year_of_era(date)
 
-  def year_of_era(%{calendar: calendar, year: year}) do
-    calendar.year_of_era(year)
+  def year_of_era(%{calendar: calendar, year: year, month: month, day: day}) do
+    # TODO: Remove me on 1.17
+    # The behaviour implementation already warns on missing callback.
+    if function_exported?(calendar, :year_of_era, 3) do
+      calendar.year_of_era(year, month, day)
+    else
+      calendar.year_of_era(year)
+    end
   end
 
   @doc """
@@ -943,6 +1037,8 @@ defmodule Date do
   """
   @doc since: "1.11.0"
   @spec beginning_of_month(Calendar.date()) :: t()
+  def beginning_of_month(date)
+
   def beginning_of_month(%{year: year, month: month, calendar: calendar}) do
     %Date{year: year, month: month, day: 1, calendar: calendar}
   end
@@ -962,6 +1058,8 @@ defmodule Date do
   """
   @doc since: "1.11.0"
   @spec end_of_month(Calendar.date()) :: t()
+  def end_of_month(date)
+
   def end_of_month(%{year: year, month: month, calendar: calendar} = date) do
     day = Date.days_in_month(date)
     %Date{year: year, month: month, day: day, calendar: calendar}
