@@ -37,6 +37,7 @@ defmodule DateTime do
   Other time zone databases can also be configured. Here are some
   available options and libraries:
 
+    * [`time_zone_info`](https://github.com/hrzndhrn/time_zone_info)
     * [`tz`](https://github.com/mathieuprog/tz)
     * [`tzdata`](https://github.com/lau/tzdata)
     * [`zoneinfo`](https://github.com/smartrent/zoneinfo) -
@@ -146,16 +147,55 @@ defmodule DateTime do
   If you want the current time in Unix seconds,
   use `System.os_time/1` instead.
 
+  You can also pass a time unit to automatically
+  truncate the resulting datetime. This is available
+  since v1.15.0.
+
   ## Examples
 
       iex> datetime = DateTime.utc_now()
       iex> datetime.time_zone
       "Etc/UTC"
 
+      iex> datetime = DateTime.utc_now(:second)
+      iex> datetime.microsecond
+      {0, 0}
+
   """
-  @spec utc_now(Calendar.calendar()) :: t
-  def utc_now(calendar \\ Calendar.ISO) do
-    System.os_time() |> from_unix!(:native, calendar)
+  @spec utc_now(Calendar.calendar() | :native | :microsecond | :millisecond | :second) :: t
+  def utc_now(calendar_or_time_unit \\ Calendar.ISO) do
+    case calendar_or_time_unit do
+      unit when unit in [:microsecond, :millisecond, :second, :native] ->
+        utc_now(unit, Calendar.ISO)
+
+      calendar ->
+        System.os_time() |> from_unix!(:native, calendar)
+    end
+  end
+
+  @doc """
+  Returns the current datetime in UTC, supporting 
+  a specific calendar and precision.
+
+  If you want the current time in Unix seconds,
+  use `System.os_time/1` instead.
+
+  ## Examples
+
+      iex> datetime = DateTime.utc_now(:microsecond, Calendar.ISO)
+      iex> datetime.time_zone
+      "Etc/UTC"
+
+      iex> datetime = DateTime.utc_now(:second, Calendar.ISO)
+      iex> datetime.microsecond
+      {0, 0}
+
+  """
+  @doc since: "1.15.0"
+  @spec utc_now(:native | :microsecond | :millisecond | :second, Calendar.calendar()) :: t
+  def utc_now(time_unit, calendar)
+      when time_unit in [:native, :microsecond, :millisecond, :second] do
+    System.os_time(time_unit) |> from_unix!(time_unit, calendar)
   end
 
   @doc """
@@ -1136,9 +1176,8 @@ defmodule DateTime do
 
   """
   @doc since: "1.4.0"
-  @spec from_iso8601(String.t(), Calendar.calendar(), :extended | :basic) ::
+  @spec from_iso8601(String.t(), Calendar.calendar() | :extended | :basic) ::
           {:ok, t, Calendar.utc_offset()} | {:error, atom}
-
   def from_iso8601(string, format_or_calendar \\ Calendar.ISO)
 
   def from_iso8601(string, format) when format in [:basic, :extended] do
@@ -1163,7 +1202,10 @@ defmodule DateTime do
       iex> {:ok, datetime, 9000} = DateTime.from_iso8601("20150123T235007.123+0230", Calendar.ISO, :basic)
       iex> datetime
       ~U[2015-01-23 21:20:07.123Z]
+
   """
+  @spec from_iso8601(String.t(), Calendar.calendar(), :extended | :basic) ::
+          {:ok, t, Calendar.utc_offset()} | {:error, atom}
   def from_iso8601(string, calendar, format) do
     with {:ok, {year, month, day, hour, minute, second, microsecond}, offset} <-
            Calendar.ISO.parse_utc_datetime(string, format) do
@@ -1384,6 +1426,44 @@ defmodule DateTime do
     end
   end
 
+  @doc """
+  Returns true if the first datetime is strictly earlier than the second.
+
+  ## Examples
+
+      iex> DateTime.before?(~U[2021-01-01 11:00:00Z], ~U[2022-02-02 11:00:00Z])
+      true
+      iex> DateTime.before?(~U[2021-01-01 11:00:00Z], ~U[2021-01-01 11:00:00Z])
+      false
+      iex> DateTime.before?(~U[2022-02-02 11:00:00Z], ~U[2021-01-01 11:00:00Z])
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec before?(Calendar.datetime(), Calendar.datetime()) :: boolean()
+  def before?(datetime1, datetime2) do
+    compare(datetime1, datetime2) == :lt
+  end
+
+  @doc """
+  Returns true if the first datetime is strictly later than the second.
+
+  ## Examples
+
+      iex> DateTime.after?(~U[2022-02-02 11:00:00Z], ~U[2021-01-01 11:00:00Z])
+      true
+      iex> DateTime.after?(~U[2021-01-01 11:00:00Z], ~U[2021-01-01 11:00:00Z])
+      false
+      iex> DateTime.after?(~U[2021-01-01 11:00:00Z], ~U[2022-02-02 11:00:00Z])
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec after?(Calendar.datetime(), Calendar.datetime()) :: boolean()
+  def after?(datetime1, datetime2) do
+    compare(datetime1, datetime2) == :gt
+  end
+
   @doc """
   Subtracts `datetime2` from `datetime1`.
 
@@ -1459,11 +1539,11 @@ defmodule DateTime do
   ignoring the wall time and timezone changes. For example, if you add
   one day when there are summer time/daylight saving time changes,
   it will also change the time forward or backward by one hour,
-  so the ellapsed time is precisely 24 hours. Similarly, adding just
+  so the elapsed time is precisely 24 hours. Similarly, adding just
   a few seconds to a datetime just before "spring forward" can cause
   wall time to increase by more than an hour.
 
-  While this means this function is precise in terms of ellapsed time,
+  While this means this function is precise in terms of elapsed time,
   its result may be misleading in certain use cases. For example, if a
   user requests a meeting to happen every day at 15:00 and you use this
   function to compute all future meetings by adding day after day, this
