@@ -92,16 +92,23 @@ defmodule NaiveDateTime do
   Prefer using `DateTime.utc_now/0` when possible as, opposite
   to `NaiveDateTime`, it will keep the time zone information.
 
+  You can also provide a time unit to automatically truncate
+  the naive datetime. This is available since v1.15.0.
+
   ## Examples
 
       iex> naive_datetime = NaiveDateTime.utc_now()
       iex> naive_datetime.year >= 2016
       true
 
+      iex> naive_datetime = NaiveDateTime.utc_now(:second)
+      iex> naive_datetime.microsecond
+      {0, 0}
+
   """
   @doc since: "1.4.0"
-  @spec utc_now(Calendar.calendar()) :: t
-  def utc_now(calendar \\ Calendar.ISO)
+  @spec utc_now(Calendar.calendar() | :native | :microsecond | :millisecond | :second) :: t
+  def utc_now(calendar_or_time_unit \\ Calendar.ISO)
 
   def utc_now(Calendar.ISO) do
     {:ok, {year, month, day}, {hour, minute, second}, microsecond} =
@@ -119,12 +126,41 @@ defmodule NaiveDateTime do
     }
   end
 
+  def utc_now(time_unit) when time_unit in [:microsecond, :millisecond, :second, :native] do
+    utc_now(time_unit, Calendar.ISO)
+  end
+
   def utc_now(calendar) do
     calendar
     |> DateTime.utc_now()
     |> DateTime.to_naive()
   end
 
+  @doc """
+  Returns the current naive datetime in UTC, supporting a specific
+  calendar and precision.
+
+  Prefer using `DateTime.utc_now/2` when possible as, opposite
+  to `NaiveDateTime`, it will keep the time zone information.
+
+  ## Examples
+
+      iex> naive_datetime = NaiveDateTime.utc_now(:second, Calendar.ISO)
+      iex> naive_datetime.year >= 2016
+      true
+
+      iex> naive_datetime = NaiveDateTime.utc_now(:second, Calendar.ISO)
+      iex> naive_datetime.microsecond
+      {0, 0}
+
+  """
+  @doc since: "1.15.0"
+  @spec utc_now(:native | :microsecond | :millisecond | :second, Calendar.calendar()) :: t
+  def utc_now(time_unit, calendar)
+      when time_unit in [:native, :microsecond, :millisecond, :second] do
+    DateTime.utc_now(time_unit, calendar) |> DateTime.to_naive()
+  end
+
   @doc """
   Returns the "local time" for the machine the Elixir program is running on.
 
@@ -361,7 +397,8 @@ defmodule NaiveDateTime do
   `t:System.time_unit/0`. It defaults to `:second`. Negative values
   will move backwards in time.
 
-  This function always considers the unit to be computed according to the `Calendar.ISO`.
+  This function always consider the unit to be computed according
+  to the `Calendar.ISO`.
 
   ## Examples
 
@@ -1035,6 +1072,44 @@ defmodule NaiveDateTime do
     end
   end
 
+  @doc """
+  Returns true if the first `NaiveDateTime` is strictly earlier than the second.
+
+  ## Examples
+
+      iex> NaiveDateTime.before?(~N[2021-01-01 11:00:00], ~N[2022-02-02 11:00:00])
+      true
+      iex> NaiveDateTime.before?(~N[2021-01-01 11:00:00], ~N[2021-01-01 11:00:00])
+      false
+      iex> NaiveDateTime.before?(~N[2022-02-02 11:00:00], ~N[2021-01-01 11:00:00])
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec before?(Calendar.naive_datetime(), Calendar.naive_datetime()) :: boolean()
+  def before?(naive_datetime1, naive_datetime2) do
+    compare(naive_datetime1, naive_datetime2) == :lt
+  end
+
+  @doc """
+  Returns true if the first `NaiveDateTime` is strictly later than the second.
+
+  ## Examples
+
+      iex> NaiveDateTime.after?(~N[2022-02-02 11:00:00], ~N[2021-01-01 11:00:00])
+      true
+      iex> NaiveDateTime.after?(~N[2021-01-01 11:00:00], ~N[2021-01-01 11:00:00])
+      false
+      iex> NaiveDateTime.after?(~N[2021-01-01 11:00:00], ~N[2022-02-02 11:00:00])
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec after?(Calendar.naive_datetime(), Calendar.naive_datetime()) :: boolean()
+  def after?(naive_datetime1, naive_datetime2) do
+    compare(naive_datetime1, naive_datetime2) == :gt
+  end
+
   @doc """
   Converts the given `naive_datetime` from one calendar to another.
 
@@ -1130,6 +1205,67 @@ defmodule NaiveDateTime do
     end
   end
 
+  @doc """
+  Calculates a `NaiveDateTime` that is the first moment for the given `NaiveDateTime`.
+
+  To calculate the beginning of day of a `DateTime`, call this function, then convert back to a `DateTime`:
+
+      datetime
+      |> NaiveDateTime.beginning_of_day()
+      |> DateTime.from_naive(datetime.timezone)
+
+  Note that the beginning of the day may not exist or be ambiguous
+  in a given timezone, so you must handle those cases accordingly.
+
+  ## Examples
+
+      iex> NaiveDateTime.beginning_of_day(~N[2000-01-01 23:00:07.123456])
+      ~N[2000-01-01 00:00:00.000000]
+
+  """
+  @doc since: "1.15.0"
+  @spec beginning_of_day(Calendar.naive_datetime()) :: t
+  def beginning_of_day(%{calendar: calendar, microsecond: {_, precision}} = naive_datetime) do
+    naive_datetime
+    |> to_iso_days()
+    |> calendar.iso_days_to_beginning_of_day()
+    |> from_iso_days(calendar, precision)
+  end
+
+  @doc """
+  Calculates a `NaiveDateTime` that is the last moment for the given `NaiveDateTime`.
+
+  To calculate the end of day of a `DateTime`, call this function, then convert back to a `DateTime`:
+
+      datetime
+      |> NaiveDateTime.end_of_day()
+      |> DateTime.from_naive(datetime.timezone)
+
+  Note that the end of the day may not exist or be ambiguous
+  in a given timezone, so you must handle those cases accordingly.
+
+  ## Examples
+
+      iex> NaiveDateTime.end_of_day(~N[2000-01-01 23:00:07.123456])
+      ~N[2000-01-01 23:59:59.999999]
+
+  """
+  @doc since: "1.15.0"
+  @spec end_of_day(Calendar.naive_datetime()) :: t
+  def end_of_day(%{calendar: calendar, microsecond: {_, precision}} = naive_datetime) do
+    end_of_day =
+      naive_datetime
+      |> to_iso_days()
+      |> calendar.iso_days_to_end_of_day()
+      |> from_iso_days(calendar, precision)
+
+    %{microsecond: {microsecond, precision}} = end_of_day
+
+    multiplier = 10 ** (6 - precision)
+
+    %{end_of_day | microsecond: {div(microsecond, multiplier) * multiplier, precision}}
+  end
+
   ## Helpers
 
   defp seconds_from_day_fraction({parts_in_day, @seconds_per_day}),
