@@ -58,7 +58,7 @@ defmodule Code do
   it can't even be loaded.
 
   When invoked, `ensure_compiled/1` and `ensure_compiled!/1` halt the
-  compilation of the caller until the module becomes available. Note
+  compilation of the caller until the module becomes available. Note that
   the distinction between `ensure_compiled/1` and `ensure_compiled!/1`
   is important: if you are using `ensure_compiled!/1`, you are
   indicating to the compiler that you can only continue if said module
@@ -88,7 +88,7 @@ defmodule Code do
 
   ## Compilation tracers
 
-  Elixir supports compilation tracers, which allows modules to observe constructs
+  Elixir supports compilation tracers, which allow modules to observe constructs
   handled by the Elixir compiler when compiling files. A tracer is a module
   that implements the `trace/2` function. The function receives the event name
   as first argument and `Macro.Env` as second and it must return `:ok`. It is
@@ -120,7 +120,8 @@ defmodule Code do
       `{:imported_macro, meta, module, name, arity}` - traced whenever an
       imported function or macro is invoked. `meta` is the call AST metadata,
       `module` is the module the import is from, followed by the `name` and `arity`
-      of the imported function/macro.
+      of the imported function/macro. A :remote_function/:remote_macro event
+      may still be emitted for the imported module/name/arity.
 
     * `{:alias, meta, alias, as, opts}` - traced whenever `alias` is aliased
       to `as`. `meta` is the alias AST metadata and `opts` are the alias options.
@@ -135,7 +136,7 @@ defmodule Code do
 
     * `{:require, meta, module, opts}` - traced whenever `module` is required.
       `meta` is the require AST metadata and `opts` are the require options.
-      If the `meta` option contains the `:from_macro`, then `require` was called
+      If the `meta` option contains the `:from_macro`, then module was called
       from within a macro and therefore must be treated as a compile-time dependency.
 
     * `{:struct_expansion, meta, module, keys}` - traced whenever `module`'s struct
@@ -157,7 +158,7 @@ defmodule Code do
       of keys to traverse in the application environment and `return` is either
       `{:ok, value}` or `:error`.
 
-    * `{:on_module, bytecode, _ignore}` - (since v1.11.0) traced whenever a module
+    * `{:on_module, bytecode, _ignore}` - (since v1.13.0) traced whenever a module
       is defined. This is equivalent to the `@after_compile` callback and invoked
       after any `@after_compile` in the given module. The third element is currently
       `:none` but it may provide more metadata in the future. It is best to ignore
@@ -193,6 +194,22 @@ defmodule Code do
   """
   @type binding :: [{atom() | tuple(), any}]
 
+  @typedoc """
+  Diagnostics returned by the compiler and code evaluation.
+  """
+  @type diagnostic(severity) :: %{
+          required(:file) => Path.t(),
+          required(:severity) => severity,
+          required(:message) => String.t(),
+          required(:position) => position,
+          required(:stacktrace) => Exception.stacktrace(),
+          optional(any()) => any()
+        }
+
+  @typedoc "The line. 0 indicates no line."
+  @type line() :: non_neg_integer()
+  @type position() :: line() | {pos_integer(), column :: non_neg_integer}
+
   @boolean_compiler_options [
     :docs,
     :debug_info,
@@ -204,7 +221,8 @@ defmodule Code do
 
   @list_compiler_options [:no_warn_undefined, :tracers, :parser_options]
 
-  @available_compiler_options @boolean_compiler_options ++ @list_compiler_options
+  @available_compiler_options @boolean_compiler_options ++
+                                @list_compiler_options ++ [:on_undefined_variable]
 
   @doc """
   Lists all required files.
@@ -269,14 +287,15 @@ defmodule Code do
   end
 
   @doc """
-  Appends a path to the end of the Erlang VM code path list.
+  Appends a path to the Erlang VM code path list.
 
   This is the list of directories the Erlang VM uses for
   finding module code. The list of files is managed per Erlang
   VM node.
 
   The path is expanded with `Path.expand/1` before being appended.
-  If this path does not exist, an error is returned.
+  It requires the path to exist. Returns a boolean indicating if
+  the path was successfully added.
 
   ## Examples
 
@@ -284,22 +303,30 @@ defmodule Code do
       #=> true
 
       Code.append_path("/does_not_exist")
-      #=> {:error, :bad_directory}
+      #=> false
+
+  ## Options
+
+    * `:cache` - (since v1.15.0) when true, the code path is cached
+      the first time it is traversed in order to reduce file system
+      operations. It requires Erlang/OTP 26, otherwise it is a no-op.
 
   """
-  @spec append_path(Path.t()) :: true | {:error, :bad_directory}
-  def append_path(path) do
-    :code.add_pathz(to_charlist(Path.expand(path)))
+  @spec append_path(Path.t(), cache: boolean()) :: true | false
+  def append_path(path, opts \\ []) do
+    apply(:code, :add_pathz, [to_charlist(Path.expand(path)) | cache(opts)]) == true
   end
 
   @doc """
-  Prepends a path to the beginning of the Erlang VM code path list.
+  Prepends a path to the Erlang VM code path list.
 
-  This is the list of directories the Erlang VM uses for finding
-  module code. The list of files is managed per Erlang VM node.
+  This is the list of directories the Erlang VM uses for
+  finding module code. The list of files is managed per Erlang
+  VM node.
 
   The path is expanded with `Path.expand/1` before being prepended.
-  If this path does not exist, an error is returned.
+  It requires the path to exist. Returns a boolean indicating if
+  the path was successfully added.
 
   ## Examples
 
@@ -307,12 +334,84 @@ defmodule Code do
       #=> true
 
       Code.prepend_path("/does_not_exist")
-      #=> {:error, :bad_directory}
+      #=> false
+
+  ## Options
+
+    * `:cache` - (since v1.15.0) when true, the code path is cached
+      the first time it is traversed in order to reduce file system
+      operations. It requires Erlang/OTP 26, otherwise it is a no-op.
 
   """
-  @spec prepend_path(Path.t()) :: true | {:error, :bad_directory}
-  def prepend_path(path) do
-    :code.add_patha(to_charlist(Path.expand(path)))
+  @spec prepend_path(Path.t(), cache: boolean()) :: boolean()
+  def prepend_path(path, opts \\ []) do
+    apply(:code, :add_patha, [to_charlist(Path.expand(path)) | cache(opts)]) == true
+  end
+
+  @doc """
+  Prepends a list of `paths` to the Erlang VM code path list.
+
+  This is the list of directories the Erlang VM uses for
+  finding module code. The list of files is managed per Erlang
+  VM node.
+
+  All paths are expanded with `Path.expand/1` before being prepended.
+  Only existing paths are prepended. This function always returns `:ok`,
+  regardless of how many paths were prepended. Use `prepend_path/1`
+  if you need more control.
+
+  ## Examples
+
+      Code.prepend_paths([".", "/does_not_exist"])
+      #=> :ok
+
+  ## Options
+
+    * `:cache` - when true, the code path is cached the first time
+      it is traversed in order to reduce file system operations.
+      It requires Erlang/OTP 26, otherwise it is a no-op.
+  """
+  @doc since: "1.15.0"
+  @spec prepend_paths([Path.t()], cache: boolean()) :: :ok
+  def prepend_paths(paths, opts \\ []) when is_list(paths) do
+    apply(:code, :add_pathsa, [Enum.map(paths, &to_charlist(Path.expand(&1))) | cache(opts)])
+  end
+
+  @doc """
+  Appends a list of `paths` to the Erlang VM code path list.
+
+  This is the list of directories the Erlang VM uses for
+  finding module code. The list of files is managed per Erlang
+  VM node.
+
+  All paths are expanded with `Path.expand/1` before being appended.
+  Only existing paths are appended. This function always returns `:ok`,
+  regardless of how many paths were appended. Use `append_path/1`
+  if you need more control.
+
+  ## Examples
+
+      Code.append_paths([".", "/does_not_exist"])
+      #=> :ok
+
+  ## Options
+
+    * `:cache` - when true, the code path is cached the first time
+      it is traversed in order to reduce file system operations.
+      It requires Erlang/OTP 26, otherwise it is a no-op.
+  """
+  @doc since: "1.15.0"
+  @spec append_paths([Path.t()], cache: boolean()) :: :ok
+  def append_paths(paths, opts \\ []) when is_list(paths) do
+    apply(:code, :add_pathsz, [Enum.map(paths, &to_charlist(Path.expand(&1))) | cache(opts)])
+  end
+
+  defp cache(opts) do
+    if function_exported?(:code, :add_path, 2) do
+      if opts[:cache], do: [:cache], else: [:nocache]
+    else
+      []
+    end
   end
 
   @doc """
@@ -346,6 +445,25 @@ defmodule Code do
     end
   end
 
+  @doc """
+  Deletes a list of paths from the Erlang VM code path list.
+
+  This is the list of directories the Erlang VM uses for finding
+  module code. The list of files is managed per Erlang VM node.
+
+  The path is expanded with `Path.expand/1` before being deleted. If the
+  path does not exist, this function returns `false`.
+  """
+  @doc since: "1.15.0"
+  @spec delete_paths([Path.t()]) :: :ok
+  def delete_paths(paths) when is_list(paths) do
+    for path <- paths do
+      _ = :code.del_path(to_charlist(Path.expand(path)))
+    end
+
+    :ok
+  end
+
   @doc """
   Evaluates the contents given by `string`.
 
@@ -432,6 +550,52 @@ defmodule Code do
     end)
   end
 
+  @doc """
+  Executes the given `fun` and capture all diagnostics.
+
+  Diagnostics are warnings and errors emitted by the compiler
+  and by functions such as `IO.warn/2`.
+
+  ## Options
+
+    * `:log` - if the diagnostics should be logged as they happen.
+      Defaults to `false`.
+
+  """
+  @doc since: "1.15.0"
+  @spec with_diagnostics(keyword(), (-> result)) :: {result, [diagnostic(:warning | :error)]}
+        when result: term()
+  def with_diagnostics(opts \\ [], fun) do
+    value = :erlang.get(:elixir_code_diagnostics)
+    log = Keyword.get(opts, :log, false)
+    :erlang.put(:elixir_code_diagnostics, {[], log})
+
+    try do
+      result = fun.()
+      {diagnostics, _log?} = :erlang.get(:elixir_code_diagnostics)
+      {result, Enum.reverse(diagnostics)}
+    after
+      if value == :undefined do
+        :erlang.erase(:elixir_code_diagnostics)
+      else
+        :erlang.put(:elixir_code_diagnostics, value)
+      end
+    end
+  end
+
+  @doc """
+  Prints a diagnostic into the standard error.
+
+  A diagnostic is either returned by `Kernel.ParallelCompiler`
+  or by `Code.with_diagnostics/2`.
+  """
+  @doc since: "1.15.0"
+  @spec print_diagnostic(diagnostic(:warning | :error)) :: :ok
+  def print_diagnostic(diagnostic) do
+    :elixir_errors.print_diagnostic(diagnostic)
+    :ok
+  end
+
   @doc ~S"""
   Formats the given code `string`.
 
@@ -471,6 +635,11 @@ defmodule Code do
       modifiers, where `<<foo::custom_type>>` becomes `<<foo::custom_type()>>`.
       Defaults to `true`. This option changes the AST.
 
+    * `:normalize_charlists_as_sigils` (since v1.15.0) - when `true`,
+      formats charlists as [`~c`](`Kernel.sigil_c/2`) sigils, for example
+      `'foo'` becomes `~c"foo"`.
+      Defaults to `true`. This option changes the AST.
+
   ## Design principles
 
   The formatter was designed under three principles.
@@ -613,8 +782,8 @@ defmodule Code do
   The formatter respects the input format in some cases. Those are
   listed below:
 
-    * Insignificant digits in numbers are kept as is. The formatter
-      however always inserts underscores for decimal numbers with more
+    * Insignificant digits in numbers are kept as is. The formatter,
+      however, always inserts underscores for decimal numbers with more
       than 5 digits and converts hexadecimal digits to uppercase
 
     * Strings, charlists, atoms and sigils are kept as is. No character
@@ -770,7 +939,7 @@ defmodule Code do
         warn_on_unnecessary_quotes: false,
         literal_encoder: &{:ok, {:__block__, &2, [&1]}},
         token_metadata: true,
-        emit_warnings: false
+        warnings: false
       ] ++ opts
 
     {forms, comments} = string_to_quoted_with_comments!(string, to_quoted_opts)
@@ -967,7 +1136,10 @@ defmodule Code do
     * syntax keywords (`fn`, `do`, `else`, and so on)
 
     * atoms containing interpolation (`:"#{1 + 1} is two"`), as these
-      atoms are constructed at runtime.
+      atoms are constructed at runtime
+
+    * atoms used to represent single-letter sigils like `:sigil_X`
+      (but multi-letter sigils like `:sigil_XYZ` are encoded).
 
   """
   @spec string_to_quoted(List.Chars.t(), keyword) ::
@@ -1110,10 +1282,10 @@ defmodule Code do
     Process.put(:code_formatter_comments, [comment | comments])
   end
 
-  defp next_eol_count('\s' ++ rest, count), do: next_eol_count(rest, count)
-  defp next_eol_count('\t' ++ rest, count), do: next_eol_count(rest, count)
-  defp next_eol_count('\n' ++ rest, count), do: next_eol_count(rest, count + 1)
-  defp next_eol_count('\r\n' ++ rest, count), do: next_eol_count(rest, count + 1)
+  defp next_eol_count([?\s | rest], count), do: next_eol_count(rest, count)
+  defp next_eol_count([?\t | rest], count), do: next_eol_count(rest, count)
+  defp next_eol_count([?\n | rest], count), do: next_eol_count(rest, count + 1)
+  defp next_eol_count([?\r, ?\n | rest], count), do: next_eol_count(rest, count + 1)
   defp next_eol_count(_, count), do: count
 
   defp previous_eol_count([{token, {_, _, count}} | _])
@@ -1176,6 +1348,9 @@ defmodule Code do
       The arity may be the atom `:*`, which implies all arities of
       that name. The formatter already includes a list of functions
       and this option augments this list.
+
+    * `:syntax_colors` - a keyword list of colors the output is colorized.
+      See `Inspect.Opts` for more information.
   """
   @doc since: "1.13.0"
   @spec quoted_to_algebra(Macro.t(), keyword) :: Inspect.Algebra.t()
@@ -1293,13 +1468,15 @@ defmodule Code do
   options and for a description of all options, see
   `put_compiler_option/2`.
 
+  Returns a map with previous values.
+
   ## Examples
 
-      Code.compiler_options()
-      #=> %{debug_info: true, docs: true, ...}
+      Code.compiler_options(warnings_as_errors: true)
+      #=> %{warnings_as_errors: false}
 
   """
-  @spec compiler_options(Enumerable.t()) :: %{optional(atom) => boolean}
+  @spec compiler_options(Enumerable.t({atom, term})) :: %{optional(atom) => term}
   def compiler_options(opts) do
     for {key, value} <- opts, into: %{} do
       previous = get_compiler_option(key)
@@ -1353,16 +1530,20 @@ defmodule Code do
       Defaults to `true`.
 
     * `:debug_info` - when `true`, retains debug information in the compiled
-      module. This enables static analysis tools as it allows developers to
+      module. Defaults to `true`.
+      This enables static analysis tools as it allows developers to
       partially reconstruct the original source code. Therefore, disabling
       `:debug_info` is not recommended as it removes the ability of the
       Elixir compiler and other tools to provide feedback. If you want to
       remove the `:debug_info` while deploying, tools like `mix release`
       already do such by default.
+      Additionally, `mix test` disables it via the `:test_elixirc_options`
+      project configuration option.
+      This option can also be overridden per module using the `@compile` directive.
 
-    * `:ignore_already_consolidated` - when `true`, does not warn when a protocol
-      has already been consolidated and a new implementation is added. Defaults
-      to `false`.
+    * `:ignore_already_consolidated` (since v1.10.0) - when `true`, does not warn
+      when a protocol has already been consolidated and a new implementation is added.
+      Defaults to `false`.
 
     * `:ignore_module_conflict` - when `true`, does not warn when a module has
       already been defined. Defaults to `false`.
@@ -1387,6 +1568,17 @@ defmodule Code do
       `string_to_quoted/2` (except by the options that change the AST itself).
       This can be used in combination with the tracer to retrieve localized
       information about events happening during compilation. Defaults to `[]`.
+      This option only affects code compilation functions, such as `compile_string/2`
+      and `compile_file/2` but not `string_to_quoted/2` and friends, as the
+      latter is used for other purposes beyond compilation.
+
+    * `:on_undefined_variable` (since v1.15.0) - either `:raise` or `:warn`.
+      When `:raise` (the default), undefined variables will trigger a compilation
+      error. You may be set it to `:warn` if you want undefined variables to
+      emit a warning and expand as to a local call to the zero-arity function
+      of the same name (for example, `node` would be expanded as `node()`).
+      This `:warn` behaviour only exists for compatibility reasons when working
+      with old dependencies.
 
   It always returns `:ok`. Raises an error for invalid options.
 
@@ -1436,6 +1628,12 @@ defmodule Code do
     :ok
   end
 
+  # TODO: Make this option have no effect on Elixir v2.0
+  def put_compiler_option(:on_undefined_variable, value) when value in [:raise, :warn] do
+    :elixir_config.put(:on_undefined_variable, value)
+    :ok
+  end
+
   def put_compiler_option(key, _value) do
     raise "unknown compiler option: #{inspect(key)}"
   end
@@ -1558,7 +1756,7 @@ defmodule Code do
 
   ## Examples
 
-      iex> Code.ensure_loaded?(Atom)
+      iex> Code.ensure_loaded?(String)
       true
 
   """
@@ -1583,6 +1781,53 @@ defmodule Code do
     end
   end
 
+  @doc """
+  Ensures the given modules are loaded.
+
+  Similar to `ensure_loaded/1`, but accepts a list of modules instead of a single
+  module, and loads all of them.
+
+  If all modules load successfully, returns `:ok`. Otherwise, returns `{:error, errors}`
+  where `errors` is a list of tuples made of the module and the reason it failed to load.
+
+  ## Examples
+
+      iex> Code.ensure_all_loaded([Atom, String])
+      :ok
+
+      iex> Code.ensure_all_loaded([Atom, DoesNotExist])
+      {:error, [{DoesNotExist, :nofile}]}
+
+  """
+  @doc since: "1.15.0"
+  @spec ensure_all_loaded([module]) :: :ok | {:error, [{module, reason}]}
+        when reason: :badfile | :nofile | :on_load_failure
+  def ensure_all_loaded(modules) when is_list(modules) do
+    :code.ensure_modules_loaded(modules)
+  end
+
+  @doc """
+  Same as `ensure_all_loaded/1` but raises if any of the modules cannot be loaded.
+  """
+  @doc since: "1.15.0"
+  @spec ensure_all_loaded!([module]) :: :ok
+  def ensure_all_loaded!(modules) do
+    case ensure_all_loaded(modules) do
+      :ok ->
+        :ok
+
+      {:error, errors} ->
+        formatted_errors =
+          errors
+          |> Enum.sort()
+          |> Enum.map_join("\n", fn {module, reason} ->
+            "  * #{inspect(module)} due to reason #{inspect(reason)}"
+          end)
+
+        raise ArgumentError, "could not load the following modules:\n\n" <> formatted_errors
+    end
+  end
+
   @doc """
   Similar to `ensure_compiled!/1` but indicates you can continue without said module.
 
@@ -1663,6 +1908,27 @@ defmodule Code do
     end
   end
 
+  @doc """
+  Returns `true` if the module is loaded.
+
+  This function doesn't attempt to load the module. For such behaviour,
+  `ensure_loaded?/1` can be used.
+
+  ## Examples
+
+      iex> Code.loaded?(Atom)
+      true
+
+      iex> Code.loaded?(NotYetLoaded)
+      false
+
+  """
+  @doc since: "1.15.0"
+  @spec loaded?(module) :: boolean
+  def loaded?(module) do
+    :erlang.module_loaded(module)
+  end
+
   @doc """
   Returns true if the current process can await for module compilation.
 
@@ -1738,8 +2004,8 @@ defmodule Code do
         end
 
       :error ->
-        case :code.which(module) do
-          :preloaded ->
+        case :code.is_loaded(module) do
+          {:file, :preloaded} ->
             # The ERTS directory is not necessarily included in releases
             # unless it is listed as an extra application.
             case :code.lib_dir(:erts) do
@@ -1770,7 +2036,7 @@ defmodule Code do
     end
   end
 
-  @docs_chunk 'Docs'
+  @docs_chunk [?D, ?o, ?c, ?s]
 
   defp fetch_docs_from_beam(bin_or_path) do
     case :beam_lib.chunks(bin_or_path, [@docs_chunk]) do
