@@ -6,6 +6,8 @@ defmodule Code.Formatter do
   @double_heredoc "\"\"\""
   @single_quote "'"
   @single_heredoc "'''"
+  @sigil_c "~c\""
+  @sigil_c_heredoc "~c\"\"\""
   @newlines 2
   @min_line 0
   @max_line 9_999_999
@@ -30,7 +32,7 @@ defmodule Code.Formatter do
   # Operators that are logical cannot be mixed without parens
   @required_parens_logical_binary_operands [:||, :|||, :or, :&&, :&&&, :and]
 
-  # Operators with next break fits. = and :: do not consider new lines though
+  # Operators with next break fits
   @next_break_fits_operators [:<-, :==, :!=, :=~, :===, :!==, :<, :>, :<=, :>=, :=, :"::"]
 
   # Operators that always require parens on operands when they are the parent
@@ -141,24 +143,6 @@ defmodule Code.Formatter do
 
   @do_end_keywords [:rescue, :catch, :else, :after]
 
-  @bitstring_modifiers [
-    :integer,
-    :float,
-    :bits,
-    :bitstring,
-    :binary,
-    :bytes,
-    :utf8,
-    :utf16,
-    :utf32,
-    :signed,
-    :unsigned,
-    :little,
-    :big,
-    :native,
-    :_
-  ]
-
   @doc """
   Converts the quoted expression into an algebra document.
   """
@@ -198,17 +182,19 @@ defmodule Code.Formatter do
     file = Keyword.get(opts, :file, nil)
     sigils = Keyword.get(opts, :sigils, [])
     normalize_bitstring_modifiers = Keyword.get(opts, :normalize_bitstring_modifiers, true)
+    normalize_charlists_as_sigils = Keyword.get(opts, :normalize_charlists_as_sigils, true)
+    syntax_colors = Keyword.get(opts, :syntax_colors, [])
 
     sigils =
       Map.new(sigils, fn {key, value} ->
         with true <- is_atom(key) and is_function(value, 2),
-             [char] <- Atom.to_charlist(key),
-             true <- char in ?A..?Z do
-          {char, value}
+             name = Atom.to_charlist(key),
+             true <- Enum.all?(name, &(&1 in ?A..?Z)) do
+          {name, value}
         else
           _ ->
             raise ArgumentError,
-                  ":sigils must be a keyword list with a single uppercased letter as key and an " <>
+                  ":sigils must be a keyword list with uppercased atoms as keys and an " <>
                     "anonymous function expecting two arguments as value, got: #{inspect(sigils)}"
         end
       end)
@@ -221,7 +207,9 @@ defmodule Code.Formatter do
       comments: comments,
       sigils: sigils,
       file: file,
-      normalize_bitstring_modifiers: normalize_bitstring_modifiers
+      normalize_bitstring_modifiers: normalize_bitstring_modifiers,
+      normalize_charlists_as_sigils: normalize_charlists_as_sigils,
+      inspect_opts: %Inspect.Opts{syntax_colors: syntax_colors}
     }
   end
 
@@ -277,7 +265,7 @@ defmodule Code.Formatter do
   end
 
   defp quoted_to_algebra({var, _meta, var_context}, _context, state) when is_atom(var_context) do
-    {var |> Atom.to_string() |> string(), state}
+    {var |> Atom.to_string() |> string() |> color(:variable, state.inspect_opts), state}
   end
 
   defp quoted_to_algebra({:<<>>, meta, entries}, _context, state) do
@@ -311,15 +299,18 @@ defmodule Code.Formatter do
         remote_to_algebra(quoted, context, state)
 
       meta[:delimiter] == ~s['''] ->
+        {opener, quotes} = get_charlist_quotes(true, state)
+
         {doc, state} =
           entries
           |> prepend_heredoc_line()
-          |> list_interpolation_to_algebra(~s['''], state, @single_heredoc, @single_heredoc)
+          |> list_interpolation_to_algebra(quotes, state, opener, quotes)
 
         {force_unfit(doc), state}
 
       true ->
-        list_interpolation_to_algebra(entries, @single_quote, state, @single_quote, @single_quote)
+        {opener, quotes} = get_charlist_quotes(false, state)
+        list_interpolation_to_algebra(entries, quotes, state, opener, quotes)
     end
   end
 
@@ -377,12 +368,14 @@ defmodule Code.Formatter do
   defp quoted_to_algebra({:__block__, meta, [list]}, _context, state) when is_list(list) do
     case meta[:delimiter] do
       ~s['''] ->
-        string = list |> List.to_string() |> escape_heredoc(~s['''])
-        {@single_heredoc |> concat(string) |> concat(@single_heredoc) |> force_unfit(), state}
+        {opener, quotes} = get_charlist_quotes(true, state)
+        string = list |> List.to_string() |> escape_heredoc(quotes)
+        {opener |> concat(string) |> concat(quotes) |> force_unfit(), state}
 
       ~s['] ->
-        string = list |> List.to_string() |> escape_string(@single_quote)
-        {@single_quote |> concat(string) |> concat(@single_quote), state}
+        {opener, quotes} = get_charlist_quotes(false, state)
+        string = list |> List.to_string() |> escape_string(quotes)
+        {opener |> concat(string) |> concat(quotes), state}
 
       _other ->
         list_to_algebra(meta, list, state)
@@ -392,24 +385,33 @@ defmodule Code.Formatter do
   defp quoted_to_algebra({:__block__, meta, [string]}, _context, state) when is_binary(string) do
     if meta[:delimiter] == ~s["""] do
       string = escape_heredoc(string, ~s["""])
-      {@double_heredoc |> concat(string) |> concat(@double_heredoc) |> force_unfit(), state}
+
+      {@double_heredoc
+       |> concat(string)
+       |> concat(@double_heredoc)
+       |> color(:string, state.inspect_opts)
+       |> force_unfit(), state}
     else
       string = escape_string(string, @double_quote)
-      {@double_quote |> concat(string) |> concat(@double_quote), state}
+
+      {@double_quote
+       |> concat(string)
+       |> concat(@double_quote)
+       |> color(:string, state.inspect_opts), state}
     end
   end
 
   defp quoted_to_algebra({:__block__, meta, [atom]}, _context, state) when is_atom(atom) do
-    {atom_to_algebra(atom, meta), state}
+    {atom_to_algebra(atom, meta, state.inspect_opts), state}
   end
 
   defp quoted_to_algebra({:__block__, meta, [integer]}, _context, state)
        when is_integer(integer) do
-    {integer_to_algebra(Keyword.fetch!(meta, :token)), state}
+    {Keyword.fetch!(meta, :token) |> integer_to_algebra(state.inspect_opts), state}
   end
 
   defp quoted_to_algebra({:__block__, meta, [float]}, _context, state) when is_float(float) do
-    {float_to_algebra(Keyword.fetch!(meta, :token)), state}
+    {Keyword.fetch!(meta, :token) |> float_to_algebra(state.inspect_opts), state}
   end
 
   defp quoted_to_algebra(
@@ -426,10 +428,10 @@ defmodule Code.Formatter do
   end
 
   defp quoted_to_algebra({:__block__, _meta, []}, _context, state) do
-    {"nil", state}
+    {color("nil", nil, state.inspect_opts), state}
   end
 
-  defp quoted_to_algebra({:__block__, meta, _} = block, _context, state) do
+  defp quoted_to_algebra({:__block__, meta, args} = block, _context, state) when is_list(args) do
     {block, state} = block_to_algebra(block, line(meta), closing_line(meta), state)
     {surround("(", block, ")"), state}
   end
@@ -442,7 +444,8 @@ defmodule Code.Formatter do
         quoted_to_algebra_with_parens_if_operator(head, context, state)
       end
 
-    {Enum.reduce(tail, doc, &concat(&2, "." <> Atom.to_string(&1))), state}
+    {Enum.reduce(tail, doc, &concat(&2, "." <> Atom.to_string(&1)))
+     |> color(:atom, state.inspect_opts), state}
   end
 
   # &1
@@ -496,7 +499,7 @@ defmodule Code.Formatter do
 
   # (left -> right)
   defp quoted_to_algebra([{:->, _, _} | _] = clauses, _context, state) do
-    type_fun_to_algebra(clauses, @max_line, @min_line, state)
+    paren_fun_to_algebra(clauses, @max_line, @min_line, state)
   end
 
   # [keyword: :list] (inner part)
@@ -524,7 +527,7 @@ defmodule Code.Formatter do
                   IO.iodata_to_binary([?", Atom.to_string(atom), ?", ?:])
                 end
 
-              {string(key), state}
+              {string(key) |> color(:atom, state.inspect_opts), state}
 
             {{:., _, [:erlang, :binary_to_atom]}, _, [{:<<>>, _, entries}, :utf8]} ->
               interpolation_to_algebra(entries, @double_quote, state, "\"", "\":")
@@ -554,8 +557,8 @@ defmodule Code.Formatter do
 
   ## Blocks
 
-  defp block_to_algebra([{:->, _, _} | _] = type_fun, min_line, max_line, state) do
-    type_fun_to_algebra(type_fun, min_line, max_line, state)
+  defp block_to_algebra([{:->, _, _} | _] = paren_fun, min_line, max_line, state) do
+    paren_fun_to_algebra(paren_fun, min_line, max_line, state)
   end
 
   defp block_to_algebra({:__block__, _, []}, min_line, max_line, state) do
@@ -619,7 +622,7 @@ defmodule Code.Formatter do
         Atom.to_string(op)
       end
 
-    {concat(op_string, doc), state}
+    {color(op_string, :operator, state.inspect_opts) |> concat(doc), state}
   end
 
   defp maybe_binary_op_to_algebra(fun, meta, args, context, state) do
@@ -725,11 +728,12 @@ defmodule Code.Formatter do
     doc =
       cond do
         op in @no_space_binary_operators ->
-          concat(concat(group(left), op_string), group(right))
+          op_doc = color(op_string, :operator, state.inspect_opts)
+          concat(concat(group(left), op_doc), group(right))
 
         op in @no_newline_binary_operators ->
-          op_string = " " <> op_string <> " "
-          concat(concat(group(left), op_string), group(right))
+          op_doc = color(" " <> op_string <> " ", :operator, state.inspect_opts)
+          concat(concat(group(left), op_doc), group(right))
 
         true ->
           eol? = eol?(meta, state)
@@ -738,8 +742,8 @@ defmodule Code.Formatter do
             op in @next_break_fits_operators and next_break_fits?(right_arg, state) and not eol?
 
           with_next_break_fits(next_break_fits?, right, fn right ->
-            op_string = " " <> op_string
-            right = nest(glue(op_string, group(right)), nesting, :break)
+            op_doc = color(" " <> op_string, :operator, state.inspect_opts)
+            right = nest(glue(op_doc, group(right)), nesting, :break)
             right = if eol?, do: force_unfit(right), else: right
             concat(group(left), group(right))
           end)
@@ -925,10 +929,10 @@ defmodule Code.Formatter do
   defp capture_to_algebra(arg, context, state) do
     {doc, state} = capture_target_to_algebra(arg, context, state)
 
-    if doc |> format_to_string() |> String.starts_with?("&") do
-      {concat("& ", doc), state}
-    else
-      {concat("&", doc), state}
+    case format_to_string(doc) do
+      <<"&", _::binary>> -> {concat("& ", doc), state}
+      <<int, _::binary>> when int in ?0..?9 -> {concat("& ", doc), state}
+      _ -> {concat("&", doc), state}
     end
   end
 
@@ -984,8 +988,11 @@ defmodule Code.Formatter do
   defp remote_to_algebra({{:., _, [target, fun]}, meta, args}, context, state)
        when is_atom(fun) do
     {target_doc, state} = remote_target_to_algebra(target, state)
-    fun = Macro.inspect_atom(:remote_call, fun)
-    remote_doc = target_doc |> concat(".") |> concat(string(fun))
+
+    fun_doc =
+      Macro.inspect_atom(:remote_call, fun) |> string() |> color(:call, state.inspect_opts)
+
+    remote_doc = target_doc |> concat(".") |> concat(fun_doc)
 
     if args == [] and not remote_target_is_a_module?(target) and not meta?(meta, :closing) do
       {remote_doc, state}
@@ -1051,6 +1058,7 @@ defmodule Code.Formatter do
       fun
       |> Atom.to_string()
       |> string()
+      |> color(:call, state.inspect_opts)
       |> concat(call_doc)
 
     doc = if wrap_in_parens?, do: wrap_in_parens(doc), else: doc
@@ -1151,6 +1159,7 @@ defmodule Code.Formatter do
               right_doc
               |> nest(2, :break)
               |> concat(break(""))
+              |> concat(")")
               |> group(:inherit)
               |> next_break_fits(:enabled)
 
@@ -1182,7 +1191,7 @@ defmodule Code.Formatter do
           if skip_parens? do
             nest(args_doc, :cursor, :break)
           else
-            nest(args_doc, 2, :break) |> concat(break(""))
+            nest(args_doc, 2, :break) |> concat(break("")) |> concat(")")
           end
 
         {args_doc, next_break_fits?, state}
@@ -1208,7 +1217,6 @@ defmodule Code.Formatter do
           |> concat(break(""))
           |> nest(2, :break)
           |> concat(args_doc)
-          |> concat(")")
           |> concat(extra)
           |> group()
       end
@@ -1329,18 +1337,18 @@ defmodule Code.Formatter do
   ## Sigils
 
   defp maybe_sigil_to_algebra(fun, meta, args, state) do
-    with <<"sigil_", name>> <- Atom.to_string(fun),
+    with <<"sigil_", name::binary>> <- Atom.to_string(fun),
          [{:<<>>, _, entries}, modifiers] when is_list(modifiers) <- args,
          opening_delimiter when not is_nil(opening_delimiter) <- meta[:delimiter] do
-      doc = <<?~, name, opening_delimiter::binary>>
+      doc = <<?~, name::binary, opening_delimiter::binary>>
 
       entries =
-        case state.sigils do
-          %{^name => callback} ->
+        case Map.fetch(state.sigils, String.to_charlist(name)) do
+          {:ok, callback} ->
             metadata = [
               file: state.file,
               line: meta[:line],
-              sigil: List.to_atom([name]),
+              sigil: String.to_atom(name),
               modifiers: modifiers,
               opening_delimiter: opening_delimiter
             ]
@@ -1354,7 +1362,7 @@ defmodule Code.Formatter do
                       "expected sigil callback to return iodata, got: #{inspect(other)}"
             end
 
-          %{} ->
+          :error ->
             entries
         end
 
@@ -1411,7 +1419,9 @@ defmodule Code.Formatter do
 
   defp bitstring_segment_to_algebra({{:"::", _, [segment, spec]}, i}, state, last) do
     {doc, state} = quoted_to_algebra(segment, :parens_arg, state)
-    {spec, state} = bitstring_spec_to_algebra(spec, state, state.normalize_bitstring_modifiers)
+
+    {spec, state} =
+      bitstring_spec_to_algebra(spec, state, state.normalize_bitstring_modifiers, :"::")
 
     spec = wrap_in_parens_if_inspected_atom(spec)
     spec = if i == last, do: bitstring_wrap_parens(spec, i, last), else: spec
@@ -1430,15 +1440,17 @@ defmodule Code.Formatter do
     {bitstring_wrap_parens(doc, i, last), state}
   end
 
-  defp bitstring_spec_to_algebra({op, _, [left, right]}, state, normalize_modifiers)
+  defp bitstring_spec_to_algebra({op, _, [left, right]}, state, normalize_modifiers, paren_op)
        when op in [:-, :*] do
     normalize_modifiers = normalize_modifiers && op != :*
-    {left, state} = bitstring_spec_to_algebra(left, state, normalize_modifiers)
+    {left, state} = bitstring_spec_to_algebra(left, state, normalize_modifiers, op)
     {right, state} = bitstring_spec_element_to_algebra(right, state, normalize_modifiers)
-    {concat(concat(left, Atom.to_string(op)), right), state}
+    doc = concat(concat(left, Atom.to_string(op)), right)
+    doc = if paren_op == :*, do: wrap_in_parens(doc), else: doc
+    {doc, state}
   end
 
-  defp bitstring_spec_to_algebra(spec, state, normalize_modifiers) do
+  defp bitstring_spec_to_algebra(spec, state, normalize_modifiers, _paren_op) do
     bitstring_spec_element_to_algebra(spec, state, normalize_modifiers)
   end
 
@@ -1456,8 +1468,14 @@ defmodule Code.Formatter do
     quoted_to_algebra_with_parens_if_operator(spec_element, :parens_arg, state)
   end
 
-  defp bitstring_spec_normalize_empty_args(atom) when atom in @bitstring_modifiers, do: nil
-  defp bitstring_spec_normalize_empty_args(_atom), do: []
+  defp bitstring_spec_normalize_empty_args(:_), do: nil
+
+  defp bitstring_spec_normalize_empty_args(atom) do
+    case :elixir_bitstring.validate_spec(atom, nil) do
+      :none -> []
+      _ -> nil
+    end
+  end
 
   defp bitstring_wrap_parens(doc, i, last) when i == 0 or i == last do
     string = format_to_string(doc)
@@ -1481,7 +1499,10 @@ defmodule Code.Formatter do
     {args_doc, _join, state} =
       args_to_algebra_with_comments(args, meta, false, :none, join, state, fun)
 
-    {surround("[", args_doc, "]"), state}
+    left_bracket = color("[", :list, state.inspect_opts)
+    right_bracket = color("]", :list, state.inspect_opts)
+
+    {surround(left_bracket, args_doc, right_bracket), state}
   end
 
   defp map_to_algebra(meta, name_doc, [{:|, _, [left, right]}], state) do
@@ -1497,8 +1518,7 @@ defmodule Code.Formatter do
       |> wrap_in_parens_if_binary_operator(left)
       |> glue(concat("| ", nest(right_doc, 2)))
 
-    name_doc = "%" |> concat(name_doc) |> concat("{")
-    {surround(name_doc, args_doc, "}"), state}
+    do_map_to_algebra(name_doc, args_doc, state)
   end
 
   defp map_to_algebra(meta, name_doc, args, state) do
@@ -1508,8 +1528,12 @@ defmodule Code.Formatter do
     {args_doc, _join, state} =
       args_to_algebra_with_comments(args, meta, false, :none, join, state, fun)
 
-    name_doc = "%" |> concat(name_doc) |> concat("{")
-    {surround(name_doc, args_doc, "}"), state}
+    do_map_to_algebra(name_doc, args_doc, state)
+  end
+
+  defp do_map_to_algebra(name_doc, args_doc, state) do
+    name_doc = "%" |> concat(name_doc) |> concat("{") |> color(:map, state.inspect_opts)
+    {surround(name_doc, args_doc, color("}", :map, state.inspect_opts)), state}
   end
 
   defp tuple_to_algebra(meta, args, join, state) do
@@ -1519,23 +1543,30 @@ defmodule Code.Formatter do
     {args_doc, join, state} =
       args_to_algebra_with_comments(args, meta, false, :none, join, state, fun)
 
+    left_bracket = color("{", :tuple, state.inspect_opts)
+    right_bracket = color("}", :tuple, state.inspect_opts)
+
     if join == :flex_break do
-      {"{" |> concat(args_doc) |> nest(1) |> concat("}") |> group(), state}
+      {left_bracket |> concat(args_doc) |> nest(1) |> concat(right_bracket) |> group(), state}
     else
-      {surround("{", args_doc, "}"), state}
+      {surround(left_bracket, args_doc, right_bracket), state}
     end
   end
 
-  defp atom_to_algebra(atom, _) when atom in [nil, true, false] do
-    Atom.to_string(atom)
+  defp atom_to_algebra(atom, _, inspect_opts) when atom in [true, false] do
+    Atom.to_string(atom) |> color(:boolean, inspect_opts)
+  end
+
+  defp atom_to_algebra(nil, _, inspect_opts) do
+    Atom.to_string(nil) |> color(nil, inspect_opts)
   end
 
   # TODO: Remove this clause in v1.16 when we no longer quote operator :..//
-  defp atom_to_algebra(:"..//", _) do
-    string(":\"..//\"")
+  defp atom_to_algebra(:"..//", _, inspect_opts) do
+    string(":\"..//\"") |> color(:atom, inspect_opts)
   end
 
-  defp atom_to_algebra(:\\, meta) do
+  defp atom_to_algebra(:\\, meta, inspect_opts) do
     # Since we parse strings without unescaping, the atoms
     # :\\ and :"\\" have the same representation, so we need
     # to check the delimiter and handle them accordingly.
@@ -1545,10 +1576,10 @@ defmodule Code.Formatter do
         _ -> ":\\\\"
       end
 
-    string(string)
+    string(string) |> color(:atom, inspect_opts)
   end
 
-  defp atom_to_algebra(atom, _) do
+  defp atom_to_algebra(atom, _, inspect_opts) do
     string = Atom.to_string(atom)
 
     iodata =
@@ -1558,10 +1589,10 @@ defmodule Code.Formatter do
         [?:, ?", String.replace(string, "\"", "\\\""), ?"]
       end
 
-    iodata |> IO.iodata_to_binary() |> string()
+    iodata |> IO.iodata_to_binary() |> string() |> color(:atom, inspect_opts)
   end
 
-  defp integer_to_algebra(text) do
+  defp integer_to_algebra(text, inspect_otps) do
     case text do
       <<?0, ?x, rest::binary>> ->
         "0x" <> String.upcase(rest)
@@ -1575,12 +1606,15 @@ defmodule Code.Formatter do
       decimal ->
         insert_underscores(decimal)
     end
+    |> color(:number, inspect_otps)
   end
 
-  defp float_to_algebra(text) do
+  defp float_to_algebra(text, inspect_otps) do
     [int_part, decimal_part] = :binary.split(text, ".")
     decimal_part = String.downcase(decimal_part)
-    insert_underscores(int_part) <> "." <> decimal_part
+
+    string = insert_underscores(int_part) <> "." <> decimal_part
+    color(string, :number, inspect_otps)
   end
 
   defp insert_underscores("-" <> digits) do
@@ -1597,7 +1631,7 @@ defmodule Code.Formatter do
         |> String.to_charlist()
         |> Enum.reverse()
         |> Enum.chunk_every(3)
-        |> Enum.intersperse('_')
+        |> Enum.intersperse(~c"_")
         |> List.flatten()
         |> Enum.reverse()
         |> List.to_string()
@@ -1772,13 +1806,13 @@ defmodule Code.Formatter do
 
   ## Type functions
 
-  # (() -> block)
-  defp type_fun_to_algebra([{:->, meta, [[], body]}] = clauses, _min_line, max_line, state) do
+  # (-> block)
+  defp paren_fun_to_algebra([{:->, meta, [[], body]}] = clauses, _min_line, max_line, state) do
     min_line = line(meta)
     {body_doc, state} = block_to_algebra(body, min_line, max_line, state)
 
     doc =
-      "(() -> "
+      "(-> "
       |> concat(nest(body_doc, :cursor))
       |> concat(")")
       |> maybe_force_clauses(clauses, state)
@@ -1790,7 +1824,7 @@ defmodule Code.Formatter do
   # (x -> y)
   # (x ->
   #    y)
-  defp type_fun_to_algebra([{:->, meta, [args, body]}] = clauses, _min_line, max_line, state) do
+  defp paren_fun_to_algebra([{:->, meta, [args, body]}] = clauses, _min_line, max_line, state) do
     min_line = line(meta)
     {args_doc, state} = clause_args_to_algebra(args, min_line, state)
     {body_doc, state} = block_to_algebra(body, min_line, max_line, state)
@@ -1815,7 +1849,7 @@ defmodule Code.Formatter do
   #   args2 ->
   #     block2
   # )
-  defp type_fun_to_algebra(clauses, min_line, max_line, state) do
+  defp paren_fun_to_algebra(clauses, min_line, max_line, state) do
     {clauses_doc, state} = clauses_to_algebra(clauses, min_line, max_line, state)
     {"(" |> line(clauses_doc) |> nest(2) |> line(")") |> force_unfit(), state}
   end
@@ -1938,16 +1972,30 @@ defmodule Code.Formatter do
         Enum.split_while(comments, fn %{line: line} -> line <= min_line end)
       end)
 
-    {docs, comments?, state} =
-      each_quoted_to_algebra_with_comments(args, acc, max_line, state, false, fun)
+    {reverse_docs, comments?, state} =
+      if state.comments == [] do
+        each_quoted_to_algebra_without_comments(args, acc, state, fun)
+      else
+        each_quoted_to_algebra_with_comments(args, acc, max_line, state, false, fun)
+      end
 
+    docs = merge_algebra_with_comments(Enum.reverse(reverse_docs), @empty)
     {docs, comments?, update_in(state.comments, &(pre_comments ++ &1))}
   end
 
+  defp each_quoted_to_algebra_without_comments([], acc, state, _fun) do
+    {acc, false, state}
+  end
+
+  defp each_quoted_to_algebra_without_comments([arg | args], acc, state, fun) do
+    {doc_triplet, state} = fun.(arg, args, state)
+    acc = [doc_triplet | acc]
+    each_quoted_to_algebra_without_comments(args, acc, state, fun)
+  end
+
   defp each_quoted_to_algebra_with_comments([], acc, max_line, state, comments?, _fun) do
     {acc, comments, comments?} = extract_comments_before(max_line, acc, state.comments, comments?)
-    args_docs = merge_algebra_with_comments(Enum.reverse(acc), @empty)
-    {args_docs, comments?, %{state | comments: comments}}
+    {acc, comments?, %{state | comments: comments}}
   end
 
   defp each_quoted_to_algebra_with_comments([arg | args], acc, max_line, state, comments?, fun) do
@@ -2008,10 +2056,11 @@ defmodule Code.Formatter do
   defp adjust_trailing_newlines(doc_triplet, _, _), do: doc_triplet
 
   defp traverse_line({expr, meta, args}, {min, max}) do
+    # This is a hot path, so use :lists.keyfind/3 instead Keyword.fetch!/2
     acc =
-      case Keyword.fetch(meta, :line) do
-        {:ok, line} -> {min(line, min), max(line, max)}
-        :error -> {min, max}
+      case :lists.keyfind(:line, 1, meta) do
+        {:line, line} -> {min(line, min), max(line, max)}
+        false -> {min, max}
       end
 
     traverse_line(args, traverse_line(expr, acc))
@@ -2353,4 +2402,20 @@ defmodule Code.Formatter do
     {left, [right]} = Enum.split(list, -1)
     {left, right}
   end
+
+  defp get_charlist_quotes(_heredoc = false, state) do
+    if state.normalize_charlists_as_sigils do
+      {@sigil_c, @double_quote}
+    else
+      {@single_quote, @single_quote}
+    end
+  end
+
+  defp get_charlist_quotes(_heredoc = true, state) do
+    if state.normalize_charlists_as_sigils do
+      {@sigil_c_heredoc, @double_heredoc}
+    else
+      {@single_heredoc, @single_heredoc}
+    end
+  end
 end
