@@ -112,6 +112,7 @@ defmodule File do
           encoding_mode()
           | :append
           | :compressed
+          | :delayed_write
           | :trim_bom
           | {:read_ahead, pos_integer | false}
           | {:delayed_write, non_neg_integer, non_neg_integer}
@@ -122,6 +123,8 @@ defmodule File do
 
   @type posix_time :: integer()
 
+  @type on_conflict_callback :: (Path.t(), Path.t() -> boolean)
+
   @doc """
   Returns `true` if the path is a regular file.
 
@@ -693,7 +696,10 @@ defmodule File do
   @spec copy(Path.t() | io_device, Path.t() | io_device, pos_integer | :infinity) ::
           {:ok, non_neg_integer} | {:error, posix}
   def copy(source, destination, bytes_count \\ :infinity) do
-    :file.copy(maybe_to_string(source), maybe_to_string(destination), bytes_count)
+    source = normalize_path_or_io_device(source)
+    destination = normalize_path_or_io_device(destination)
+
+    :file.copy(source, destination, bytes_count)
   end
 
   @doc """
@@ -711,8 +717,8 @@ defmodule File do
         raise File.CopyError,
           reason: reason,
           action: "copy",
-          source: maybe_to_string(source),
-          destination: maybe_to_string(destination)
+          source: normalize_path_or_io_device(source),
+          destination: normalize_path_or_io_device(destination)
     end
   end
 
@@ -740,6 +746,8 @@ defmodule File do
   @doc since: "1.1.0"
   @spec rename(Path.t(), Path.t()) :: :ok | {:error, posix}
   def rename(source, destination) do
+    source = IO.chardata_to_string(source)
+    destination = IO.chardata_to_string(destination)
     :file.rename(source, destination)
   end
 
@@ -764,15 +772,11 @@ defmodule File do
   end
 
   @doc """
-  Copies the contents in `source_file` to `destination_file` preserving its modes.
+  Copies the contents of `source_file` to `destination_file` preserving its modes.
 
-  `source_file` and `destination_file` must be a file or a symbolic link to one,
-  or in the case of destination, a path to a non-existent file. If either one of
-  them is a directory, `{:error, :eisdir}` will be returned.
-
-  If a file already exists in the destination, it invokes a
-  callback which should return `true` if the existing file
-  should be overwritten, `false` otherwise. The callback defaults to return `true`.
+  `source_file` must be a file or a symbolic link to one. `destination_file` must
+  be a path to a non-existent file. If either is a directory, `{:error, :eisdir}`
+  will be returned.
 
   The function returns `:ok` in case of success. Otherwise, it returns
   `{:error, reason}`.
@@ -785,13 +789,30 @@ defmodule File do
   whether the destination is an existing directory or not. We have chosen to
   explicitly disallow copying to a destination which is a directory,
   and an error will be returned if tried.
+
+  ## Options
+
+    * `:on_conflict` - (since v1.14.0) Invoked when a file already exists in the destination.
+      The function receives arguments for `source_file` and `destination_file`. It should
+      return `true` if the existing file should be overwritten, `false` if otherwise.
+      The default callback returns `true`. On earlier versions, this callback could be
+      given as third argument, but such behaviour is now deprecated.
+
   """
-  @spec cp(Path.t(), Path.t(), (Path.t(), Path.t() -> boolean)) :: :ok | {:error, posix}
-  def cp(source_file, destination_file, callback \\ fn _, _ -> true end) do
+  @spec cp(Path.t(), Path.t(), on_conflict: on_conflict_callback) :: :ok | {:error, posix}
+  def cp(source_file, destination_file, options \\ [])
+
+  # TODO: Deprecate me on Elixir v1.19
+  def cp(source_file, destination_file, callback) when is_function(callback, 2) do
+    cp(source_file, destination_file, on_conflict: callback)
+  end
+
+  def cp(source_file, destination_file, options) when is_list(options) do
+    on_conflict = Keyword.get(options, :on_conflict, fn _, _ -> true end)
     source_file = IO.chardata_to_string(source_file)
     destination_file = IO.chardata_to_string(destination_file)
 
-    case do_cp_file(source_file, destination_file, callback, []) do
+    case do_cp_file(source_file, destination_file, on_conflict, []) do
       {:error, reason, _} -> {:error, reason}
       _ -> :ok
     end
@@ -807,9 +828,9 @@ defmodule File do
   The same as `cp/3`, but raises a `File.CopyError` exception if it fails.
   Returns `:ok` otherwise.
   """
-  @spec cp!(Path.t(), Path.t(), (Path.t(), Path.t() -> boolean)) :: :ok
-  def cp!(source_file, destination_file, callback \\ fn _, _ -> true end) do
-    case cp(source_file, destination_file, callback) do
+  @spec cp!(Path.t(), Path.t(), on_conflict: on_conflict_callback) :: :ok
+  def cp!(source_file, destination_file, options \\ []) do
+    case cp(source_file, destination_file, options) do
       :ok ->
         :ok
 
@@ -832,18 +853,15 @@ defmodule File do
   If `source` is a directory, or a symbolic link to it, then `destination` must
   be an existent `directory` or a symbolic link to one, or a path to a non-existent directory.
 
-  If the source is a file, it copies `source` to
-  `destination`. If the `source` is a directory, it copies
-  the contents inside source into the `destination` directory.
+  If the source is a file, it copies `source` to `destination`. If the `source`
+  is a directory, it copies the contents inside source into the `destination` directory.
 
-  If a file already exists in the destination, it invokes `callback`.
-  `callback` must be a function that takes two arguments: `source` and `destination`.
-  The callback should return `true` if the existing file should be overwritten and `false` otherwise.
+  If a file already exists in the destination, it invokes the optional `on_conflict`
+  callback given as an option. See "Options" for more information.
 
-  This function may fail while copying files,
-  in such cases, it will leave the destination
-  directory in a dirty state, where file which have already been copied
-  won't be removed.
+  This function may fail while copying files, in such cases, it will leave the
+  destination directory in a dirty state, where file which have already been
+  copied won't be removed.
 
   The function returns `{:ok, files_and_directories}` in case of
   success, `files_and_directories` lists all files and directories copied in no
@@ -854,6 +872,19 @@ defmodule File do
   explicitly disallow this behaviour. If `source` is a `file` and `destination`
   is a directory, `{:error, :eisdir}` will be returned.
 
+  ## Options
+
+    * `:on_conflict` - (since v1.14.0) Invoked when a file already exists in the destination.
+      The function receives arguments for `source` and `destination`. It should return
+      `true` if the existing file should be overwritten, `false` if otherwise. The default
+      callback returns `true`. On earlier versions, this callback could be given as third
+      argument, but such behaviour is now deprecated.
+
+    * `:dereference_symlinks` - (since v1.14.0) By default, this function will copy symlinks
+      by creating symlinks that point to the same location. This option forces symlinks to be
+      dereferenced and have their contents copied instead when set to `true`. If the dereferenced
+      files do not exist, than the operation fails. The default is `false`.
+
   ## Examples
 
       # Copies file "a.txt" to "b.txt"
@@ -863,14 +894,28 @@ defmodule File do
       File.cp_r("samples", "tmp")
 
       # Same as before, but asks the user how to proceed in case of conflicts
-      File.cp_r("samples", "tmp", fn source, destination ->
+      File.cp_r("samples", "tmp", on_conflict: fn source, destination ->
         IO.gets("Overwriting #{destination} by #{source}. Type y to confirm. ") == "y\n"
       end)
 
   """
-  @spec cp_r(Path.t(), Path.t(), (Path.t(), Path.t() -> boolean)) ::
+  @spec cp_r(Path.t(), Path.t(),
+          on_conflict: on_conflict_callback,
+          dereference_symlinks: boolean()
+        ) ::
           {:ok, [binary]} | {:error, posix, binary}
-  def cp_r(source, destination, callback \\ fn _, _ -> true end) when is_function(callback, 2) do
+
+  def cp_r(source, destination, options \\ [])
+
+  # TODO: Deprecate me on Elixir v1.19
+  def cp_r(source, destination, callback) when is_function(callback, 2) do
+    cp_r(source, destination, on_conflict: callback)
+  end
+
+  def cp_r(source, destination, options) when is_list(options) do
+    on_conflict = Keyword.get(options, :on_conflict, fn _, _ -> true end)
+    dereference? = Keyword.get(options, :dereference_symlinks, false)
+
     source =
       source
       |> IO.chardata_to_string()
@@ -881,7 +926,7 @@ defmodule File do
       |> IO.chardata_to_string()
       |> assert_no_null_byte!("File.cp_r/3")
 
-    case do_cp_r(source, destination, callback, []) do
+    case do_cp_r(source, destination, on_conflict, dereference?, []) do
       {:error, _, _} = error -> error
       res -> {:ok, res}
     end
@@ -891,9 +936,12 @@ defmodule File do
   The same as `cp_r/3`, but raises a `File.CopyError` exception if it fails.
   Returns the list of copied files otherwise.
   """
-  @spec cp_r!(Path.t(), Path.t(), (Path.t(), Path.t() -> boolean)) :: [binary]
-  def cp_r!(source, destination, callback \\ fn _, _ -> true end) do
-    case cp_r(source, destination, callback) do
+  @spec cp_r!(Path.t(), Path.t(),
+          on_conflict: on_conflict_callback,
+          dereference_symlinks: boolean()
+        ) :: [binary]
+  def cp_r!(source, destination, options \\ []) do
+    case cp_r(source, destination, options) do
       {:ok, files} ->
         files
 
@@ -907,15 +955,21 @@ defmodule File do
     end
   end
 
-  defp do_cp_r(src, dest, callback, acc) when is_list(acc) do
+  defp do_cp_r(src, dest, on_conflict, dereference?, acc) when is_list(acc) do
     case :elixir_utils.read_link_type(src) do
       {:ok, :regular} ->
-        do_cp_file(src, dest, callback, acc)
+        do_cp_file(src, dest, on_conflict, acc)
 
       {:ok, :symlink} ->
         case :file.read_link(src) do
-          {:ok, link} -> do_cp_link(link, src, dest, callback, acc)
-          {:error, reason} -> {:error, reason, src}
+          {:ok, link} when dereference? ->
+            do_cp_r(Path.expand(link, Path.dirname(src)), dest, on_conflict, dereference?, acc)
+
+          {:ok, link} ->
+            do_cp_link(link, src, dest, on_conflict, acc)
+
+          {:error, reason} ->
+            {:error, reason, src}
         end
 
       {:ok, :directory} ->
@@ -924,7 +978,7 @@ defmodule File do
             case mkdir(dest) do
               success when success in [:ok, {:error, :eexist}] ->
                 Enum.reduce(files, [dest | acc], fn x, acc ->
-                  do_cp_r(Path.join(src, x), Path.join(dest, x), callback, acc)
+                  do_cp_r(Path.join(src, x), Path.join(dest, x), on_conflict, dereference?, acc)
                 end)
 
               {:error, reason} ->
@@ -943,29 +997,41 @@ defmodule File do
     end
   end
 
-  # If we reach this clause, there was an error while
-  # processing a file.
-  defp do_cp_r(_, _, _, acc) do
+  # If we reach this clause, there was an error while processing a file.
+  defp do_cp_r(_, _, _, _, acc) do
     acc
   end
 
-  defp copy_file_mode!(src, dest) do
-    write_stat!(dest, %{stat!(dest) | mode: stat!(src).mode})
+  defp copy_file_mode(src, dest) do
+    with {:ok, dest_fileinfo} <- stat(dest),
+         {:ok, src_fileinfo} <- stat(src) do
+      write_stat(dest, %{dest_fileinfo | mode: src_fileinfo.mode})
+    end
   end
 
   # Both src and dest are files.
-  defp do_cp_file(src, dest, callback, acc) do
+  defp do_cp_file(src, dest, on_conflict, acc) do
     case :file.copy(src, {dest, [:exclusive]}) do
       {:ok, _} ->
-        copy_file_mode!(src, dest)
-        [dest | acc]
+        case copy_file_mode(src, dest) do
+          :ok ->
+            [dest | acc]
+
+          {:error, reason} ->
+            {:error, reason, src}
+        end
 
       {:error, :eexist} ->
-        if path_differs?(src, dest) and callback.(src, dest) do
+        if path_differs?(src, dest) and on_conflict.(src, dest) do
           case copy(src, dest) do
             {:ok, _} ->
-              copy_file_mode!(src, dest)
-              [dest | acc]
+              case copy_file_mode(src, dest) do
+                :ok ->
+                  [dest | acc]
+
+                {:error, reason} ->
+                  {:error, reason, src}
+              end
 
             {:error, reason} ->
               {:error, reason, src}
@@ -980,13 +1046,13 @@ defmodule File do
   end
 
   # Both src and dest are files.
-  defp do_cp_link(link, src, dest, callback, acc) do
+  defp do_cp_link(link, src, dest, on_conflict, acc) do
     case :file.make_symlink(link, dest) do
       :ok ->
         [dest | acc]
 
       {:error, :eexist} ->
-        if path_differs?(src, dest) and callback.(src, dest) do
+        if path_differs?(src, dest) and on_conflict.(src, dest) do
           # If rm/1 fails, :file.make_symlink/2 will fail
           _ = rm(dest)
 
@@ -1043,9 +1109,7 @@ defmodule File do
   """
   @spec write!(Path.t(), iodata, [mode]) :: :ok
   def write!(path, content, modes \\ []) do
-    modes = normalize_modes(modes, false)
-
-    case :file.write_file(path, content, modes) do
+    case write(path, content, modes) do
       :ok ->
         :ok
 
@@ -1193,82 +1257,83 @@ defmodule File do
   """
   @spec rm_rf(Path.t()) :: {:ok, [binary]} | {:error, posix, binary}
   def rm_rf(path) do
+    {major, _} = :os.type()
+
     path
     |> IO.chardata_to_string()
     |> assert_no_null_byte!("File.rm_rf/1")
-    |> do_rm_rf({:ok, []})
+    |> do_rm_rf([], major)
   end
 
-  defp do_rm_rf(path, {:ok, _} = entry) do
-    case safe_list_dir(path) do
+  defp do_rm_rf(path, acc, major) do
+    case safe_list_dir(path, major) do
       {:ok, files} when is_list(files) ->
-        res =
-          Enum.reduce(files, entry, fn file, tuple ->
-            do_rm_rf(Path.join(path, file), tuple)
+        acc =
+          Enum.reduce(files, acc, fn file, acc ->
+            # In case we can't delete, continue anyway, we might succeed
+            # to delete it on Windows due to how they handle symlinks.
+            case do_rm_rf(Path.join(path, file), acc, major) do
+              {:ok, acc} -> acc
+              {:error, _, _} -> acc
+            end
           end)
 
-        case res do
-          {:ok, acc} ->
-            case rmdir(path) do
-              :ok -> {:ok, [path | acc]}
-              {:error, :enoent} -> res
-              {:error, reason} -> {:error, reason, path}
-            end
-
-          reason ->
-            reason
+        case rmdir(path) do
+          :ok -> {:ok, [path | acc]}
+          {:error, :enoent} -> {:ok, acc}
+          {:error, reason} -> {:error, reason, path}
         end
 
       {:ok, :directory} ->
-        do_rm_directory(path, entry)
+        do_rm_directory(path, acc)
 
       {:ok, :regular} ->
-        do_rm_regular(path, entry)
+        do_rm_regular(path, acc)
 
       {:error, reason} when reason in [:enoent, :enotdir] ->
-        entry
+        {:ok, acc}
 
       {:error, reason} ->
         {:error, reason, path}
     end
   end
 
-  defp do_rm_rf(_, reason) do
-    reason
-  end
-
-  defp do_rm_regular(path, {:ok, acc} = entry) do
+  defp do_rm_regular(path, acc) do
     case rm(path) do
       :ok -> {:ok, [path | acc]}
-      {:error, :enoent} -> entry
+      {:error, :enoent} -> {:ok, acc}
       {:error, reason} -> {:error, reason, path}
     end
   end
 
   # On Windows, symlinks are treated as directory and must be removed
-  # with rmdir/1. But on Unix-like systems, we remove them via rm/1. So we first try
-  # to remove it as a directory and, if we get :enotdir, we fall back to
-  # a file removal.
-  defp do_rm_directory(path, {:ok, acc} = entry) do
+  # with rmdir/1. But on Unix-like systems, we remove them via rm/1.
+  # So we first try to remove it as a directory and, if we get :enotdir,
+  # we fall back to a file removal.
+  defp do_rm_directory(path, acc) do
     case rmdir(path) do
       :ok -> {:ok, [path | acc]}
-      {:error, :enotdir} -> do_rm_regular(path, entry)
-      {:error, :enoent} -> entry
+      {:error, :enotdir} -> do_rm_regular(path, acc)
+      {:error, :enoent} -> {:ok, acc}
       {:error, reason} -> {:error, reason, path}
     end
   end
 
-  defp safe_list_dir(path) do
+  defp safe_list_dir(path, major) do
     case :elixir_utils.read_link_type(path) do
-      {:ok, :symlink} ->
+      {:ok, :directory} ->
+        # If we cannot read the files, try to delete it anyway
+        case :file.list_dir_all(path) do
+          {:ok, files} -> {:ok, files}
+          {:error, _} -> {:ok, :directory}
+        end
+
+      {:ok, :symlink} when major == :win32 ->
         case :elixir_utils.read_file_type(path) do
           {:ok, :directory} -> {:ok, :directory}
           _ -> {:ok, :regular}
         end
 
-      {:ok, :directory} ->
-        :file.list_dir(path)
-
       {:ok, _} ->
         {:ok, :regular}
 
@@ -1321,7 +1386,7 @@ defmodule File do
     * `:write` - the file is opened for writing. It is created if it does not
       exist.
 
-      If the file does exists, and if write is not combined with read, the file
+      If the file does exist, and if write is not combined with read, the file
       will be truncated.
 
     * `:append` - the file will be opened for writing, and it will be created
@@ -1541,10 +1606,16 @@ defmodule File do
   executes the given function and then reverts back
   to the previous path regardless of whether there is an exception.
 
+  The current working directory is temporarily set for the BEAM globally. This
+  can lead to race conditions if multiple processes are changing the current
+  working directory concurrently. To run an external command in a given
+  directory without changing the global current working directory, use the
+  `:cd` option of `System.cmd/3` and `Port.open/2`.
+
   Raises an error if retrieving or changing the current
   directory fails.
   """
-  @spec cd!(Path.t(), (() -> res)) :: res when res: var
+  @spec cd!(Path.t(), (-> res)) :: res when res: var
   def cd!(path, function) do
     old = cwd!()
     cd!(path)
@@ -1600,14 +1671,21 @@ defmodule File do
     :file.close(io_device)
   end
 
-  @doc """
+  @doc ~S"""
   Returns a `File.Stream` for the given `path` with the given `modes`.
 
   The stream implements both `Enumerable` and `Collectable` protocols,
   which means it can be used both for read and write.
 
   The `line_or_bytes` argument configures how the file is read when
-  streaming, by `:line` (default) or by a given number of bytes.
+  streaming, by `:line` (default) or by a given number of bytes. When
+  using the `:line` option, CRLF line breaks (`"\r\n"`) are normalized
+  to LF (`"\n"`).
+
+  Similar to other file operations, a stream can be created in one node
+  and forwarded to another node. Once the stream is opened in another node,
+  a request will be sent to the creator node to spawn a process for file
+  streaming.
 
   Operating the stream can fail on open for the same reasons as
   `File.open!/2`. Note that the file is automatically opened each time streaming
@@ -1619,12 +1697,13 @@ defmodule File do
   Since Elixir controls when the streamed file is opened, the underlying
   device cannot be shared and as such it is convenient to open the file
   in raw mode for performance reasons. Therefore, Elixir **will** open
-  streams in `:raw` mode with the `:read_ahead` option unless an encoding
-  is specified. This means any data streamed into the file must be
-  converted to `t:iodata/0` type. If you pass, for example, `[encoding: :utf8]`
-  or `[encoding: {:utf16, :little}]` in the modes parameter,
-  the underlying stream will use `IO.write/2` and the `String.Chars` protocol
-  to convert the data. See `IO.binwrite/2` and `IO.write/2` .
+  streams in `:raw` mode with the `:read_ahead` option if the stream is
+  open in the same node as it is created and no encoding has been specified.
+  This means any data streamed into the file must be converted to `t:iodata/0`
+  type. If you pass, for example, `[encoding: :utf8]` or
+  `[encoding: {:utf16, :little}]` in the modes parameter, the underlying stream
+  will use `IO.write/2` and the `String.Chars` protocol to convert the data.
+  See `IO.binwrite/2` and `IO.write/2` .
 
   One may also consider passing the `:delayed_write` option if the stream
   is meant to be written to under a tight loop.
@@ -1634,8 +1713,8 @@ defmodule File do
   If you pass `:trim_bom` in the modes parameter, the stream will
   trim UTF-8, UTF-16 and UTF-32 byte order marks when reading from file.
 
-  Note that this function does not try to discover the file encoding basing
-  on BOM.
+  Note that this function does not try to discover the file encoding
+  based on BOM.
 
   ## Examples
 
@@ -1796,7 +1875,8 @@ defmodule File do
   defp normalize_modes([], true), do: [:binary]
   defp normalize_modes([], false), do: []
 
-  defp maybe_to_string(path) when is_list(path), do: IO.chardata_to_string(path)
-  defp maybe_to_string(path) when is_binary(path), do: path
-  defp maybe_to_string(path), do: path
+  defp normalize_path_or_io_device(path) when is_list(path), do: IO.chardata_to_string(path)
+  defp normalize_path_or_io_device(path) when is_binary(path), do: path
+  defp normalize_path_or_io_device(io_device) when is_pid(io_device), do: io_device
+  defp normalize_path_or_io_device(io_device = {:file_descriptor, _, _}), do: io_device
 end
