@@ -119,7 +119,7 @@ defprotocol Inspect do
   Elixir will raise an `ArgumentError` error and will automatically fall back
   to a raw representation for printing the structure.
 
-  You can however access the underlying error by invoking the `Inspect`
+  You can, however, access the underlying error by invoking the `Inspect`
   implementation directly. For example, to test `Inspect.MapSet` above,
   you can invoke it as:
 
@@ -158,8 +158,8 @@ defimpl Inspect, for: BitString do
   def inspect(term, opts) when is_binary(term) do
     %Inspect.Opts{binaries: bins, base: base, printable_limit: printable_limit} = opts
 
-    if base == :decimal and
-         (bins == :as_strings or (bins == :infer and String.printable?(term, printable_limit))) do
+    if bins == :as_strings or
+         (bins == :infer and String.printable?(term, printable_limit) and base == :decimal) do
       inspected =
         case Identifier.escape(term, ?", printable_limit) do
           {escaped, ""} -> [?", escaped, ?"]
@@ -208,7 +208,7 @@ defimpl Inspect, for: BitString do
 
   defp each_bit(bitstring, _counter, opts) do
     size = bit_size(bitstring)
-    <<h::size(size)>> = bitstring
+    <<h::size(^size)>> = bitstring
     concat(Inspect.Integer.inspect(h, opts), "::size(" <> Integer.to_string(size) <> ")")
   end
 
@@ -257,9 +257,9 @@ defimpl Inspect, for: List do
     cond do
       lists == :as_charlists or (lists == :infer and List.ascii_printable?(term, printable_limit)) ->
         inspected =
-          case Identifier.escape(IO.chardata_to_string(term), ?', printable_limit) do
-            {escaped, ""} -> [?', escaped, ?']
-            {escaped, _} -> [?', escaped, ?', " ++ ..."]
+          case Identifier.escape(IO.chardata_to_string(term), ?", printable_limit) do
+            {escaped, ""} -> [?~, ?c, ?", escaped, ?"]
+            {escaped, _} -> [?~, ?c, ?", escaped, ?", " ++ ..."]
           end
 
         color(IO.iodata_to_binary(inspected), :charlist, opts)
@@ -281,7 +281,7 @@ defimpl Inspect, for: List do
   @doc false
   def keyword?([{key, _value} | rest]) when is_atom(key) do
     case Atom.to_charlist(key) do
-      'Elixir.' ++ _ -> false
+      [?E, ?l, ?i, ?x, ?i, ?r, ?.] ++ _ -> false
       _ -> keyword?(rest)
     end
   end
@@ -402,7 +402,7 @@ defimpl Inspect, for: Regex do
       |> normalize(<<>>)
       |> Identifier.escape(?/, :infinity, &escape_map/1)
 
-    source = IO.iodata_to_binary(['~r/', escaped, ?/, regex.opts])
+    source = IO.iodata_to_binary([?~, ?r, ?/, escaped, ?/, regex.opts])
     color(source, :regex, opts)
   end
 
@@ -411,16 +411,18 @@ defimpl Inspect, for: Regex do
   defp normalize(<<char, rest::binary>>, acc), do: normalize(rest, <<acc::binary, char>>)
   defp normalize(<<>>, acc), do: acc
 
-  defp escape_map(?\a), do: '\\a'
-  defp escape_map(?\f), do: '\\f'
-  defp escape_map(?\n), do: '\\n'
-  defp escape_map(?\r), do: '\\r'
-  defp escape_map(?\t), do: '\\t'
-  defp escape_map(?\v), do: '\\v'
+  defp escape_map(?\a), do: [?\\, ?a]
+  defp escape_map(?\f), do: [?\\, ?f]
+  defp escape_map(?\n), do: [?\\, ?n]
+  defp escape_map(?\r), do: [?\\, ?r]
+  defp escape_map(?\t), do: [?\\, ?t]
+  defp escape_map(?\v), do: [?\\, ?v]
   defp escape_map(_), do: false
 end
 
 defimpl Inspect, for: Function do
+  @elixir_compiler :binary.bin_to_list("elixir_compiler_")
+
   def inspect(function, _opts) do
     fun_info = Function.info(function)
     mod = fun_info[:module]
@@ -435,7 +437,7 @@ defimpl Inspect, for: Function do
         inspected_as_function = Macro.inspect_atom(:remote_call, name)
         "&#{inspected_as_atom}.#{inspected_as_function}/#{fun_info[:arity]}"
 
-      match?('elixir_compiler_' ++ _, Atom.to_charlist(mod)) ->
+      match?(@elixir_compiler ++ _, Atom.to_charlist(mod)) ->
         if function_exported?(mod, :__RELATIVE__, 0) do
           "#Function<#{uniq(fun_info)} in file:#{mod.__RELATIVE__}>"
         else
@@ -516,7 +518,7 @@ end
 
 defimpl Inspect, for: Reference do
   def inspect(ref, _opts) do
-    '#Ref' ++ rest = :erlang.ref_to_list(ref)
+    [?#, ?R, ?e, ?f] ++ rest = :erlang.ref_to_list(ref)
     "#Reference" <> IO.iodata_to_binary(rest)
   end
 end
