@@ -34,7 +34,7 @@ defmodule IO do
   IO data is a data type that can be used as a more efficient alternative to binaries
   in certain situations.
 
-  A term of type **IO data** is a binary or a list containing bytes (integers in `0..255`)
+  A term of type **IO data** is a binary or a list containing bytes (integers within the `0..255` range)
   or nested IO data. The type is recursive. Let's see an example of one of
   the possible IO data representing the binary `"hello"`:
 
@@ -99,8 +99,8 @@ defmodule IO do
   Erlang and Elixir also have the idea of `t:chardata/0`. Chardata is very
   similar to IO data: the only difference is that integers in IO data represent
   bytes while integers in chardata represent Unicode code points. Bytes
-  (`t:byte/0`) are integers in the `0..255` range, while Unicode code points
-  (`t:char/0`) are integers in the range `0..0x10FFFF`. The `IO` module provides
+  (`t:byte/0`) are integers within the `0..255` range, while Unicode code points
+  (`t:char/0`) are integers within the `0..0x10FFFF` range. The `IO` module provides
   the `chardata_to_string/1` function for chardata as the "counter-part" of the
   `iodata_to_binary/1` function for IO data.
 
@@ -108,8 +108,8 @@ defmodule IO do
   argument error. For example, let's try to put a code point that is not
   representable with one byte, like `?π`, inside IO data:
 
-      iex> IO.iodata_to_binary(["The symbol for pi is: ", ?π])
-      ** (ArgumentError) argument error
+      IO.iodata_to_binary(["The symbol for pi is: ", ?π])
+      #=> ** (ArgumentError) argument error
 
   If we use chardata instead, it will work as expected:
 
@@ -122,14 +122,14 @@ defmodule IO do
   @type nodata :: {:error, term} | :eof
   @type chardata :: String.t() | maybe_improper_list(char | chardata, String.t() | [])
 
+  defguardp is_device(term) when is_atom(term) or is_pid(term)
   defguardp is_iodata(data) when is_list(data) or is_binary(data)
 
   @doc """
   Reads from the IO `device`.
 
-  The `device` is iterated by the given number of characters or line by line if
-  `:line` is given.
-  Alternatively, if `:all` is given, then whole `device` is returned.
+  The `device` is iterated by the given number of characters, line by line if
+  `:line` is given, or until `:eof`.
 
   It returns:
 
@@ -141,53 +141,48 @@ defmodule IO do
       for instance, `{:error, :estale}` if reading from an
       NFS volume
 
-  If `:all` is given, `:eof` is never returned, but an
-  empty string in case the device has reached EOF.
   """
-  @spec read(device, :all | :line | non_neg_integer) :: chardata | nodata
+  @spec read(device, :eof | :line | non_neg_integer) :: chardata | nodata
   def read(device \\ :stdio, line_or_chars)
 
+  # TODO: Deprecate me on v1.17
   def read(device, :all) do
-    do_read_all(map_dev(device), :empty)
+    with :eof <- read(device, :eof) do
+      with [_ | _] = opts <- :io.getopts(device),
+           false <- Keyword.get(opts, :binary, true) do
+        ~c""
+      else
+        _ -> ""
+      end
+    end
+  end
+
+  def read(device, :eof) do
+    getn(device, ~c"", :eof)
   end
 
   def read(device, :line) do
-    :io.get_line(map_dev(device), '')
+    :io.get_line(map_dev(device), ~c"")
   end
 
   def read(device, count) when is_integer(count) and count >= 0 do
-    :io.get_chars(map_dev(device), '', count)
+    :io.get_chars(map_dev(device), ~c"", count)
   end
 
-  defp do_read_all(mapped_dev, acc) do
-    case :io.get_line(mapped_dev, "") do
-      line when is_binary(line) or is_list(line) -> do_read_all(mapped_dev, concat(acc, line))
-      :eof -> read_eof(mapped_dev, acc)
-      other -> other
-    end
-  end
-
-  defp concat(:empty, line), do: line
-  defp concat(acc, line) when is_binary(acc), do: acc <> line
-  defp concat(acc, line) when is_list(acc), do: acc ++ line
-
-  defp read_eof(device, :empty) do
-    with [_ | _] = opts <- :io.getopts(device),
-         false <- Keyword.get(opts, :binary, true) do
-      ''
-    else
-      _ -> ""
-    end
-  end
-
-  defp read_eof(_device, acc), do: acc
-
   @doc """
   Reads from the IO `device`. The operation is Unicode unsafe.
 
-  The `device` is iterated by the given number of bytes or line by line if
-  `:line` is given.
-  Alternatively, if `:all` is given, then whole `device` is returned.
+  The `device` is iterated as specified by the `line_or_chars` argument:
+
+    * if `line_or_chars` is an integer, it represents a number of bytes. The device is
+      iterated by that number of bytes.
+
+    * if `line_or_chars` is `:line`, the device is iterated line by line.
+
+    * if `line_or_chars` is `:eof`, the device is iterated until `:eof`. `line_or_chars`
+      can only be `:eof` since Elixir 1.13.0. `:eof` replaces the deprecated `:all`,
+      with the difference that `:all` returns `""` on end of file, while `:eof` returns
+      `:eof` itself.
 
   It returns:
 
@@ -199,17 +194,19 @@ defmodule IO do
       for instance, `{:error, :estale}` if reading from an
       NFS volume
 
-  If `:all` is given, `:eof` is never returned, but an
-  empty string in case the device has reached EOF.
-
   Note: do not use this function on IO devices in Unicode mode
   as it will return the wrong result.
   """
-  @spec binread(device, :all | :line | non_neg_integer) :: iodata | nodata
+  @spec binread(device, :eof | :line | non_neg_integer) :: iodata | nodata
   def binread(device \\ :stdio, line_or_chars)
 
+  # TODO: Deprecate me on v1.17
   def binread(device, :all) do
-    do_binread_all(map_dev(device), "")
+    with :eof <- binread(device, :eof), do: ""
+  end
+
+  def binread(device, :eof) do
+    binread_eof(map_dev(device), "")
   end
 
   def binread(device, :line) do
@@ -227,10 +224,10 @@ defmodule IO do
   end
 
   @read_all_size 4096
-  defp do_binread_all(mapped_dev, acc) do
+  defp binread_eof(mapped_dev, acc) do
     case :file.read(mapped_dev, @read_all_size) do
-      {:ok, data} -> do_binread_all(mapped_dev, acc <> data)
-      :eof -> acc
+      {:ok, data} -> binread_eof(mapped_dev, acc <> data)
+      :eof -> if acc == "", do: :eof, else: acc
       other -> other
     end
   end
@@ -293,19 +290,29 @@ defmodule IO do
 
   """
   @spec puts(device, chardata | String.Chars.t()) :: :ok
-  def puts(device \\ :stdio, item) do
+  def puts(device \\ :stdio, item) when is_device(device) do
     :io.put_chars(map_dev(device), [to_chardata(item), ?\n])
   end
 
   @doc """
-  Writes a `message` to stderr, along with the given `stacktrace`.
+  Writes a `message` to stderr, along with the given `stacktrace_info`.
+
+  The `stacktrace_info` must be one of:
+
+    * a `__STACKTRACE__`, where all entries in the stacktrace will be
+      included in the error message
+
+    * a `Macro.Env` structure (since v1.14.0), where a single stacktrace
+      entry from the compilation environment will be used
+
+    * a keyword list with at least the `:file` option representing
+      a single stacktrace entry (since v1.14.0). The `:line`, `:module`,
+      `:function` options are also supported
 
   This function also notifies the compiler a warning was printed
   (in case --warnings-as-errors was enabled). It returns `:ok`
   if it succeeds.
 
-  An empty list can be passed to avoid stacktrace printing.
-
   ## Examples
 
       stacktrace = [{MyApp, :main, 1, [file: 'my_app.ex', line: 4]}]
@@ -314,24 +321,41 @@ defmodule IO do
       #=>   my_app.ex:4: MyApp.main/1
 
   """
-  @spec warn(chardata | String.Chars.t(), Exception.stacktrace()) :: :ok
+  @spec warn(chardata | String.Chars.t(), Exception.stacktrace() | keyword() | Macro.Env.t()) ::
+          :ok
+  def warn(message, stacktrace_info)
+
+  def warn(message, %Macro.Env{} = env) do
+    warn(message, Macro.Env.stacktrace(env))
+  end
+
   def warn(message, []) do
-    message = [to_chardata(message), ?\n]
-    :elixir_errors.io_warn(0, nil, message, message)
+    :elixir_errors.emit_diagnostic(:warning, 0, nil, to_chardata(message), [])
+  end
+
+  def warn(message, [{_, _} | _] = keyword) do
+    if file = keyword[:file] do
+      warn(
+        message,
+        %{
+          __ENV__
+          | module: keyword[:module],
+            function: keyword[:function],
+            line: keyword[:line],
+            file: file
+        }
+      )
+    else
+      warn(message, [])
+    end
   end
 
   def warn(message, [{_, _, _, opts} | _] = stacktrace) do
     message = to_chardata(message)
-    formatted_trace = Enum.map_join(stacktrace, "\n  ", &Exception.format_stacktrace_entry(&1))
     line = opts[:line]
     file = opts[:file]
-
-    :elixir_errors.io_warn(
-      line || 0,
-      file && List.to_string(file),
-      message,
-      [message, ?\n, "  ", formatted_trace, ?\n]
-    )
+    file = file && List.to_string(file)
+    :elixir_errors.emit_diagnostic(:warning, line || 0, file, message, stacktrace)
   end
 
   @doc false
@@ -428,9 +452,9 @@ defmodule IO do
   See `inspect/2` for a full list of options.
   """
   @spec inspect(device, item, keyword) :: item when item: var
-  def inspect(device, item, opts) when is_list(opts) do
+  def inspect(device, item, opts) when is_device(device) and is_list(opts) do
     label = if label = opts[:label], do: [to_chardata(label), ": "], else: []
-    opts = struct(Inspect.Opts, opts)
+    opts = Inspect.Opts.new(opts)
     doc = Inspect.Algebra.group(Inspect.Algebra.to_doc(item, opts))
     chardata = Inspect.Algebra.format(doc, opts.width)
     puts(device, [label, chardata])
@@ -445,12 +469,18 @@ defmodule IO do
   Otherwise, `count` is the number of raw bytes to be retrieved.
 
   See `IO.getn/3` for a description of return values.
-
   """
-  @spec getn(device | chardata | String.Chars.t(), pos_integer | chardata | String.Chars.t()) ::
+  @spec getn(
+          device | chardata | String.Chars.t(),
+          pos_integer | :eof | chardata | String.Chars.t()
+        ) ::
           chardata | nodata
   def getn(prompt, count \\ 1)
 
+  def getn(prompt, :eof) do
+    getn(:stdio, prompt, :eof)
+  end
+
   def getn(prompt, count) when is_integer(count) and count > 0 do
     getn(:stdio, prompt, count)
   end
@@ -477,11 +507,27 @@ defmodule IO do
       NFS volume
 
   """
-  @spec getn(device, chardata | String.Chars.t(), pos_integer) :: chardata | nodata
+  @spec getn(device, chardata | String.Chars.t(), pos_integer | :eof) :: chardata | nodata
+  def getn(device, prompt, :eof) do
+    getn_eof(map_dev(device), to_chardata(prompt), [])
+  end
+
   def getn(device, prompt, count) when is_integer(count) and count > 0 do
     :io.get_chars(map_dev(device), to_chardata(prompt), count)
   end
 
+  defp getn_eof(device, prompt, acc) do
+    case :io.get_line(device, prompt) do
+      line when is_binary(line) or is_list(line) -> getn_eof(device, ~c"", [line | acc])
+      :eof -> wrap_eof(:lists.reverse(acc))
+      other -> other
+    end
+  end
+
+  defp wrap_eof([h | _] = acc) when is_binary(h), do: IO.iodata_to_binary(acc)
+  defp wrap_eof([h | _] = acc) when is_list(h), do: :lists.flatten(acc)
+  defp wrap_eof([]), do: :eof
+
   @doc ~S"""
   Reads a line from the IO `device`.
 
@@ -508,6 +554,17 @@ defmodule IO do
     :io.get_line(map_dev(device), to_chardata(prompt))
   end
 
+  @doc """
+  Returns a line-based `IO.Stream` on `:stdio`.
+
+  This is equivalent to:
+
+      IO.stream(:stdio, :line)
+
+  """
+  @doc since: "1.12.0"
+  def stream, do: stream(:stdio, :line)
+
   @doc """
   Converts the IO `device` into an `IO.Stream`.
 
@@ -524,6 +581,9 @@ defmodule IO do
   Note that an IO stream has side effects and every time
   you go over the stream you may get different results.
 
+  `stream/0` has been introduced in Elixir v1.12.0,
+  while `stream/2` has been available since v1.0.0.
+
   ## Examples
 
   Here is an example on how we mimic an echo server
@@ -531,14 +591,33 @@ defmodule IO do
 
       Enum.each(IO.stream(:stdio, :line), &IO.write(&1))
 
+  Another example where you might want to collect a user input
+  every new line and break on an empty line, followed by removing
+  redundant new line characters (`"\\n"`):
+
+      IO.stream(:stdio, :line)
+      |> Enum.take_while(&(&1 != "\\n"))
+      |> Enum.map(&String.replace(&1, "\\n", ""))
+
   """
   @spec stream(device, :line | pos_integer) :: Enumerable.t()
-  def stream(device, line_or_codepoints)
+  def stream(device \\ :stdio, line_or_codepoints)
       when line_or_codepoints == :line
       when is_integer(line_or_codepoints) and line_or_codepoints > 0 do
     IO.Stream.__build__(map_dev(device), false, line_or_codepoints)
   end
 
+  @doc """
+  Returns a raw, line-based `IO.Stream` on `:stdio`. The operation is Unicode unsafe.
+
+  This is equivalent to:
+
+      IO.binstream(:stdio, :line)
+
+  """
+  @doc since: "1.12.0"
+  def binstream, do: binstream(:stdio, :line)
+
   @doc """
   Converts the IO `device` into an `IO.Stream`. The operation is Unicode unsafe.
 
@@ -547,8 +626,7 @@ defmodule IO do
   and write.
 
   The `device` is iterated by the given number of bytes or line by line if
-  `:line` is given.
-  This reads from the IO device as a raw binary.
+  `:line` is given. This reads from the IO device as a raw binary.
 
   Note that an IO stream has side effects and every time
   you go over the stream you may get different results.
@@ -556,9 +634,11 @@ defmodule IO do
   Finally, do not use this function on IO devices in Unicode
   mode as it will return the wrong result.
 
+  `binstream/0` has been introduced in Elixir v1.12.0,
+  while `binstream/2` has been available since v1.0.0.
   """
   @spec binstream(device, :line | pos_integer) :: Enumerable.t()
-  def binstream(device, line_or_bytes)
+  def binstream(device \\ :stdio, line_or_bytes)
       when line_or_bytes == :line
       when is_integer(line_or_bytes) and line_or_bytes > 0 do
     IO.Stream.__build__(map_dev(device), true, line_or_bytes)
@@ -586,6 +666,8 @@ defmodule IO do
 
   """
   @spec chardata_to_string(chardata) :: String.t()
+  def chardata_to_string(chardata)
+
   def chardata_to_string(string) when is_binary(string) do
     string
   end
