@@ -182,7 +182,7 @@ defmodule Kernel.SpecialForms do
       <<1, 2, 3>>
 
   Elixir also accepts by default the segment to be a literal
-  string or a literal charlist, which are by default expanded to integers:
+  string which expands to integers:
 
       iex> <<0, "foo">>
       <<0, 102, 111, 111>>
@@ -201,7 +201,7 @@ defmodule Kernel.SpecialForms do
       iex> <<"foo"::utf32>>
       <<0, 0, 0, 102, 0, 0, 0, 111, 0, 0, 0, 111>>
 
-  Otherwise we get an `ArgumentError` when construcing the binary:
+  Otherwise we get an `ArgumentError` when constructing the binary:
 
       rest = "oo"
       <<102, rest>>
@@ -233,9 +233,9 @@ defmodule Kernel.SpecialForms do
 
   Sizes for types are a bit more nuanced. The default size for integers is 8.
 
-  For floats, it is 64. For floats, `size * unit` must result in 32 or 64,
+  For floats, it is 64. For floats, `size * unit` must result in 16, 32, or 64,
   corresponding to [IEEE 754](https://en.wikipedia.org/wiki/IEEE_floating_point)
-  binary32 and binary64, respectively.
+  binary16, binary32, and binary64, respectively.
 
   For binaries, the default is the size of the binary. Only the last binary in a
   match can use the default size. All others must have their size specified
@@ -246,20 +246,20 @@ defmodule Kernel.SpecialForms do
       iex> {name, species}
       {"Frank", "Walrus"}
 
-  The size can be a variable:
+  The size can be a variable or any valid guard expression:
 
       iex> name_size = 5
-      iex> <<name::binary-size(name_size), " the ", species::binary>> = <<"Frank the Walrus">>
+      iex> <<name::binary-size(^name_size), " the ", species::binary>> = <<"Frank the Walrus">>
       iex> {name, species}
       {"Frank", "Walrus"}
 
-  And the variable can be defined in the match itself (prior to its use):
+  The size can access prior variables defined in the binary itself:
 
       iex> <<name_size::size(8), name::binary-size(name_size), " the ", species::binary>> = <<5, "Frank the Walrus">>
       iex> {name, species}
       {"Frank", "Walrus"}
 
-  However, the size cannot be defined in the match outside the binary/bitstring match:
+  However, it cannot access variables defined in the match outside of the binary/bitstring:
 
       {name_size, <<name::binary-size(name_size), _rest::binary>>} = {5, <<"Frank the Walrus">>}
       ** (CompileError): undefined variable "name_size" in bitstring segment
@@ -365,8 +365,8 @@ defmodule Kernel.SpecialForms do
       ERL_COMPILER_OPTIONS=bin_opt_info mix compile
 
   To learn more about specific optimizations and performance considerations,
-  check out
-  [Erlang's Efficiency Guide on handling binaries](http://www.erlang.org/doc/efficiency_guide/binaryhandling.html).
+  check out the
+  ["Constructing and matching binaries" chapter of the Erlang's Efficiency Guide](https://www.erlang.org/doc/efficiency_guide/binaryhandling.html).
   """
   defmacro unquote(:<<>>)(args), do: error!([args])
 
@@ -604,11 +604,12 @@ defmodule Kernel.SpecialForms do
 
       import List
 
-  A developer can filter to import only macros or functions via
-  the only option:
+  A developer can filter to import only functions, macros, or sigils
+  (which can be functions or macros) via the `:only` option:
 
       import List, only: :functions
       import List, only: :macros
+      import Kernel, only: :sigils
 
   Alternatively, Elixir allows a developer to pass pairs of
   name/arities to `:only` or `:except` as a fine grained control
@@ -617,9 +618,11 @@ defmodule Kernel.SpecialForms do
       import List, only: [flatten: 1]
       import String, except: [split: 2]
 
-  Note that calling `except` is always exclusive on a previously
-  declared `import/2`. If there is no previous import, then it applies
-  to all functions and macros in the module. For example:
+  Importing the same module again will erase the previous imports,
+  except when the `except` option is used, which is always exclusive
+  on a previously declared `import/2`. If there is no previous import,
+  then it applies to all functions and macros in the module. For
+  example:
 
       import List, only: [flatten: 1, keyfind: 4]
       import List, except: [flatten: 1]
@@ -776,7 +779,7 @@ defmodule Kernel.SpecialForms do
 
       <<int::integer-little, rest::bits>> = bits
 
-  Read the documentation on the `Typespec` page and
+  Read the documentation on the [Typespecs page](typespecs.md) and
   `<<>>/1` for more information on typespecs and
   bitstrings respectively.
   """
@@ -834,8 +837,8 @@ defmodule Kernel.SpecialForms do
     * `:context` - sets the resolution context.
 
     * `:generated` - marks the given chunk as generated so it does not emit warnings.
-      Currently it only works on special forms (for example, you can annotate a `case`
-      but not an `if`).
+      It is also useful to avoid dialyzer reporting errors when macros generate
+      unused clauses.
 
     * `:file` - sets the quoted expressions to have the given file.
 
@@ -957,7 +960,7 @@ defmodule Kernel.SpecialForms do
       import Math
       squared(5)
       x
-      ** (CompileError) undefined variable x or undefined function x/0
+      ** (CompileError) undefined variable "x"
 
   We can see that `x` did not leak to the user context. This happens
   because Elixir macros are hygienic, a topic we will discuss at length
@@ -1045,6 +1048,17 @@ defmodule Kernel.SpecialForms do
       ContextHygiene.read()
       #=> 1
 
+  The contexts of a variable is identified by the third element of the tuple.
+  The default context is `nil` and `quote` assigns another context to all
+  variables within:
+
+      quote(do: var)
+      #=> {:var, [], Elixir}
+
+  In case of variables returned by macros, there may also be a `:counter` key
+  in the metadata, which is used to further refine its contexts and guarantee
+  isolation between macro invocations as seen in the previous example.
+
   ## Hygiene in aliases
 
   Aliases inside quote are hygienic by default.
@@ -1209,6 +1223,14 @@ defmodule Kernel.SpecialForms do
   reported to where `defadd` was invoked. `location: :keep` affects
   only definitions inside the quote.
 
+  > #### `location: :keep` and unquote {: .warning}
+  >
+  > Do not use `location: :keep` if the function definition
+  > also `unquote`s some of the macro arguments. If you do so, Elixir
+  > will store the file definition of the current location but the
+  > unquoted arguments may contain line information of the macro caller,
+  > leading to erroneous stacktraces.
+
   ## Binding and unquote fragments
 
   Elixir quote/unquote mechanisms provide a functionality called
@@ -1307,11 +1329,13 @@ defmodule Kernel.SpecialForms do
         sum(1, value, 3)
       end
 
-  Which would then return:
+
+  Which the argument for the `:sum` function call is not the
+  expected result:
 
       {:sum, [], [1, {:value, [], Elixir}, 3]}
 
-  Which is not the expected result. For this, we use `unquote`:
+  For this, we use `unquote`:
 
       iex> value =
       ...>   quote do
@@ -1361,8 +1385,10 @@ defmodule Kernel.SpecialForms do
       iex> for n <- [1, 2, 3, 4], do: n * 2
       [2, 4, 6, 8]
 
-  A comprehension accepts many generators and filters. Enumerable
-  generators are defined using `<-`:
+  A comprehension accepts many generators and filters. `for` uses
+  the `<-` operator to extract values from the enumerable on its
+  right side and match them against the pattern on the left.
+  We call them generators:
 
       # A list generator:
       iex> for n <- [1, 2, 3, 4], do: n * 2
@@ -1378,6 +1404,10 @@ defmodule Kernel.SpecialForms do
       iex> for n <- [1, 2, 3, 4, 5, 6], rem(n, 2) == 0, do: n
       [2, 4, 6]
 
+  Filters must evaluate to truthy values (everything but `nil`
+  and `false`). If a filter is falsy, then the current value is
+  discarded.
+
   Generators can also be used to filter as it removes any value
   that doesn't match the pattern on the left side of `<-`:
 
@@ -1398,6 +1428,35 @@ defmodule Kernel.SpecialForms do
   filters or inside the block, are not reflected outside of the
   comprehension.
 
+  Variable assignments inside filters must still return a truthy value,
+  otherwise values are discarded. Let's see an example. Imagine you have
+  a keyword list where the key is a programming language and the value
+  is its direct parent. Then let's try to compute the grandparent of each
+  language. You could try this:
+
+      iex> languages = [elixir: :erlang, erlang: :prolog, prolog: nil]
+      iex> for {language, parent} <- languages, grandparent = languages[parent], do: {language, grandparent}
+      [elixir: :prolog]
+
+  Given the grandparents of Erlang and Prolog were nil, those values were
+  filtered out. If you don't want this behaviour, a simple option is to
+  move the filter inside the do-block:
+
+      iex> languages = [elixir: :erlang, erlang: :prolog, prolog: nil]
+      iex> for {language, parent} <- languages do
+      ...>   grandparent = languages[parent]
+      ...>   {language, grandparent}
+      ...> end
+      [elixir: :prolog, erlang: nil, prolog: nil]
+
+  However, such option is not always available, as you may have further
+  filters. An alternative is to convert the filter into a generator by
+  wrapping the right side of `=` in a list:
+
+      iex> languages = [elixir: :erlang, erlang: :prolog, prolog: nil]
+      iex> for {language, parent} <- languages, grandparent <- [languages[parent]], do: {language, grandparent}
+      [elixir: :prolog, erlang: nil, prolog: nil]
+
   ## The `:into` and `:uniq` options
 
   In the examples above, the result returned by the comprehension was
@@ -1414,7 +1473,7 @@ defmodule Kernel.SpecialForms do
   The `IO` module provides streams, that are both `Enumerable` and
   `Collectable`, here is an upcase echo server using comprehensions:
 
-      for line <- IO.stream(:stdio, :line), into: IO.stream(:stdio, :line) do
+      for line <- IO.stream(), into: IO.stream() do
         String.upcase(line)
       end
 
@@ -1473,9 +1532,48 @@ defmodule Kernel.SpecialForms do
   defmacro for(args), do: error!([args])
 
   @doc """
-  Used to combine matching clauses.
+  Combine matching clauses.
 
-  Let's start with an example:
+  One of the ways to understand with is to show which code
+  patterns it improves. Imagine you have a map where the fields
+  `width` and `height` are optional and you want to compute its
+  area, as `{:ok, area}` or return `:error`. We could implement
+  this function as:
+
+      def area(map) do
+        case Map.fetch(opts, :width) do
+          {:ok, width} ->
+            case Map.fetch(opts, :height) do
+              {:ok, height} -> {:ok, width * height}
+              :error -> :error
+            end
+
+          :error ->
+            :error
+        end
+      end
+
+  when called as `area(%{width: 10, height: 15})`, it should return
+  `{:ok, 150}`. If any of the fields are missing, it returns `:error`.
+
+  While the code above works, it is quite verbose. Using `with`,
+  we could rewrite it as:
+
+      def area(map) do
+        with {:ok, width} <- Map.fetch(opts, :width),
+             {:ok, height} <- Map.fetch(opts, :height) do
+          {:ok, width * height}
+        end
+      end
+
+  Instead of defining nested `case`s with clauses, we use `with`
+  alongside the `PATTERN <- EXPRESSION` operator to match
+  expressions on its right side against the pattern on the left.
+  Consider `<-` as a sibling to `=`, except that, while `=` raises
+  in case of not matches, `<-` will simply abort the `with` chain
+  and return the non-matched value.
+
+  Let's give it a try on IEx:
 
       iex> opts = %{width: 10, height: 15}
       iex> with {:ok, width} <- Map.fetch(opts, :width),
@@ -1502,7 +1600,9 @@ defmodule Kernel.SpecialForms do
       ...> end
       {:ok, "admin"}
 
-  As in `for/1`, variables bound inside `with/1` won't leak.
+  As in `for/1`, variables bound inside `with/1` won't be accessible
+  outside of `with/1`.
+
   Expressions without `<-` may also be used in clauses. For instance,
   you can perform regular matches with the `=` operator:
 
@@ -1517,15 +1617,15 @@ defmodule Kernel.SpecialForms do
       iex> width
       nil
 
-  The behaviour of any expression in a clause is the same as outside.
-  For example, `=` will raise a `MatchError` instead of returning the
-  non-matched value:
+  The behaviour of any expression in a clause is the same as if it was
+  written outside of `with`. For example, `=` will raise a `MatchError`
+  instead of returning the non-matched value:
 
       with :foo = :bar, do: :ok
       ** (MatchError) no match of right hand side value: :bar
 
   As with any other function or macro call in Elixir, explicit parens can
-  also be used around the arguments before the `do`/`end` block:
+  also be used around the arguments before the `do`-`end` block:
 
       iex> opts = %{width: 10, height: 15}
       iex> with(
@@ -1538,6 +1638,8 @@ defmodule Kernel.SpecialForms do
 
   The choice between parens and no parens is a matter of preference.
 
+  ## Else clauses
+
   An `else` option can be given to modify what is being returned from
   `with` in the case of a failed match:
 
@@ -1548,11 +1650,61 @@ defmodule Kernel.SpecialForms do
       ...> else
       ...>   :error ->
       ...>     {:error, :wrong_data}
+      ...>
+      ...>   _other_error ->
+      ...>     :unexpected_error
       ...> end
       {:error, :wrong_data}
 
+  The `else` block works like a `case` clause: it can have multiple clauses,
+  and the first match will be used. Variables bound inside `with` (such as
+  `width` in this example) are not available in the `else` block.
+
   If an `else` block is used and there are no matching clauses, a `WithClauseError`
   exception is raised.
+
+  ### Beware!
+
+  Keep in mind that, one of potential drawback of `with` is that all
+  failure clauses are flattened into a single `else` block. For example,
+  take this code that checks if a given path points to an Elixir file
+  and that it exists before creating a backup copy:
+
+      with ".ex" <- Path.extname(path),
+           true <- File.exists?(path) do
+        backup_path = path <> ".backup"
+        File.cp!(path, backup_path)
+        {:ok, backup_path}
+      else
+        binary when is_binary(binary) ->
+          {:error, :invalid_extension}
+
+        false ->
+          {:error, :missing_file}
+      end
+
+  Note how we are having to reconstruct the result types of `Path.extname/1`
+  and `File.exists?/1` to build error messages. In this case, it is better
+  to refactor the code so each `<-` already return the desired format in case
+  of errors, like this:
+
+      with :ok <- validate_extension(path),
+           :ok <- validate_exists(path) do
+        backup_path = path <> ".backup"
+        File.cp!(path, backup_path)
+        {:ok, backup_path}
+      end
+
+      defp validate_extension(path) do
+        if Path.extname(path) == ".ex", do: :ok, else: {:error, :invalid_extension}
+      end
+
+      defp validate_exists(path) do
+        if File.exists?(path), do: :ok, else: {:error, :missing_file}
+      end
+
+  Note how the code above is better organized and clearer once we
+  make sure each `<-` in `with` returns a normalized format.
   """
   defmacro with(args), do: error!([args])
 
@@ -1598,7 +1750,7 @@ defmodule Kernel.SpecialForms do
   defmacro unquote(:__block__)(args), do: error!([args])
 
   @doc """
-  Caputure operator. Captures or creates an anonymous function.
+  Capture operator. Captures or creates an anonymous function.
 
   ## Capture
 
@@ -1721,19 +1873,21 @@ defmodule Kernel.SpecialForms do
 
   ## Examples
 
-      case thing do
-        {:selector, i, value} when is_integer(i) ->
-          value
-        value ->
-          value
+      case File.read(file) do
+        {:ok, contents} when is_binary(contents) ->
+          String.split(contents, "\n")
+
+        {:error, _reason} ->
+          Logger.warning "could not find #{file}, assuming empty..."
+          []
       end
 
-  In the example above, we match `thing` against each clause "head"
-  and execute the clause "body" corresponding to the first clause
-  that matches.
+  In the example above, we match the result of `File.read/1`
+  against each clause "head" and execute the clause "body"
+  corresponding to the first clause that matches.
 
-  If no clause matches, an error is raised.
-  For this reason, it may be necessary to add a final catch-all clause (like `_`)
+  If no clause matches, an error is raised. For this reason,
+  it may be necessary to add a final catch-all clause (like `_`)
   which will always match.
 
       x = 10
@@ -1741,6 +1895,7 @@ defmodule Kernel.SpecialForms do
       case x do
         0 ->
           "This clause won't match"
+
         _ ->
           "This clause would match any value (x = #{x})"
       end
@@ -1758,8 +1913,7 @@ defmodule Kernel.SpecialForms do
       value
       #=> unbound variable value
 
-  When binding variables with the same names as variables in the outer context,
-  the variables in the outer context are not affected.
+  Variables in the outer context cannot be overridden either:
 
       value = 7
 
@@ -1786,6 +1940,19 @@ defmodule Kernel.SpecialForms do
       end
       #=> "Will match"
 
+  ## Using guards to match against multiple values
+
+  While it is not possible to match against multiple patterns in a single
+  clause, it's possible to match against multiple values by using guards:
+
+      case data do
+        value when value in [:one, :two] ->
+          "#{value} has been matched"
+
+        :three ->
+          "three has been matched"
+      end
+
   """
   defmacro case(condition, clauses), do: error!([condition, clauses])
 
@@ -1983,6 +2150,25 @@ defmodule Kernel.SpecialForms do
         File.rm("tmp/story.txt")
       end
 
+  Although `after` clauses are invoked whether or not there was an error, they do not
+  modify the return value. All of the following examples return `:return_me`:
+
+      try do
+        :return_me
+      after
+        IO.puts("I will be printed")
+        :not_returned
+      end
+
+      try do
+        raise "boom"
+      rescue
+        _ -> :return_me
+      after
+        IO.puts("I will be printed")
+        :not_returned
+      end
+
   ## `else` clauses
 
   `else` clauses allow the result of the body passed to `try/1` to be pattern
