@@ -56,41 +56,61 @@ defmodule Macro do
   To learn more about Elixir's AST and how to build them programmatically,
   see `quote/2`.
 
-  > Note: the functions in this module do not evaluate code. In fact,
+  > #### Evaluating code {: .tip}
+  >
+  > The functions in this module do not evaluate code. In fact,
   > evaluating code from macros is often an anti-pattern. For code
   > evaluation, see the `Code` module.
 
   ## Custom Sigils
 
-  Macros are also commonly used to implement custom sigils. To create a custom
-  sigil, define a macro with the name `sigil_{identifier}` that takes two
-  arguments. The first argument will be the string, the second will be a charlist
-  containing any modifiers. If the sigil is lower case (such as `sigil_x`) then
-  the string argument will allow interpolation. If the sigil is upper case
-  (such as `sigil_X`) then the string will not be interpolated.
+  Macros are also commonly used to implement custom sigils.
 
-  Valid modifiers include only lower and upper case letters. Other characters
-  will cause a syntax error.
+  Sigils start with `~` and are followed by one lowercase letter or by one
+  or more uppercase letters, and then a separator
+  (see the [Syntax Reference](syntax-reference.md)). One example is
+  `~D[2020-10-13]` to define a date.
+
+  To create a custom sigil, define a macro with the name `sigil_{identifier}`
+  that takes two arguments. The first argument will be the string, the second
+  will be a charlist containing any modifiers. If the sigil is lower case
+  (such as `sigil_x`) then the string argument will allow interpolation.
+  If the sigil is one or more upper case letters (such as `sigil_X` and
+  `sigil_EXAMPLE`) then the string will not be interpolated.
+
+  Valid modifiers are ASCII letters and digits. Any other character will
+  cause a syntax error.
+
+  Single-letter sigils are typically reserved to the language. Multi-letter
+  sigils are uppercased and extensively used by the community to embed
+  alternative markups and data-types within Elixir source code.
 
   The module containing the custom sigil must be imported before the sigil
   syntax can be used.
 
   ### Examples
 
+  As an example, let's define a sigil `~x` and sigil `~X` which
+  return its contents as a string. However, if the `r` modifier
+  is given, it reverses the string instead:
+
       defmodule MySigils do
         defmacro sigil_x(term, [?r]) do
           quote do
             unquote(term) |> String.reverse()
           end
         end
+
         defmacro sigil_x(term, _modifiers) do
           term
         end
+
         defmacro sigil_X(term, [?r]) do
           quote do
             unquote(term) |> String.reverse()
           end
         end
+
         defmacro sigil_X(term, _modifiers) do
           term
         end
@@ -99,16 +119,16 @@ defmodule Macro do
       import MySigils
 
       ~x(with #{"inter" <> "polation"})
-      #=>"with interpolation"
+      #=> "with interpolation"
 
       ~x(with #{"inter" <> "polation"})r
-      #=>"noitalopretni htiw"
+      #=> "noitalopretni htiw"
 
       ~X(without #{"interpolation"})
-      #=>"without \#{"interpolation"}"
+      #=> "without \#{"interpolation"}"
 
       ~X(without #{"interpolation"})r
-      #=>"}\"noitalopretni\"{# tuohtiw"
+      #=> "}\"noitalopretni\"{# tuohtiw"
 
   """
 
@@ -156,18 +176,27 @@ defmodule Macro do
       For example, `quote/2` will include the module calling `quote/2`
       as the context. This is often used to distinguish regular code from code
       generated by a macro or by `quote/2`.
+
     * `:counter` - The variable counter used for variable hygiene. In terms of
       the compiler, each variable is identified by the combination of either
       `name` and `metadata[:counter]`, or `name` and `context`.
+
     * `:generated` - Whether the code should be considered as generated by
       the compiler or not. This means the compiler and tools like Dialyzer may not
       emit certain warnings.
+
     * `:if_undefined` - How to expand a variable that is undefined. Set it to
-      `:apply` if you want a variable to become a nullary call without warning.
+      `:apply` if you want a variable to become a nullary call without warning
+      or `:raise`
+
     * `:keep` - Used by `quote/2` with the option `location: :keep` to annotate
       the file and the line number of the quoted source.
+
     * `:line` - The line number of the AST node.
 
+    * `:from_brackets` - Used to determine whether a call to `Access.get/3` is from
+      bracket syntax or a function call.
+
   The following metadata keys are enabled by `Code.string_to_quoted/2`:
 
     * `:closing` - contains metadata about the closing pair, such as a `}`
@@ -312,6 +341,30 @@ defmodule Macro do
             "Kernel.#{op}(#{to_string(arg)}), instead of #{op}#{to_string(arg)}"
   end
 
+  # Piping to an Access.get/2,3 call in the form of brackets
+  # (foo |> bar[]) raises a nice error.
+  def pipe(
+        expr,
+        {{_, meta, [Access, :get] = op}, _meta, [first, second]} = _op_args,
+        integer
+      ) do
+    if {:from_brackets, true} in meta do
+      raise ArgumentError, """
+      wrong operator precedence when piping into bracket-based access
+
+       Instead of:
+
+           #{to_string(expr)} |> #{to_string(first)}[#{to_string(second)}]
+
+       You should write:
+
+           (#{to_string(expr)} |> #{to_string(first)})[#{to_string(second)}]
+      """
+    else
+      {op, meta, List.insert_at([first, second], integer, expr)}
+    end
+  end
+
   def pipe(expr, {op, line, args} = op_args, integer) when is_list(args) do
     cond do
       is_atom(op) and operator?(op, 1) ->
@@ -842,7 +895,11 @@ defmodule Macro do
   def struct!(module, env) when is_atom(module) do
     if module == env.module do
       Module.get_attribute(module, :__struct__)
-    end || :elixir_map.load_struct([line: env.line], module, [], [], env)
+    end ||
+      case :elixir_map.maybe_load_struct([line: env.line], module, [], [], env) do
+        {:ok, struct} -> struct
+        {:error, desc} -> raise ArgumentError, List.to_string(:elixir_map.format_error(desc))
+      end
   end
 
   @doc """
@@ -1338,8 +1395,8 @@ defmodule Macro do
           "[]"
 
         :io_lib.printable_list(list) ->
-          {escaped, _} = Identifier.escape(IO.chardata_to_string(list), ?')
-          IO.iodata_to_binary([?', escaped, ?'])
+          {escaped, _} = Identifier.escape(IO.chardata_to_string(list), ?")
+          IO.iodata_to_binary([?~, ?c, ?", escaped, ?"])
 
         Inspect.List.keyword?(list) ->
           "[" <> kw_list_to_string(list, fun) <> "]"
@@ -1503,11 +1560,18 @@ defmodule Macro do
     {left, right} = delimiter_pair(delimiter)
 
     case Atom.to_string(sigil) do
-      <<"sigil_", name>> when name >= ?A and name <= ?Z ->
-        args = sigil_args(args, fun)
-        {:<<>>, _, [binary]} = parts
-        formatted = <<?~, name, left::binary, binary::binary, right::binary, args::binary>>
-        {:ok, fun.(ast, formatted)}
+      <<"sigil_", first, rest::binary>> when first >= ?A and first <= ?Z ->
+        if upcase_letters?(rest) do
+          args = sigil_args(args, fun)
+          {:<<>>, _, [binary]} = parts
+
+          formatted =
+            <<?~, first, rest::binary, left::binary, binary::binary, right::binary, args::binary>>
+
+          {:ok, fun.(ast, formatted)}
+        else
+          :error
+        end
 
       <<"sigil_", name>> when name >= ?a and name <= ?z ->
         args = sigil_args(args, fun)
@@ -1523,6 +1587,15 @@ defmodule Macro do
     :error
   end
 
+  defp upcase_letters?(<<letter, rest::binary>>) when letter >= ?A and letter <= ?Z,
+    do: upcase_letters?(rest)
+
+  defp upcase_letters?(<<_>>),
+    do: false
+
+  defp upcase_letters?(<<>>),
+    do: true
+
   defp delimiter_pair("["), do: {"[", "]"}
   defp delimiter_pair("{"), do: {"{", "}"}
   defp delimiter_pair("("), do: {"(", ")"}
@@ -1972,10 +2045,35 @@ defmodule Macro do
   def quoted_literal?({:%{}, _, args}), do: quoted_literal?(args)
   def quoted_literal?({:{}, _, args}), do: quoted_literal?(args)
   def quoted_literal?({:__MODULE__, _, ctx}) when is_atom(ctx), do: true
+  def quoted_literal?({:<<>>, _, segments}), do: Enum.all?(segments, &quoted_bitstring_segment?/1)
   def quoted_literal?({left, right}), do: quoted_literal?(left) and quoted_literal?(right)
   def quoted_literal?(list) when is_list(list), do: :lists.all(&quoted_literal?/1, list)
   def quoted_literal?(term), do: is_atom(term) or is_number(term) or is_binary(term)
 
+  defp quoted_bitstring_segment?(term) when is_integer(term) or is_binary(term), do: true
+
+  defp quoted_bitstring_segment?({:"::", _, [term, modifier]})
+       when is_integer(term) or is_binary(term),
+       do: quoted_bitstring_modifier?(modifier)
+
+  defp quoted_bitstring_segment?(_other), do: false
+
+  defp quoted_bitstring_modifier?({:-, _, [left, right]}),
+    do: quoted_bitstring_modifier?(left) and quoted_bitstring_modifier?(right)
+
+  defp quoted_bitstring_modifier?({atom, _, [size]})
+       when atom in [:size, :unit] and is_integer(size),
+       do: true
+
+  defp quoted_bitstring_modifier?({:*, _, [left, right]})
+       when is_integer(left) and is_integer(right),
+       do: true
+
+  defp quoted_bitstring_modifier?({modifier, _, ctx}) when is_atom(ctx) or ctx == [],
+    do: :elixir_bitstring.validate_spec(modifier, nil) != :none
+
+  defp quoted_bitstring_modifier?(_other), do: false
+
   @doc false
   @deprecated "Use Macro.expand_literals/2 instead"
   def expand_literal(ast, env) do
@@ -1992,7 +2090,8 @@ defmodule Macro do
       Macro.expand_literals(ast, %{env | function: {:my_code, 1}})
 
   At the moment, the only expandable literal nodes in an AST are
-  aliases, so this function only expands aliases.
+  aliases, so this function only expands aliases (and it does so
+  anywhere in a literal).
 
   However, be careful when removing compile-time dependencies between
   modules. If you remove them but you still invoke the module at
@@ -2451,7 +2550,7 @@ defmodule Macro do
     end
   end
 
-  defp valid_alias?('Elixir' ++ rest), do: valid_alias_piece?(rest)
+  defp valid_alias?([?E, ?l, ?i, ?x, ?i, ?r] ++ rest), do: valid_alias_piece?(rest)
   defp valid_alias?(_other), do: false
 
   defp valid_alias_piece?([?., char | rest]) when char >= ?A and char <= ?Z,
@@ -2486,10 +2585,27 @@ defmodule Macro do
 
   You can call this function directly to build `Kernel.dbg/2` backends that fall back
   to this function.
+
+  This function raises if the context of the given `env` is `:match` or `:guard`.
   """
   @doc since: "1.14.0"
   @spec dbg(t, t, Macro.Env.t()) :: t
   def dbg(code, options, %Macro.Env{} = env) do
+    case env.context do
+      :match ->
+        raise ArgumentError,
+              "invalid expression in match, dbg is not allowed in patterns " <>
+                "such as function clauses, case clauses or on the left side of the = operator"
+
+      :guard ->
+        raise ArgumentError,
+              "invalid expression in guard, dbg is not allowed in guards. " <>
+                "To learn more about guards, visit: https://hexdocs.pm/elixir/patterns-and-guards.html"
+
+      _ ->
+        :ok
+    end
+
     header = dbg_format_header(env)
 
     quote do
@@ -2500,14 +2616,12 @@ defmodule Macro do
 
   # Pipelines.
   defp dbg_ast_to_debuggable({:|>, _meta, _args} = pipe_ast) do
-    value_var = Macro.unique_var(:value, __MODULE__)
-    values_acc_var = Macro.unique_var(:values, __MODULE__)
+    value_var = unique_var(:value, __MODULE__)
+    values_acc_var = unique_var(:values, __MODULE__)
 
     [start_ast | rest_asts] = asts = for {ast, 0} <- unpipe(pipe_ast), do: ast
     rest_asts = Enum.map(rest_asts, &pipe(value_var, &1, 0))
 
-    string_asts = Enum.map(asts, &to_string/1)
-
     initial_acc =
       quote do
         unquote(value_var) = unquote(start_ast)
@@ -2526,13 +2640,95 @@ defmodule Macro do
 
     quote do
       unquote(values_ast)
-      {:pipe, unquote(string_asts), Enum.reverse(unquote(values_acc_var))}
+
+      {:pipe, unquote(escape(asts)), Enum.reverse(unquote(values_acc_var))}
+    end
+  end
+
+  dbg_decomposed_binary_operators = [:&&, :||, :and, :or]
+
+  # Logic operators.
+  defp dbg_ast_to_debuggable({op, _meta, [_left, _right]} = ast)
+       when op in unquote(dbg_decomposed_binary_operators) do
+    acc_var = unique_var(:acc, __MODULE__)
+    result_var = unique_var(:result, __MODULE__)
+
+    quote do
+      unquote(acc_var) = []
+      unquote(dbg_boolean_tree(ast, acc_var, result_var))
+      {:logic_op, Enum.reverse(unquote(acc_var))}
+    end
+  end
+
+  defp dbg_ast_to_debuggable({:case, _meta, [expr, [do: clauses]]} = ast) do
+    clauses_returning_index =
+      Enum.with_index(clauses, fn {:->, meta, [left, right]}, index ->
+        {:->, meta, [left, {right, index}]}
+      end)
+
+    quote do
+      expr = unquote(expr)
+
+      {result, clause_index} =
+        case expr do
+          unquote(clauses_returning_index)
+        end
+
+      {:case, unquote(escape(ast)), expr, clause_index, result}
+    end
+  end
+
+  defp dbg_ast_to_debuggable({:cond, _meta, [[do: clauses]]} = ast) do
+    modified_clauses =
+      Enum.with_index(clauses, fn {:->, _meta, [[left], right]}, index ->
+        hd(
+          quote do
+            clause_value = unquote(left) ->
+              {unquote(escape(left)), clause_value, unquote(index), unquote(right)}
+          end
+        )
+      end)
+
+    quote do
+      {clause_ast, clause_value, clause_index, value} =
+        cond do
+          unquote(modified_clauses)
+        end
+
+      {:cond, unquote(escape(ast)), clause_ast, clause_value, clause_index, value}
     end
   end
 
   # Any other AST.
   defp dbg_ast_to_debuggable(ast) do
-    quote do: {:value, unquote(to_string(ast)), unquote(ast)}
+    quote do: {:value, unquote(escape(ast)), unquote(ast)}
+  end
+
+  # This is a binary operator. We replace the left side with a recursive call to
+  # this function to decompose it, and then execute the operation and add it to the acc.
+  defp dbg_boolean_tree({op, _meta, [left, right]} = ast, acc_var, result_var)
+       when op in unquote(dbg_decomposed_binary_operators) do
+    replaced_left = dbg_boolean_tree(left, acc_var, result_var)
+
+    quote do
+      unquote(result_var) = unquote(op)(unquote(replaced_left), unquote(right))
+
+      unquote(acc_var) = [
+        {unquote(escape(ast)), unquote(result_var)} | unquote(acc_var)
+      ]
+
+      unquote(result_var)
+    end
+  end
+
+  # This is finally an expression, so we assign "result = expr", add it to the acc, and
+  # return the result.
+  defp dbg_boolean_tree(ast, acc_var, result_var) do
+    quote do
+      unquote(result_var) = unquote(ast)
+      unquote(acc_var) = [{unquote(escape(ast)), unquote(result_var)} | unquote(acc_var)]
+      unquote(result_var)
+    end
   end
 
   # Made public to be called from Macro.dbg/3, so that we generate as little code
@@ -2560,8 +2756,8 @@ defmodule Macro do
 
   defp dbg_format_ast_to_debug({:pipe, code_asts, values}, options) do
     result = List.last(values)
-    [{first_ast, first_value} | asts_with_values] = Enum.zip(code_asts, values)
-
+    code_strings = Enum.map(code_asts, &to_string_with_colors(&1, options))
+    [{first_ast, first_value} | asts_with_values] = Enum.zip(code_strings, values)
     first_formatted = [dbg_format_ast(first_ast), " ", inspect(first_value, options), ?\n]
 
     rest_formatted =
@@ -2572,8 +2768,63 @@ defmodule Macro do
     {[first_formatted | rest_formatted], result}
   end
 
+  defp dbg_format_ast_to_debug({:logic_op, components}, options) do
+    {_ast, final_value} = List.last(components)
+
+    formatted =
+      Enum.map(components, fn {ast, value} ->
+        [dbg_format_ast(to_string_with_colors(ast, options)), " ", inspect(value, options), ?\n]
+      end)
+
+    {formatted, final_value}
+  end
+
+  defp dbg_format_ast_to_debug({:case, ast, expr_value, clause_index, value}, options) do
+    {:case, _meta, [expr_ast, _]} = ast
+
+    formatted = [
+      dbg_maybe_underline("Case argument", options),
+      ":\n",
+      dbg_format_ast_with_value(expr_ast, expr_value, options),
+      ?\n,
+      dbg_maybe_underline("Case expression", options),
+      " (clause ##{clause_index + 1} matched):\n",
+      dbg_format_ast_with_value(ast, value, options)
+    ]
+
+    {formatted, value}
+  end
+
+  defp dbg_format_ast_to_debug(
+         {:cond, ast, clause_ast, clause_value, clause_index, value},
+         options
+       ) do
+    formatted = [
+      dbg_maybe_underline("Cond clause", options),
+      " (clause ##{clause_index + 1} matched):\n",
+      dbg_format_ast_with_value(clause_ast, clause_value, options),
+      ?\n,
+      dbg_maybe_underline("Cond expression", options),
+      ":\n",
+      dbg_format_ast_with_value(ast, value, options)
+    ]
+
+    {formatted, value}
+  end
+
   defp dbg_format_ast_to_debug({:value, code_ast, value}, options) do
-    {[dbg_format_ast(code_ast), " ", inspect(value, options), ?\n], value}
+    {dbg_format_ast_with_value(code_ast, value, options), value}
+  end
+
+  defp dbg_format_ast_with_value(ast, value, options) do
+    [dbg_format_ast(to_string_with_colors(ast, options)), " ", inspect(value, options), ?\n]
+  end
+
+  defp to_string_with_colors(ast, options) do
+    options = Keyword.take(options, [:syntax_colors])
+
+    algebra = Code.quoted_to_algebra(ast, options)
+    IO.iodata_to_binary(Inspect.Algebra.format(algebra, 98))
   end
 
   defp dbg_format_header(env) do
@@ -2582,6 +2833,14 @@ defmodule Macro do
     "[" <> Exception.format_stacktrace_entry(stacktrace_entry) <> "]"
   end
 
+  defp dbg_maybe_underline(string, options) do
+    if options[:syntax_colors] != [] do
+      IO.ANSI.format([:underline, string, :reset])
+    else
+      string
+    end
+  end
+
   defp dbg_format_ast(ast) do
     [ast, :faint, " #=>", :reset]
   end
