@@ -46,7 +46,8 @@ defmodule System do
       mode](https://www.erlang.org/doc/apps/erts/time_correction.html#Time_Warp_Modes);
 
     * `monotonic_time/0` - a monotonically increasing time provided
-      by the Erlang VM.
+      by the Erlang VM. This is not strictly monotonically increasing. Multiple
+      sequential calls of the function may return the same value.
 
   The time functions in this module work in the `:native` unit
   (unless specified otherwise), which is operating system dependent. Most of
@@ -141,11 +142,11 @@ defmodule System do
   defmacrop get_revision do
     null =
       case :os.type() do
-        {:win32, _} -> 'NUL'
-        _ -> '/dev/null'
+        {:win32, _} -> ~c"NUL"
+        _ -> ~c"/dev/null"
       end
 
-    'git rev-parse --short=7 HEAD 2> '
+    ~c"git rev-parse --short=7 HEAD 2> "
     |> Kernel.++(null)
     |> :os.cmd()
     |> strip
@@ -157,7 +158,7 @@ defmodule System do
   # Follows https://reproducible-builds.org/specs/source-date-epoch/
   defmacrop get_date do
     unix_epoch =
-      if source_date_epoch = :os.getenv('SOURCE_DATE_EPOCH') do
+      if source_date_epoch = :os.getenv(~c"SOURCE_DATE_EPOCH") do
         try do
           List.to_integer(source_date_epoch)
         rescue
@@ -298,6 +299,22 @@ defmodule System do
     :elixir_config.get(:no_halt)
   end
 
+  @doc """
+  Waits until the system boots.
+
+  Calling this function blocks until all of ARGV is processed.
+  Inside a release, this means the boot script and then ARGV
+  have been processed. This is only useful for those implementing
+  custom shells/consoles on top of Elixir.
+
+  However, be careful to not invoke this command from within
+  the process that is processing the command line arguments,
+  as doing so would lead to a deadlock.
+  """
+  @doc since: "1.15.0"
+  @spec wait_until_booted() :: :ok
+  defdelegate wait_until_booted(), to: :elixir_config
+
   @doc """
   Current working directory.
 
@@ -374,8 +391,8 @@ defmodule System do
   """
   @spec tmp_dir() :: String.t() | nil
   def tmp_dir do
-    write_env_tmp_dir('TMPDIR') || write_env_tmp_dir('TEMP') || write_env_tmp_dir('TMP') ||
-      write_tmp_dir('/tmp') || write_cwd_tmp_dir()
+    write_env_tmp_dir(~c"TMPDIR") || write_env_tmp_dir(~c"TEMP") || write_env_tmp_dir(~c"TMP") ||
+      write_tmp_dir(~c"/tmp") || write_cwd_tmp_dir()
   end
 
   defp write_cwd_tmp_dir do
@@ -435,7 +452,7 @@ defmodule System do
   The function must receive the exit status code as an argument.
 
   If the VM terminates programmatically, via `System.stop/1`, `System.halt/1`,
-  or exit signals, the `at_exit/1` callbacks are not executed.
+  or exit signals, the `at_exit/1` callbacks are not guaranteed to be executed.
   """
   @spec at_exit((non_neg_integer -> any)) :: :ok
   def at_exit(fun) when is_function(fun, 1) do
@@ -472,8 +489,10 @@ defmodule System do
   @doc """
   Traps the given `signal` to execute the `fun`.
 
-  > **Important**: Trapping signals may have strong implications
-  > on how a system shuts down and behave in production and
+  > #### Avoid setting traps in libraries {: .warning}
+  >
+  > Trapping signals may have strong implications
+  > on how a system shuts down and behaves in production and
   > therefore it is extremely discouraged for libraries to
   > set their own traps. Instead, they should redirect users
   > to configure them themselves. The only cases where it is
@@ -528,8 +547,8 @@ defmodule System do
   it may disable Elixir traps (or Elixir may override your configuration).
   """
   @doc since: "1.12.0"
-  @spec trap_signal(signal, (() -> :ok)) :: {:ok, reference()} | {:error, :not_sup}
-  @spec trap_signal(signal, id, (() -> :ok)) ::
+  @spec trap_signal(signal, (-> :ok)) :: {:ok, reference()} | {:error, :not_sup}
+  @spec trap_signal(signal, id, (-> :ok)) ::
           {:ok, id} | {:error, :already_registered} | {:error, :not_sup}
         when id: term()
   def trap_signal(signal, id \\ make_ref(), fun)
@@ -606,9 +625,6 @@ defmodule System do
     end
   end
 
-  # TODO: Remove this once we require Erlang/OTP 24+
-  @compile {:no_warn_undefined, {:os, :env, 0}}
-
   @doc """
   Returns all system environment variables.
 
@@ -617,17 +633,9 @@ defmodule System do
   """
   @spec get_env() :: %{optional(String.t()) => String.t()}
   def get_env do
-    if function_exported?(:os, :env, 0) do
-      Map.new(:os.env(), fn {k, v} ->
-        {IO.chardata_to_string(k), IO.chardata_to_string(v)}
-      end)
-    else
-      Enum.into(:os.getenv(), %{}, fn var ->
-        var = IO.chardata_to_string(var)
-        [k, v] = String.split(var, "=", parts: 2)
-        {k, v}
-      end)
-    end
+    Map.new(:os.env(), fn {k, v} ->
+      {IO.chardata_to_string(k), IO.chardata_to_string(v)}
+    end)
   end
 
   @doc """
@@ -651,7 +659,8 @@ defmodule System do
 
   """
   @doc since: "1.9.0"
-  @spec get_env(String.t(), String.t() | nil) :: String.t() | nil
+  @spec get_env(String.t(), String.t()) :: String.t()
+  @spec get_env(String.t(), nil) :: String.t() | nil
   def get_env(varname, default \\ nil)
       when is_binary(varname) and
              (is_binary(default) or is_nil(default)) do
@@ -697,7 +706,7 @@ defmodule System do
       "4000"
 
       iex> System.fetch_env!("NOT_SET")
-      ** (ArgumentError) could not fetch environment variable "NOT_SET" because it is not set
+      ** (System.EnvError) could not fetch environment variable "NOT_SET" because it is not set
 
   """
   @doc since: "1.9.0"
@@ -862,7 +871,7 @@ defmodule System do
   defdelegate restart(), to: :init
 
   @doc """
-  Carefully stops the Erlang runtime system.
+  Asynchronously and carefully stops the Erlang runtime system.
 
   All applications are taken down smoothly, all code is unloaded, and all ports
   are closed before the system terminates by calling `halt/1`.
@@ -870,13 +879,16 @@ defmodule System do
   `status` must be a non-negative integer or a binary.
 
     * If an integer, the runtime system exits with the integer value which is
-      returned to the operating system.
+      returned to the operating system. On many platforms, only the status codes
+      0-255 are supported by the operating system.
 
     * If a binary, an Erlang crash dump is produced with status as slogan, and
       then the runtime system exits with status code 1.
 
-  Note that on many platforms, only the status codes 0-255 are supported
-  by the operating system.
+  Note this function is asynchronous and the current process will continue
+  executing after this function is invoked. In case you want to block the
+  current process until the system effectively shuts down, you can invoke
+  `Process.sleep(:infinity)`.
 
   ## Examples
 
@@ -885,14 +897,16 @@ defmodule System do
 
   """
   @doc since: "1.5.0"
-  @spec stop(non_neg_integer | binary) :: no_return
+  @spec stop(non_neg_integer | binary) :: :ok
   def stop(status \\ 0)
 
   def stop(status) when is_integer(status) do
+    at_exit(fn _ -> Process.sleep(:infinity) end)
     :init.stop(status)
   end
 
   def stop(status) when is_binary(status) do
+    at_exit(fn _ -> Process.sleep(:infinity) end)
     :init.stop(String.to_charlist(status))
   end
 
@@ -901,7 +915,9 @@ defmodule System do
 
   It uses `sh` for Unix-like systems and `cmd` for Windows.
 
-  > **Important:**: Use this function with care. In particular, **never
+  > #### Watch out {: .warning}
+  >
+  > Use this function with care. In particular, **never
   > pass untrusted user input to this function**, as the user would be
   > able to perform "command injection attacks" by executing any code
   > directly on the machine. Generally speaking, prefer to use `cmd/3`
@@ -947,9 +963,9 @@ defmodule System do
 
         command =
           case {System.get_env("COMSPEC"), osname} do
-            {nil, :windows} -> 'command.com /s /c ' ++ command
-            {nil, _} -> 'cmd /s /c ' ++ command
-            {cmd, _} -> '#{cmd} /s /c ' ++ command
+            {nil, :windows} -> ~c"command.com /s /c " ++ command
+            {nil, _} -> ~c"cmd /s /c " ++ command
+            {cmd, _} -> ~c"#{cmd} /s /c " ++ command
           end
 
         do_cmd({:spawn, command}, [], opts)
@@ -994,23 +1010,39 @@ defmodule System do
       hello
       {%IO.Stream{}, 0}
 
+  If you want to read lines:
+
+      iex> System.cmd("echo", ["hello\nworld"], into: [], lines: 1024)
+      {["hello", "world"], 0}
+
   ## Options
 
     * `:into` - injects the result into the given collectable, defaults to `""`
+
+    * `:lines` - (since v1.15.0) reads the output by lines instead of in bytes. It expects a
+      number of maximum bytes to buffer internally (1024 is a reasonable default).
+      The collectable will be called with each finished line (regardless of buffer
+      size) and without the EOL character
+
     * `:cd` - the directory to run the command in
+
     * `:env` - an enumerable of tuples containing environment key-value as
       binary. The child process inherits all environment variables from its
       parent process, the Elixir application, except those overwritten or
       cleared using this option. Specify a value of `nil` to clear (unset) an
       environment variable, which is useful for preventing credentials passed
-      to the application from leaking into child processes.
+      to the application from leaking into child processes
+
     * `:arg0` - sets the command arg0
+
     * `:stderr_to_stdout` - redirects stderr to stdout when `true`
+
     * `:parallelism` - when `true`, the VM will schedule port tasks to improve
       parallelism in the system. If set to `false`, the VM will try to perform
       commands immediately, improving latency at the expense of parallelism.
-      The default can be set on system startup by passing the "+spp" argument
-      to `--erl`.
+      The default is `false`, and can be set on system startup by passing the 
+      [`+spp`](https://www.erlang.org/doc/man/erl.html#+spp) flag to `--erl`. 
+      Use `:erlang.system_info(:port_parallelism)` to check if enabled.
 
   ## Error reasons
 
@@ -1064,11 +1096,16 @@ defmodule System do
   end
 
   defp do_cmd(port_init, base_opts, opts) do
-    {into, opts} = cmd_opts(opts, [:use_stdio, :exit_status, :binary, :hide] ++ base_opts, "")
+    {into, line, opts} =
+      cmd_opts(opts, [:use_stdio, :exit_status, :binary, :hide] ++ base_opts, "", false)
+
     {initial, fun} = Collectable.into(into)
 
     try do
-      do_port(Port.open(port_init, opts), initial, fun)
+      case line do
+        true -> do_port_line(Port.open(port_init, opts), initial, fun, [])
+        false -> do_port_byte(Port.open(port_init, opts), initial, fun)
+      end
     catch
       kind, reason ->
         fun.(initial, :halt)
@@ -1078,42 +1115,67 @@ defmodule System do
     end
   end
 
-  defp do_port(port, acc, fun) do
+  defp do_port_byte(port, acc, fun) do
     receive do
       {^port, {:data, data}} ->
-        do_port(port, fun.(acc, {:cont, data}), fun)
+        do_port_byte(port, fun.(acc, {:cont, data}), fun)
 
       {^port, {:exit_status, status}} ->
         {acc, status}
     end
   end
 
-  defp cmd_opts([{:into, any} | t], opts, _into),
-    do: cmd_opts(t, opts, any)
+  defp do_port_line(port, acc, fun, buffer) do
+    receive do
+      {^port, {:data, {:noeol, data}}} ->
+        do_port_line(port, acc, fun, [data | buffer])
 
-  defp cmd_opts([{:cd, bin} | t], opts, into) when is_binary(bin),
-    do: cmd_opts(t, [{:cd, bin} | opts], into)
+      {^port, {:data, {:eol, data}}} ->
+        data = [data | buffer] |> Enum.reverse() |> IO.iodata_to_binary()
+        do_port_line(port, fun.(acc, {:cont, data}), fun, [])
 
-  defp cmd_opts([{:arg0, bin} | t], opts, into) when is_binary(bin),
-    do: cmd_opts(t, [{:arg0, bin} | opts], into)
+      {^port, {:exit_status, status}} ->
+        # Data may arrive after exit status on line mode
+        receive do
+          {^port, {:data, {_, data}}} ->
+            data = [data | buffer] |> Enum.reverse() |> IO.iodata_to_binary()
+            {fun.(acc, {:cont, data}), status}
+        after
+          0 -> {acc, status}
+        end
+    end
+  end
 
-  defp cmd_opts([{:stderr_to_stdout, true} | t], opts, into),
-    do: cmd_opts(t, [:stderr_to_stdout | opts], into)
+  defp cmd_opts([{:into, any} | t], opts, _into, line),
+    do: cmd_opts(t, opts, any, line)
 
-  defp cmd_opts([{:stderr_to_stdout, false} | t], opts, into),
-    do: cmd_opts(t, opts, into)
+  defp cmd_opts([{:cd, bin} | t], opts, into, line) when is_binary(bin),
+    do: cmd_opts(t, [{:cd, bin} | opts], into, line)
 
-  defp cmd_opts([{:parallelism, bool} | t], opts, into) when is_boolean(bool),
-    do: cmd_opts(t, [{:parallelism, bool} | opts], into)
+  defp cmd_opts([{:arg0, bin} | t], opts, into, line) when is_binary(bin),
+    do: cmd_opts(t, [{:arg0, bin} | opts], into, line)
 
-  defp cmd_opts([{:env, enum} | t], opts, into),
-    do: cmd_opts(t, [{:env, validate_env(enum)} | opts], into)
+  defp cmd_opts([{:stderr_to_stdout, true} | t], opts, into, line),
+    do: cmd_opts(t, [:stderr_to_stdout | opts], into, line)
 
-  defp cmd_opts([{key, val} | _], _opts, _into),
+  defp cmd_opts([{:stderr_to_stdout, false} | t], opts, into, line),
+    do: cmd_opts(t, opts, into, line)
+
+  defp cmd_opts([{:parallelism, bool} | t], opts, into, line) when is_boolean(bool),
+    do: cmd_opts(t, [{:parallelism, bool} | opts], into, line)
+
+  defp cmd_opts([{:env, enum} | t], opts, into, line),
+    do: cmd_opts(t, [{:env, validate_env(enum)} | opts], into, line)
+
+  defp cmd_opts([{:lines, max_line_length} | t], opts, into, _line)
+       when is_integer(max_line_length) and max_line_length > 0,
+       do: cmd_opts(t, [{:line, max_line_length} | opts], into, true)
+
+  defp cmd_opts([{key, val} | _], _opts, _into, _line),
     do: raise(ArgumentError, "invalid option #{inspect(key)} with value #{inspect(val)}")
 
-  defp cmd_opts([], opts, into),
-    do: {into, opts}
+  defp cmd_opts([], opts, into, line),
+    do: {into, line, opts}
 
   defp validate_env(enum) do
     Enum.map(enum, fn
@@ -1132,7 +1194,8 @@ defmodule System do
   Returns the current monotonic time in the `:native` time unit.
 
   This time is monotonically increasing and starts in an unspecified
-  point in time.
+  point in time. This is not strictly monotonically increasing. Multiple
+  sequential calls of the function may return the same value.
 
   Inlined by the compiler.
   """
@@ -1251,7 +1314,7 @@ defmodule System do
   This time may be adjusted forwards or backwards in time
   with no limitation and is not monotonic.
   """
-  @spec os_time(time_unit) :: integer
+  @spec os_time(time_unit | :native) :: integer
   @doc since: "1.3.0"
   def os_time(unit) do
     :os.system_time(normalize_time_unit(unit))
