@@ -1,7 +1,7 @@
 % Holds the logic responsible for function definitions (def(p) and defmacro(p)).
 -module(elixir_def).
--export([setup/1, reset_last/1, local_for/4,
-  take_definition/2, store_definition/5, store_definition/9,
+-export([setup/1, reset_last/1, local_for/5, external_for/5,
+  take_definition/2, store_definition/3, store_definition/9,
   fetch_definitions/2, format_error/1]).
 -include("elixir.hrl").
 -define(last_def, {elixir, last_def}).
@@ -16,18 +16,32 @@ reset_last({DataSet, _DataBag}) ->
 reset_last(Module) when is_atom(Module) ->
   reset_last(elixir_module:data_tables(Module)).
 
-local_for(Module, Name, Arity, Kinds) ->
+%% Finds the local definition for the current module.
+local_for(Meta, Name, Arity, Kinds, E) ->
+  External = fun({Mod, Fun}, Args) ->
+    invoke_external([{from_macro, true} | Meta], Mod, Fun, Args, E)
+  end,
+  fun_for(Meta, ?key(E, module), Name, Arity, Kinds, {value, External}).
+
+%% Finds the local definition for an external module.
+external_for(Meta, Module, Name, Arity, Kinds) ->
+  fun_for(Meta, Module, Name, Arity, Kinds, none).
+
+fun_for(Meta, Module, Name, Arity, Kinds, External) ->
   Tuple = {Name, Arity},
 
   try
     {Set, Bag} = elixir_module:data_tables(Module),
     {ets:lookup(Set, {def, Tuple}), ets:lookup(Bag, {clauses, Tuple})}
   of
-    {[{_, Kind, Meta, _, _, _}], Clauses} ->
+    {[{_, Kind, LocalMeta, _, _, _}], ClausesPairs} ->
       case (Kinds == all) orelse (lists:member(Kind, Kinds)) of
-        true -> elixir_erl:definition_to_anonymous(Module, Kind, Meta,
-                                                   [Clause || {_, Clause} <- Clauses]);
-        false -> false
+        true ->
+          Local = {value, fun(Fun, Args) -> invoke_local(Meta, Module, Fun, Args, External) end},
+          Clauses = [Clause || {_, Clause} <- ClausesPairs],
+          elixir_erl:definition_to_anonymous(Kind, LocalMeta, Clauses, Local, External);
+        false ->
+          false
       end;
     {[], _} ->
       false
@@ -35,6 +49,21 @@ local_for(Module, Name, Arity, Kinds) ->
     _:_ -> false
   end.
 
+invoke_local(Meta, Module, ErlName, Args, External) ->
+  {Name, Arity} = elixir_utils:erl_fa_to_elixir_fa(ErlName, length(Args)),
+
+  case fun_for(Meta, Module, Name, Arity, all, External) of
+    false ->
+      {current_stacktrace, [_ | T]} = erlang:process_info(self(), current_stacktrace),
+      erlang:raise(error, undef, [{Module, Name, Arity, []} | T]);
+    Fun ->
+      apply(Fun, Args)
+  end.
+
+invoke_external(Meta, Mod, Name, Args, E) ->
+  is_map(E) andalso elixir_env:trace({require, Meta, Mod, []}, E),
+  apply(Mod, Name, Args).
+
 %% Take a definition out of the table
 
 take_definition(Module, {Name, Arity} = Tuple) ->
@@ -50,7 +79,7 @@ take_definition(Module, {Name, Arity} = Tuple) ->
 
 %% Fetch all available definitions
 
-fetch_definitions(File, Module) ->
+fetch_definitions(Module, E) ->
   {Set, Bag} = elixir_module:data_tables(Module),
 
   Entries = try
@@ -59,13 +88,9 @@ fetch_definitions(File, Module) ->
     error:badarg -> []
   end,
 
-  {All, Private} = fetch_definition(Entries, File, Module, Set, Bag, [], []),
-  Unreachable = elixir_locals:warn_unused_local(File, Module, All, Private),
-  elixir_locals:ensure_no_undefined_local(File, Module, All),
-  elixir_locals:ensure_no_import_conflict(File, Module, All),
-  {All, Unreachable}.
+  fetch_definition(Entries, E, Module, Set, Bag, [], []).
 
-fetch_definition([Tuple | T], File, Module, Set, Bag, All, Private) ->
+fetch_definition([Tuple | T], E, Module, Set, Bag, All, Private) ->
   [{_, Kind, Meta, _, Check, {MaxDefaults, _, Defaults}}] = ets:lookup(Set, {def, Tuple}),
 
   try ets:lookup_element(Bag, {clauses, Tuple}, 2) of
@@ -80,14 +105,14 @@ fetch_definition([Tuple | T], File, Module, Set, Bag, All, Private) ->
           false ->
             Private
         end,
-      fetch_definition(T, File, Module, Set, Bag, NewAll, NewPrivate)
+      fetch_definition(T, E, Module, Set, Bag, NewAll, NewPrivate)
   catch
     error:badarg ->
-      check_bodiless_function(Check, Meta, File, Module, Kind, Tuple),
-      fetch_definition(T, File, Module, Set, Bag, All, Private)
+      elixir_errors:module_error(Meta, E, ?MODULE, {function_head, Kind, Tuple}),
+      fetch_definition(T, E, Module, Set, Bag, All, Private)
   end;
 
-fetch_definition([], _File, _Module, _Set, _Bag, All, Private) ->
+fetch_definition([], _E, _Module, _Set, _Bag, All, Private) ->
   {All, Private}.
 
 add_defaults_to_meta(0, Meta) -> Meta;
@@ -102,14 +127,21 @@ head_and_definition_meta(false, _Meta, _HeadDefaults, _All) ->
 
 %% Section for storing definitions
 
-store_definition(Kind, CheckClauses, Call, Body, Pos) ->
-  #{line := Line} = E = elixir_locals:get_cached_env(Pos),
+store_definition(Kind, {Call, Body}, Pos) ->
+  E = elixir_locals:get_cached_env(Pos),
+  store_definition(Kind, false, Call, Body, E);
+store_definition(Kind, Key, Pos) ->
+  #{module := Module} = E = elixir_locals:get_cached_env(Pos),
+  {Call, Body} = elixir_module:read_cache(Module, Key),
+  store_definition(Kind, true, Call, Body, E).
+
+store_definition(Kind, CheckClauses, Call, Body, #{line := Line} = E) ->
   {NameAndArgs, Guards} = elixir_utils:extract_guards(Call),
 
   {Name, Args} = case NameAndArgs of
     {N, _, A} when is_atom(N), is_atom(A) -> {N, []};
     {N, _, A} when is_atom(N), is_list(A) -> {N, A};
-    _ -> elixir_errors:form_error([{line, Line}], E, ?MODULE, {invalid_def, Kind, NameAndArgs})
+    _ -> elixir_errors:file_error([{line, Line}], E, ?MODULE, {invalid_def, Kind, NameAndArgs})
   end,
 
   %% Now that we have verified the call format,
@@ -150,8 +182,8 @@ store_definition(Kind, CheckClauses, Call, Body, Pos) ->
 
   {File, DefMeta} =
     case retrieve_location(Location, ?key(E, module)) of
-      {F, L} ->
-        {F, [{line, Line}, {file, {F, L}} | Generated]};
+      {AF, RF, L} ->
+        {AF, [{line, Line}, {file, {RF, L}} | Generated]};
       nil ->
         {nil, [{line, Line} | Generated]}
     end,
@@ -166,10 +198,10 @@ store_definition(Kind, CheckClauses, Call, Body, Pos) ->
 store_definition(Meta, Kind, CheckClauses, Name, Arity, DefaultsArgs, Guards, Body, File, ER) ->
   Module = ?key(ER, module),
   Tuple = {Name, Arity},
-  E = env_for_expansion(Kind, Tuple, ER),
+  {S, E} = env_for_expansion(Kind, Tuple, ER),
 
-  {Args, Defaults} = unpack_defaults(Kind, Meta, Name, DefaultsArgs, E),
-  Clauses = [elixir_clauses:def(Clause, E) ||
+  {Args, Defaults} = unpack_defaults(Kind, Meta, Name, DefaultsArgs, S, E),
+  Clauses = [elixir_clauses:def(Clause, S, E) ||
              Clause <- def_to_clauses(Kind, Meta, Args, Guards, Body, E)],
 
   DefaultsLength = length(Defaults),
@@ -185,24 +217,25 @@ store_definition(Meta, Kind, CheckClauses, Name, Arity, DefaultsArgs, Guards, Bo
   Tuple.
 
 env_for_expansion(Kind, Tuple, E) when Kind =:= defmacro; Kind =:= defmacrop ->
-  E#{function := Tuple, contextual_vars := ['__CALLER__']};
+  S = elixir_env:env_to_ex(E),
+  {S#elixir_ex{caller=true}, E#{function := Tuple}};
 env_for_expansion(_Kind, Tuple, E) ->
-  E#{function := Tuple, contextual_vars := []}.
+  {elixir_env:env_to_ex(E), E#{function := Tuple}}.
 
 retrieve_location(Location, Module) ->
   {Set, _} = elixir_module:data_tables(Module),
   case ets:take(Set, file) of
     [] when is_tuple(Location) ->
       {File, Line} = Location,
-      {elixir_utils:relative_to_cwd(File), Line};
+      {filename:absname(File), elixir_utils:relative_to_cwd(File), Line};
     [] ->
       nil;
-    [{file, File, _}] when is_binary(File) ->
+    [{file, File, _, _}] when is_binary(File) ->
       'Elixir.Module':delete_attribute(Module, file),
-      {elixir_utils:relative_to_cwd(File), 0};
-    [{file, {File, Line}, _}] when is_binary(File) andalso is_integer(Line) ->
+      {filename:absname(File), elixir_utils:relative_to_cwd(File), 0};
+    [{file, {File, Line}, _, _}] when is_binary(File) andalso is_integer(Line) ->
       'Elixir.Module':delete_attribute(Module, file),
-      {elixir_utils:relative_to_cwd(File), Line}
+      {filename:absname(File), elixir_utils:relative_to_cwd(File), Line}
   end.
 
 run_with_location_change(nil, E, Callback) ->
@@ -222,7 +255,7 @@ def_to_clauses(Kind, Meta, Args, Guards, Body, E) ->
     {do, _} ->
       [{Meta, Args, Guards, {'try', [{origin,  Kind} | Meta], [Body]}}];
     false ->
-      elixir_errors:form_error(Meta, E, elixir_expand, {missing_option, Kind, [do]})
+      elixir_errors:file_error(Meta, E, elixir_expand, {missing_option, Kind, [do]})
   end.
 
 run_on_definition_callbacks(Kind, Module, Name, Args, Guards, Body, E) ->
@@ -263,27 +296,27 @@ store_definition(Check, Kind, Meta, Name, Arity, File, Module, Defaults, Clauses
 
 %% Handling of defaults
 
-unpack_defaults(Kind, Meta, Name, Args, E) ->
-  Expanded = expand_defaults(Args, E#{context := nil}),
-  unpack_defaults(Kind, Meta, Name, Expanded, [], []).
+unpack_defaults(Kind, Meta, Name, Args, S, E) ->
+  {Expanded, #elixir_ex{unused={_, VersionOffset}}} = expand_defaults(Args, S, E#{context := nil}, []),
+  unpack_expanded(Kind, Meta, Name, Expanded, VersionOffset, [], []).
 
-unpack_defaults(Kind, Meta, Name, [{'\\\\', DefaultMeta, [Expr, _]} | T] = List, Acc, Clauses) ->
-  Base = match_defaults(Acc, length(Acc), []),
-  {Args, Invoke} = extract_defaults(List, length(Base), [], []),
+unpack_expanded(Kind, Meta, Name, [{'\\\\', DefaultMeta, [Expr, _]} | T] = List, VersionOffset, Acc, Clauses) ->
+  Base = match_defaults(Acc, length(Acc) + VersionOffset, []),
+  {Args, Invoke} = extract_defaults(List, length(Base) + VersionOffset, [], []),
   Clause = {Meta, Base ++ Args, [], {super, [{super, {Kind, Name}} | DefaultMeta], Base ++ Invoke}},
-  unpack_defaults(Kind, Meta, Name, T, [Expr | Acc], [Clause | Clauses]);
-unpack_defaults(Kind, Meta, Name, [H | T], Acc, Clauses) ->
-  unpack_defaults(Kind, Meta, Name, T, [H | Acc], Clauses);
-unpack_defaults(_Kind, _Meta, _Name, [], Acc, Clauses) ->
+  unpack_expanded(Kind, Meta, Name, T, VersionOffset, [Expr | Acc], [Clause | Clauses]);
+unpack_expanded(Kind, Meta, Name, [H | T], VersionOffset, Acc, Clauses) ->
+  unpack_expanded(Kind, Meta, Name, T, VersionOffset, [H | Acc], Clauses);
+unpack_expanded(_Kind, _Meta, _Name, [], _VersionOffset, Acc, Clauses) ->
   {lists:reverse(Acc), lists:reverse(Clauses)}.
 
-expand_defaults([{'\\\\', Meta, [Expr, Default]} | Args], E) ->
-  {ExpandedDefault, _} = elixir_expand:expand(Default, E),
-  [{'\\\\', Meta, [Expr, ExpandedDefault]} | expand_defaults(Args, E)];
-expand_defaults([Arg | Args], E) ->
-  [Arg | expand_defaults(Args, E)];
-expand_defaults([], _E) ->
-  [].
+expand_defaults([{'\\\\', Meta, [Expr, Default]} | Args], S, E, Acc) ->
+  {ExpandedDefault, SE, _} = elixir_expand:expand(Default, S, E),
+  expand_defaults(Args, SE, E, [{'\\\\', Meta, [Expr, ExpandedDefault]} | Acc]);
+expand_defaults([Arg | Args], S, E, Acc) ->
+   expand_defaults(Args, S, E, [Arg | Acc]);
+expand_defaults([], S, _E, Acc) ->
+  {lists:reverse(Acc), S}.
 
 extract_defaults([{'\\\\', _, [_Expr, Default]} | T], Counter, NewArgs, NewInvoke) ->
   extract_defaults(T, Counter, NewArgs, [Default | NewInvoke]);
@@ -293,7 +326,7 @@ extract_defaults([_ | T], Counter, NewArgs, NewInvoke) ->
 extract_defaults([], _Counter, NewArgs, NewInvoke) ->
   {lists:reverse(NewArgs), lists:reverse(NewInvoke)}.
 
-match_defaults([], 0, Acc) ->
+match_defaults([], _Counter, Acc) ->
   Acc;
 match_defaults([_ | T], Counter, Acc) ->
   NewCounter = Counter - 1,
@@ -306,7 +339,7 @@ default_var(Counter) ->
 
 check_valid_kind(_Meta, _File, _Name, _Arity, Kind, Kind, _StoredFile, _StoredMeta) -> ok;
 check_valid_kind(Meta, File, Name, Arity, Kind, StoredKind, StoredFile, StoredMeta) ->
-  elixir_errors:form_error(Meta, File, ?MODULE,
+  elixir_errors:file_error(Meta, File, ?MODULE,
     {changed_kind, {Name, Arity, StoredKind, Kind, StoredFile, ?line(StoredMeta)}}).
 
 check_valid_clause(Meta, File, Name, Arity, Kind, Set, StoredMeta, StoredFile, Clauses) ->
@@ -314,43 +347,37 @@ check_valid_clause(Meta, File, Name, Arity, Kind, Set, StoredMeta, StoredFile, C
     none ->
       ok;
     {Name, Arity} when Clauses == [] ->
-      elixir_errors:form_warn(Meta, File, ?MODULE,
+      elixir_errors:file_warn(Meta, File, ?MODULE,
         {late_function_head, {Kind, Name, Arity}});
     {Name, Arity} ->
       ok;
     {Name, _} ->
       Relative = elixir_utils:relative_to_cwd(StoredFile),
-      elixir_errors:form_warn(Meta, File, ?MODULE,
+      elixir_errors:file_warn(Meta, File, ?MODULE,
         {ungrouped_name, {Kind, Name, Arity, ?line(StoredMeta), Relative}});
     _ ->
       Relative = elixir_utils:relative_to_cwd(StoredFile),
-      elixir_errors:form_warn(Meta, File, ?MODULE,
+      elixir_errors:file_warn(Meta, File, ?MODULE,
         {ungrouped_arity, {Kind, Name, Arity, ?line(StoredMeta), Relative}})
   end.
 
 % Clause with defaults after clause with defaults
 check_valid_defaults(Meta, File, Name, Arity, Kind, Defaults, StoredDefaults, _, _, _)
     when Defaults > 0, StoredDefaults > 0 ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {duplicate_defaults, {Kind, Name, Arity}});
+  elixir_errors:file_error(Meta, File, ?MODULE, {duplicate_defaults, {Kind, Name, Arity}});
 % Clause with defaults after clause without defaults
 check_valid_defaults(Meta, File, Name, Arity, Kind, Defaults, 0, _, _, _) when Defaults > 0 ->
-  elixir_errors:form_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
+  elixir_errors:file_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
 % Clause without defaults directly after clause with defaults (bodiless does not count)
 check_valid_defaults(Meta, File, Name, Arity, Kind, 0, _, LastDefaults, true, true) when LastDefaults > 0 ->
-  elixir_errors:form_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
+  elixir_errors:file_warn(Meta, File, ?MODULE, {mixed_defaults, {Kind, Name, Arity}});
 % Clause without defaults
 check_valid_defaults(_Meta, _File, _Name, _Arity, _Kind, 0, _StoredDefaults, _LastDefaults, _HasBody, _LastHasBody) ->
   ok.
 
-check_bodiless_function(Check, _Meta, _File, Module, _Kind, _Tuple)
-    when Check == false; Module == 'Elixir.Module' ->
-  ok;
-check_bodiless_function(_Check, Meta, File, _Module, Kind, Tuple) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {function_head, Kind, Tuple}).
-
 check_args_for_function_head(Meta, Args, E) ->
   [begin
-     elixir_errors:form_error(Meta, E, ?MODULE, invalid_args_for_function_head)
+     elixir_errors:module_error(Meta, E, ?MODULE, invalid_args_for_function_head)
    end || Arg <- Args, invalid_arg(Arg)].
 
 invalid_arg({Name, _, Kind}) when is_atom(Name), is_atom(Kind) -> false;
@@ -360,7 +387,7 @@ check_previous_defaults(Meta, Module, Name, Arity, Kind, Defaults, E) ->
   {_Set, Bag} = elixir_module:data_tables(Module),
   Matches = ets:lookup(Bag, {default, Name}),
   [begin
-     elixir_errors:form_error(Meta, E, ?MODULE,
+     elixir_errors:file_error(Meta, E, ?MODULE,
        {defs_with_defaults, Kind, Name, Arity, A})
    end || {_, A, D} <- Matches, A /= Arity, D /= 0, defaults_conflict(A, D, Arity, Defaults)].
 
@@ -369,18 +396,18 @@ defaults_conflict(A, D, Arity, Defaults) ->
     ((A >= (Arity - Defaults)) andalso (A < Arity)).
 
 assert_no_aliases_name(Meta, '__aliases__', [Atom], #{file := File}) when is_atom(Atom) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {no_alias, Atom});
+  elixir_errors:file_error(Meta, File, ?MODULE, {no_alias, Atom});
 assert_no_aliases_name(_Meta, _Aliases, _Args, _S) ->
   ok.
 
 assert_valid_name(Meta, Kind, '__info__', [_], #{file := File}) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {'__info__', Kind});
+  elixir_errors:file_error(Meta, File, ?MODULE, {'__info__', Kind});
 assert_valid_name(Meta, Kind, 'module_info', [], #{file := File}) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {module_info, Kind, 0});
+  elixir_errors:file_error(Meta, File, ?MODULE, {module_info, Kind, 0});
 assert_valid_name(Meta, Kind, 'module_info', [_], #{file := File}) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {module_info, Kind, 1});
+  elixir_errors:file_error(Meta, File, ?MODULE, {module_info, Kind, 1});
 assert_valid_name(Meta, Kind, is_record, [_, _], #{file := File}) when Kind == defp; Kind == def ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {is_record, Kind});
+  elixir_errors:file_error(Meta, File, ?MODULE, {is_record, Kind});
 assert_valid_name(_Meta, _Kind, _Name, _Args, _S) ->
   ok.
 
