@@ -340,7 +340,7 @@ validate_behaviour_info_and_attributes(#{definitions := Defs} = Map, AllCallback
     {false, _} ->
       ok;
     {_, [{Kind, {Name, Arity}, _, _} | _]} when Kind == callback; Kind == macrocallback ->
-      form_error(Map, {callbacks_but_also_behaviour_info, {Kind, Name, Arity}});
+      file_error(Map, {callbacks_but_also_behaviour_info, {Kind, Name, Arity}});
     {_, _} ->
       ok
   end.
@@ -349,16 +349,16 @@ validate_optional_callbacks(Map, AllCallbacks, Optional) ->
   lists:foldl(fun(Callback, Acc) ->
     case Callback of
       {Name, Arity} when is_atom(Name) and is_integer(Arity) -> ok;
-      _ -> form_error(Map, {ill_defined_optional_callback, Callback})
+      _ -> file_error(Map, {ill_defined_optional_callback, Callback})
     end,
 
     case lists:keyfind(Callback, 2, AllCallbacks) of
-      false -> form_error(Map, {unknown_callback, Callback});
+      false -> file_error(Map, {unknown_callback, Callback});
       _ -> ok
     end,
 
     case Acc of
-      #{Callback := _} -> form_error(Map, {duplicate_optional_callback, Callback});
+      #{Callback := _} -> file_error(Map, {duplicate_optional_callback, Callback});
       _ -> ok
     end,
 
@@ -424,7 +424,7 @@ validate_spec_for_existing_function(ModuleMap, NameAndArity, Line) ->
 
   case lists:keymember(NameAndArity, 1, Defs) of
     true -> ok;
-    false -> form_error(#{line => Line, file => File}, {spec_for_undefined_function, NameAndArity})
+    false -> file_error(#{line => Line, file => File}, {spec_for_undefined_function, NameAndArity})
   end.
 
 % Attributes
@@ -510,12 +510,15 @@ get_moduledoc_meta(Set) ->
 
 get_docs(Set, Module, Definitions, Kind) ->
   [{Key,
-    erl_anno:new(Line),
+    maybe_generated(erl_anno:new(Line), Ctx),
     [signature_to_binary(Module, Name, Signature)],
     doc_value(Doc, Name),
     Meta
    } || {Name, Arity} <- Definitions,
-        {Key, _Ctx, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
+        {Key, Ctx, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
+
+maybe_generated(Ann, nil) -> Ann;
+maybe_generated(Ann, _Ctx) -> erl_anno:set_generated(true, Ann).
 
 get_callback_docs(Set, Callbacks) ->
   [{Key,
@@ -556,7 +559,7 @@ signature_to_binary(_, Name, Signature) ->
   Doc = 'Elixir.Inspect.Algebra':format('Elixir.Code':quoted_to_algebra(Quoted), infinity),
   'Elixir.IO':iodata_to_binary(Doc).
 
-checker_chunk(#{definitions := Definitions, deprecated := Deprecated, is_behaviour := IsBehaviour}) ->
+checker_chunk(#{definitions := Definitions, deprecated := Deprecated, defines_behaviour := DefinesBehaviour}) ->
   DeprecatedMap = maps:from_list(Deprecated),
 
   Exports =
@@ -571,7 +574,7 @@ checker_chunk(#{definitions := Definitions, deprecated := Deprecated, is_behavio
     end, [], Definitions),
 
   Contents = #{
-    exports => lists:sort(behaviour_info_exports(IsBehaviour) ++ Exports)
+    exports => lists:sort(behaviour_info_exports(DefinesBehaviour) ++ Exports)
   },
 
   [{<<"ExCk">>, erlang:term_to_binary({elixir_checker_v1, Contents})}].
@@ -581,8 +584,8 @@ behaviour_info_exports(false) -> [].
 
 %% Errors
 
-form_error(#{line := Line, file := File}, Error) ->
-  elixir_errors:form_error([{line, Line}], File, ?MODULE, Error).
+file_error(#{line := Line, file := File}, Error) ->
+  elixir_errors:file_error([{line, Line}], File, ?MODULE, Error).
 
 format_error({ill_defined_optional_callback, Callback}) ->
   io_lib:format("invalid optional callback ~ts. @optional_callbacks expects a "
