@@ -1,6 +1,6 @@
 -module(elixir_expand).
 -export([expand/3, expand_args/3, expand_arg/3, format_error/1]).
--import(elixir_errors, [form_error/4]).
+-import(elixir_errors, [file_error/4, module_error/4, function_error/4]).
 -include("elixir.hrl").
 
 %% =
@@ -27,8 +27,14 @@ expand({'%', Meta, [Left, Right]}, S, E) ->
 expand({'<<>>', Meta, Args}, S, E) ->
   elixir_bitstring:expand(Meta, Args, S, E, false);
 
-expand({'->', Meta, _Args}, _S, E) ->
-  form_error(Meta, E, ?MODULE, unhandled_arrow_op);
+expand({'->', Meta, [_, _]}, _S, E) ->
+  file_error(Meta, E, ?MODULE, unhandled_arrow_op);
+
+expand({'::', Meta, [_, _]}, _S, E) ->
+  file_error(Meta, E, ?MODULE, unhandled_type_op);
+
+expand({'|', Meta, [_, _]}, _S, E) ->
+  file_error(Meta, E, ?MODULE, unhandled_cons_op);
 
 %% __block__
 
@@ -53,12 +59,8 @@ expand({Kind, Meta, [{{'.', _, [Base, '{}']}, _, Refs} | Rest]}, S, E)
     [] ->
       expand_multi_alias_call(Kind, Meta, Base, Refs, [], S, E);
     [Opts] ->
-      case lists:keymember(as, 1, Opts) of
-        true ->
-          form_error(Meta, E, ?MODULE, as_in_multi_alias_call);
-        false ->
-          expand_multi_alias_call(Kind, Meta, Base, Refs, Opts, S, E)
-      end
+      lists:keymember(as, 1, Opts) andalso file_error(Meta, E, ?MODULE, as_in_multi_alias_call),
+      expand_multi_alias_call(Kind, Meta, Base, Refs, Opts, S, E)
   end;
 expand({alias, Meta, [Ref]}, S, E) ->
   expand({alias, Meta, [Ref, []]}, S, E);
@@ -67,14 +69,14 @@ expand({alias, Meta, [Ref, Opts]}, S, E) ->
   {ERef, SR, ER} = expand_without_aliases_report(Ref, S, E),
   {EOpts, ST, ET} = expand_opts(Meta, alias, [as, warn], no_alias_opts(Opts), SR, ER),
   case lists:member(ERef, ['Elixir.True', 'Elixir.False', 'Elixir.Nil']) of
-    true ->  elixir_errors:form_warn(Meta, E, ?MODULE, {commonly_mistaken_alias, Ref});
+    true ->  elixir_errors:file_warn(Meta, E, ?MODULE, {commonly_mistaken_alias, Ref});
     false -> ok
   end,
   if
     is_atom(ERef) ->
       {ERef, ST, expand_alias(Meta, true, ERef, EOpts, ET)};
     true ->
-      form_error(Meta, E, ?MODULE, {expected_compile_time_module, alias, Ref})
+      file_error(Meta, E, ?MODULE, {expected_compile_time_module, alias, Ref})
   end;
 
 expand({require, Meta, [Ref]}, S, E) ->
@@ -90,7 +92,7 @@ expand({require, Meta, [Ref, Opts]}, S, E) ->
       elixir_aliases:ensure_loaded(Meta, ERef, ET),
       {ERef, ST, expand_require(Meta, ERef, EOpts, ET)};
     true ->
-      form_error(Meta, E, ?MODULE, {expected_compile_time_module, require, Ref})
+      file_error(Meta, E, ?MODULE, {expected_compile_time_module, require, Ref})
   end;
 
 expand({import, Meta, [Left]}, S, E) ->
@@ -107,7 +109,7 @@ expand({import, Meta, [Ref, Opts]}, S, E) ->
       {Functions, Macros} = elixir_import:import(Meta, ERef, EOpts, ET),
       {ERef, ST, expand_require(Meta, ERef, EOpts, ET#{functions := Functions, macros := Macros})};
     true ->
-      form_error(Meta, E, ?MODULE, {expected_compile_time_module, import, Ref})
+      file_error(Meta, E, ?MODULE, {expected_compile_time_module, import, Ref})
   end;
 
 %% Compilation environment macros
@@ -118,16 +120,12 @@ expand({'__DIR__', _, Atom}, S, E) when is_atom(Atom) ->
   {filename:dirname(?key(E, file)), S, E};
 expand({'__CALLER__', Meta, Atom} = Caller, S, E) when is_atom(Atom) ->
   assert_no_match_scope(Meta, "__CALLER__", E),
-  case S#elixir_ex.caller of
-    true -> {Caller, S, E};
-    false -> form_error(Meta, E, ?MODULE, caller_not_allowed)
-  end;
+  (not S#elixir_ex.caller) andalso function_error(Meta, E, ?MODULE, caller_not_allowed),
+  {Caller, S, E};
 expand({'__STACKTRACE__', Meta, Atom} = Stacktrace, S, E) when is_atom(Atom) ->
   assert_no_match_scope(Meta, "__STACKTRACE__", E),
-  case S#elixir_ex.stacktrace of
-    true -> {Stacktrace, S, E};
-    false -> form_error(Meta, E, ?MODULE, stacktrace_not_allowed)
-  end;
+  (not S#elixir_ex.stacktrace) andalso function_error(Meta, E, ?MODULE, stacktrace_not_allowed),
+  {Stacktrace, S, E};
 expand({'__ENV__', Meta, Atom}, S, E) when is_atom(Atom) ->
   assert_no_match_scope(Meta, "__ENV__", E),
   {escape_map(escape_env_entries(Meta, S, E)), S, E};
@@ -143,24 +141,24 @@ expand({{'.', DotMeta, [{'__ENV__', Meta, Atom}, Field]}, CallMeta, []}, S, E)
 %% Quote
 
 expand({Unquote, Meta, [_]}, _S, E) when Unquote == unquote; Unquote == unquote_splicing ->
-  form_error(Meta, E, ?MODULE, {unquote_outside_quote, Unquote});
+  file_error(Meta, E, ?MODULE, {unquote_outside_quote, Unquote});
 
 expand({quote, Meta, [Opts]}, S, E) when is_list(Opts) ->
   case lists:keyfind(do, 1, Opts) of
     {do, Do} ->
       expand({quote, Meta, [lists:keydelete(do, 1, Opts), [{do, Do}]]}, S, E);
     false ->
-      form_error(Meta, E, ?MODULE, {missing_option, 'quote', [do]})
+      file_error(Meta, E, ?MODULE, {missing_option, 'quote', [do]})
   end;
 
 expand({quote, Meta, [_]}, _S, E) ->
-  form_error(Meta, E, ?MODULE, {invalid_args, 'quote'});
+  file_error(Meta, E, ?MODULE, {invalid_args, 'quote'});
 
 expand({quote, Meta, [Opts, Do]}, S, E) when is_list(Do) ->
   Exprs =
     case lists:keyfind(do, 1, Do) of
       {do, Expr} -> Expr;
-      false -> form_error(Meta, E, ?MODULE, {missing_option, 'quote', [do]})
+      false -> file_error(Meta, E, ?MODULE, {missing_option, 'quote', [do]})
     end,
 
   ValidOpts = [context, location, line, file, unquote, bind_quoted, generated],
@@ -183,7 +181,7 @@ expand({quote, Meta, [Opts, Do]}, S, E) when is_list(Do) ->
       case is_list(BQ) andalso
             lists:all(fun({Key, _}) when is_atom(Key) -> true; (_) -> false end, BQ) of
         true -> {BQ, false};
-        false -> form_error(Meta, E, ?MODULE, {invalid_bind_quoted_for_quote, BQ})
+        false -> file_error(Meta, E, ?MODULE, {invalid_bind_quoted_for_quote, BQ})
       end;
     false ->
       {[], true}
@@ -197,7 +195,7 @@ expand({quote, Meta, [Opts, Do]}, S, E) when is_list(Do) ->
   expand(Quoted, ST, ET);
 
 expand({quote, Meta, [_, _]}, _S, E) ->
-  form_error(Meta, E, ?MODULE, {invalid_args, 'quote'});
+  file_error(Meta, E, ?MODULE, {invalid_args, 'quote'});
 
 %% Functions
 
@@ -272,31 +270,28 @@ expand({super, Meta, Args}, S, E) when is_list(Args) ->
 
 %% Vars
 
-expand({'^', Meta, [Arg]}, S, #{context := match} = E) ->
-  #elixir_ex{prematch={Prematch, _}, vars={_Read, Write}} = S,
-
-  %% We need to rollback to a no match context
-  NoMatchE = E#{context := nil},
+expand({'^', Meta, [Arg]}, #elixir_ex{prematch={Prematch, _, _}, vars={_, Write}} = S, E) ->
   NoMatchS = S#elixir_ex{prematch=pin, vars={Prematch, Write}},
 
-  case expand(Arg, NoMatchS, NoMatchE) of
+  case expand(Arg, NoMatchS, E#{context := nil}) of
     {{Name, _, Kind} = Var, #elixir_ex{unused=Unused}, _} when is_atom(Name), is_atom(Kind) ->
       {{'^', Meta, [Var]}, S#elixir_ex{unused=Unused}, E};
 
     _ ->
-      form_error(Meta, E, ?MODULE, {invalid_arg_for_pin, Arg})
+      function_error(Meta, E, ?MODULE, {invalid_arg_for_pin, Arg}),
+      {{'^', Meta, [Arg]}, S, E}
   end;
-expand({'^', Meta, [Arg]}, _S, E) ->
-  form_error(Meta, E, ?MODULE, {pin_outside_of_match, Arg});
+expand({'^', Meta, [Arg]}, S, E) ->
+  function_error(Meta, E, ?MODULE, {pin_outside_of_match, Arg}),
+  {{'^', Meta, [Arg]}, S, E};
 
-expand({'_', _Meta, Kind} = Var, S, #{context := match} = E) when is_atom(Kind) ->
+expand({'_', Meta, Kind} = Var, S, #{context := Context} = E) when is_atom(Kind) ->
+  (Context /= match) andalso function_error(Meta, E, ?MODULE, unbound_underscore),
   {Var, S, E};
-expand({'_', Meta, Kind}, _S, E) when is_atom(Kind) ->
-  form_error(Meta, E, ?MODULE, unbound_underscore);
 
 expand({Name, Meta, Kind}, S, #{context := match} = E) when is_atom(Name), is_atom(Kind) ->
   #elixir_ex{
-    prematch={_, PrematchVersion},
+    prematch={_, PrematchVersion, _},
     unused={Unused, Version},
     vars={Read, Write}
   } = S,
@@ -336,10 +331,19 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
     case Read of
       #{Pair := CurrentVersion} ->
         case Prematch of
-          {bitsize, Pre, Original} ->
+          {Pre, _Counter, {bitsize, Original}} ->
             if
-              is_map_key(Pair, Pre); not is_map_key(Pair, Original) -> {ok, CurrentVersion};
-              true -> raise
+              map_get(Pair, Pre) /= CurrentVersion ->
+                {ok, CurrentVersion};
+              is_map_key(Pair, Pre) ->
+                %% TODO: Enable this warning on Elixir v1.19
+                %% TODO: Remove me on Elixir 2.0
+                %% elixir_errors:file_warn(Meta, E, ?MODULE, {unpinned_bitsize_var, Name, Kind}),
+                {ok, CurrentVersion};
+              not is_map_key(Pair, Original) ->
+                {ok, CurrentVersion};
+              true ->
+                raise
             end;
 
           _ ->
@@ -358,23 +362,26 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
 
     Error ->
       case lists:keyfind(if_undefined, 1, Meta) of
-        %% TODO: Remove this clause on v2.0 as we will always raise by default
-        {if_undefined, raise} ->
-          form_error(Meta, E, ?MODULE, {undefined_var, Name, Kind});
-
         {if_undefined, apply} ->
           expand({Name, Meta, []}, S, E);
 
-        %% TODO: Remove this clause on v2.0
+        %% TODO: Remove this clause on v2.0 as we will raise by default
+        {if_undefined, raise} ->
+          function_error(Meta, E, ?MODULE, {undefined_var, Name, Kind}),
+          {{Name, Meta, Kind}, S, E};
+
+        %% TODO: Remove this clause on v2.0 as we will no longer support warn
         _ when Error == warn ->
-          elixir_errors:form_warn(Meta, E, ?MODULE, {unknown_variable, Name}),
+          elixir_errors:file_warn(Meta, E, ?MODULE, {undefined_var_to_call, Name}),
           expand({Name, [{if_undefined, warn} | Meta], []}, S, E);
 
         _ when Error == pin ->
-          form_error(Meta, E, ?MODULE, {undefined_var_pin, Name, Kind});
+          function_error(Meta, E, ?MODULE, {undefined_var_pin, Name, Kind}),
+          {{Name, Meta, Kind}, S, E};
 
         _ ->
-          form_error(Meta, E, ?MODULE, {undefined_var, Name, Kind})
+          function_error(Meta, E, ?MODULE, {undefined_var, Name, Kind}),
+          {{Name, Meta, Kind}, S, E}
       end
   end;
 
@@ -401,19 +408,14 @@ expand({{'.', DotMeta, [Left, Right]}, Meta, Args}, S, E)
 
 expand({{'.', DotMeta, [Expr]}, Meta, Args}, S, E) when is_list(Args) ->
   assert_no_match_or_guard_scope(Meta, "anonymous call", S, E),
-
-  case expand_args([Expr | Args], S, E) of
-    {[EExpr | _], _, _} when is_atom(EExpr) ->
-      form_error(Meta, E, ?MODULE, {invalid_function_call, EExpr});
-
-    {[EExpr | EArgs], SA, EA} ->
-      {{{'.', DotMeta, [EExpr]}, Meta, EArgs}, SA, EA}
-  end;
+  {[EExpr | EArgs], SA, EA} = expand_args([Expr | Args], S, E),
+  is_atom(EExpr) andalso function_error(Meta, E, ?MODULE, {invalid_function_call, EExpr}),
+  {{{'.', DotMeta, [EExpr]}, Meta, EArgs}, SA, EA};
 
 %% Invalid calls
 
 expand({_, Meta, Args} = Invalid, _S, E) when is_list(Meta) and is_list(Args) ->
-  form_error(Meta, E, ?MODULE, {invalid_call, Invalid});
+  file_error(Meta, E, ?MODULE, {invalid_call, Invalid});
 
 %% Literals
 
@@ -436,7 +438,7 @@ expand(Function, S, E) when is_function(Function) ->
     true ->
       {elixir_quote:fun_to_quoted(Function), S, E};
     false ->
-      form_error([{line, 0}], ?key(E, file), ?MODULE, {invalid_quoted_expr, Function})
+      file_error([{line, 0}], ?key(E, file), ?MODULE, {invalid_quoted_expr, Function})
   end;
 
 expand(Pid, S, E) when is_pid(Pid) ->
@@ -445,7 +447,7 @@ expand(Pid, S, E) when is_pid(Pid) ->
       {Pid, S, E};
     Function ->
       %% TODO: Make me an error on v2.0
-      elixir_errors:form_warn([], E, ?MODULE, {invalid_pid_in_function, Pid, Function}),
+      elixir_errors:file_warn([], E, ?MODULE, {invalid_pid_in_function, Pid, Function}),
       {Pid, E}
   end;
 
@@ -453,7 +455,7 @@ expand(Other, S, E) when is_number(Other); is_atom(Other); is_binary(Other) ->
   {Other, S, E};
 
 expand(Other, _S, E) ->
-  form_error([{line, 0}], ?key(E, file), ?MODULE, {invalid_quoted_expr, Other}).
+  file_error([{line, 0}], ?key(E, file), ?MODULE, {invalid_quoted_expr, Other}).
 
 %% Helpers
 
@@ -465,7 +467,7 @@ escape_env_entries(Meta, #elixir_ex{vars={Read, _}}, Env0) ->
 
   Env1#{versioned_vars := escape_map(Read), line := ?line(Meta)}.
 
-escape_map(Map) -> {'%{}', [], maps:to_list(Map)}.
+escape_map(Map) -> {'%{}', [], lists:sort(maps:to_list(Map))}.
 
 expand_multi_alias_call(Kind, Meta, Base, Refs, Opts, S, E) ->
   {BaseRef, SB, EB} = expand_without_aliases_report(Base, S, E),
@@ -478,7 +480,7 @@ expand_multi_alias_call(Kind, Meta, Base, Refs, Opts, S, E) ->
       expand({Kind, Meta, [elixir_aliases:concat([BaseRef, Ref]), Opts]}, SR, ER);
 
     (Other, _SR, _ER) ->
-      form_error(Meta, E, ?MODULE, {expected_compile_time_module, Kind, Other})
+      file_error(Meta, E, ?MODULE, {expected_compile_time_module, Kind, Other})
   end,
 
   mapfold(Fun, SB, EB, Refs).
@@ -494,7 +496,7 @@ resolve_super(Meta, Arity, E) ->
       {Kind, Name, SuperMeta};
 
     _ ->
-      form_error(Meta, E, ?MODULE, wrong_number_of_args_for_super)
+      file_error(Meta, E, ?MODULE, wrong_number_of_args_for_super)
   end.
 
 expand_fn_capture(Meta, Arg, S, E) ->
@@ -505,7 +507,7 @@ expand_fn_capture(Meta, Arg, S, E) ->
       AttachedMeta = attach_context_module(Remote, Meta, E),
       {{'&', AttachedMeta, [{'/', [], [{{'.', [], [Remote, Fun]}, [], []}, Arity]}]}, SE, EE};
     {{local, Fun, Arity}, _LocalMeta, _SE, #{function := nil}} ->
-      form_error(Meta, E, ?MODULE, {undefined_local_capture, Fun, Arity});
+      file_error(Meta, E, ?MODULE, {undefined_local_capture, Fun, Arity});
     {{local, Fun, Arity}, _LocalMeta, SE, EE} ->
       {{'&', Meta, [{'/', [], [{Fun, [], nil}, Arity]}]}, SE, EE};
     {expand, Expr, SE, EE} ->
@@ -547,7 +549,7 @@ expand_block([H | T], Acc, Meta, S, E) ->
   %% simply returning something as replacement.
   case is_useless_building(H, EH, Meta) of
     {UselessMeta, UselessTerm} ->
-      elixir_errors:form_warn(UselessMeta, E, ?MODULE, UselessTerm);
+      elixir_errors:file_warn(UselessMeta, E, ?MODULE, UselessTerm);
 
     false ->
       ok
@@ -609,7 +611,7 @@ mapfold(_Fun, S, E, [], Acc) ->
 
 var_unused({_, Kind} = Pair, Meta, Version, Unused, Override) ->
   case (Kind == nil) andalso should_warn(Meta) of
-    true -> Unused#{{Pair, Version} => {?line(Meta), Override}};
+    true -> Unused#{{Pair, Version} => {Meta, Override}};
     false -> Unused
   end.
 
@@ -625,7 +627,7 @@ var_used(Meta, {_, Kind} = Pair, Version, Unused) ->
 maybe_warn_underscored_var_repeat(Meta, Name, Kind, E) ->
   case should_warn(Meta) andalso atom_to_list(Name) of
     "_" ++ _ ->
-      elixir_errors:form_warn(Meta, E, ?MODULE, {underscored_var_repeat, Name, Kind});
+      elixir_errors:file_warn(Meta, E, ?MODULE, {underscored_var_repeat, Name, Kind});
     _ ->
       ok
   end.
@@ -633,7 +635,7 @@ maybe_warn_underscored_var_repeat(Meta, Name, Kind, E) ->
 maybe_warn_underscored_var_access(Meta, Name, Kind, E) ->
   case (Kind == nil) andalso should_warn(Meta) andalso atom_to_list(Name) of
     "_" ++ _ ->
-      elixir_errors:form_warn(Meta, E, ?MODULE, {underscored_var_access, Name});
+      elixir_errors:file_warn(Meta, E, ?MODULE, {underscored_var_access, Name});
     _ ->
       ok
   end.
@@ -647,7 +649,7 @@ maybe_warn_deprecated_super_in_gen_server_callback(Meta, Function, SuperMeta, E)
           ok;
 
         _ ->
-          elixir_errors:form_warn(Meta, E, ?MODULE, {super_in_genserver, Function})
+          elixir_errors:file_warn(Meta, E, ?MODULE, {super_in_genserver, Function})
       end;
 
     _ ->
@@ -715,20 +717,7 @@ generated_case_clauses([{do, Clauses}]) ->
 
 expand_for({for, Meta, [_ | _] = Args}, S, E, Return) ->
   assert_no_match_or_guard_scope(Meta, "for", S, E),
-
-  {Cases, Block} =
-    case elixir_utils:split_last(Args) of
-      {OuterCases, OuterOpts} when is_list(OuterOpts) ->
-        case elixir_utils:split_last(OuterCases) of
-          {InnerCases, InnerOpts} when is_list(InnerOpts) ->
-            {InnerCases, InnerOpts ++ OuterOpts};
-          _ ->
-            {OuterCases, OuterOpts}
-        end;
-      _ ->
-        {Args, []}
-    end,
-
+  {Cases, Block} = elixir_utils:split_opts(Args),
   validate_opts(Meta, for, [do, into, uniq, reduce], Block, E),
 
   {Expr, Opts} =
@@ -736,7 +725,7 @@ expand_for({for, Meta, [_ | _] = Args}, S, E, Return) ->
       {value, {do, Do}, DoOpts} ->
         {Do, DoOpts};
       false ->
-        form_error(Meta, E, ?MODULE, {missing_option, for, [do]})
+        file_error(Meta, E, ?MODULE, {missing_option, for, [do]})
     end,
 
   {EOpts, SO, EO} = expand(Opts, elixir_env:reset_unused_vars(S), E),
@@ -746,7 +735,7 @@ expand_for({for, Meta, [_ | _] = Args}, S, E, Return) ->
   {{EExpr, SE, EE}, NormalizedOpts} =
     case validate_for_options(EOpts, false, false, false, Return, Meta, E, []) of
       {ok, MaybeReduce, NOpts} -> {expand_for_do_block(Meta, Expr, SC, EC, MaybeReduce), NOpts};
-      {error, Error} -> {form_error(Meta, E, ?MODULE, Error), EOpts}
+      {error, Error} -> {file_error(Meta, E, ?MODULE, Error), EOpts}
     end,
 
   {{for, Meta, ECases ++ [[{do, EExpr} | NormalizedOpts]]},
@@ -767,14 +756,14 @@ validate_for_options([], _Into = false, Uniq, Reduce = false, Return = true, Met
   Pair = {into, []},
   validate_for_options([Pair], Pair, Uniq, Reduce, Return, Meta, E, Acc);
 validate_for_options([], Into = false, {uniq, true}, Reduce = false, Return = false, Meta, E, Acc) ->
-  elixir_errors:form_warn(Meta, E, ?MODULE, for_with_unused_uniq),
+  elixir_errors:file_warn(Meta, E, ?MODULE, for_with_unused_uniq),
   AccWithoutUniq = lists:keydelete(uniq, 1, Acc),
   validate_for_options([], Into, false, Reduce, Return, Meta, E, AccWithoutUniq);
 validate_for_options([], _Into, _Uniq, Reduce, _Return, _Meta, _E, Acc) ->
   {ok, Reduce, lists:reverse(Acc)}.
 
 expand_for_do_block(Meta, [{'->', _, _} | _], _S, E, false) ->
-  form_error(Meta, E, ?MODULE, for_without_reduce_bad_block);
+  file_error(Meta, E, ?MODULE, for_without_reduce_bad_block);
 expand_for_do_block(_Meta, Expr, S, E, false) ->
   expand(Expr, S, E);
 expand_for_do_block(Meta, [{'->', _, _} | _] = Clauses, S, E, {reduce, _}) ->
@@ -788,13 +777,13 @@ expand_for_do_block(Meta, [{'->', _, _} | _] = Clauses, S, E, {reduce, _}) ->
       {EClause, elixir_env:merge_and_check_unused_vars(SAcc, SA, EAcc)};
 
     (_, _) ->
-      form_error(Meta, E, ?MODULE, for_with_reduce_bad_block)
+      file_error(Meta, E, ?MODULE, for_with_reduce_bad_block)
   end,
 
   {Do, SA} = lists:mapfoldl(Transformer, S, Clauses),
   {Do, SA, E};
 expand_for_do_block(Meta, _Expr, _S, E, {reduce, _}) ->
-  form_error(Meta, E, ?MODULE, for_with_reduce_bad_block).
+  file_error(Meta, E, ?MODULE, for_with_reduce_bad_block).
 
 %% Locals
 
@@ -804,7 +793,7 @@ assert_no_ambiguous_op(Name, Meta, [Arg], S, E) ->
       Pair = {Name, Kind},
       case S#elixir_ex.vars of
         {#{Pair := _}, _} ->
-          form_error(Meta, E, ?MODULE, {op_ambiguity, Name, Arg});
+          file_error(Meta, E, ?MODULE, {op_ambiguity, Name, Arg});
         _ ->
           ok
       end;
@@ -822,29 +811,36 @@ assert_no_clauses(Name, Meta, Args, E) ->
 assert_arg_with_no_clauses(Name, Meta, [{Key, Value} | Rest], E) when is_atom(Key) ->
   case Value of
     [{'->', _, _} | _] ->
-      form_error(Meta, E, ?MODULE, {invalid_clauses, Name});
+      file_error(Meta, E, ?MODULE, {invalid_clauses, Name});
     _ ->
       assert_arg_with_no_clauses(Name, Meta, Rest, E)
   end;
 assert_arg_with_no_clauses(_Name, _Meta, _Arg, _E) ->
   ok.
 
-expand_local(Meta, Name, Args, S, #{module := Module, function := Function, context := nil} = E)
+expand_local(Meta, Name, Args, S, #{module := Module, function := Function, context := Context} = E)
     when Function /= nil ->
   assert_no_clauses(Name, Meta, Args, E),
-  Arity = length(Args),
-  elixir_env:trace({local_function, Meta, Name, Arity}, E),
-  elixir_locals:record_local({Name, Arity}, Module, Function, Meta, false),
+
+  %% In case we have the wrong context, we log a module error
+  %% so we can print multiple entries at the same time.
+  case Context of
+    match ->
+      module_error(Meta, E, ?MODULE, {invalid_local_invocation, match, {Name, Meta, Args}});
+
+    guard ->
+      module_error(Meta, E, ?MODULE, {invalid_local_invocation, guard_context(S), {Name, Meta, Args}});
+
+    nil ->
+      Arity = length(Args),
+      elixir_env:trace({local_function, Meta, Name, Arity}, E),
+      elixir_locals:record_local({Name, Arity}, Module, Function, Meta, false)
+  end,
+
   {EArgs, SA, EA} = expand_args(Args, S, E),
   {{Name, Meta, EArgs}, SA, EA};
-expand_local(Meta, Name, Args, _S, E) when Name == '|'; Name == '::' ->
-  form_error(Meta, E, ?MODULE, {undefined_function, Name, Args});
 expand_local(Meta, Name, Args, _S, #{function := nil} = E) ->
-  form_error(Meta, E, ?MODULE, {undefined_function, Name, Args});
-expand_local(Meta, Name, Args, _S, #{context := match} = E) ->
-  form_error(Meta, E, ?MODULE, {invalid_local_invocation, match, {Name, Meta, Args}});
-expand_local(Meta, Name, Args, S, #{context := guard} = E) ->
-  form_error(Meta, E, ?MODULE, {invalid_local_invocation, guard_context(S), {Name, Meta, Args}}).
+  file_error(Meta, E, ?MODULE, {undefined_function, Name, Args}).
 
 %% Remote
 
@@ -852,31 +848,28 @@ expand_remote(Receiver, DotMeta, Right, Meta, Args, S, SL, #{context := Context}
   assert_no_clauses(Right, Meta, Args, E),
 
   case {Context, lists:keyfind(no_parens, 1, Meta)} of
-    {guard, {no_parens, true}} when is_tuple(Receiver) ->
-      {{{'.', DotMeta, [Receiver, Right]}, Meta, []}, SL, E};
+    {guard, NoParens} when is_tuple(Receiver) ->
+      (NoParens /= {no_parens, true}) andalso
+        function_error(Meta, E, ?MODULE, {parens_map_lookup, Receiver, Right, guard_context(S)}),
 
-    {guard, _} when is_tuple(Receiver) ->
-      form_error(Meta, E, ?MODULE, {parens_map_lookup, Receiver, Right, guard_context(S)});
+      {{{'.', DotMeta, [Receiver, Right]}, Meta, []}, SL, E};
 
     _ ->
       AttachedDotMeta = attach_context_module(Receiver, DotMeta, E),
-
-      is_atom(Receiver) andalso
-        elixir_env:trace({remote_function, Meta, Receiver, Right, length(Args)}, E),
-
       {EArgs, {SA, _}, EA} = mapfold(fun expand_arg/3, {SL, S}, E, Args),
 
       case rewrite(Context, Receiver, AttachedDotMeta, Right, Meta, EArgs, S) of
         {ok, Rewritten} ->
           maybe_warn_comparison(Rewritten, Args, E),
           {Rewritten, elixir_env:close_write(SA, S), EA};
+
         {error, Error} ->
-          form_error(Meta, E, elixir_rewrite, Error)
+          file_error(Meta, E, elixir_rewrite, Error)
       end
   end;
 expand_remote(Receiver, DotMeta, Right, Meta, Args, _, _, E) ->
   Call = {{'.', DotMeta, [Receiver, Right]}, Meta, Args},
-  form_error(Meta, E, ?MODULE, {invalid_call, Call}).
+  file_error(Meta, E, ?MODULE, {invalid_call, Call}).
 
 attach_context_module(_Receiver, Meta, #{function := nil}) ->
   Meta;
@@ -906,10 +899,10 @@ maybe_warn_comparison({{'.', _, [erlang, Op]}, Meta, [ELeft, ERight]}, [Left, Ri
       case is_nested_comparison(Op, ELeft, ERight, Left, Right) of
         false -> ok;
         CompExpr ->
-          elixir_errors:form_warn(Meta, E, ?MODULE, {nested_comparison, CompExpr})
+          elixir_errors:file_warn(Meta, E, ?MODULE, {nested_comparison, CompExpr})
       end;
     StructExpr ->
-      elixir_errors:form_warn(Meta, E, ?MODULE, {struct_comparison, StructExpr})
+      elixir_errors:file_warn(Meta, E, ?MODULE, {struct_comparison, StructExpr})
   end;
 maybe_warn_comparison(_, _, _) ->
   ok.
@@ -957,11 +950,11 @@ expand_opts(Meta, Kind, Allowed, Opts, S, E) ->
 
 validate_opts(Meta, Kind, Allowed, Opts, E) when is_list(Opts) ->
   [begin
-    form_error(Meta, E, ?MODULE, {unsupported_option, Kind, Key})
+    file_error(Meta, E, ?MODULE, {unsupported_option, Kind, Key})
   end || {Key, _} <- Opts, not lists:member(Key, Allowed)];
 
 validate_opts(Meta, Kind, _Allowed, Opts, E) ->
-  form_error(Meta, E, ?MODULE, {options_are_not_keyword, Kind, Opts}).
+  file_error(Meta, E, ?MODULE, {options_are_not_keyword, Kind, Opts}).
 
 no_alias_opts(Opts) when is_list(Opts) ->
   case lists:keyfind(as, 1, Opts) of
@@ -1004,22 +997,22 @@ expand_as({as, Atom}, Meta, _IncludeByDefault, _Ref, E) when is_atom(Atom), not
         [_] ->
           Atom;
         _ ->
-          form_error(Meta, E, ?MODULE, {invalid_alias_for_as, nested_alias, Atom})
+          file_error(Meta, E, ?MODULE, {invalid_alias_for_as, nested_alias, Atom})
       end;
     _ ->
-      form_error(Meta, E, ?MODULE, {invalid_alias_for_as, not_alias, Atom})
+      file_error(Meta, E, ?MODULE, {invalid_alias_for_as, not_alias, Atom})
   end;
 expand_as(false, Meta, IncludeByDefault, Ref, E) ->
   if
     IncludeByDefault ->
       case elixir_aliases:last(Ref) of
         {ok, NewRef} -> NewRef;
-        error -> form_error(Meta, E, ?MODULE, {invalid_alias_module, Ref})
+        error -> file_error(Meta, E, ?MODULE, {invalid_alias_module, Ref})
       end;
     true -> Ref
   end;
 expand_as({as, Other}, Meta, _IncludeByDefault, _Ref, E) ->
-  form_error(Meta, E, ?MODULE, {invalid_alias_for_as, not_alias, Other}).
+  file_error(Meta, E, ?MODULE, {invalid_alias_for_as, not_alias, Other}).
 
 %% Aliases
 
@@ -1044,7 +1037,7 @@ expand_aliases({'__aliases__', Meta, _} = Alias, S, E, Report) ->
           {Receiver, SA, EA};
 
         false ->
-          form_error(Meta, E, ?MODULE, {invalid_alias, Alias})
+          file_error(Meta, E, ?MODULE, {invalid_alias, Alias})
       end
   end.
 
@@ -1075,12 +1068,12 @@ assert_generator_start(_, [{'<-', _, [_, _]} | _], _) ->
 assert_generator_start(_, [{'<<>>', _, [{'<-', _, [_, _]}]} | _], _) ->
   ok;
 assert_generator_start(Meta, _, E) ->
-  elixir_errors:form_error(Meta, E, ?MODULE, for_generator_start).
+  elixir_errors:file_error(Meta, E, ?MODULE, for_generator_start).
 
 %% Assertions
 
 refute_parallel_bitstring_match({'<<>>', _, _}, {'<<>>', Meta, _} = Arg, E, true) ->
-  form_error(Meta, E, ?MODULE, {parallel_bitstring_match, Arg});
+  file_error(Meta, E, ?MODULE, {parallel_bitstring_match, Arg});
 refute_parallel_bitstring_match(Left, {'=', _Meta, [MatchLeft, MatchRight]}, E, Parallel) ->
   refute_parallel_bitstring_match(Left, MatchLeft, E, true),
   refute_parallel_bitstring_match(Left, MatchRight, E, Parallel);
@@ -1119,27 +1112,27 @@ refute_parallel_bitstring_match_map_field(_Args1, _Args2, _E, _Parallel) ->
   ok.
 
 assert_module_scope(Meta, Kind, #{module := nil, file := File}) ->
-  form_error(Meta, File, ?MODULE, {invalid_expr_in_scope, "module", Kind});
+  file_error(Meta, File, ?MODULE, {invalid_expr_in_scope, "module", Kind});
 assert_module_scope(_Meta, _Kind, #{module:=Module}) -> Module.
 
 assert_function_scope(Meta, Kind, #{function := nil, file := File}) ->
-  form_error(Meta, File, ?MODULE, {invalid_expr_in_scope, "function", Kind});
+  file_error(Meta, File, ?MODULE, {invalid_expr_in_scope, "function", Kind});
 assert_function_scope(_Meta, _Kind, #{function := Function}) -> Function.
 
 assert_no_match_or_guard_scope(Meta, Kind, S, E) ->
   assert_no_match_scope(Meta, Kind, E),
   assert_no_guard_scope(Meta, Kind, S, E).
 assert_no_match_scope(Meta, Kind, #{context := match, file := File}) ->
-  form_error(Meta, File, ?MODULE, {invalid_pattern_in_match, Kind});
-assert_no_match_scope(_Meta, _Kind, _E) -> [].
+  file_error(Meta, File, ?MODULE, {invalid_pattern_in_match, Kind});
+assert_no_match_scope(_Meta, _Kind, _E) -> ok.
 assert_no_guard_scope(Meta, Kind, S, #{context := guard, file := File}) ->
   Key =
     case S#elixir_ex.prematch of
-      {bitsize, _, _}  -> invalid_expr_in_bitsize;
+      {_, _, {bitsize, _}}  -> invalid_expr_in_bitsize;
       _ -> invalid_expr_in_guard
     end,
-  form_error(Meta, File, ?MODULE, {Key, Kind});
-assert_no_guard_scope(_Meta, _Kind, _S, _E) -> [].
+  file_error(Meta, File, ?MODULE, {Key, Kind});
+assert_no_guard_scope(_Meta, _Kind, _S, _E) -> ok.
 
 %% Here we look into the Clauses "optimistically", that is, we don't check for
 %% multiple "do"s and similar stuff. After all, the error we're gonna give here
@@ -1150,7 +1143,7 @@ assert_no_guard_scope(_Meta, _Kind, _S, _E) -> [].
 assert_no_underscore_clause_in_cond([{do, Clauses}], E) when is_list(Clauses) ->
   case lists:last(Clauses) of
     {'->', Meta, [[{'_', _, Atom}], _]} when is_atom(Atom) ->
-      form_error(Meta, E, ?MODULE, underscore_in_cond);
+      file_error(Meta, E, ?MODULE, underscore_in_cond);
     _Other ->
       ok
   end;
@@ -1159,7 +1152,7 @@ assert_no_underscore_clause_in_cond(_Other, _E) ->
 
 %% Errors
 
-guard_context(#elixir_ex{prematch={bitsize, _, _}}) -> "bitstring size specifier";
+guard_context(#elixir_ex{prematch={_, _, {bitsize, _}}}) -> "bitstring size specifier";
 guard_context(_) -> "guards".
 
 format_error({useless_literal, Term}) ->
@@ -1192,7 +1185,21 @@ format_error(for_generator_start) ->
 format_error(for_with_unused_uniq) ->
   "the :uniq option has no effect since the result of the for comprehension is not used";
 format_error(unhandled_arrow_op) ->
-  "unhandled operator ->";
+  "misplaced operator ->\n\n"
+  "This typically means invalid syntax or a macro is not available in scope";
+format_error(unhandled_cons_op) ->
+  "misplaced operator |/2\n\n"
+  "The | operator is typically used between brackets as the cons operator:\n\n"
+  "    [head | tail]\n\n"
+  "where head is a sequence of elements separated by commas and the tail\n"
+  "is the remaining of a list.\n\n"
+  "It is also used to update maps and structs, via the %{map | key: value} notation,\n"
+  "and in typespecs, such as @type and @spec, to express the union of two types";
+format_error(unhandled_type_op) ->
+  "misplaced operator ::/2\n\n"
+  "The :: operator is typically used in bitstrings to specify types and sizes of segments:\n\n"
+  "    <<size::32-integer, letter::utf8, rest::binary>>\n\n"
+  "It is also used in typespecs, such as @type and @spec, to describe inputs and outputs";
 format_error(as_in_multi_alias_call) ->
   ":as option is not supported by multi-alias call";
 format_error({invalid_alias_module, Ref}) ->
@@ -1216,9 +1223,14 @@ format_error({invalid_arg_for_pin, Arg}) ->
   io_lib:format("invalid argument for unary operator ^, expected an existing variable, got: ^~ts",
                 ['Elixir.Macro':to_string(Arg)]);
 format_error({pin_outside_of_match, Arg}) ->
-  io_lib:format("cannot use ^~ts outside of match clauses", ['Elixir.Macro':to_string(Arg)]);
+  io_lib:format(
+    "misplaced operator ^~ts\n\n"
+    "The pin operator ^ is supported only inside matches or inside custom macros. "
+    "Make sure you are inside a match or all necessary macros have been required",
+    ['Elixir.Macro':to_string(Arg)]
+  );
 format_error(unbound_underscore) ->
-  "invalid use of _. \"_\" represents a value to be ignored in a pattern and cannot be used in expressions";
+  "invalid use of _. _ can only be used inside patterns to ignore values and cannot be used in expressions. Make sure you are inside a pattern or change it accordingly";
 format_error({undefined_var, Name, Kind}) ->
   io_lib:format("undefined variable \"~ts\"~ts", [Name, context_info(Kind)]);
 format_error({undefined_var_pin, Name, Kind}) ->
@@ -1299,20 +1311,12 @@ format_error({unsupported_option, Kind, Key}) ->
 format_error({options_are_not_keyword, Kind, Opts}) ->
   io_lib:format("invalid options for ~s, expected a keyword list, got: ~ts",
                 [Kind, 'Elixir.Macro':to_string(Opts)]);
-format_error({undefined_function, '|', [_, _]}) ->
-  "misplaced operator |/2\n\n"
-  "The | operator is typically used between brackets as the cons operator:\n\n"
-  "    [head | tail]\n\n"
-  "where head is a single element and the tail is the remaining of a list.\n"
-  "It is also used to update maps and structs, via the %{map | key: value} notation,\n"
-  "and in typespecs, such as @type and @spec, to express the union of two types";
-format_error({undefined_function, '::', [_, _]}) ->
-  "misplaced operator ::/2\n\n"
-  "The :: operator is typically used in bitstrings to specify types and sizes of segments:\n\n"
-  "    <<size::32-integer, letter::utf8, rest::binary>>\n\n"
-  "It is also used in typespecs, such as @type and @spec, to describe inputs and outputs";
 format_error({undefined_function, Name, Args}) ->
   io_lib:format("undefined function ~ts/~B (there is no such import)", [Name, length(Args)]);
+format_error({unpinned_bitsize_var, Name, Kind}) ->
+  io_lib:format("the variable \"~ts\"~ts is accessed inside size(...) of a bitstring "
+                "but it was defined outside of the match. You must preceed it with the "
+                "pin operator", [Name, context_info(Kind)]);
 format_error({underscored_var_repeat, Name, Kind}) ->
   io_lib:format("the underscored variable \"~ts\"~ts appears more than once in a "
                 "match. This means the pattern will only match if all \"~ts\" bind "
@@ -1347,7 +1351,7 @@ format_error(caller_not_allowed) ->
   "__CALLER__ is available only inside defmacro and defmacrop";
 format_error(stacktrace_not_allowed) ->
   "__STACKTRACE__ is available only inside catch and rescue clauses of try expressions";
-format_error({unknown_variable, Name}) ->
+format_error({undefined_var_to_call, Name}) ->
   io_lib:format("variable \"~ts\" does not exist and is being expanded to \"~ts()\","
                 " please use parentheses to remove the ambiguity or change the variable name", [Name, Name]);
 format_error({parens_map_lookup, Map, Field, Context}) ->
