@@ -23,7 +23,7 @@ import(Meta, Ref, Opts, E) ->
         {Added2, Macs} = import_macros(false, Meta, Ref, Opts, E),
         {Funs, Macs, Added1 or Added2};
       {only, Other} ->
-        elixir_errors:form_error(Meta, E, ?MODULE, {invalid_option, only, Other});
+        elixir_errors:file_error(Meta, E, ?MODULE, {invalid_option, only, Other});
       false ->
         {Added1, Funs} = import_functions(Meta, Ref, Opts, E),
         {Added2, Macs} = import_macros(false, Meta, Ref, Opts, E),
@@ -44,7 +44,7 @@ import_macros(Force, Meta, Ref, Opts, E) ->
       {ok, Macros} ->
         Macros;
       error when Force ->
-        elixir_errors:form_error(Meta, E, ?MODULE, {no_macros, Ref});
+        elixir_errors:file_error(Meta, E, ?MODULE, {no_macros, Ref});
       error ->
         []
     end
@@ -65,40 +65,49 @@ import_sigil_macros(Meta, Ref, Opts, E) ->
     end
   end).
 
-filter_sigils(Key) ->
-  lists:filter(fun({Atom, _}) ->
-    case atom_to_list(Atom) of
-      "sigil_" ++ [L] when L >= $a, L =< $z; L >= $A, L =< $Z -> true;
-      _ -> false
-    end
-  end, Key).
+filter_sigils(Funs) ->
+  lists:filter(fun is_sigil/1, Funs).
+
+is_sigil({Name, 2}) ->
+  case atom_to_list(Name) of
+    "sigil_" ++ Letters ->
+      case Letters of
+        [L] when L >= $a, L =< $z -> true;
+        [] -> false;
+        Letters -> lists:all(fun(L) -> L >= $A andalso L =< $Z end, Letters)
+      end;
+    _ ->
+      false
+  end;
+is_sigil(_) ->
+  false.
 
 %% Calculates the imports based on only and except
 
 calculate(Meta, Key, Opts, Old, File, Existing) ->
   New = case keyfind(only, Opts) of
-    {only, Only} when is_list(Only) ->
-      ensure_keyword_list(Meta, File, Only, only),
-      ensure_no_duplicates(Meta, File, Only, only),
+    {only, DupOnly} when is_list(DupOnly) ->
+      ensure_keyword_list(Meta, File, DupOnly, only),
+      Only = ensure_no_duplicates(Meta, File, DupOnly, only),
+
       case keyfind(except, Opts) of
-        false ->
-          ok;
-        _ ->
-          elixir_errors:form_error(Meta, File, ?MODULE, only_and_except_given)
+        false -> ok;
+        _ -> elixir_errors:file_error(Meta, File, ?MODULE, only_and_except_given)
       end,
-      case Only -- get_exports(Key) of
-        [{Name, Arity} | _] ->
-          elixir_errors:form_error(Meta, File, ?MODULE, {invalid_import, {Key, Name, Arity}});
-        _ ->
-          intersection(Only, Existing())
-      end;
+
+      [elixir_errors:file_warn(Meta, File, ?MODULE, {invalid_import, {Key, Name, Arity}}) ||
+       {Name, Arity} <- Only -- get_exports(Key)],
+
+      intersection(Only, Existing());
+
     _ ->
       case keyfind(except, Opts) of
         false ->
           remove_underscored(Existing());
-        {except, Except} when is_list(Except) ->
-          ensure_keyword_list(Meta, File, Except, except),
-          ensure_no_duplicates(Meta, File, Except, except),
+
+        {except, DupExcept} when is_list(DupExcept) ->
+          ensure_keyword_list(Meta, File, DupExcept, except),
+          Except = ensure_no_duplicates(Meta, File, DupExcept, except),
           %% We are not checking existence of exports listed in :except option
           %% on purpose: to support backwards compatible code.
           %% For example, "import String, except: [trim: 1]"
@@ -107,21 +116,19 @@ calculate(Meta, Key, Opts, Old, File, Existing) ->
             false -> remove_underscored(Existing()) -- Except;
             {Key, OldImports} -> OldImports -- Except
           end;
+
         {except, Other} ->
-          elixir_errors:form_error(Meta, File, ?MODULE, {invalid_option, except, Other})
+          elixir_errors:file_error(Meta, File, ?MODULE, {invalid_option, except, Other})
       end
   end,
 
   %% Normalize the data before storing it
-  Set   = ordsets:from_list(New),
-  Final = remove_internals(Set),
-
-  case Final of
+  case ordsets:from_list(New) of
     [] ->
       {false, keydelete(Key, Old)};
-    _  ->
-      ensure_no_special_form_conflict(Meta, File, Key, Final),
-      {true, [{Key, Final} | keydelete(Key, Old)]}
+    Set  ->
+      ensure_no_special_form_conflict(Meta, File, Key, Set),
+      {true, [{Key, Set} | keydelete(Key, Old)]}
   end.
 
 %% Retrieve functions and macros from modules
@@ -133,7 +140,7 @@ get_functions(Module) ->
   try
     Module:'__info__'(functions)
   catch
-    error:undef -> Module:module_info(exports)
+    error:undef -> remove_internals(Module:module_info(exports))
   end.
 
 get_macros(Module) ->
@@ -156,7 +163,7 @@ fetch_macros(Module) ->
 ensure_no_special_form_conflict(Meta, File, Key, [{Name, Arity} | T]) ->
   case special_form(Name, Arity) of
     true  ->
-      elixir_errors:form_error(Meta, File, ?MODULE, {special_form_conflict, {Key, Name, Arity}});
+      elixir_errors:file_error(Meta, File, ?MODULE, {special_form_conflict, {Key, Name, Arity}});
     false ->
       ensure_no_special_form_conflict(Meta, File, Key, T)
   end;
@@ -169,13 +176,14 @@ ensure_keyword_list(Meta, File, [{Key, Value} | Rest], Kind) when is_atom(Key),
   ensure_keyword_list(Meta, File, Rest, Kind);
 
 ensure_keyword_list(Meta, File, _Other, Kind) ->
-  elixir_errors:form_error(Meta, File, ?MODULE, {invalid_option, Kind}).
+  elixir_errors:file_error(Meta, File, ?MODULE, {invalid_option, Kind}).
 
 ensure_no_duplicates(Meta, File, Option, Kind) ->
   lists:foldl(fun({Name, Arity}, Acc) ->
     case lists:member({Name, Arity}, Acc) of
       true ->
-        elixir_errors:form_error(Meta, File, ?MODULE, {duplicated_import, {Kind, Name, Arity}});
+        elixir_errors:file_warn(Meta, File, ?MODULE, {duplicated_import, {Kind, Name, Arity}}),
+        Acc;
       false ->
         [{Name, Arity} | Acc]
     end
@@ -241,8 +249,7 @@ remove_underscored(List) ->
   end, List).
 
 remove_internals(Set) ->
-  ordsets:del_element({module_info, 1},
-    ordsets:del_element({module_info, 0}, Set)).
+  Set -- [{behaviour_info, 1}, {module_info, 1}, {module_info, 0}].
 
 %% Special forms
 
