@@ -39,7 +39,7 @@ extract([$\\, Last, Last, Last | Rest], Buffer, Output, Line, Column, Scope, Int
   extract(Rest, [Last, Last, Last | Buffer], Output, Line, Column+4, Scope, Interpol, All);
 
 extract([$\\, $#, ${ | Rest], Buffer, Output, Line, Column, Scope, true, Last) ->
-  extract(Rest, [${, $#, $\\ | Buffer], Output, Line, Column+1, Scope, true, Last);
+  extract(Rest, [${, $#, $\\ | Buffer], Output, Line, Column+3, Scope, true, Last);
 
 extract([$#, ${ | Rest], Buffer, Output, Line, Column, Scope, true, Last) ->
   Output1 = build_string(Buffer, Output),
@@ -51,7 +51,7 @@ extract([$#, ${ | Rest], Buffer, Output, Line, Column, Scope, true, Last) ->
     {error, Reason, _, _, _} ->
       {error, Reason};
     {ok, EndLine, EndColumn, Warnings, Tokens} when Scope#elixir_tokenizer.cursor_completion /= false ->
-      NewScope = Scope#elixir_tokenizer{warnings=Warnings, cursor_completion=terminators},
+      NewScope = Scope#elixir_tokenizer{warnings=Warnings, cursor_completion=noprune},
       Output2 = build_interpol(Line, Column, EndLine, EndColumn, Tokens, Output1),
       extract([], [], Output2, EndLine, EndColumn, NewScope, true, Last);
     {ok, _, _, _, _} ->
@@ -71,17 +71,18 @@ extract(Rest, Buffer, Output, Line, Column, Scope, Interpol, Last) ->
   extract_char(Rest, Buffer, Output, Line, Column, Scope, Interpol, Last).
 
 extract_char(Rest, Buffer, Output, Line, Column, Scope, Interpol, Last) ->
-  [Char | NewRest] = unicode_util:gc(Rest),
-
-  if
-    ?bidi(Char) ->
+  case unicode_util:gc(Rest) of
+    [Char | _] when ?bidi(Char) ->
       Token = io_lib:format("\\u~4.16.0B", [Char]),
       Pre = "invalid bidirectional formatting character in string: ",
       Pos = io_lib:format(". If you want to use such character, use it in its escaped ~ts form instead", [Token]),
       {error, {Line, Column, {Pre, Pos}, Token}};
 
-    true ->
-      extract(NewRest, [Char | Buffer], Output, Line, Column + 1, Scope, Interpol, Last)
+    [Char | NewRest] ->
+      extract(NewRest, [Char | Buffer], Output, Line, Column + 1, Scope, Interpol, Last);
+
+    [] ->
+      extract([], Buffer, Output, Line, Column, Scope, Interpol, Last)
   end.
 
 %% Handle newlines. Heredocs require special attention
