@@ -9,6 +9,11 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string({:{}, [], nil}) == "{}"
     end
 
+    test "variable with colors" do
+      opts = [syntax_colors: [variable: :blue]]
+      assert quoted_to_string(quote(do: foo), opts) == "\e[34mfoo\e[0m"
+    end
+
     test "local call" do
       assert quoted_to_string(quote(do: foo(1, 2, 3))) == "foo(1, 2, 3)"
       assert quoted_to_string(quote(do: foo([1, 2, 3]))) == "foo([1, 2, 3])"
@@ -25,6 +30,13 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quote(line: __ENV__.line, do: foo(1, b))) == "foo(1, b)"
     end
 
+    test "local call with colors" do
+      opts = [syntax_colors: [call: :blue, number: :yellow, variable: :red]]
+
+      assert quoted_to_string(quote(do: foo(1, a)), opts) ==
+               "\e[34mfoo\e[0m(\e[33m1\e[0m, \e[31ma\e[0m)"
+    end
+
     test "local call no parens" do
       assert quoted_to_string({:def, [], [1, 2]}) == "def 1, 2"
       assert quoted_to_string({:def, [closing: []], [1, 2]}) == "def(1, 2)"
@@ -77,6 +89,19 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quote(do: Foo.bar(<<>>, []))) == "Foo.bar(<<>>, [])"
     end
 
+    test "remote call with colors" do
+      opts = [syntax_colors: [call: :blue, number: :yellow, variable: :red, atom: :green]]
+
+      assert quoted_to_string(quote(do: foo.bar(1, 2)), opts) ==
+               "\e[31mfoo\e[0m.\e[34mbar\e[0m(\e[33m1\e[0m, \e[33m2\e[0m)"
+
+      assert quoted_to_string(quote(do: :foo.bar(1, 2)), opts) ==
+               "\e[32m:foo\e[0m.\e[34mbar\e[0m(\e[33m1\e[0m, \e[33m2\e[0m)"
+
+      assert quoted_to_string(quote(do: Foo.Bar.bar(1, 2)), opts) ==
+               "\e[32mFoo.Bar\e[0m.\e[34mbar\e[0m(\e[33m1\e[0m, \e[33m2\e[0m)"
+    end
+
     test "keyword call" do
       assert quoted_to_string(quote(do: Foo.bar(foo: :bar))) == "Foo.bar(foo: :bar)"
       assert quoted_to_string(quote(do: Foo.bar("Elixir.Foo": :bar))) == "Foo.bar([{Foo, :bar}])"
@@ -154,6 +179,13 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quote do: {1, 2, 3, foo: :bar}) == "{1, 2, 3, foo: :bar}"
     end
 
+    test "tuple with colors" do
+      opts = [syntax_colors: [tuple: :blue, number: :yellow]]
+
+      assert quoted_to_string(quote(do: {1, 2, 3}), opts) ==
+               "\e[34m{\e[0m\e[33m1\e[0m, \e[33m2\e[0m, \e[33m3\e[0m\e[34m}\e[0m"
+    end
+
     test "tuple call" do
       assert quoted_to_string(quote(do: alias(Foo.{Bar, Baz, Bong}))) ==
                "alias Foo.{Bar, Baz, Bong}"
@@ -194,6 +226,11 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quoted) <> "\n" == expected
     end
 
+    test "invalid block" do
+      assert quoted_to_string({:__block__, [], {:bar, [], []}}) ==
+               "{:__block__, [], {:bar, [], []}}"
+    end
+
     test "not in" do
       assert quoted_to_string(quote(do: false not in [])) == "false not in []"
     end
@@ -368,7 +405,7 @@ defmodule Code.Normalizer.QuotedASTTest do
     end
 
     test "when" do
-      assert quoted_to_string(quote(do: (() -> x))) == "(() -> x)"
+      assert quoted_to_string(quote(do: (-> x))) == "(-> x)"
       assert quoted_to_string(quote(do: (x when y -> z))) == "(x when y -> z)"
       assert quoted_to_string(quote(do: (x, y when z -> w))) == "(x, y when z -> w)"
       assert quoted_to_string(quote(do: (x, y when z -> w))) == "(x, y when z -> w)"
@@ -463,6 +500,12 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quote(do: identity(&1))) == "identity(&1)"
     end
 
+    test "operators with colors" do
+      opts = [syntax_colors: [operator: :blue, number: :yellow]]
+      assert quoted_to_string(quote(do: !!1), opts) == "\e[34m!\e[0m\e[34m!\e[0m\e[33m1\e[0m"
+      assert quoted_to_string(quote(do: 1 + 2), opts) == "\e[33m1\e[0m\e[34m +\e[0m \e[33m2\e[0m"
+    end
+
     test "access" do
       assert quoted_to_string(quote(do: a[b])) == "a[b]"
       assert quoted_to_string(quote(do: a[1 + 2])) == "a[1 + 2]"
@@ -483,6 +526,13 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quote(do: [{binary(), integer()}])) == "[{binary(), integer()}]"
     end
 
+    test "keyword list with colors" do
+      opts = [syntax_colors: [list: :blue, atom: :green, number: :yellow]]
+
+      assert quoted_to_string(quote(do: [a: 1, b: 2]), opts) ==
+               "\e[34m[\e[0m\e[32ma:\e[0m \e[33m1\e[0m, \e[32mb:\e[0m \e[33m2\e[0m\e[34m]\e[0m"
+    end
+
     test "interpolation" do
       assert quoted_to_string(quote(do: "foo#{bar}baz")) == ~S["foo#{bar}baz"]
     end
@@ -510,11 +560,18 @@ defmodule Code.Normalizer.QuotedASTTest do
     test "integer/float" do
       assert quoted_to_string(1) == "1"
       assert quoted_to_string({:__block__, [], [1]}) == "1"
+      assert quoted_to_string(1.23) == "1.23"
+    end
+
+    test "integer/float with colors" do
+      opts = [syntax_colors: [number: :yellow]]
+      assert quoted_to_string(1, opts) == "\e[33m1\e[0m"
+      assert quoted_to_string(1.23, opts) == "\e[33m1.23\e[0m"
     end
 
     test "charlist" do
       assert quoted_to_string(quote(do: [])) == "[]"
-      assert quoted_to_string(quote(do: 'abc')) == "'abc'"
+      assert quoted_to_string(quote(do: ~c"abc")) == ~S/~c"abc"/
 
       # False positive
       assert quoted_to_string(
@@ -534,6 +591,11 @@ defmodule Code.Normalizer.QuotedASTTest do
       assert quoted_to_string(quote(do: "#{"abc"}")) == ~S/"#{"abc"}"/
     end
 
+    test "string with colors" do
+      opts = [syntax_colors: [string: :green]]
+      assert quoted_to_string(quote(do: "abc"), opts) == "\e[32m\"abc\"\e[0m"
+    end
+
     test "catch-all" do
       assert quoted_to_string(quote do: {unquote(self())}) == "{#{inspect(self())}}"
       assert quoted_to_string(quote do: foo(unquote(self()))) == "foo(#{inspect(self())})"
@@ -602,32 +664,44 @@ defmodule Code.Normalizer.QuotedASTTest do
     end
 
     test "charlists with slash escapes" do
-      assert quoted_to_string(quote(do: '\a\b\e\n\r\t\v'), escape: false) ==
-               ~s/'\a\b\e\n\r\t\v'/
+      assert quoted_to_string(~c"\a\b\e\n\r\t\v", escape: false) ==
+               ~s/~c"\a\b\e\n\r\t\v"/
 
-      assert quoted_to_string(quote(do: '\a\b\e\n\r\t\v')) ==
-               ~s/'\\a\\b\\e\\n\\r\\t\\v'/
+      assert quoted_to_string(~c"\a\b\e\n\r\t\v") ==
+               ~s/~c"\\a\\b\\e\\n\\r\\t\\v"/
 
-      assert quoted_to_string({:__block__, [], ['\a\b\e\n\r\t\v']}, escape: false) ==
-               ~s/'\a\b\e\n\r\t\v'/
+      assert quoted_to_string({:__block__, [], [~c"\a\b\e\n\r\t\v"]}, escape: false) ==
+               ~s/~c"\a\b\e\n\r\t\v"/
 
-      assert quoted_to_string({:__block__, [], ['\a\b\e\n\r\t\v']}) ==
-               ~s/'\\a\\b\\e\\n\\r\\t\\v'/
+      assert quoted_to_string({:__block__, [], [~c"\a\b\e\n\r\t\v"]}) ==
+               ~s/~c"\\a\\b\\e\\n\\r\\t\\v"/
     end
 
     test "charlists with non printable characters" do
-      assert quoted_to_string(quote(do: '\x00\x01\x10'), escape: false) == ~S/[0, 1, 16]/
-      assert quoted_to_string(quote(do: '\x00\x01\x10')) == ~S/[0, 1, 16]/
+      assert quoted_to_string(~c"\x00\x01\x10", escape: false) == ~S/[0, 1, 16]/
+      assert quoted_to_string(~c"\x00\x01\x10") == ~S/[0, 1, 16]/
     end
 
     test "charlists with interpolations" do
-      assert quoted_to_string(quote(do: 'one #{2} three'), escape: false) == ~S/'one #{2} three'/
-      assert quoted_to_string(quote(do: 'one #{2} three')) == ~S/'one #{2} three'/
+      # using string_to_quoted to avoid the formatter fixing the charlists
 
-      assert quoted_to_string(quote(do: 'one\n\'#{2}\'\nthree'), escape: false) ==
-               ~s['one\n\\'\#{2}\\'\nthree']
+      assert Code.string_to_quoted!(~S/'one #{2} three'/) |> quoted_to_string(escape: false) ==
+               ~S/~c"one #{2} three"/
 
-      assert quoted_to_string(quote(do: 'one\n\'#{2}\'\nthree')) == ~S['one\n\'#{2}\'\nthree']
+      assert Code.string_to_quoted!(~S/'one #{2} three'/) |> quoted_to_string() ==
+               ~S/~c"one #{2} three"/
+
+      assert Code.string_to_quoted!(~S/'one\n\'#{2}\'\nthree'/) |> quoted_to_string(escape: false) ==
+               ~s[~c"one\n'\#{2}'\nthree"]
+
+      assert Code.string_to_quoted!(~S/'one\n\'#{2}\'\nthree'/) |> quoted_to_string() ==
+               ~S[~c"one\n'#{2}'\nthree"]
+
+      assert Code.string_to_quoted!(~S/'one\n"#{2}"\nthree'/) |> quoted_to_string(escape: false) ==
+               ~s[~c"one\n\\"\#{2}\\"\nthree"]
+
+      assert Code.string_to_quoted!(~S/'one\n"#{2}"\nthree'/) |> quoted_to_string() ==
+               ~S[~c"one\n\"#{2}\"\nthree"]
     end
 
     test "atoms" do
