@@ -227,7 +227,7 @@ defmodule Kernel.QuoteTest do
                 () -> 1
               end)
 
-    assert [{:->, _, [[], 1]}] = quote(do: (() -> 1))
+    assert [{:->, _, [[], 1]}] = quote(do: (-> 1))
   end
 
   test "empty block" do
@@ -322,9 +322,21 @@ defmodule Kernel.QuoteTest do
                 end
               end)
   end
+
+  test "capture" do
+    assert Code.string_to_quoted!("&1[:foo]") == Code.string_to_quoted!("(&1)[:foo]")
+    assert Code.string_to_quoted!("&1 [:foo]") == Code.string_to_quoted!("(&1)[:foo]")
+    assert Code.string_to_quoted!("& 1[:foo]") == Code.string_to_quoted!("&(1[:foo])")
+  end
+
+  test "not and ! as rearrange ops" do
+    assert {:__block__, _, [{:not, [line: 1], [true]}]} = Code.string_to_quoted!("(not true)")
+
+    assert {:fn, _, [{:->, _, [[], {:not, _, [true]}]}]} =
+             Code.string_to_quoted!("fn -> not true end")
+  end
 end
 
-# DO NOT MOVE THIS LINE
 defmodule Kernel.QuoteTest.Errors do
   def line, do: __ENV__.line + 4
 
@@ -562,25 +574,25 @@ defmodule Kernel.QuoteTest.ImportsHygieneTest do
 
   defmacrop get_list_length do
     quote do
-      length('hello')
+      length(~c"hello")
     end
   end
 
   defmacrop get_list_length_with_pipe do
     quote do
-      'hello' |> length()
+      ~c"hello" |> length()
     end
   end
 
   defmacrop get_list_length_with_partial do
     quote do
-      (&length(&1)).('hello')
+      (&length(&1)).(~c"hello")
     end
   end
 
   defmacrop get_list_length_with_function do
     quote do
-      (&length/1).('hello')
+      (&length/1).(~c"hello")
     end
   end
 
@@ -619,7 +631,7 @@ defmodule Kernel.QuoteTest.ImportsHygieneTest do
     quote do
       import Kernel, except: [length: 1]
       import String, only: [length: 1]
-      length('hello')
+      length(~c"hello")
     end
   end
 
