@@ -44,7 +44,7 @@ defmodule MacroTest do
 
     test "escapes bitstring" do
       assert {:<<>>, [], args} = Macro.escape(<<300::12>>)
-      assert [{:"::", [], [1, {:size, [], [4]}]}, {:"::", [], [",", {:binary, [], []}]}] = args
+      assert [{:"::", [], [1, {:size, [], [4]}]}, {:"::", [], [",", {:binary, [], nil}]}] = args
     end
 
     test "escapes recursively" do
@@ -402,6 +402,136 @@ defmodule MacroTest do
       assert formatted =~ ~r/[^\n]\n\n$/
     end
 
+    test "with simple boolean expressions" do
+      {result, formatted} = dbg_format(:rand.uniform() < 0.0 and length([]) == 0)
+      assert result == false
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             :rand.uniform() < 0.0 #=> false
+             :rand.uniform() < 0.0 and length([]) == 0 #=> false
+             """
+    end
+
+    test "with left-associative operators" do
+      {result, formatted} = dbg_format(List.first([]) || "yes" || raise("foo"))
+      assert result == "yes"
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             List.first([]) #=> nil
+             List.first([]) || "yes" #=> "yes"
+             List.first([]) || "yes" || raise "foo" #=> "yes"
+             """
+    end
+
+    test "with composite boolean expressions" do
+      true1 = length([]) == 0
+      true2 = length([]) == 0
+      {result, formatted} = dbg_format((true1 and true2) or (List.first([]) || true1))
+
+      assert result == true
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             true1 #=> true
+             true1 and true2 #=> true
+             (true1 and true2) or (List.first([]) || true1) #=> true
+             """
+    end
+
+    test "with case" do
+      list = [1, 2, 3]
+
+      {result, formatted} =
+        dbg_format(
+          case list do
+            [] -> nil
+            _ -> Enum.sum(list)
+          end
+        )
+
+      assert result == 6
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             Case argument:
+             list #=> [1, 2, 3]
+
+             Case expression (clause #2 matched):
+             case list do
+               [] -> nil
+               _ -> Enum.sum(list)
+             end #=> 6
+             """
+    end
+
+    test "with case - guard" do
+      {result, formatted} =
+        dbg_format(
+          case 0..100//5 do
+            %{first: first, last: last, step: step} when last > first ->
+              count = div(last - first, step)
+              {:ok, count}
+
+            _ ->
+              :error
+          end
+        )
+
+      assert result == {:ok, 20}
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             Case argument:
+             0..100//5 #=> 0..100//5
+
+             Case expression (clause #1 matched):
+             case 0..100//5 do
+               %{first: first, last: last, step: step} when last > first ->
+                 count = div(last - first, step)
+                 {:ok, count}
+
+               _ ->
+                 :error
+             end #=> {:ok, 20}
+             """
+    end
+
+    test "with cond" do
+      map = %{b: 5}
+
+      {result, formatted} =
+        dbg_format(
+          cond do
+            a = map[:a] -> a + 1
+            b = map[:b] -> b * 2
+            true -> nil
+          end
+        )
+
+      assert result == 10
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             Cond clause (clause #2 matched):
+             b = map[:b] #=> 5
+
+             Cond expression:
+             cond do
+               a = map[:a] -> a + 1
+               b = map[:b] -> b * 2
+               true -> nil
+             end #=> 10
+             """
+    end
+
     test "with \"syntax_colors: []\" it doesn't print any color sequences" do
       {_result, formatted} = dbg_format("hello")
       refute formatted =~ "\e["
@@ -413,7 +543,7 @@ defmodule MacroTest do
     end
 
     test "forwards options to the underlying inspect calls" do
-      value = 'hello'
+      value = ~c"hello"
       assert {^value, formatted} = dbg_format(value, syntax_colors: [], charlists: :as_lists)
       assert formatted =~ "value #=> [104, 101, 108, 108, 111]\n"
     end
@@ -439,6 +569,32 @@ defmodule MacroTest do
       assert Macro.to_string(576_460_752_303_423_455) == "576_460_752_303_423_455"
       assert Macro.to_string(-576_460_752_303_423_455) == "-576_460_752_303_423_455"
     end
+
+    defmodule HTML do
+      defstruct [:string]
+
+      defimpl Inspect do
+        def inspect(%{string: string}, _) do
+          "~HTML[#{string}]"
+        end
+      end
+    end
+
+    defmacro sigil_HTML({:<<>>, _, [string]}, []) do
+      Macro.escape(%HTML{string: string})
+    end
+
+    test "sigils" do
+      assert Macro.to_string(quote(do: ~HTML[hi])) == ~S/~HTML[hi]/
+
+      assert Macro.to_string(
+               quote do
+                 ~HTML"""
+                 hi
+                 """
+               end
+             ) == ~s[~HTML"""\nhi\n"""]
+    end
   end
 
   describe "to_string/2" do
@@ -530,6 +686,8 @@ defmodule MacroTest do
       assert macro_to_string(quote(do: ~S["'(123)'"])) == ~S/~S["'(123)'"]/
       assert macro_to_string(quote(do: ~s"#{"foo"}")) == ~S/~s"#{"foo"}"/
 
+      assert macro_to_string(quote(do: ~HTML[hi])) == ~S/~HTML[hi]/
+
       assert macro_to_string(
                quote do
                  ~s"""
@@ -569,6 +727,14 @@ defmodule MacroTest do
                  """
                end
              ) == ~s[~S"""\n"123"\n"""]
+
+      assert macro_to_string(
+               quote do
+                 ~HTML"""
+                 "123"
+                 """
+               end
+             ) == ~s[~HTML"""\n"123"\n"""]
     end
 
     test "tuple call" do
@@ -769,7 +935,7 @@ defmodule MacroTest do
     end
 
     test "when" do
-      assert macro_to_string(quote(do: (() -> x))) == "(() -> x)"
+      assert macro_to_string(quote(do: (-> x))) == "(() -> x)"
       assert macro_to_string(quote(do: (x when y -> z))) == "(x when y -> z)"
       assert macro_to_string(quote(do: (x, y when z -> w))) == "((x, y) when z -> w)"
       assert macro_to_string(quote(do: (x, y when z -> w))) == "((x, y) when z -> w)"
@@ -891,7 +1057,8 @@ defmodule MacroTest do
 
     test "charlist" do
       assert macro_to_string(quote(do: [])) == "[]"
-      assert macro_to_string(quote(do: 'abc')) == "'abc'"
+      assert macro_to_string(quote(do: ~c"abc")) == ~S/~c"abc"/
+      assert macro_to_string(quote(do: [?a, ?b, ?c])) == ~S/~c"abc"/
     end
 
     test "string" do
@@ -977,15 +1144,18 @@ defmodule MacroTest do
       env = %{__ENV__ | file: "foo", line: 12}
 
       assert Macro.Env.stacktrace(env) ==
-               [{__MODULE__, :"test env stacktrace", 1, [file: 'foo', line: 12]}]
+               [{__MODULE__, :"test env stacktrace", 1, [file: ~c"foo", line: 12]}]
 
       env = %{env | function: nil}
-      assert Macro.Env.stacktrace(env) == [{__MODULE__, :__MODULE__, 0, [file: 'foo', line: 12]}]
+
+      assert Macro.Env.stacktrace(env) == [
+               {__MODULE__, :__MODULE__, 0, [file: ~c"foo", line: 12]}
+             ]
 
       env = %{env | module: nil}
 
       assert Macro.Env.stacktrace(env) ==
-               [{:elixir_compiler, :__FILE__, 1, [file: 'foo', line: 12]}]
+               [{:elixir_compiler, :__FILE__, 1, [file: ~c"foo", line: 12]}]
     end
 
     test "context modules" do
@@ -1025,6 +1195,9 @@ defmodule MacroTest do
     assert Macro.pipe(1, quote(do: foo), -1) == quote(do: foo(1))
     assert Macro.pipe(2, quote(do: foo(1)), -1) == quote(do: foo(1, 2))
 
+    assert Macro.pipe(quote(do: %{foo: "bar"}), quote(do: Access.get(:foo)), 0) ==
+             quote(do: Access.get(%{foo: "bar"}, :foo))
+
     assert_raise ArgumentError, ~r"cannot pipe 1 into 2", fn ->
       Macro.pipe(1, 2, 0)
     end
@@ -1062,6 +1235,12 @@ defmodule MacroTest do
     assert_raise ArgumentError, message, fn ->
       Macro.pipe(:foo, quote(do: fn x -> x end), 0)
     end
+
+    message = ~r"wrong operator precedence when piping into bracket-based access"
+
+    assert_raise ArgumentError, message, fn ->
+      Macro.pipe(:foo, quote(do: %{foo: bar}[:foo]), 0)
+    end
   end
 
   test "unpipe/1" do
@@ -1260,7 +1439,16 @@ defmodule MacroTest do
     assert Macro.quoted_literal?(quote(do: {"foo", 1}))
     assert Macro.quoted_literal?(quote(do: %{foo: "bar"}))
     assert Macro.quoted_literal?(quote(do: %URI{path: "/"}))
+    assert Macro.quoted_literal?(quote(do: <<>>))
+    assert Macro.quoted_literal?(quote(do: <<1, "foo", "bar"::utf16>>))
+    assert Macro.quoted_literal?(quote(do: <<1000::size(8)-unit(4)>>))
+    assert Macro.quoted_literal?(quote(do: <<1000::8*4>>))
+    assert Macro.quoted_literal?(quote(do: <<102::unsigned-big-integer-size(8)>>))
     refute Macro.quoted_literal?(quote(do: {"foo", var}))
+    refute Macro.quoted_literal?(quote(do: <<"foo"::size(name_size)>>))
+    refute Macro.quoted_literal?(quote(do: <<"foo"::binary-size(name_size)>>))
+    refute Macro.quoted_literal?(quote(do: <<"foo"::custom_modifier()>>))
+    refute Macro.quoted_literal?(quote(do: <<102, rest::binary>>))
   end
 
   test "underscore/1" do
