@@ -126,7 +126,6 @@ defmodule ProtocolTest do
 
     write_beam(
       defprotocol SampleDocsProto do
-        @type t :: any
         @doc "Ok"
         @deprecated "Reason"
         @spec ok(t) :: boolean
@@ -136,6 +135,9 @@ defmodule ProtocolTest do
 
     {:docs_v1, _, _, _, _, _, docs} = Code.fetch_docs(SampleDocsProto)
 
+    assert {{:type, :t, 0}, _, [], %{"en" => type_doc}, _} = List.keyfind(docs, {:type, :t, 0}, 0)
+    assert type_doc =~ "All the types that implement this protocol"
+
     assert {{:function, :ok, 1}, _, ["ok(term)"], %{"en" => "Ok"}, _} =
              List.keyfind(docs, {:function, :ok, 1}, 0)
 
@@ -159,6 +161,10 @@ defmodule ProtocolTest do
     assert args == [{:type, 23, :product, [{:user_type, 23, :t, []}]}, {:type, 23, :term, []}]
   end
 
+  test "protocol defines t/0 type with documentation" do
+    assert {:type, {:t, {_, _, :any, []}, []}} = get_type(@sample_binary, :t, 0)
+  end
+
   test "protocol defines functions and attributes" do
     assert Sample.__protocol__(:module) == Sample
     assert Sample.__protocol__(:functions) == [ok: 1]
@@ -230,6 +236,15 @@ defmodule ProtocolTest do
     List.keyfind(callbacks, {name, arity}, 0) |> elem(1)
   end
 
+  defp get_type(beam, name, arity) do
+    {:ok, types} = Code.Typespec.fetch_types(beam)
+
+    assert {:value, value} =
+             :lists.search(&match?({_, {^name, _, args}} when length(args) == arity, &1), types)
+
+    value
+  end
+
   test "derives protocol implicitly" do
     struct = %ImplStruct{a: 1, b: 1}
     assert WithAny.ok(struct) == {:ok, struct}
