@@ -55,7 +55,7 @@ defmodule PathHelpers do
   end
 
   defp run_cmd(executable, args) do
-    '#{executable} #{IO.chardata_to_string(args)}#{redirect_std_err_on_win()}'
+    ~c"#{executable} #{IO.chardata_to_string(args)}#{redirect_std_err_on_win()}"
     |> :os.cmd()
     |> :binary.list_to_bin()
   end
@@ -98,9 +98,19 @@ os_exclude = if PathHelpers.windows?(), do: [unix: true], else: [windows: true]
 {line_exclude, line_include} =
   if line = System.get_env("LINE"), do: {[:test], [line: line]}, else: {[], []}
 
+distributed_exclude =
+  if Code.ensure_loaded?(:peer) and Node.alive?() do
+    {:ok, _pid, node} = :peer.start(%{name: :secondary})
+    true = :erpc.call(node, :code, :set_path, [:code.get_path()])
+    {:ok, _} = :erpc.call(node, :application, :ensure_all_started, [:elixir])
+    []
+  else
+    [distributed: true]
+  end
+
 ExUnit.start(
   trace: !!System.get_env("TRACE"),
   assert_receive_timeout: assert_timeout,
-  exclude: epmd_exclude ++ os_exclude ++ line_exclude,
+  exclude: epmd_exclude ++ os_exclude ++ line_exclude ++ distributed_exclude,
   include: line_include
 )
