@@ -28,7 +28,7 @@ defmodule ExUnit do
   command line. Assuming you named the file `assertion_test.exs`,
   you can run it as:
 
-      elixir assertion_test.exs
+      $ elixir assertion_test.exs
 
   ## Case, Callbacks and Assertions
 
@@ -128,12 +128,13 @@ defmodule ExUnit do
       * `:tests` - all tests in this module
 
     """
-    defstruct [:file, :name, :state, tests: []]
+    defstruct [:file, :name, :state, tags: %{}, tests: []]
 
     @type t :: %__MODULE__{
             file: binary(),
             name: module,
             state: ExUnit.state(),
+            tags: map,
             tests: [ExUnit.Test.t()]
           }
   end
@@ -235,6 +236,14 @@ defmodule ExUnit do
 
     * `:colors` - a keyword list of color options to be used by some formatters:
       * `:enabled` - boolean option to enable colors, defaults to `IO.ANSI.enabled?/0`;
+
+      * `:success` - success message (defaults to `:green`)
+      * `:invalid` - invalid test message (defaults to `:yellow`)
+      * `:skipped` - skipped test message (defaults to `:yellow`)
+      * `:failure` - failed test message (defaults to `:red`)
+      * `:error_info` - display of actual error (defaults to `:red`)
+      * `:extra_info` - additional information (defaults to `:cyan`)
+      * `:location_info` - filename and tags (defaults to `[:bright, :black]`)
       * `:diff_insert` - color of the insertions on diffs, defaults to `:green`;
       * `:diff_insert_whitespace` - color of the whitespace insertions on diffs,
         defaults to `IO.ANSI.color_background(2, 0, 0)`;
@@ -361,7 +370,7 @@ defmodule ExUnit do
   Returns a map containing the total number of tests, the number
   of failures, the number of excluded tests and the number of skipped tests.
   """
-  @spec run(list(atom)) :: suite_result()
+  @spec run([module()]) :: suite_result()
   def run(additional_modules \\ []) do
     for module <- additional_modules do
       module_attributes = module.__info__(:attributes)
@@ -387,8 +396,9 @@ defmodule ExUnit do
   @doc since: "1.12.0"
   @spec async_run() :: Task.t()
   def async_run() do
+    options = persist_defaults(configuration())
+
     Task.async(fn ->
-      options = persist_defaults(configuration())
       ExUnit.Runner.run(options, nil)
     end)
   end
