@@ -17,9 +17,10 @@ defmodule ExUnit.Callbacks do
   test is run. All `setup` callbacks are run before each test. No callback
   is run if the test case has no tests or all tests have been filtered out.
 
-  `setup` and `setup_all` callbacks can be defined by a block, by passing
-  an atom naming a one-arity function, or by passing a list of such
-  atoms. Both can opt to receive the current context by specifying it
+  `setup` and `setup_all` callbacks can be defined by either a block, an atom
+  naming a local function, a `{module, function}` tuple, or a list of atoms/tuples.
+
+  Both can opt to receive the current context by specifying it
   as parameter if defined by a block. Functions used to define a test
   setup must accept the context as single argument.
 
@@ -124,13 +125,13 @@ defmodule ExUnit.Callbacks do
 
   It is also common to define your setup as a series of functions,
   which are put together by calling `setup` or `setup_all` with a
-  list of atoms. Each of these functions receive the context and can
+  list of function names. Each of these functions receive the context and can
   return any of the values allowed in `setup` blocks:
 
       defmodule ExampleContextTest do
         use ExUnit.Case
 
-        setup [:step1, :step2, :step3]
+        setup [:step1, :step2, :step3, {OtherModule, :step4}]
 
         defp step1(_context), do: [step_one: true]
         defp step2(_context), do: {:ok, step_two: true} # return values with shape of {:ok, keyword() | map()} allowed
@@ -183,8 +184,12 @@ defmodule ExUnit.Callbacks do
   @doc """
   Defines a callback to be run before each test in a case.
 
-  Accepts a block or the name of a one-arity function in the form of an atom,
-  or a list of such atoms.
+  Accepts one of these:
+
+    * a block
+    * an atom naming a local function
+    * a `{module, function}` tuple
+    * a list of atoms and `{module, function}` tuples
 
   Can return values to be merged into the context, to set up the state for
   tests. For more details, see the "Context" section shown above.
@@ -193,24 +198,28 @@ defmodule ExUnit.Callbacks do
 
   ## Examples
 
-      def clean_up_tmp_directory(context) do
+      defp clean_up_tmp_directory(context) do
         # perform setup
         :ok
       end
 
       setup :clean_up_tmp_directory
 
+      setup [:clean_up_tmp_directory, :another_setup]
+
       setup do
         [conn: Plug.Conn.build_conn()]
       end
 
+      setup {MyModule, :my_setup_function}
+
   """
-  defmacro setup(block) do
-    if Keyword.keyword?(block) do
-      do_setup(quote(do: _), block)
+  defmacro setup(block_or_functions) do
+    if Keyword.keyword?(block_or_functions) do
+      do_setup(quote(do: _), block_or_functions)
     else
       quote do
-        ExUnit.Callbacks.__setup__(__MODULE__, unquote(block))
+        ExUnit.Callbacks.__setup__(__MODULE__, unquote(block_or_functions))
       end
     end
   end
@@ -218,11 +227,10 @@ defmodule ExUnit.Callbacks do
   @doc """
   Defines a callback to be run before each test in a case.
 
-  Accepts a block or the name of a one-arity function in the form of an atom,
-  or a list of such atoms.
+  This is similar to `setup/1`, but the first argument is the context.
+  The `block` argument can only be a block.
 
-  Can return values to be merged into the `context`, to set up the state for
-  tests. For more details, see the "Context" section shown above.
+  For more details, see the "Context" section shown above.
 
   ## Examples
 
@@ -232,6 +240,11 @@ defmodule ExUnit.Callbacks do
 
   """
   defmacro setup(context, block) do
+    unless Keyword.keyword?(block) and Keyword.has_key?(block, :do) do
+      raise ArgumentError,
+            "setup/2 requires a block as the second argument after the context, got: #{Macro.to_string(block)}"
+    end
+
     do_setup(context, block)
   end
 
@@ -245,7 +258,12 @@ defmodule ExUnit.Callbacks do
   @doc false
   def __setup__(module, callbacks) do
     setup = Module.get_attribute(module, :ex_unit_setup)
-    Module.put_attribute(module, :ex_unit_setup, Enum.reverse(callbacks(callbacks), setup))
+
+    Module.put_attribute(
+      module,
+      :ex_unit_setup,
+      Enum.reverse(validate_callbacks!(callbacks), setup)
+    )
   end
 
   @doc false
@@ -265,8 +283,12 @@ defmodule ExUnit.Callbacks do
   @doc """
   Defines a callback to be run before all tests in a case.
 
-  Accepts a block or the name of a one-arity function in the form of an atom,
-  or a list of such atoms.
+  Accepts one of these:
+
+    * a block
+    * an atom naming a local function
+    * a `{module, function}` tuple
+    * a list of atoms and `{module, function}` tuples
 
   Can return values to be merged into the `context`, to set up the state for
   tests. For more details, see the "Context" section shown above.
@@ -288,12 +310,18 @@ defmodule ExUnit.Callbacks do
       # One-arity function name
       setup_all :clean_up_tmp_directory
 
-      def clean_up_tmp_directory(_context) do
+      # A module and function
+      setup_all {MyModule, :my_setup_function}
+
+      # A list of one-arity functions and module/function tuples
+      setup_all [:clean_up_tmp_directory, {MyModule, :my_setup_function}]
+
+      defp clean_up_tmp_directory(_context) do
         # perform setup
         :ok
       end
 
-      # Block
+      # A block
       setup_all do
         [conn: Plug.Conn.build_conn()]
       end
@@ -336,8 +364,8 @@ defmodule ExUnit.Callbacks do
   @doc """
   Defines a callback to be run before all tests in a case.
 
-  Same as `setup_all/1` but also takes a context. See
-  the "Context" section in the module documentation.
+  Similar as `setup_all/1` but also takes a context. The second argument
+  must be a block. See the "Context" section in the module documentation.
 
   ## Examples
 
@@ -347,6 +375,11 @@ defmodule ExUnit.Callbacks do
 
   """
   defmacro setup_all(context, block) do
+    unless Keyword.keyword?(block) and Keyword.has_key?(block, :do) do
+      raise ArgumentError,
+            "setup_all/2 requires a block as the second argument after the context, got: #{Macro.to_string(block)}"
+    end
+
     do_setup_all(context, block)
   end
 
@@ -365,7 +398,7 @@ defmodule ExUnit.Callbacks do
     Module.put_attribute(
       module,
       :ex_unit_setup_all,
-      Enum.reverse(callbacks(callbacks), setup_all)
+      Enum.reverse(validate_callbacks!(callbacks), setup_all)
     )
   end
 
@@ -385,15 +418,21 @@ defmodule ExUnit.Callbacks do
     end
   end
 
-  defp callbacks(callbacks) do
+  defp validate_callbacks!(callbacks) do
     for k <- List.wrap(callbacks) do
-      if not is_atom(k) do
-        raise ArgumentError,
-              "setup/setup_all expect a callback name as an atom or " <>
-                "a list of callback names, got: #{inspect(k)}"
-      end
+      case k do
+        {mod, fun} when is_atom(mod) and is_atom(fun) ->
+          {mod, fun}
 
-      k
+        name when is_atom(name) ->
+          name
+
+        invalid ->
+          raise ArgumentError,
+                "setup/setup_all expect a callback as an atom, " <>
+                  "a {module, function} tuple or a list of callbacks, got: " <>
+                  inspect(invalid)
+      end
     end
   end
 
@@ -448,7 +487,7 @@ defmodule ExUnit.Callbacks do
   indirection. However, it can be useful in some cases or for library
   authors, for example.
   """
-  @spec on_exit(term, (() -> term)) :: :ok
+  @spec on_exit(term, (-> term)) :: :ok
   def on_exit(name_or_ref \\ make_ref(), callback) when is_function(callback, 0) do
     case ExUnit.OnExitHandler.add(self(), name_or_ref, callback) do
       :ok ->
@@ -510,14 +549,7 @@ defmodule ExUnit.Callbacks do
       end
 
     child_spec = Supervisor.child_spec(child_spec_or_module, opts)
-
-    case Supervisor.start_child(sup, child_spec) do
-      {:error, {:already_started, _pid}} ->
-        {:error, {:duplicate_child_name, child_spec.id}}
-
-      other ->
-        other
-    end
+    Supervisor.start_child(sup, child_spec)
   end
 
   @doc """
@@ -646,51 +678,51 @@ defmodule ExUnit.Callbacks do
   end
 
   @doc false
-  def __merge__(_mod, context, :ok) do
+  def __merge__(_mod, _kind, context, :ok) do
     context
   end
 
-  def __merge__(mod, context, {:ok, value}) do
-    unwrapped_merge(mod, context, value, {:ok, value})
+  def __merge__(mod, kind, context, {:ok, value}) do
+    unwrapped_merge(mod, kind, context, value, {:ok, value})
   end
 
-  def __merge__(mod, context, value) do
-    unwrapped_merge(mod, context, value, value)
+  def __merge__(mod, kind, context, value) do
+    unwrapped_merge(mod, kind, context, value, value)
   end
 
-  defp unwrapped_merge(mod, _context, %_{}, original_value) do
-    raise_merge_failed!(mod, original_value)
+  defp unwrapped_merge(mod, kind, _context, %_{}, original_value) do
+    raise_merge_failed!(mod, kind, original_value)
   end
 
-  defp unwrapped_merge(mod, context, data, _original_value) when is_list(data) do
-    context_merge(mod, context, Map.new(data))
+  defp unwrapped_merge(mod, kind, context, data, _original_value) when is_list(data) do
+    context_merge(mod, kind, context, Map.new(data))
   end
 
-  defp unwrapped_merge(mod, context, data, _original_value) when is_map(data) do
-    context_merge(mod, context, data)
+  defp unwrapped_merge(mod, kind, context, data, _original_value) when is_map(data) do
+    context_merge(mod, kind, context, data)
   end
 
-  defp unwrapped_merge(mod, _, _return_value, original_value) do
-    raise_merge_failed!(mod, original_value)
+  defp unwrapped_merge(mod, kind, _, _return_value, original_value) do
+    raise_merge_failed!(mod, kind, original_value)
   end
 
-  defp context_merge(mod, context, data) do
+  defp context_merge(mod, kind, context, data) do
     Map.merge(context, data, fn
       k, v1, v2 when k in @reserved ->
-        if v1 == v2, do: v1, else: raise_merge_reserved!(mod, k, v2)
+        if v1 == v2, do: v1, else: raise_merge_reserved!(mod, kind, k, v2)
 
       _, _, v ->
         v
     end)
   end
 
-  defp raise_merge_failed!(mod, return_value) do
-    raise "expected ExUnit callback in #{inspect(mod)} to return :ok | keyword | map, " <>
-            "got #{inspect(return_value)} instead"
+  defp raise_merge_failed!(mod, kind, return_value) do
+    raise "expected ExUnit #{kind} callback in #{inspect(mod)} to " <>
+            "return :ok | keyword | map, got #{inspect(return_value)} instead"
   end
 
-  defp raise_merge_reserved!(mod, key, value) do
-    raise "ExUnit callback in #{inspect(mod)} is trying to set " <>
+  defp raise_merge_reserved!(mod, kind, key, value) do
+    raise "ExUnit #{kind} callback in #{inspect(mod)} is trying to set " <>
             "reserved field #{inspect(key)} to #{inspect(value)}"
   end
 
@@ -744,8 +776,11 @@ defmodule ExUnit.Callbacks do
   def __describe__(module, message, used_describes) do
     {name, body} =
       case Module.get_attribute(module, :ex_unit_setup) do
-        [] -> {nil, nil}
-        callbacks -> {:"__ex_unit_describe_#{map_size(used_describes)}", compile_setup(callbacks)}
+        [] ->
+          {nil, nil}
+
+        callbacks ->
+          {:"__ex_unit_describe_#{map_size(used_describes)}", compile_setup(callbacks, :setup)}
       end
 
     used_describes = Map.put(used_describes, message, name)
@@ -757,7 +792,7 @@ defmodule ExUnit.Callbacks do
     calls =
       env.module
       |> Module.get_attribute(:"ex_unit_#{kind}")
-      |> compile_setup()
+      |> compile_setup(kind)
 
     describe_clauses =
       for {describe, callback} <- describes,
@@ -786,24 +821,53 @@ defmodule ExUnit.Callbacks do
     end
   end
 
-  defp compile_setup([]) do
+  defp compile_setup([], _kind) do
     quote(do: var!(context))
   end
 
-  defp compile_setup(callbacks) do
+  defp compile_setup(callbacks, kind) do
     [h | t] = Enum.reverse(callbacks)
 
-    Enum.reduce(t, compile_setup_call(h), fn callback, acc ->
+    Enum.reduce(t, compile_setup_call(h, kind), fn callback, acc ->
       quote do
         var!(context) = unquote(acc)
-        unquote(compile_setup_call(callback))
+        unquote(compile_setup_call(callback, kind))
       end
     end)
   end
 
-  defp compile_setup_call(callback) do
+  defp compile_setup_call(callback, kind) when is_atom(callback) do
     quote do
-      unquote(__MODULE__).__merge__(__MODULE__, var!(context), unquote(callback)(var!(context)))
+      result =
+        unquote(__MODULE__).__merge__(
+          __MODULE__,
+          unquote(kind),
+          var!(context),
+          unquote(callback)(var!(context))
+        )
+
+      ExUnit.Callbacks.__noop__()
+      result
     end
   end
+
+  defp compile_setup_call({module, callback}, kind) do
+    quote do
+      result =
+        unquote(__MODULE__).__merge__(
+          __MODULE__,
+          unquote(kind),
+          var!(context),
+          unquote(module).unquote(callback)(var!(context))
+        )
+
+      ExUnit.Callbacks.__noop__()
+      result
+    end
+  end
+
+  # We inject this into compiled setup/setup_all calls, to avoid tail-call optimization and get
+  # better stacktraces. See the comments in https://github.com/elixir-lang/elixir/pull/12382.
+  @doc false
+  def __noop__, do: :noop
 end
