@@ -353,8 +353,8 @@ defmodule ExUnit.CLIFormatter do
 
   # Color styles
 
-  defp colorize(escape, string, %{colors: colors}) do
-    if colors[:enabled] do
+  defp colorize(key, string, %{colors: colors}) do
+    if escape = colors[:enabled] && colors[key] do
       [escape, string, :reset]
       |> IO.ANSI.format_fragment(true)
       |> IO.iodata_to_binary()
@@ -372,48 +372,47 @@ defmodule ExUnit.CLIFormatter do
   end
 
   defp success(msg, config) do
-    colorize(:green, msg, config)
+    colorize(:success, msg, config)
   end
 
   defp invalid(msg, config) do
-    colorize(:yellow, msg, config)
+    colorize(:invalid, msg, config)
   end
 
   defp skipped(msg, config) do
-    colorize(:yellow, msg, config)
+    colorize(:skipped, msg, config)
   end
 
   defp failure(msg, config) do
-    colorize(:red, msg, config)
+    colorize(:failure, msg, config)
   end
 
-  defp formatter(:diff_enabled?, _, %{colors: colors}), do: colors[:enabled]
+  # Diff formatting
 
-  defp formatter(:error_info, msg, config), do: colorize(:red, msg, config)
+  defp formatter(:diff_enabled?, _, %{colors: colors}),
+    do: colors[:enabled]
 
-  defp formatter(:extra_info, msg, config), do: colorize(:cyan, msg, config)
-
-  defp formatter(:location_info, msg, config), do: colorize([:bright, :black], msg, config)
-
-  defp formatter(:diff_delete, doc, config), do: colorize_doc(:diff_delete, doc, config)
+  defp formatter(:diff_delete, doc, config),
+    do: colorize_doc(:diff_delete, doc, config)
 
   defp formatter(:diff_delete_whitespace, doc, config),
     do: colorize_doc(:diff_delete_whitespace, doc, config)
 
-  defp formatter(:diff_insert, doc, config), do: colorize_doc(:diff_insert, doc, config)
+  defp formatter(:diff_insert, doc, config),
+    do: colorize_doc(:diff_insert, doc, config)
 
   defp formatter(:diff_insert_whitespace, doc, config),
     do: colorize_doc(:diff_insert_whitespace, doc, config)
 
   defp formatter(:blame_diff, msg, %{colors: colors} = config) do
     if colors[:enabled] do
-      colorize(:red, msg, config)
+      colorize(:diff_delete, msg, config)
     else
       "-" <> msg <> "-"
     end
   end
 
-  defp formatter(_, msg, _config), do: msg
+  defp formatter(key, msg, config), do: colorize(key, msg, config)
 
   defp pluralize(1, singular, _plural), do: singular
   defp pluralize(_, _singular, plural), do: plural
@@ -429,7 +428,16 @@ defmodule ExUnit.CLIFormatter do
     diff_delete: :red,
     diff_delete_whitespace: IO.ANSI.color_background(2, 0, 0),
     diff_insert: :green,
-    diff_insert_whitespace: IO.ANSI.color_background(0, 2, 0)
+    diff_insert_whitespace: IO.ANSI.color_background(0, 2, 0),
+
+    # CLI formatter
+    success: :green,
+    invalid: :yellow,
+    skipped: :yellow,
+    failure: :red,
+    error_info: :red,
+    extra_info: :cyan,
+    location_info: [:bright, :black]
   ]
 
   defp colors(opts) do
