@@ -28,9 +28,12 @@ defmodule ExUnit.FormatterTest do
     false
   end
 
-  defp formatter(_kind, message) do
-    message
-  end
+  defp formatter(_key, value), do: value
+
+  defp diff_formatter(:diff_enabled?, _default), do: true
+  defp diff_formatter(_key, value), do: value
+
+  defp kw_to_string(kw), do: for({k, v} <- kw, do: {k, IO.iodata_to_binary(v)})
 
   test "formats test case filters" do
     filters = [run: true, slow: false]
@@ -153,8 +156,9 @@ defmodule ExUnit.FormatterTest do
       end
 
     failure = [{{:EXIT, self()}, {error, stack}, []}]
+    format = trim_multiline_whitespace(format_test_failure(test(), failure, 1, 80, &formatter/2))
 
-    assert trim_multiline_whitespace(format_test_failure(test(), failure, 1, 80, &formatter/2)) =~
+    assert format =~
              """
                1) world (Hello)
                   test/ex_unit/formatter_test.exs:1
@@ -174,6 +178,8 @@ defmodule ExUnit.FormatterTest do
 
                            def fetch(%module{} = container, key)
              """
+
+    assert format =~ ~r"lib/access.ex:\d+: Access.fetch/2"
   end
 
   test "formats test EXITs with assertion errors" do
@@ -238,7 +244,8 @@ defmodule ExUnit.FormatterTest do
   end
 
   test "formats assertions" do
-    failure = [{:error, catch_assertion(assert ExUnit.FormatterTest.falsy()), []}]
+    assertion_error = catch_assertion(assert ExUnit.FormatterTest.falsy())
+    failure = [{:error, assertion_error, []}]
 
     assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
              1) world (Hello)
@@ -246,6 +253,89 @@ defmodule ExUnit.FormatterTest do
                 Expected truthy, got false
                 code: assert ExUnit.FormatterTest.falsy()
            """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &formatter/2) == []
+  end
+
+  test "formats assertions with patterns and values" do
+    assertion_error = catch_assertion(assert {1, 2, 3} > {1, 2, 3})
+    failure = [{:error, assertion_error, []}]
+
+    assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
+             1) world (Hello)
+                test/ex_unit/formatter_test.exs:1
+                Assertion with > failed, both sides are exactly equal
+                code: assert {1, 2, 3} > {1, 2, 3}
+                left: {1, 2, 3}
+           """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &diff_formatter/2)
+           |> kw_to_string() ==
+             [left: "{1, 2, 3}"]
+
+    assertion_error = catch_assertion(assert {3, 2, 1} = {1, 2, 3})
+    failure = [{:error, assertion_error, []}]
+
+    assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
+             1) world (Hello)
+                test/ex_unit/formatter_test.exs:1
+                match (=) failed
+                code:  assert {3, 2, 1} = {1, 2, 3}
+                left:  {3, 2, 1}
+                right: {1, 2, 3}
+           """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &diff_formatter/2)
+           |> kw_to_string() ==
+             [{:left, "{3, 2, 1}"}, {:right, "{1, 2, 3}"}]
+  end
+
+  nfc_hello = String.normalize("héllo", :nfc)
+  nfd_hello = String.normalize("héllo", :nfd)
+
+  test "formats assertions with hints" do
+    assertion_error = catch_assertion(assert unquote(nfc_hello) == unquote(nfd_hello))
+    failure = [{:error, assertion_error, []}]
+
+    assert format_test_failure(test(), failure, 1, 80, &diff_formatter/2) =~ """
+             1) world (Hello)
+                test/ex_unit/formatter_test.exs:1
+                Assertion with == failed
+                code:  assert "#{unquote(nfc_hello)}" == "#{unquote(nfd_hello)}"
+                left:  "#{unquote(nfc_hello)}"
+                right: "#{unquote(nfd_hello)}"
+                hint:  you are comparing strings that have the same visual representation but are made of different Unicode codepoints
+           """
+
+    assert format_assertion_diff(assertion_error, 0, :infinity, &diff_formatter/2)
+           |> kw_to_string() ==
+             [
+               left: inspect(unquote(nfc_hello)),
+               right: inspect(unquote(nfd_hello)),
+               hint:
+                 "you are comparing strings that have the same visual representation but are made of different Unicode codepoints"
+             ]
+  end
+
+  test "formats match error between pinned struct type and a non-struct" do
+    failure = [
+      {:error,
+       catch_assertion do
+         expected_module = ExUnit.TestModule
+         assert %^expected_module{} = nil
+       end, []}
+    ]
+
+    assert format_test_failure(test(), failure, 1, 80, &diff_formatter/2) =~ """
+             1) world (Hello)
+                test/ex_unit/formatter_test.exs:1
+                match (=) failed
+                The following variables were pinned:
+                  expected_module = ExUnit.TestModule
+                code:  assert %^expected_module{} = nil
+                left:  %^expected_module{}
+                right: nil
+           """
   end
 
   test "formats multiple assertions" do
@@ -345,7 +435,9 @@ defmodule ExUnit.FormatterTest do
              1) Hello: failure on setup_all callback, all tests have been invalidated
                 Assertion with == failed
                 code:  assert [1, 2, 3] == [4, 5, 6]
-                left:  [1, 2, 3]
+                left:  [1,
+                        2,
+                        3]
                 right: [4,
                         5,
                         6]
@@ -399,21 +491,25 @@ defmodule ExUnit.FormatterTest do
   test "inspect failure" do
     failure = [{:error, catch_assertion(assert :will_fail == %BadInspect{}), []}]
 
-    message =
-      "got FunctionClauseError with message \"no function clause matching " <>
-        "in Inspect.ExUnit.FormatterTest.BadInspect.inspect/2\" while inspecting " <>
-        "%{__struct__: ExUnit.FormatterTest.BadInspect, key: 0}"
-
-    assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
+    assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ ~s'''
              1) world (Hello)
                 test/ex_unit/formatter_test.exs:1
                 Assertion with == failed
                 code:  assert :will_fail == %BadInspect{}
                 left:  :will_fail
-                right: %Inspect.Error{
-                         message: #{inspect(message)}
-                       }
-           """
+                right: #Inspect.Error<
+             got FunctionClauseError with message:
+
+                 """
+                 no function clause matching in Inspect.ExUnit.FormatterTest.BadInspect.inspect/2
+                 """
+
+             while inspecting:
+
+                 #{inspect(%BadInspect{}, structs: false)}
+
+             Stacktrace:
+           '''
   end
 
   defmodule BadMessage do
@@ -430,7 +526,7 @@ defmodule ExUnit.FormatterTest do
 
     message =
       "got RuntimeError with message \"oops\" while retrieving Exception.message/1 " <>
-        "for %ExUnit.FormatterTest.BadMessage{key: 0}"
+        "for %ExUnit.FormatterTest.BadMessage{key: 0}. Stacktrace:"
 
     assert format_test_failure(test(), failure, 1, 80, &formatter/2) =~ """
              1) world (Hello)
