@@ -163,6 +163,24 @@ defmodule ExUnitTest do
            end) =~ "\n1 test, 1 failure\n"
   end
 
+  test "reports capture log crashes" do
+    defmodule CaptureLogTest do
+      use ExUnit.Case
+
+      test "capture log crash because logger stopped" do
+        Logger.App.stop()
+        Logger.App.start()
+        flunk("this should fail")
+      end
+    end
+
+    configure_and_reload_on_exit([])
+
+    assert capture_io(fn ->
+             assert ExUnit.run() == %{failures: 1, skipped: 0, total: 1, excluded: 0}
+           end) =~ "\n1 test, 1 failure\n"
+  end
+
   test "supports timeouts" do
     defmodule TimeoutTest do
       use ExUnit.Case
@@ -298,14 +316,8 @@ defmodule ExUnitTest do
   test "log capturing" do
     defmodule LogCapturingTest do
       use ExUnit.Case
-
       require Logger
 
-      setup_all do
-        :ok = Logger.remove_backend(:console)
-        on_exit(fn -> Logger.add_backend(:console, flush: true) end)
-      end
-
       @tag :capture_log
       test "one" do
         Logger.debug("one")
@@ -323,18 +335,12 @@ defmodule ExUnitTest do
         Logger.debug("three")
         assert 1 == 2
       end
-
-      test "four" do
-        Logger.debug("four")
-        assert 1 == 2
-      end
     end
 
     output = capture_io(&ExUnit.run/0)
-    assert output =~ "[debug] two\n"
     refute output =~ "[debug] one\n"
+    assert output =~ "[debug] two\n"
     assert output =~ "[debug] three\n"
-    refute output =~ "[debug] four\n"
   end
 
   test "supports multi errors" do
