@@ -19,24 +19,34 @@ defmodule IEx do
 
       Enum.
 
-  A module may export functions that are not meant to be used directly: these
-  functions won't be autocompleted by IEx. IEx will not autocomplete functions
-  annotated with `@doc false`, `@impl true`, or functions that aren't explicitly
-  documented and where the function name is in the form of `__foo__`.
+  A module may export functions that are not meant to be used directly:
+  these functions won't be autocompleted by IEx. IEx will not autocomplete
+  functions annotated with `@doc false`, `@impl true`, or functions that
+  aren't explicitly documented and where the function name is in the form
+  of `__foo__`.
 
-  Autocomplete may not be available on some Windows shells. You may need
-  to pass the `--werl` option when starting IEx, as in `iex --werl` for it
-  to work. `--werl` may be permanently enabled by setting the `IEX_WITH_WERL`
-  environment variable.
+  Autocomplete is available by default on Windows shells from Erlang/OTP 26.
+  In earlier versions, you may need to pass the `--werl` option when starting
+  IEx, such as `iex --werl` (or `iex.bat --werl` if using PowerShell).
+  `--werl` may be permanently enabled by setting the `IEX_WITH_WERL`
+  environment variable to `1`.
 
-  ## Coloring
+  ## Encoding and coloring
 
-  Coloring is enabled by default on most Unix terminals. They are also
-  available on Windows consoles from Windows 10, although it must be
-  explicitly enabled for the current user in the registry by running
+  IEx expects inputs and outputs to be in UTF-8 encoding. This is the
+  default for most Unix terminals but it may not be the case on Windows.
+  If you are running on Windows and you see incorrect values printed,
+  you may need change the encoding of your current session by running
+  `chcp 65001` before calling `iex` (or before calling `iex.bat` if using
+  PowerShell).
+
+  Similarly, ANSI coloring is enabled by default on most Unix terminals.
+  They are also available on Windows consoles from Windows 10 and on
+  Erlang/OTP 26 or later. For earlier Erlang/OTP versions, you can
+  explicitly enable it for the current user in the registry by running
   the following command:
 
-      reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
+      $ reg add HKCU\Console /v VirtualTerminalLevel /t REG_DWORD /d 1
 
   After running the command above, you must restart your current console.
 
@@ -45,7 +55,7 @@ defmodule IEx do
   It is possible to get shell history by passing some options that enable it
   in the VM. This can be done on a per-need basis when starting IEx:
 
-      iex --erl "-kernel shell_history enabled"
+      $ iex --erl "-kernel shell_history enabled"
 
   If you would rather enable it on your system as a whole, you can use
   the `ERL_AFLAGS` environment variable and make sure that it is set
@@ -53,15 +63,15 @@ defmodule IEx do
 
   On Unix-like / Bash:
 
-      export ERL_AFLAGS="-kernel shell_history enabled"
+      $ export ERL_AFLAGS="-kernel shell_history enabled"
 
   On Windows:
 
-      set ERL_AFLAGS "-kernel shell_history enabled"
+      $ set ERL_AFLAGS "-kernel shell_history enabled"
 
   On Windows 10 / PowerShell:
 
-      $env:ERL_AFLAGS = "-kernel shell_history enabled"
+      $ $env:ERL_AFLAGS = "-kernel shell_history enabled"
 
   ## Expressions in IEx
 
@@ -98,42 +108,53 @@ defmodule IEx do
 
   ## Pasting multiline expressions into IEx
 
-  IEx evaluates its input line by line in an eagerly fashion which means
-  that if at the end of a line the code seen so far is a complete expression
-  IEx will evaluate it at that point. This behaviour may produce errors for
-  expressions that have been formatted across multiple lines which is often
-  the case for piped expressions. Consider the following expression using
-  the `|>/2` operator:
+  IEx evaluates its input line by line in an eager fashion. If at the end of a
+  line the code seen so far is a complete expression, IEx will evaluate it at
+  that point.
 
-      iex(1)> [1, [2], 3] |> List.flatten()
-      [1, 2, 3]
+      iex(1)> [1, [2], 3]
+      [1, [2], 3]
 
-  When written in multiline form and pasted into IEx this valid expression
-  produces a syntax error:
+  To prevent this behaviour breaking valid code where the subsequent line
+  begins with a binary operator, such as `|>/2` or `++/2` , IEx automatically
+  treats such lines as if they were prepended with `IEx.Helpers.v/0`, which
+  returns the value of the previous expression, if available.
 
       iex(1)> [1, [2], 3]
       [1, [2], 3]
       iex(2)> |> List.flatten()
-      ** (SyntaxError) iex:2: syntax error before: '|>'
-
-  As IEx evaluates its input line by line, it will first encounter
-  `[1, [2], 3]`. As a list is a valid expression, IEx will evaluate
-  it immediately before looking at the next input line. Only then
-  will IEx attempt to evaluate the now incomplete expression
-  `|> List.flatten()`, which on its own is missing its left operand.
-  The evaluation thus fails with the above syntax error.
-
-  In order to help IEx understand that an expression consists of multiple
-  lines we can wrap it into parentheses:
-
-      iex(1)> (
-      ...(1)> [1, [2], 3]
-      ...(1)> |> List.flatten()
-      ...(1)> )
       [1, 2, 3]
 
-  Note that this not only works with single expressions but also with
-  arbitrary code blocks.
+  The above is equivalent to:
+
+      iex(1)> [1, [2], 3]
+      [1, [2], 3]
+      iex(2)> v() |> List.flatten()
+      [1, 2, 3]
+
+  If there are no previous expressions in the history, the pipe operator will
+  fail:
+
+      iex(1)> |> List.flatten()
+      ** (RuntimeError) v(-1) is out of bounds
+
+  If the previous expression was a match operation, the pipe operator will also
+  fail, to prevent an unsolicited break of the match:
+
+      iex(1)> x = 42
+      iex(2)> |> IO.puts()
+      ** (SyntaxError) iex:2:1: pipe shorthand is not allowed immediately after a match expression in IEx. To make it work, surround the whole pipeline with parentheses ('|>')
+          |
+        2 | |> IO.puts()
+          | ^
+
+  Note, however, the above does not work for `+/2` and `-/2`, as they
+  are ambiguous with the unary `+/1` and `-/1`:
+
+      iex(1)> 1
+      1
+      iex(2)> + 2
+      2
 
   ## The BREAK menu
 
@@ -151,27 +172,49 @@ defmodule IEx do
 
   If you are connected to remote shell, it remains alive after disconnection.
 
-  ## Prying and breakpoints
+  ## `dbg` and breakpoints
 
-  IEx also has the ability to set breakpoints on Elixir code and
-  "pry" into running processes. This allows the developer to have
-  an IEx session run inside a given function.
+  IEx integrates with `Kernel.dbg/2` and introduces a backend that
+  can pause code execution. To enable it, you must pass `--dbg pry`:
 
-  `IEx.pry/0` can be used when you are able to modify the source
-  code directly and recompile it:
+      $ iex --dbg pry
+
+  For example, take the following function:
 
       def my_fun(arg1, arg2) do
-        require IEx; IEx.pry()
+        dbg(arg1 + arg2)
         ... implementation ...
       end
 
-  When the code is executed, it will ask you for permission to be
-  introspected.
+  When the code is executed with `iex` (most often by calling
+  `iex --dbg pry -S mix`), it will ask you permission to use "pry".
+  If you agree, it will start an IEx shell in the context of the function
+  above, with access to its variables, imports, and aliases. However,
+  you can only access existing values, it is not possible to access
+  private functions nor change the execution itself (hence the name
+  "pry").
 
-  Alternatively, you can use `IEx.break!/4` to setup a breakpoint
-  on a given module, function and arity you have no control of.
-  While `IEx.break!/4` is more flexible, it does not contain
-  information about imports and aliases from the source code.
+  When using `|> dbg()` at the end of a pipeline, you can pry each
+  step of the pipeline. You can type `n` whenever you want to jump
+  into the next pipe. Type `continue` when you want to execute all
+  of the steps but stay within the pried process. Type `respawn` when
+  you want to leave the pried process and start a new shell.
+
+  Alternatively, you can start a pry session directly, without `dbg/2`
+  by calling `IEx.pry/0`.
+
+  IEx also allows you to set breakpoints to start pry sessions
+  on a given module, function, and arity you have no control of
+  via `IEx.break!/4`. Similar to pipelines in `dbg()`, `IEx.break!/4`
+  allows you to debug a function line by line and access its variables.
+  However, breakpoints do not contain information about imports and
+  aliases from the source code.
+
+  When using `dbg` or breakpoints with tests, remember to pass the
+  `--trace` to `mix test` to avoid running into timeouts:
+
+      $ iex -S mix test --trace
+      $ iex -S mix test path/to/file:line --trace
 
   ## The User switch command
 
@@ -222,7 +265,7 @@ defmodule IEx do
   First of all, we can only connect to a shell if we give names
   both to the current shell and the shell we want to connect to.
 
-  Let's give it a try. First start a new shell:
+  Let's give it a try. First, start a new shell:
 
       $ iex --sname foo
       iex(foo@HOST)1>
@@ -247,7 +290,7 @@ defmodule IEx do
       $ iex --sname bar
       iex(bar@HOST)1>
 
-  If we try to dispatch to `Hello.world`, it won't be available
+  If we try to dispatch to `Hello.world/0`, it won't be available
   as it was defined only in the other shell:
 
       iex(bar@HOST)1> Hello.world()
@@ -283,15 +326,14 @@ defmodule IEx do
   ## The .iex.exs file
 
   When starting, IEx looks for a local `.iex.exs` file (located in the current
-  working directory), then a global one (located at `~/.iex.exs`) and loads the
-  first one it finds (if any). Note the location of the `.iex.exs` files, both
-  in the current directory and the global one, are taken relative to the user
-  that started the application, not to the user that is connecting to the node in
-  case of remote IEx connections.
+  working directory), then for a global `.iex.exs` file located inside the
+  directory pointed by the `IEX_HOME` environment variable (which defaults
+  to `~`) and loads the first one it finds (if any).
 
-  The code in the loaded `.iex.exs` file is evaluated in the shell's context.
-  For instance, any modules that are loaded or variables that are bound in the
-  `.iex.exs` file will be available in the shell after it has booted.
+  The code in the chosen `.iex.exs` file is evaluated line by line in the shell's
+  context, as if each line were being typed in the shell. For instance, any modules
+  that are loaded or variables that are bound in the `.iex.exs` file will be available
+  in the shell after it has booted.
 
   Take the following `.iex.exs` file:
 
@@ -311,15 +353,19 @@ defmodule IEx do
   results in:
 
       $ iex
-      Erlang/OTP 21 [...]
+      Erlang/OTP 24 [...]
 
       hello world
       Interactive Elixir - press Ctrl+C to exit (type h() ENTER for help)
       iex(1)> value
       13
 
-  It is possible to load another file by supplying the `--dot-iex`
-  option to IEx. See `iex --help`.
+  It is possible to load another file by supplying the `--dot-iex` option
+  to IEx. See `iex --help`.
+
+  In case of remote nodes, the location of the `.iex.exs` files are taken
+  relative to the user that started the application, not to the user that
+  is connecting to the node in case of remote IEx connections.
 
   ## Configuring the shell
 
@@ -335,7 +381,7 @@ defmodule IEx do
   Now run the shell:
 
       $ iex
-      Erlang/OTP 21 [...]
+      Erlang/OTP 24 [...]
 
       Interactive Elixir - press Ctrl+C to exit (type h() ENTER for help)
       iex(1)> [1, 2, 3, 4, 5]
@@ -356,6 +402,7 @@ defmodule IEx do
     * `:continuation_prompt`
     * `:alive_prompt`
     * `:alive_continuation_prompt`
+    * `:parser`
 
   They are discussed individually in the sections below.
 
@@ -393,21 +440,18 @@ defmodule IEx do
 
       IEx.configure(colors: [syntax_colors: false])
 
-  You can also configure the syntax colors, however, as desired:
+  You can also configure the syntax colors, however, as desired.
+  The below will format atoms in red and remove the coloring for
+  all other data types:
 
       IEx.configure(colors: [syntax_colors: [atom: :red]])
 
-  Configuration for most built-in data types are supported: `:atom`,
-  `:string`, `:binary`, `:list`, `:number`, `:boolean`, `:nil`, and others.
-  The default is:
-
-      [number: :magenta, atom: :cyan, string: :green,
-       boolean: :magenta, nil: :magenta]
+  The default values can be found in `IO.ANSI.syntax_colors/0`.
 
   ## Inspect
 
   A keyword list containing inspect options used by the shell
-  when printing results of expression evaluation. Default to
+  when printing results of expression evaluation. Defaults to
   pretty formatting with a limit of 50 entries.
 
   To show all entries, configure the limit to `:infinity`:
@@ -452,6 +496,18 @@ defmodule IEx do
     * `%prefix`  - a prefix given by `IEx.Server`
     * `%node`    - the name of the local node
 
+  ## Parser
+
+  This is an option determining the parser to use for IEx.
+
+  The parser is a "mfargs", which is a tuple with three elements:
+  the module name, the function name, and extra arguments to
+  be appended. The parser receives at least three arguments, the
+  current input as a string, the parsing options as a keyword list,
+  and the buffer as a string. It must return `{:ok, expr, buffer}`
+  or `{:incomplete, buffer}`.
+
+  If the parser raises, the buffer is reset to an empty string.
   """
   @spec configure(keyword()) :: :ok
   def configure(options) do
@@ -466,24 +522,23 @@ defmodule IEx do
     IEx.Config.configuration()
   end
 
-  @doc """
-  Registers a function to be invoked after the IEx process is spawned.
-  """
-  @spec after_spawn(fun()) :: :ok
+  @doc false
+  @deprecated "The functionality is no longer supported"
   def after_spawn(fun) when is_function(fun) do
     IEx.Config.after_spawn(fun)
   end
 
-  @doc """
-  Returns registered `after_spawn` callbacks.
-  """
-  @spec after_spawn() :: [fun()]
+  @doc false
+  @deprecated "The functionality is no longer supported"
   def after_spawn do
     IEx.Config.after_spawn()
   end
 
   @doc """
   Returns `true` if IEx was started, `false` otherwise.
+
+  This means the IEx application was started, but not
+  that its CLI interface is running.
   """
   @spec started?() :: boolean()
   def started? do
@@ -527,7 +582,7 @@ defmodule IEx do
   @doc """
   Pries into the process environment.
 
-  This is useful for debugging a particular chunk of code
+  This function is useful for debugging a particular chunk of code
   when executed by a particular process. The process becomes
   the evaluator of IEx commands and is temporarily changed to
   have a custom group leader. Those values are reverted by
@@ -536,14 +591,16 @@ defmodule IEx do
 
   When a process is pried, all code runs inside IEx and has
   access to all imports and aliases from the original code.
-  However, the code is evaluated and therefore cannot access
-  private functions of the module being pried. Module functions
-  still need to be accessed via `Mod.fun(args)`.
+  However, you cannot change the execution of the code nor
+  access private functions of the module being pried. Module
+  functions still need to be accessed via `Mod.fun(args)`.
 
-  Alternatively, you can use `IEx.break!/4` to setup a breakpoint
-  on a given module, function and arity you have no control of.
-  While `IEx.break!/4` is more flexible,  it does not contain
-  information about imports and aliases from the source code.
+  See also `break!/4` for others ways to pry.
+
+  > #### `dbg/0` integration
+  >
+  > By calling `iex --dbg pry`, `iex` will set this function
+  > as the default backend for `dbg/0` calls.
 
   ## Examples
 
@@ -567,7 +624,7 @@ defmodule IEx do
   the shell will be reset and you gain access to all variables
   and the lexical scope from above:
 
-      pry(1)> map([a, b, c], &IO.inspect(&1))
+      iex(1)> map([a, b, c], &IO.inspect(&1))
       1
       2
       3
@@ -577,7 +634,7 @@ defmodule IEx do
   process can be freed by calling [`respawn/0`](`IEx.Helpers.respawn/0`), which starts a
   new IEx evaluation cycle, letting this one go:
 
-      pry(2)> respawn()
+      iex(2)> respawn()
       true
 
       Interactive Elixir - press Ctrl+C to exit (type h() ENTER for help)
@@ -611,8 +668,8 @@ defmodule IEx do
   the `iex` command and pass the `--trace` to `mix test` to avoid running
   into timeouts:
 
-      iex -S mix test --trace
-      iex -S mix test path/to/file:line --trace
+      $ iex -S mix test --trace
+      $ iex -S mix test path/to/file:line --trace
 
   """
   defmacro pry() do
@@ -680,8 +737,9 @@ defmodule IEx do
   the given number of `stops`.
 
   This function will instrument the given module and load a new
-  version in memory with breakpoints at the given function and
-  arity. If the module is recompiled, all breakpoints are lost.
+  version in memory with line by line breakpoints at the given
+  function and arity. If the module is recompiled, all breakpoints
+  are lost.
 
   When a breakpoint is reached, IEx will ask if you want to `pry`
   the given function and arity. In other words, this works similar
@@ -699,6 +757,8 @@ defmodule IEx do
     * `IEx.Helpers.break!/4` - sets up a breakpoint for the given module, function, arity
     * `IEx.Helpers.breaks/0` - prints all breakpoints and their IDs
     * `IEx.Helpers.continue/0` - continues until the next breakpoint in the same shell
+    * `IEx.Helpers.n/0` - goes to the next line of the current breakpoint
+    * `IEx.Helpers.next/0` - same as above
     * `IEx.Helpers.open/0` - opens editor on the current breakpoint
     * `IEx.Helpers.remove_breaks/0` - removes all breakpoints in all modules
     * `IEx.Helpers.remove_breaks/1` - removes all breakpoints in a given module
@@ -718,10 +778,11 @@ defmodule IEx do
   `IEx.Helpers.remove_breaks/1` and on all modules by calling
   `IEx.Helpers.remove_breaks/0`.
 
-  To exit a breakpoint, the developer can either invoke `continue()`,
-  which will block the shell until the next breakpoint is found or
-  the process terminates, or invoke `respawn()`, which starts a new IEx
-  shell, freeing up the pried one.
+  Within a breakpoint, you can call `n` to jump to the next line.
+  To exit a breakpoint, you can either invoke `continue`, which will
+  block the shell until the next breakpoint is found or the process
+  terminates, or invoke `respawn`, which starts a new IEx shell,
+  freeing up the pried one.
 
   ## Examples
 
@@ -729,23 +790,23 @@ defmodule IEx do
   a breakpoint directly from your IEx shell. But you can set up a break
   from anywhere by using the fully qualified name `IEx.break!`.
 
-  The following sets up a breakpoint on `URI.decode_query/2`:
+  The following sets up a breakpoint on `URI.parse/1`:
 
-      break! URI, :decode_query, 2
+      break! URI, :parse, 1
 
   This call will setup a breakpoint that stops once.
   To set a breakpoint that will stop 10 times:
 
-      break! URI, :decode_query, 2, 10
+      break! URI, :parse, 1, 10
 
   `IEx.break!/2` is a convenience macro that allows breakpoints
   to be given in the `Mod.fun/arity` format:
 
-      break! URI.decode_query/2
+      break! URI.parse/1
 
   Or to set a breakpoint that will stop 10 times:
 
-      break! URI.decode_query/2, 10
+      break! URI.parse/1, 10
 
   This function returns the breakpoint ID and will raise if there
   is an error setting up the breakpoint.
@@ -754,23 +815,22 @@ defmodule IEx do
 
   `IEx.break!/2` allows patterns to be given, triggering the
   breakpoint only in some occasions. For example, to trigger
-  the breakpoint only when the first argument is the "foo=bar"
-  string:
+  the breakpoint only when the first argument starts with the
+  "https" string:
 
-      break! URI.decode_query("foo=bar", _)
-
-  Or to trigger it whenever the second argument is a map with
-  more than one element:
-
-      break! URI.decode_query(_, map) when map_size(map) > 0
+      break! URI.parse("https" <> _, _)
 
   Only a single break point can be set per function. So if you call
   `IEx.break!` multiple times with different patterns, only the last
   pattern is kept.
 
-  Note that, while patterns may be given to macros, remember that
-  macros receive ASTs as arguments, and not values. For example, if
-  you try to break on a macro with the following pattern:
+  ## Macros
+
+  While it is possible to set breakpoint in macros, remember that macros
+  are generally expanded at compilation time, and therefore they may never
+  be invoked during runtime. Similarly, while patterns may be given to
+  macros, macros receive ASTs as arguments, and not values. For example,
+  if you try to break on a macro with the following pattern:
 
       break! MyModule.some_macro(pid) when pid == self()
 
@@ -785,62 +845,120 @@ defmodule IEx do
   the `iex` command and pass the `--trace` to `mix test` to avoid running
   into timeouts:
 
-      iex -S mix test --trace
-      iex -S mix test path/to/file:line --trace
+      $ iex -S mix test --trace
+      $ iex -S mix test path/to/file:line --trace
 
   """
   @doc since: "1.5.0"
   @spec break!(module, atom, arity, non_neg_integer) :: IEx.Pry.id()
   defdelegate break!(module, function, arity, stops \\ 1), to: IEx.Pry
 
-  ## Callbacks
-
-  # This is a callback invoked by Erlang shell utilities
-  # when someone presses Ctrl+G and adds 's Elixir.IEx'.
-  @doc false
-  def start(opts \\ [], mfa \\ {IEx, :dont_display_result, []}) do
-    spawn(fn ->
-      case :init.notify_when_started(self()) do
-        :started -> :ok
-        _ -> :init.wait_until_started()
-      end
-
-      :ok = start_iex()
-      :ok = set_expand_fun()
-      :ok = run_after_spawn()
-      IEx.Server.run_from_shell(opts, mfa)
-    end)
-  end
-
   @doc false
   def dont_display_result, do: :"do not show this result in output"
 
-  ## Helpers
+  ## CLI
 
-  defp start_iex() do
-    {:ok, _} = Application.ensure_all_started(:iex)
-    :ok
+  @compile {:no_warn_undefined, {:shell, :start_interactive, 1}}
+
+  # This is a callback invoked by Erlang shell utilities
+  # when someone presses Ctrl+G and adds `s 'Elixir.IEx'`.
+  # Let's consider exposing this as iex:start() and rename
+  # elixir:start_iex() to iex:cli().
+  @doc false
+  def start(opts \\ [], mfa \\ {IEx, :dont_display_result, []}) do
+    # TODO: Keep only this branch, delete optional args and mfa,
+    # and delete IEx.Server.run_from_shell/2 on Erlang/OTP 26+
+    if Code.ensure_loaded?(:prim_tty) do
+      spawn(fn ->
+        {:ok, _} = Application.ensure_all_started(:iex)
+        :ok = :io.setopts(binary: true, encoding: :unicode)
+        _ = for fun <- Enum.reverse(after_spawn()), do: fun.()
+        IEx.Server.run([register: false] ++ opts)
+      end)
+    else
+      spawn(fn ->
+        case :init.notify_when_started(self()) do
+          :started -> :ok
+          _ -> :init.wait_until_started()
+        end
+
+        {:ok, _} = Application.ensure_all_started(:iex)
+        :ok = :io.setopts(binary: true, encoding: :unicode)
+        _ = for fun <- Enum.reverse(after_spawn()), do: fun.()
+        IEx.Server.run_from_shell(opts, mfa)
+      end)
+    end
   end
 
-  defp set_expand_fun do
-    gl = Process.group_leader()
+  # Manual tests for changing the CLI boot.
+  #
+  # 1. In some situations, we cannot read inputs as IEx boots:
+  #
+  #      $ iex -e ":io.get_line(:foo)"
+  #
+  # 2. In some situations, connecting to a remote node via --remsh
+  #    is not possible. This can be tested by starting two IEx nodes:
+  #
+  #      $ iex --sname foo
+  #      $ iex --sname bar --remsh foo
+  #
+  # 3. When still using --remsh, we need to guarantee the arguments
+  #    are processed on the local node and not the remote one. For such,
+  #    one can replace the last line above by:
+  #
+  #      $ iex --sname bar --remsh foo -e 'IO.inspect node()'
+  #
+  #    And verify that the local node name is printed.
+  #
+  # 4. Finally, in some other circumstances, printing messages may become
+  #    borked. This can be verified with:
+  #
+  #      $ iex -e ":logger.info('foo~nbar', [])"
+  #
+  # By the time those instructions have been written, all tests above pass.
+  @doc false
+  def cli(opts \\ []) do
+    args = :init.get_plain_arguments()
+    opts = opts ++ [dot_iex_path: find_dot_iex(args), on_eof: :halt]
 
-    expand_fun =
-      if node(gl) != node() do
-        IEx.Autocomplete.remsh(node())
+    ref = make_ref()
+    mfa = {__MODULE__, :__cli__, [self(), ref, opts]}
+
+    shell =
+      if remote = get_remsh(args) do
+        {:remote, remote, mfa}
       else
-        &IEx.Autocomplete.expand/1
+        mfa
       end
 
-    # expand_fun is not supported by a shell variant
-    # on Windows, so we do two IO calls, not caring
-    # about the result of the expand_fun one.
-    _ = :io.setopts(gl, expand_fun: expand_fun)
-    :io.setopts(gl, binary: true, encoding: :unicode)
+    case :shell.start_interactive(shell) do
+      :ok ->
+        receive do
+          {^ref, shell} -> shell
+        after
+          15_000 ->
+            IO.puts(:stderr, "Could not start IEx CLI due to reason: :boot_timeout")
+            System.halt(1)
+        end
+
+      {:error, reason} ->
+        IO.puts(:stderr, "Could not start IEx CLI due to reason: #{inspect(reason)}")
+        System.halt(1)
+    end
   end
 
-  defp run_after_spawn do
-    _ = for fun <- Enum.reverse(after_spawn()), do: fun.()
-    :ok
+  @doc false
+  def __cli__(parent, ref, opts) do
+    pid = start(opts)
+    send(parent, {ref, pid})
+    pid
   end
+
+  defp find_dot_iex([~c"--dot-iex", h | _]), do: List.to_string(h)
+  defp find_dot_iex([_ | t]), do: find_dot_iex(t)
+  defp find_dot_iex([]), do: nil
+
+  defp get_remsh([~c"--remsh", h | _]), do: h
+  defp get_remsh([_ | t]), do: get_remsh(t)
+  defp get_remsh([]), do: nil
 end
