@@ -35,6 +35,15 @@ defmodule IEx.Introspection do
       {mod, fun, []} ->
         {mod, fun}
 
+      {maybe_sigil, [_, _]} ->
+        case Atom.to_string(maybe_sigil) do
+          "sigil_" <> _ ->
+            {:{}, [], [find_decompose_fun_arity(maybe_sigil, 2, context), maybe_sigil, 2]}
+
+          _ ->
+            call
+        end
+
       {fun, []} ->
         {find_decompose_fun(fun, context), fun}
 
@@ -284,6 +293,7 @@ defmodule IEx.Introspection do
               module.module_info(:exports)
           end
           |> Enum.sort()
+          |> Enum.dedup()
 
         result =
           for {^function, arity} <- exports,
@@ -376,8 +386,9 @@ defmodule IEx.Introspection do
         :type_found
 
       is_nil(docs) and spec != [] ->
-        message = %{"en" => "Module was compiled without docs. Showing only specs."}
-        print_doc(["#{inspect(mod)}.#{fun}/#{arity}"], spec, "text/markdown", message, %{})
+        en = "Module was compiled without docs. Showing only specs."
+        formatted = Exception.format_mfa(mod, fun, arity)
+        print_doc([formatted], spec, "text/markdown", %{"en" => en}, %{})
         :ok
 
       is_nil(docs) ->
@@ -420,14 +431,13 @@ defmodule IEx.Introspection do
   defp extract_name_and_arity({{_, name, arity}, _, _, _, _}), do: {name, arity}
 
   defp find_doc_with_content(docs, function, arity) do
-    doc = find_doc(docs, function, arity)
-    if doc != nil and has_content?(doc), do: doc
+    case find_doc(docs, function, arity) do
+      {_, _, _, :hidden, _} -> nil
+      {_, _, _, _, _} = doc -> doc
+      _ -> nil
+    end
   end
 
-  defp has_content?({_, _, _, :hidden, _}), do: false
-  defp has_content?({{_, name, _}, _, _, :none, _}), do: hd(Atom.to_charlist(name)) != ?_
-  defp has_content?({_, _, _, _, _}), do: true
-
   defp find_doc(nil, _fun, _arity) do
     nil
   end
@@ -454,12 +464,24 @@ defmodule IEx.Introspection do
          {{kind, fun, arity}, _line, signature, doc, metadata},
          spec
        ) do
-    if callback_module = doc == :none and callback_module(mod, fun, arity) do
+    callback_module = doc == :none and callback_module(mod, fun, arity)
+
+    if callback_module do
       filter = &match?({_, ^fun, ^arity}, elem(&1, 0))
 
       case get_callback_docs(callback_module, filter) do
-        :no_beam -> nil
-        callback_docs -> Enum.each(callback_docs, &print_typespec/1)
+        :no_beam ->
+          nil
+
+        callback_docs ->
+          en =
+            "#{Exception.format_mfa(mod, fun, arity)} has no docs but is a callback for behaviour " <>
+              "#{inspect(callback_module)}. Showing callback docs instead."
+
+          format_signature(language, kind, signature)
+          |> print_doc(spec, format, %{"en" => en}, metadata)
+
+          Enum.each(callback_docs, &print_typespec/1)
       end
     else
       print_doc(format_signature(language, kind, signature), spec, format, doc, metadata)
@@ -734,15 +756,10 @@ defmodule IEx.Introspection do
   ## Helpers
 
   defp format_typespec(definition, kind, nesting) do
-    string = "@#{kind} #{Macro.to_string(definition)}"
-
-    try do
-      string
-      |> Code.format_string!(line_length: IEx.width() - 2 * nesting)
-      |> IO.iodata_to_binary()
-    rescue
-      _ -> string
-    end
+    {:@, [], [{kind, [], [definition]}]}
+    |> Code.quoted_to_algebra()
+    |> Inspect.Algebra.format(IEx.Config.width())
+    |> IO.iodata_to_binary()
     |> color_prefix_with_line()
     |> indent(nesting)
   end
@@ -779,9 +796,7 @@ defmodule IEx.Introspection do
   end
 
   defp translate_doc(%{"en" => doc}), do: doc
-  defp translate_doc(%{}), do: nil
-  defp translate_doc(:none), do: nil
-  defp translate_doc(:hidden), do: nil
+  defp translate_doc(_), do: nil
 
   defp no_beam(module) do
     case Code.ensure_loaded(module) do
