@@ -21,7 +21,7 @@ defmodule IEx.InteractionTest do
 
   test "invalid input" do
     assert capture_iex("if true do ) false end") =~
-             "** (SyntaxError) iex:1:12: unexpected token: ). The \"do\" at line 1 is missing terminator \"end\""
+             "** (SyntaxError) iex:1:12: unexpected token: )"
   end
 
   test "multiple vars" do
@@ -124,15 +124,18 @@ defmodule IEx.InteractionTest do
   end
 
   test "inspect opts" do
-    opts = [inspect: [binaries: :as_binaries, charlists: :as_lists, structs: false, limit: 4]]
+    opts = [
+      inspect: [
+        binaries: :as_binaries,
+        charlists: :as_lists,
+        structs: false,
+        limit: 4,
+        custom_options: [sort_maps: true]
+      ]
+    ]
 
-    assert "<<45, 46, 47>>\n[45, 46, 47]\n%{" <> map =
-             capture_iex("<<45, 46, 47>>\n[45, 46, 47]\n%IO.Stream{}", opts)
-
-    assert map =~ "__struct__: IO.Stream"
-    assert map =~ "device: nil"
-    assert map =~ "line_or_bytes: :line"
-    assert map =~ "raw: true"
+    assert capture_iex("<<45, 46, 47>>\n[45, 46, 47]\n%IO.Stream{}", opts) ==
+             "<<45, 46, 47>>\n[45, 46, 47]\n%{__struct__: IO.Stream, device: nil, line_or_bytes: :line, raw: true}"
   end
 
   test "exception" do
@@ -147,7 +150,9 @@ defmodule IEx.InteractionTest do
   test "exception while invoking conflicting helpers" do
     import File, only: [open: 1], warn: false
 
-    assert capture_iex("open('README.md')", [], env: __ENV__) =~
+    assert capture_io(:stderr, fn ->
+             capture_iex("open('README.md')", [], env: __ENV__)
+           end) =~
              ~r"function open/1 imported from both File and IEx.Helpers"
   end
 
@@ -203,10 +208,8 @@ defmodule IEx.InteractionTest do
 
   describe ".iex" do
     test "no .iex" do
-      capture_io(:stderr, fn ->
-        assert "** (CompileError) iex:1: undefined function my_variable/0" <> _ =
-                 capture_iex("my_variable")
-      end)
+      assert capture_io(:stderr, fn -> capture_iex("my_variable") end) =~
+               "undefined variable \"my_variable\""
     end
 
     @tag :tmp_dir
