@@ -108,6 +108,17 @@ defmodule Mix do
 
       {:some_test_dependency, "~> 1.0", only: :test}
 
+  When running Mix via the command line, you can configure the default
+  environment or the preferred environment per task via the `def cli`
+  function in your `mix.exs`. For example:
+
+      def cli do
+        [
+          default_env: :local,
+          preferred_envs: [docs: :docs]
+        ]
+      end
+
   The environment can be read via `Mix.env/0`.
 
   ## Targets
@@ -116,6 +127,18 @@ defmodule Mix do
   project needs to compile to different architectures and some of the
   dependencies are only available to some of them. By default, the target
   is `:host` but it can be set via the `MIX_TARGET` environment variable.
+
+  When running Mix via the command line, you can configure the default
+  target or the preferred target per task via the `def cli` function
+  in your `mix.exs`. For example:
+
+      def cli do
+        [
+          default_target: :local,
+          preferred_targets: [docs: :docs]
+        ]
+      end
+
   The target can be read via `Mix.target/0`.
 
   ## Configuration
@@ -125,7 +148,9 @@ defmodule Mix do
   the application environment. On this section, we will focus on how to configure
   it at two distinct moments: build-time and runtime.
 
-  > Note: The application environment is discouraged for libraries. See Elixir's
+  > #### Avoiding the application environment {: .warning}
+  >
+  > The application environment is discouraged for libraries. See Elixir's
   > [Library Guidelines](https://hexdocs.pm/elixir/library-guidelines.html) for
   > more information.
 
@@ -151,8 +176,11 @@ defmodule Mix do
   machine before your code compiles. This can be an issue if the machine
   compiling your code does not have access to all environment variables
   used to run your code, as loading the config above will fail due to the
-  missing environment variable. Luckily, Mix also provides runtime
-  configuration, which should be preferred and we will see next.
+  missing environment variable. Furthermore, even if the environment variable
+  is set, changing the environment variable will require a full recompilation
+  of your application by calling `mix compile --force` (otherwise your project
+  won't start). Luckily, Mix also provides runtime configuration, which is
+  preferred in such cases and we will see next.
 
   ### Runtime configuration
 
@@ -189,19 +217,31 @@ defmodule Mix do
         defp aliases do
           [
             c: "compile",
-            hello: &hello/1
+            hello: &hello/1,
+            paid_task: &paid_task/1
           ]
         end
 
         defp hello(_) do
           Mix.shell().info("Hello world")
         end
+
+        defp paid_task(_) do
+          Mix.Task.run("paid.task", [
+            "first_arg",
+            "second_arg",
+            "--license-key",
+            System.fetch_env!("SOME_LICENSE_KEY")
+          ])
+        end
       end
 
-  In the example above, we have defined two aliases. One is `mix c`
-  which is a shortcut for `mix compile`. The other is named
-  `mix hello`, which is the equivalent to the `Mix.Tasks.Hello`
-  we have defined in the [Mix.Task section](#module-mix-task).
+  In the example above, we have defined three aliases. One is `mix c`
+  which is a shortcut for `mix compile`. Another is named
+  `mix hello` and the third is named `mix paid_task`, which executes
+  the code inside a custom function to invoke the `paid.task` task
+  with several arguments, including one pulled from an environment
+  variable.
 
   Aliases may also be lists, specifying multiple tasks to be run
   consecutively:
@@ -221,19 +261,13 @@ defmodule Mix do
   Where `&clean_extra/1` would be a function in your `mix.exs`
   with extra cleanup logic.
 
-  Arguments given to the alias will be appended to the arguments of
-  the last task in the list. Except when overriding an existing task.
-  In this case, the arguments will be given to the original task,
-  in order to preserve semantics. For example, in the `:clean` alias
-  above, the arguments given to the alias will be passed to "clean"
-  and not to `clean_extra/1`.
+  If the alias is overriding an existing task, the arguments given
+  to the alias will be forwarded to the original task in order to
+  preserve semantics. Otherwise arguments given to the alias are
+  appended to the arguments of the last task in the list.
 
-  Aliases defined in the current project do not affect its dependencies
-  and aliases defined in dependencies are not accessible from the
-  current project.
-
-  Aliases can be used very powerfully to also run Elixir scripts and
-  shell commands, for example:
+  Another use case of aliases is to run Elixir scripts and shell
+  commands, for example:
 
       # priv/hello1.exs
       IO.puts("Hello One")
@@ -257,10 +291,15 @@ defmodule Mix do
   then `mix cmd` to execute a command line shell script. This shows how
   powerful aliases mixed with Mix tasks can be.
 
-  Mix tasks are designed to run only once. This prevents the same task
-  from being executed multiple times. For example, if there are several tasks
-  depending on `mix compile`, the code will be compiled once. Tasks can
-  be executed again if they are explicitly reenabled using `Mix.Task.reenable/1`:
+  One commit pitfall of aliases comes when trying to invoke the same task
+  multiple times. Mix tasks are designed to run only once. This prevents
+  the same task from being executed multiple times. For example, if there
+  are several tasks depending on `mix compile`, the code will be compiled
+  only once.
+
+  Similary, `mix format` can only be invoked once. So if you have an alias
+  that attempts to invoke `mix format` multiple times, it won't work unless
+  it is explicitly reenabled using `Mix.Task.reenable/1`:
 
       another_alias: [
         "format --check-formatted priv/hello1.exs",
@@ -270,15 +309,14 @@ defmodule Mix do
       ]
 
   Some tasks are automatically reenabled though, as they are expected to
-  be invoked multiple times. They are: `mix cmd`, `mix do`, `mix loadconfig`,
-  `mix profile.cprof`, `mix profile.eprof`, `mix profile.fprof`, `mix run`,
-  and `mix xref`.
+  be invoked multiple times, such as: `mix cmd`, `mix do`, `mix xref`, etc.
 
-  It is worth mentioning that some tasks, such as in the case of the
-  `mix format` command in the example above, can accept multiple files so it
-  could be rewritten as:
-
-      another_alias: ["format --check-formatted priv/hello1.exs priv/hello2.exs"]
+  Finally, aliases defined in the current project do not affect its
+  dependencies and aliases defined in dependencies are not accessible
+  from the current project, with the exception of umbrella projects.
+  Umbrella projects will run the aliases of its children when the
+  umbrella project itself does not define said alias and there is no
+  task with said name.
 
   ## Environment variables
 
@@ -288,26 +326,43 @@ defmodule Mix do
 
     * `MIX_ARCHIVES` - specifies the directory into which the archives should be installed
       (default: `~/.mix/archives`)
-    * `MIX_BUILD_ROOT` - sets the root directory where build artifacts
-      should be written to. For example, "_build". If `MIX_BUILD_PATH` is set, this
-      option is ignored.
-    * `MIX_BUILD_PATH` - sets the project `Mix.Project.build_path/0` config. This option
-      must always point to a subdirectory inside a temporary directory. For instance,
-      never "/tmp" or "_build" but "_build/PROD" or "/tmp/PROD", as required by Mix
-    * `MIX_DEPS_PATH` - sets the project `Mix.Project.deps_path/0` config for the current project (default: `deps`)
+
+    * `MIX_BUILD_PATH` - sets the project `Mix.Project.build_path/0` config.
+      This option must always point to a subdirectory inside a temporary directory.
+      For instance, never "/tmp" or "_build" but "_build/PROD" or "/tmp/PROD", as
+      required by Mix. This environment variable is used mostly by external build
+      tools. For your CI servers, you likely want to use `MIX_BUILD_ROOT` below.
+
+    * `MIX_BUILD_ROOT` - sets the root directory where build artifacts should be
+      written to. For example, "_build". If `MIX_BUILD_PATH` is set, this option
+      is ignored.
+
     * `MIX_DEBUG` - outputs debug information about each task before running it
+
+    * `MIX_DEPS_PATH` - sets the project `Mix.Project.deps_path/0` config for the
+      current project (default: `deps`)
+
     * `MIX_ENV` - specifies which environment should be used. See [Environments](#module-environments)
-    * `MIX_TARGET` - specifies which target should be used. See [Targets](#module-targets)
+
     * `MIX_EXS` - changes the full path to the `mix.exs` file
+
     * `MIX_HOME` - path to Mix's home directory, stores configuration files and scripts used by Mix
       (default: `~/.mix`)
+
     * `MIX_INSTALL_DIR` - (since v1.12.0) specifies directory where `Mix.install/2` keeps
        install cache
+
     * `MIX_INSTALL_FORCE` - (since v1.13.0) runs `Mix.install/2` with empty install cache
+
     * `MIX_PATH` - appends extra code paths
+
     * `MIX_QUIET` - does not print information messages to the terminal
+
     * `MIX_REBAR3` - path to rebar3 command that overrides the one Mix installs
       (default: `~/.mix/rebar3`)
+
+    * `MIX_TARGET` - specifies which target should be used. See [Targets](#module-targets)
+
     * `MIX_XDG` - asks Mix to follow the [XDG Directory Specification](https://specifications.freedesktop.org/basedir-spec/basedir-spec-latest.html)
       for its home directory and configuration files. This behaviour needs to
       be opt-in due to backwards compatibility. `MIX_HOME` has higher preference
@@ -332,13 +387,22 @@ defmodule Mix do
     :ok
   end
 
-  @doc false
+  @impl true
   def start(_type, []) do
+    Mix.Local.append_archives()
+    Mix.Local.append_paths()
     children = [Mix.State, Mix.TasksServer, Mix.ProjectStack]
     opts = [strategy: :one_for_one, name: Mix.Supervisor, max_restarts: 0]
     Supervisor.start_link(children, opts)
   end
 
+  @impl true
+  def stop(_data) do
+    Mix.Local.remove_archives()
+    Mix.Local.remove_paths()
+    :ok
+  end
+
   @doc """
   Returns the current Mix environment.
 
@@ -526,6 +590,42 @@ defmodule Mix do
     Path.join(Mix.Utils.mix_home(), "escripts")
   end
 
+  @doc """
+  Ensures the given application from Erlang/OTP or Elixir and its dependencies
+  are available in the path.
+
+  This is mostly used for Mix internal needs. In your own tasks, you should
+  list the Erlang application dependencies under the `:extra_applications`
+  section of your `mix.exs`.
+  """
+  @doc since: "1.15.0"
+  def ensure_application!(app) when is_atom(app) do
+    ensure_application!(app, Mix.State.builtin_apps(), [])
+    :ok
+  end
+
+  defp ensure_application!(app, builtin_apps, optional) do
+    case builtin_apps do
+      %{^app => path} ->
+        Code.prepend_path(path, cache: true)
+        Application.load(app)
+        optional = List.wrap(Application.spec(app, :optional_applications))
+
+        Application.spec(app, :applications)
+        |> List.wrap()
+        |> Enum.each(&ensure_application!(&1, builtin_apps, optional))
+
+      %{} ->
+        unless app in optional do
+          Mix.raise(
+            "The application \"#{app}\" could not be found. This may happen if your " <>
+              "Operating System broke Erlang into multiple packages and may be fixed " <>
+              "by installing the missing \"erlang-dev\" and \"erlang-#{app}\" packages"
+          )
+        end
+    end
+  end
+
   @doc """
   Installs and starts dependencies.
 
@@ -565,10 +665,11 @@ defmodule Mix do
       names with respective values as binaries. The system environment is made part
       of the `Mix.install/2` cache, so different configurations will lead to different apps
 
-    * `:config` (since v1.13.0) - a keyword list of keyword lists with application
-      configuration to be set before the apps loaded. The configuration is part of
-      the `Mix.install/2` cache, so different configurations will lead to different
-      apps
+    * `:config` (since v1.13.0) - a keyword list of keyword lists of compile-time
+      configuration. The configuration is part of the `Mix.install/2` cache, so
+      different configurations will lead to different apps. For this reason, you
+      want to minimize the amount of configuration set through this option.
+      Use `Application.put_all_env/4` for setting other runtime configuration.
 
     * `:config_path` (since v1.14.0) - path to a configuration file. If a `runtime.exs`
       file exists in the same directory as the given path, it is loaded too.
@@ -576,6 +677,9 @@ defmodule Mix do
     * `:lockfile` (since v1.14.0) - path to a lockfile to be used as a basis of
       dependency resolution.
 
+    * `:start_applications` (since v1.15.3) - if `true`, ensures that installed app
+      and its dependencies are started after install (Default: `true`)
+
   ## Examples
 
   Installing `:decimal` and `:jason`:
@@ -585,7 +689,7 @@ defmodule Mix do
         {:jason, "~> 1.0"}
       ])
 
-  Installing `:nx` & `:exla`, and configuring the underlying applications
+  Installing `:nx` and `:exla`, and configuring the underlying applications
   and environment variables:
 
       Mix.install(
@@ -696,6 +800,7 @@ defmodule Mix do
     config_path = expand_path(opts[:config_path], deps, :config_path, "config/config.exs")
     system_env = Keyword.get(opts, :system_env, [])
     consolidate_protocols? = Keyword.get(opts, :consolidate_protocols, true)
+    start_applications? = Keyword.get(opts, :start_applications, true)
 
     id =
       {deps, config, system_env, consolidate_protocols?}
@@ -729,15 +834,15 @@ defmodule Mix do
           deps_path: "deps",
           deps: deps,
           app: :mix_install,
-          erlc_paths: ["src"],
-          elixirc_paths: ["lib"],
+          erlc_paths: [],
+          elixirc_paths: [],
           compilers: [],
           consolidate_protocols: consolidate_protocols?,
-          config_path: config_path
+          config_path: config_path,
+          prune_code_paths: false
         ]
 
         started_apps = Application.started_applications()
-        :ok = Mix.Local.append_archives()
         :ok = Mix.ProjectStack.push(@mix_install_project, config, "nofile")
         build_dir = Path.join(install_dir, "_build")
         external_lockfile = expand_path(opts[:lockfile], deps, :lockfile, "mix.lock")
@@ -767,7 +872,7 @@ defmodule Mix do
                   lockfile = Path.join(install_dir, "mix.lock")
                   old_lock = Mix.Dep.Lock.read(lockfile)
                   new_lock = Mix.Dep.Lock.read(external_lockfile)
-                  Mix.Dep.Lock.write(lockfile, Map.merge(old_lock, new_lock))
+                  Mix.Dep.Lock.write(Map.merge(old_lock, new_lock), file: lockfile)
                   File.write!(md5_path, Base.encode64(new_md5))
                   Mix.Task.rerun("deps.get")
                 end
@@ -776,6 +881,9 @@ defmodule Mix do
                 Mix.Task.rerun("deps.get")
 
               true ->
+                # We already have a cache. If the user by any chance uninstalled Hex,
+                # we make sure it is installed back (which mix deps.get would do anyway)
+                Mix.Hex.ensure_installed?(true)
                 :ok
             end
 
@@ -792,9 +900,11 @@ defmodule Mix do
             end
           end)
 
-          for %{app: app, opts: opts} <- Mix.Dep.cached(),
-              Keyword.get(opts, :runtime, true) and Keyword.get(opts, :app, true) do
-            Application.ensure_all_started(app)
+          if start_applications? do
+            for %{app: app, opts: opts} <- Mix.Dep.cached(),
+                Keyword.get(opts, :runtime, true) and Keyword.get(opts, :app, true) do
+              Application.ensure_all_started(app)
+            end
           end
 
           Mix.State.put(:installed, id)
