@@ -9,11 +9,17 @@ defmodule Mix.Tasks.Compile do
   It simply runs the compilers registered in your project and returns
   a tuple with the compilation status and a list of diagnostics.
 
-  Before compiling code, it loads the code in all dependencies and
-  perform a series of checks to ensure the project is up to date.
+  Before compiling code, it performs a series of checks to ensure all
+  dependencies are compiled and the project is up to date. Then the
+  code path of your Elixir system is pruned to only contain the dependencies
+  and applications that you have explicitly listed in your `mix.exs`.
 
   ## Configuration
 
+    * `:build_embedded` - this option was used to copy all code and
+      priv content to the `_build` directory. However, this option no
+      longer has an effect as Elixir will now copy those at release time
+
     * `:compilers` - compilers to run, defaults to `Mix.compilers/0`,
       which are `[:yecc, :leex, :erlang, :elixir, :app]`.
 
@@ -21,9 +27,6 @@ defmodule Mix.Tasks.Compile do
       consolidation via the `mix compile.protocols` task. The default
       value is `true`.
 
-    * `:build_embedded` - when `true`, embeds all code and priv
-      content in the `_build` directory instead of using symlinks.
-
     * `:build_path` - the directory where build artifacts
       should be written to. This option is intended only for
       child apps within a larger umbrella application so that
@@ -37,19 +40,19 @@ defmodule Mix.Tasks.Compile do
   To see documentation for each specific compiler, you must
   invoke `help` directly for the compiler command:
 
-      mix help compile.elixir
-      mix help compile.erlang
+      $ mix help compile.elixir
+      $ mix help compile.erlang
 
   You can get a list of all compilers by running:
 
-      mix compile --list
+      $ mix compile --list
 
   ## Command line options
 
-    * `--all-warnings` - prints warnings even from files that do not need to be recompiled
     * `--erl-config` - path to an Erlang term file that will be loaded as Mix config
     * `--force` - forces compilation
     * `--list` - lists all enabled compilers
+    * `--no-all-warnings` - prints only warnings from files currently compiled (instead of all)
     * `--no-app-loading` - does not load .app resource file after compilation
     * `--no-archives-check` - skips checking of archives
     * `--no-compile` - does not actually compile, only loads code and perform checks
@@ -58,6 +61,8 @@ defmodule Mix.Tasks.Compile do
     * `--no-optional-deps` - does not compile or load optional deps. Useful for testing
       if a library still successfully compiles without optional dependencies (which is the
       default case with dependencies)
+    * `--no-prune-code-paths` - do not prune code paths before compilation, this keeps
+      the entirety of Erlang/OTP available on the project starts
     * `--no-protocol-consolidation` - skips protocol consolidation
     * `--no-validate-compile-env` - does not validate the application compile environment
     * `--return-errors` - returns error status and diagnostics instead of exiting on error
@@ -112,7 +117,7 @@ defmodule Mix.Tasks.Compile do
     sorted = Enum.sort(docs)
 
     Enum.each(sorted, fn {task, doc} ->
-      shell.info(format('mix ~-#{max}s # ~ts', [task, doc]))
+      shell.info(format(~c"mix ~-#{max}s # ~ts", [task, doc]))
     end)
 
     consolidate_protocols? = Mix.Project.config()[:consolidate_protocols]
@@ -124,13 +129,14 @@ defmodule Mix.Tasks.Compile do
   @impl true
   def run(args) do
     Mix.Project.get!()
-    Mix.Task.run("loadpaths", args)
+
+    # We run loadpaths to perform checks but we don't bother setting
+    # up the load paths because compile.all will manage them anyway.
+    Mix.Task.run("loadpaths", ["--no-deps-loading" | args])
 
     {opts, _, _} = OptionParser.parse(args, switches: [erl_config: :string])
     load_erl_config(opts)
 
-    # We invoke compile.all even on --no-compile because compile.all
-    # loads the apps, and we still want to do that.
     {res, diagnostics} =
       Mix.Task.run("compile.all", args)
       |> List.wrap()
@@ -139,6 +145,18 @@ defmodule Mix.Tasks.Compile do
 
     config = Mix.Project.config()
 
+    # If we are in an umbrella project, now load paths from all children.
+    if apps_paths = Mix.Project.apps_paths(config) do
+      loaded_paths =
+        (Mix.Tasks.Compile.All.project_apps(config) ++ Map.keys(apps_paths))
+        |> Mix.AppLoader.load_apps(Mix.Dep.cached(), config, [], fn
+          {_app, path}, acc -> if path, do: [path | acc], else: acc
+        end)
+
+      # We don't cache umbrella paths as we may write to them
+      Code.prepend_paths(loaded_paths -- :code.get_path())
+    end
+
     consolidate_protocols? =
       config[:consolidate_protocols] and "--no-protocol-consolidation" not in args
 
@@ -159,6 +177,7 @@ defmodule Mix.Tasks.Compile do
     with true <- consolidate_protocols?,
          path = Mix.Project.consolidation_path(config),
          {:ok, protocols} <- File.ls(path) do
+      # We don't cache consolidation path as we may write to it
       Code.prepend_path(path)
       Enum.each(protocols, &load_protocol/1)
     end
