@@ -6,8 +6,9 @@ defmodule Mix.Tasks.Deps.Compile do
   @moduledoc """
   Compiles dependencies.
 
-  By default, compile all dependencies. A list of dependencies
-  can be given to compile multiple dependencies in order.
+  By default, this task attempts to compile all dependencies.
+  A list of dependencies can be given to compile multiple
+  dependencies in order.
 
   This task attempts to detect if the project contains one of
   the following files and act accordingly:
@@ -56,10 +57,7 @@ defmodule Mix.Tasks.Deps.Compile do
 
     case OptionParser.parse(args, switches: @switches) do
       {opts, [], _} ->
-        # Because this command may be invoked explicitly with
-        # deps.compile, we simply try to compile any available
-        # dependency.
-        compile(Enum.filter(deps, &Mix.Dep.available?/1), opts)
+        compile(filter_available_and_local_deps(deps), opts)
 
       {opts, tail, _} ->
         compile(Mix.Dep.filter_by_name(tail, deps, opts), opts)
@@ -77,7 +75,7 @@ defmodule Mix.Tasks.Deps.Compile do
       |> reject_umbrella_children(options)
       |> reject_local_deps(options)
       |> Enum.map(fn %Mix.Dep{app: app, status: status, opts: opts, scm: scm} = dep ->
-        check_unavailable!(app, status)
+        check_unavailable!(app, scm, status)
         maybe_clean(dep, options)
 
         compiled? =
@@ -91,9 +89,6 @@ defmodule Mix.Tasks.Deps.Compile do
             Mix.Dep.make?(dep) ->
               do_make(dep, config)
 
-            dep.manager == :rebar ->
-              do_rebar(dep, config)
-
             dep.manager == :rebar3 ->
               do_rebar3(dep, config)
 
@@ -106,8 +101,6 @@ defmodule Mix.Tasks.Deps.Compile do
               false
           end
 
-        unless Mix.Dep.mix?(dep), do: build_structure(dep, config)
-
         # We should touch fetchable dependencies even if they
         # did not compile otherwise they will always be marked
         # as stale, even when there is nothing to do.
@@ -138,14 +131,22 @@ defmodule Mix.Tasks.Deps.Compile do
     end
   end
 
-  defp check_unavailable!(app, {:unavailable, _}) do
-    Mix.raise(
-      "Cannot compile dependency #{inspect(app)} because " <>
-        "it isn't available, run \"mix deps.get\" first"
-    )
+  defp check_unavailable!(app, scm, {:unavailable, path}) do
+    if scm.fetchable? do
+      Mix.raise(
+        "Cannot compile dependency #{inspect(app)} because " <>
+          "it isn't available, run \"mix deps.get\" first"
+      )
+    else
+      Mix.raise(
+        "Cannot compile dependency #{inspect(app)} because " <>
+          "it isn't available, please ensure the dependency is at " <>
+          inspect(Path.relative_to_cwd(path))
+      )
+    end
   end
 
-  defp check_unavailable!(_, _) do
+  defp check_unavailable!(_, _, _) do
     :ok
   end
 
@@ -161,14 +162,7 @@ defmodule Mix.Tasks.Deps.Compile do
       end
 
       try do
-        options = [
-          "--no-deps-loading",
-          "--no-apps-loading",
-          "--no-archives-check",
-          "--no-elixir-version-check",
-          "--no-warnings-as-errors"
-        ]
-
+        options = ["--from-mix-deps-compile", "--no-warnings-as-errors", "--no-code-path-pruning"]
         res = Mix.Task.run("compile", options)
         match?({:ok, _}, res)
       catch
@@ -177,8 +171,7 @@ defmodule Mix.Tasks.Deps.Compile do
 
           Mix.shell().error(
             "could not compile dependency #{inspect(app)}, \"mix compile\" failed. " <>
-              "You can recompile this dependency with \"mix deps.compile #{app}\", update it " <>
-              "with \"mix deps.update #{app}\" or clean it with \"mix deps.clean #{app}\""
+              deps_compile_feedback(app)
           )
 
           :erlang.raise(kind, reason, __STACKTRACE__)
@@ -186,23 +179,45 @@ defmodule Mix.Tasks.Deps.Compile do
     end)
   end
 
-  defp do_rebar(dep, config) do
-    lib_path = Path.join(config[:env_path], "lib")
-    cmd = "#{rebar_cmd(dep)} compile skip_deps=true deps_dir=#{inspect(lib_path)}"
-    do_command(dep, config, cmd, false)
-  end
-
   defp do_rebar3(%Mix.Dep{opts: opts} = dep, config) do
-    dep_path = opts[:build]
-    config_path = Path.join(dep_path, "mix.rebar.config")
-    lib_path = Path.join(config[:env_path], "lib/*/ebin")
+    dep_path = opts[:dest]
+    build_path = opts[:build]
+    File.mkdir_p!(build_path)
 
-    env = [{"REBAR_CONFIG", config_path}, {"TERM", "dumb"}]
-    cmd = "#{rebar_cmd(dep)} bare compile --paths=#{inspect(lib_path)}"
+    # For Rebar3, we need to copy the source/ebin to the target/ebin
+    # before we run the command given that REBAR_BARE_COMPILER_OUTPUT_DIR
+    # writes directly to _build.
+    File.cp_r(Path.join(dep_path, "ebin"), Path.join(build_path, "ebin"))
 
-    File.mkdir_p!(dep_path)
+    # Now establish symlinks to the remaining sources
+    for dir <- ~w(include priv src) do
+      Mix.Utils.symlink_or_copy(Path.join(dep_path, dir), Path.join(build_path, dir))
+    end
+
+    # Build the rebar config and setup the command line
+    config_path = Path.join(build_path, "mix.rebar.config")
+    lib_path = Path.join(config[:deps_build_path], "lib/*/ebin")
     File.write!(config_path, rebar_config(dep))
+
+    env = [
+      # REBAR_BARE_COMPILER_OUTPUT_DIR is only honored by rebar3 >= 3.14
+      {"REBAR_BARE_COMPILER_OUTPUT_DIR", build_path},
+      {"REBAR_CONFIG", config_path},
+      {"REBAR_PROFILE", "prod"},
+      {"TERM", "dumb"}
+    ]
+
+    cmd = "#{rebar_cmd(dep)} bare compile --paths #{lib_path}"
     do_command(dep, config, cmd, false, env)
+
+    # Check if we have any new symlinks after compilation
+    for dir <- ~w(include priv src),
+        File.exists?(Path.join(dep_path, dir)) and not File.exists?(Path.join(build_path, dir)) do
+      Mix.Utils.symlink_or_copy(Path.join(dep_path, dir), Path.join(build_path, dir))
+    end
+
+    Code.prepend_path(Path.join(build_path, "ebin"), cache: true)
+    true
   end
 
   defp rebar_config(dep) do
@@ -222,8 +237,6 @@ defmodule Mix.Tasks.Deps.Compile do
       "Could not find \"#{manager}\", which is needed to build dependency #{inspect(app)}"
     )
 
-    shell.info("I can install a local copy which is just used by Mix")
-
     install_question =
       "Shall I install #{manager}? (if running non-interactively, " <>
         "use \"mix local.rebar --force\")"
@@ -236,13 +249,15 @@ defmodule Mix.Tasks.Deps.Compile do
       Mix.raise(error_message)
     end
 
-    (Mix.Tasks.Local.Rebar.run([]) && Mix.Rebar.local_rebar_cmd(manager)) ||
-      Mix.raise("\"#{manager}\" installation failed")
+    Mix.Tasks.Local.Rebar.run(["--force"])
+    Mix.Rebar.local_rebar_cmd(manager) || Mix.raise("\"#{manager}\" installation failed")
   end
 
   defp do_make(dep, config) do
     command = make_command(dep)
     do_command(dep, config, command, true, [{"IS_DEP", "1"}])
+    build_structure(dep, config)
+    true
   end
 
   defp make_command(dep) do
@@ -270,6 +285,8 @@ defmodule Mix.Tasks.Deps.Compile do
   defp do_compile(%Mix.Dep{opts: opts} = dep, config) do
     if command = opts[:compile] do
       do_command(dep, config, command, true)
+      build_structure(dep, config)
+      true
     else
       false
     end
@@ -278,36 +295,22 @@ defmodule Mix.Tasks.Deps.Compile do
   defp do_command(dep, config, command, print_app?, env \\ []) do
     %Mix.Dep{app: app, system_env: system_env, opts: opts} = dep
 
-    File.cd!(opts[:dest], fn ->
-      env = [{"ERL_LIBS", Path.join(config[:env_path], "lib")} | system_env] ++ env
+    env = [{"ERL_LIBS", Path.join(config[:deps_build_path], "lib")} | system_env] ++ env
 
-      if Mix.shell().cmd(command, env: env, print_app: print_app?) != 0 do
-        Mix.raise(
-          "Could not compile dependency #{inspect(app)}, \"#{command}\" command failed. " <>
-            "You can recompile this dependency with \"mix deps.compile #{app}\", update it " <>
-            "with \"mix deps.update #{app}\" or clean it with \"mix deps.clean #{app}\""
-        )
-      end
-    end)
+    if Mix.shell().cmd(command, env: env, print_app: print_app?, cd: opts[:dest]) != 0 do
+      Mix.raise(
+        "Could not compile dependency #{inspect(app)}, \"#{command}\" command failed. " <>
+          deps_compile_feedback(app)
+      )
+    end
 
     true
   end
 
-  defp build_structure(%Mix.Dep{opts: opts} = dep, config) do
-    build_path = Path.dirname(opts[:build])
-
-    Enum.each(Mix.Dep.source_paths(dep), fn {source, base} ->
-      app = Path.join(build_path, base)
-      build_structure(source, app, config)
-      Code.prepend_path(Path.join(app, "ebin"))
-    end)
-  end
-
-  defp build_structure(dest, build, config) do
-    File.cd!(dest, fn ->
-      config = Keyword.put(config, :app_path, build)
-      Mix.Project.build_structure(config, symlink_ebin: true)
-    end)
+  defp build_structure(%Mix.Dep{opts: opts}, config) do
+    config = Keyword.put(config, :deps_app_path, opts[:build])
+    Mix.Project.build_structure(config, symlink_ebin: true, source: opts[:dest])
+    Code.prepend_path(Path.join(opts[:build], "ebin"), cache: true)
   end
 
   defp old_elixir_req(config) do
@@ -334,6 +337,12 @@ defmodule Mix.Tasks.Deps.Compile do
     end
   end
 
+  defp filter_available_and_local_deps(deps) do
+    Enum.filter(deps, fn dep ->
+      Mix.Dep.available?(dep) or not dep.scm.fetchable?
+    end)
+  end
+
   defp reject_local_deps(deps, options) do
     if options[:skip_local_deps] do
       Enum.filter(deps, fn %{scm: scm} -> scm.fetchable? end)
@@ -341,4 +350,15 @@ defmodule Mix.Tasks.Deps.Compile do
       deps
     end
   end
+
+  defp deps_compile_feedback(app) do
+    if Mix.install?() do
+      "Errors may have been logged above. You may run Mix.install/2 to try again or " <>
+        "change the arguments to Mix.install/2 to try another version"
+    else
+      "Errors may have been logged above. You can recompile this dependency with " <>
+        "\"mix deps.compile #{app} --force\", update it with \"mix deps.update #{app}\" or " <>
+        "clean it with \"mix deps.clean #{app}\""
+    end
+  end
 end
