@@ -1,7 +1,7 @@
 defmodule Mix.Tasks.Deps do
   use Mix.Task
 
-  import Mix.Dep, only: [load_on_environment: 1, format_dep: 1, format_status: 1, check_lock: 1]
+  import Mix.Dep, only: [format_dep: 1, format_status: 1, check_lock: 1]
 
   @shortdoc "Lists dependencies and their status"
 
@@ -72,7 +72,8 @@ defmodule Mix.Tasks.Deps do
       other project that depends on the current project won't be forced to
       use the optional dependency. However, if the other project includes
       the optional dependency on its own, the requirements and options
-      specified here will also be applied.
+      specified here will also be applied. Optional dependencies will _not_
+      be started by the application.
 
     * `:only` - the dependency is made available only in the given environments,
       useful when declaring dev- or test-only dependencies; by default the
@@ -88,15 +89,15 @@ defmodule Mix.Tasks.Deps do
     * `:override` - if set to `true` the dependency will override any other
       definitions of itself by other dependencies
 
-    * `:manager` - Mix can also compile Rebar, Rebar3 and makefile projects
-      and can fetch sub dependencies of Rebar and Rebar3 projects. Mix will
+    * `:manager` - Mix can also compile Rebar3 and makefile projects
+      and can fetch sub dependencies of Rebar3 projects. Mix will
       try to infer the type of project but it can be overridden with this
-      option by setting it to `:mix`, `:rebar3`, `:rebar` or `:make`. In case
+      option by setting it to `:mix`, `:rebar3`, or `:make`. In case
       there are conflicting definitions, the first manager in the list above
       will be picked up. For example, if a dependency is found with `:rebar3`
-      and `:rebar` managers in different part of the trees, `:rebar3` will
-      be automatically picked. You can find the manager by running `mix deps`
-      and override it by setting the `:override` option in a top-level project.
+      as a manager in different part of the trees, `:rebar3` will be automatically
+      picked. You can find the manager by running `mix deps` and override it by
+      setting the `:override` option in a top-level project.
 
     * `:runtime` - whether the dependency is part of runtime applications.
       If the `:applications` key is not provided in `def application` in your
@@ -116,12 +117,15 @@ defmodule Mix.Tasks.Deps do
     * `:submodules` - when `true`, initialize submodules for the repo
     * `:sparse` - checkout a single directory inside the Git repository and use it
       as your Mix dependency. Search "sparse Git checkouts" for more information.
+    * `:subdir` - (since v1.13.0) search for the project in the given directory
+      relative to the git checkout. This is similar to `:sparse` option but instead
+      of a doing a sparse checkout it does a full checkout.
 
   If your Git repository requires authentication, such as basic username:password
   HTTP authentication via URLs, it can be achieved via Git configuration, keeping
   the access rules outside of source control.
 
-      git config --global url."https://YOUR_USER:YOUR_PASS@example.com/".insteadOf "https://example.com/"
+      $ git config --global url."https://YOUR_USER:YOUR_PASS@example.com/".insteadOf "https://example.com/"
 
   For more information, see the `git config` documentation:
   https://git-scm.com/docs/git-config#git-config-urlltbasegtinsteadOf
@@ -144,6 +148,11 @@ defmodule Mix.Tasks.Deps do
       [locked at REF]
       STATUS
 
+  For dependencies satisfied by Hex, `REF` is the package checksum.
+
+  For dependencies satisfied by git, `REF` is the commit object name,
+  and may include branch or tag information.
+
   It supports the following options:
 
     * `--all` - lists all dependencies, regardless of specified environment
@@ -158,7 +167,7 @@ defmodule Mix.Tasks.Deps do
 
     shell = Mix.shell()
 
-    load_on_environment(loaded_opts)
+    Mix.Dep.Converger.converge(loaded_opts)
     |> Enum.sort_by(& &1.app)
     |> Enum.each(fn dep ->
       %Mix.Dep{scm: scm, manager: manager} = dep
