@@ -38,12 +38,13 @@ defmodule Mix.Tasks.Release do
 
   If the given name does not exist, an error is raised.
 
-  If `mix release`, without a name, is invoked and there are multiple names,
-  an error will be raised unless you set `default_release: NAME` at the root
-  of your project configuration.
+  If `mix release` is invoked, without specifying a release name, and
+  there are multiple releases configured, an error will be raised
+  unless you set `default_release: NAME` at the root of your project
+  configuration.
 
-  If `mix release` is invoked and there are no names, a release using the
-  application name and default values is assembled.
+  If `mix release` is invoked and there are no releases configured, a
+  release is assembled using the application name and default values.
 
   ## Why releases?
 
@@ -54,14 +55,12 @@ defmodule Mix.Tasks.Release do
       interactive and embedded. By default, it runs in the interactive
       mode which dynamically loads modules when they are used for the
       first time. The first time your application calls `Enum.map/2`,
-      the VM will find the `Enum` module and load it. There’s a downside.
-      When you start a new server in production, it may need to load
+      the VM will find the `Enum` module and load it. There's a downside:
+      when you start a new server in production, it may need to load
       many other modules, causing the first requests to have an unusual
-      spike in response time. When running in Erlang/OTP earlier than 23,
-      the system always runs in embedded mode. When using Erlang/OTP 23+,
-      they run in interactive mode while being configured and then it
-      swaps to embedded mode, guaranteeing your system is ready to handle
-      requests after booting.
+      spike in response time. With releases, the system preloads
+      all modules and guarantees your system is ready to handle requests
+      after booting.
 
     * Configuration and customization. Releases give developers fine
       grained control over system configuration and the VM flags used
@@ -171,8 +170,8 @@ defmodule Mix.Tasks.Release do
       $ bin/RELEASE_NAME daemon
 
   In daemon mode, the system is started on the background via
-  [`run_erl`](https://erlang.org/doc/man/run_erl.html). You may also
-  want to enable [`heart`](https://erlang.org/doc/man/heart.html)
+  [`run_erl`](https://www.erlang.org/doc/man/run_erl.html). You may also
+  want to enable [`heart`](https://www.erlang.org/doc/man/heart.html)
   in daemon mode so it automatically restarts the system in case
   of crashes. See the generated `releases/RELEASE_VSN/env.sh` file.
 
@@ -195,7 +194,7 @@ defmodule Mix.Tasks.Release do
 
   While daemons are not available on Windows, it is possible to install a
   released system as a service on Windows with the help of
-  [`erlsrv`](https://erlang.org/doc/man/erlsrv.html). This can be done by
+  [`erlsrv`](https://www.erlang.org/doc/man/erlsrv.html). This can be done by
   running:
 
       $ bin/RELEASE_NAME install
@@ -293,7 +292,7 @@ defmodule Mix.Tasks.Release do
   Alternatively, you can also bundle the compiled object files in the release,
   as long as they were compiled for the same target. If doing so, you need to
   update `LD_LIBRARY_PATH` environment variable with the paths containing the
-  bundled objects on Unix-like systems or the `PATH` environment variable on
+  bundled objects on Unix-like systems or the `$PATH` environment variable on
   Windows systems.
 
   Currently, there is no official way to cross-compile a release from one
@@ -385,10 +384,18 @@ defmodule Mix.Tasks.Release do
         * `:none` - the application is part of the release but it is neither
           loaded nor started
 
+      If you change the mode of an application, the mode will apply to all its child
+      applications. However, if an application has two parents, the mode of the parent
+      with highest priority wins (where `:permanent` has the highest priority, according
+      to the list above).
+
     * `:strip_beams` - controls if BEAM files should have their debug information,
       documentation chunks, and other non-essential metadata removed. Defaults to
       `true`. May be set to `false` to disable stripping. Also accepts
-      `[keep: ["Docs", "Dbgi"]]` to keep certain chunks that are usually stripped.
+     `[keep: ["Docs", "Dbgi"]]` to keep certain chunks that are usually stripped.
+     You can also set the `:compress` option to true to enable individual
+     compression of BEAM files, although it is typically preferred to compress
+     the whole release instead.
 
     * `:cookie` - a string representing the Erlang Distribution cookie. If this
       option is not set, a random cookie is written to the `releases/COOKIE` file
@@ -532,7 +539,7 @@ defmodule Mix.Tasks.Release do
   `@release`, with the `Mix.Release` struct.
 
   The `vm.args` and `remote.vm.args` files may contain any of the VM flags
-  accepted by the [`erl` command](https://erlang.org/doc/man/erl.html).
+  accepted by the [`erl` command](https://www.erlang.org/doc/man/erl.html).
 
   The `env.sh` and `env.bat` is used to set environment variables.
   In there, you can set vars such as `RELEASE_NODE`, `RELEASE_COOKIE`,
@@ -617,9 +624,7 @@ defmodule Mix.Tasks.Release do
 
   If a `config/runtime.exs` exists, it will be copied to your release
   and executed early in the boot process, when only Elixir and Erlang's
-  main applications have been started. Once the configuration is loaded,
-  the Erlang system will be restarted (within the same Operating System
-  process) and the new configuration will take place.
+  main applications have been started.
 
   You can change the path to the runtime configuration file by setting
   `:runtime_config_path` inside each release configuration. This path is
@@ -632,13 +637,8 @@ defmodule Mix.Tasks.Release do
         ]
       ]
 
-  Finally, in order for runtime configuration to work properly (as well
-  as any other "Config provider" as defined next), it needs to be able
-  to persist the newly computed configuration to disk. The computed config
-  file will be written to "tmp" directory inside the release every time
-  the system boots. You can configure the "tmp" directory by setting the
-  `RELEASE_TMP` environment variable, either explicitly or inside your
-  `releases/RELEASE_VSN/env.sh` (or `env.bat` on Windows).
+  By setting `:runtime_config_path` to `false` it can be used to prevent
+  a runtime configuration file to be included in the release.
 
   ### Config providers
 
@@ -652,13 +652,17 @@ defmodule Mix.Tasks.Release do
   The following options can be set inside your releases key in your `mix.exs`
   to control how config providers work:
 
-    * `:reboot_system_after_config` - every time your release is configured,
-      the system is rebooted to allow the new configuration to take place.
-      You can set this option to `false` to disable the rebooting for applications
-      that are sensitive to boot time but, in doing so, note you won't be able
-      to configure system applications, such as `:kernel` and `:stdlib`.
-      Defaults to `true` if using the deprecated `config/releases.exs`,
-      `false` otherwise.
+    * `:reboot_system_after_config` - reboot the system after configuration
+      so you can configure system applications, such as `:kernel` and `:stdlib`,
+      in your `config/runtime.exs`. Generally speaking, it is best to configure
+      `:kernel` and `:stdlib` using the `vm.args` file but this option is available
+      for those who need more complex configuration. When set to `true`, the
+      release will first boot in interactive mode to compute a config file and
+      write it to the "tmp" directory. Then it reboots in the configured `RELEASE_MODE`.
+      You can configure the "tmp" directory by setting the `RELEASE_TMP` environment
+      variable, either explicitly or inside your `releases/RELEASE_VSN/env.sh`
+      (or `env.bat` on Windows). Defaults to `true` if using the deprecated
+      `config/releases.exs`, `false` otherwise.
 
     * `:prune_runtime_sys_config_after_boot` - if `:reboot_system_after_config`
       is set, every time your system boots, the release will write a config file
@@ -795,9 +799,11 @@ defmodule Mix.Tasks.Release do
       files to. It can be set to a custom directory. It defaults to
       `$RELEASE_ROOT/tmp`
 
-    * `RELEASE_MODE` - if the release should start in embedded or
-      interactive mode. Defaults to "embedded". It applies only to
-      start/daemon/install commands
+    * `RELEASE_MODE` - if the release should load code on demand (interactive)
+      or preload it (embedded). Defaults to "embedded", which increases boot
+      time but it means the runtime will respond faster as it doesn't have to
+      load code. Choose interactive if you need to decrease boot time and reduce
+      memory usage on boot. It applies only to start/daemon/install commands
 
     * `RELEASE_DISTRIBUTION` - how do we want to run the distribution.
       May be `name` (long names), `sname` (short names) or `none`
@@ -981,13 +987,13 @@ defmodule Mix.Tasks.Release do
   hot code upgrade it. This is one of the many steps necessary
   to perform hot code upgrades and it must be taken into account by
   every process and application being upgraded in the system.
-  The [`.appup` cookbook](https://erlang.org/doc/design_principles/appup_cookbook.html)
+  The [`.appup` cookbook](https://www.erlang.org/doc/design_principles/appup_cookbook.html)
   provides a good reference and more examples.
 
   Once `.appup`s are created, the next step is to create a `.relup`
   file with all instructions necessary to update the release itself.
   Erlang documentation does provide a chapter on
-  [Creating and upgrading a target system](https://erlang.org/doc/system_principles/create_target.html).
+  [Creating and upgrading a target system](https://www.erlang.org/doc/system_principles/create_target.html).
   [Learn You Some Erlang has a chapter on hot code upgrades](https://learnyousomeerlang.com/relups).
 
   Overall, there are many steps, complexities and assumptions made
@@ -1032,6 +1038,8 @@ defmodule Mix.Tasks.Release do
   def run(args) do
     Mix.Project.get!()
     Mix.Task.run("compile", args)
+    Mix.ensure_application!(:sasl)
+    Mix.ensure_application!(:crypto)
 
     config = Mix.Project.config()
 
@@ -1093,7 +1101,7 @@ defmodule Mix.Tasks.Release do
     # releases/
     #   COOKIE
     #   start_erl.data
-    consolidation_path = build_rel(release, config)
+    {consolidation_path, release} = build_rel(release, config)
 
     [
       # erts-VSN/
@@ -1198,7 +1206,7 @@ defmodule Mix.Tasks.Release do
          :ok <- Mix.Release.make_sys_config(release, sys_config, config_provider_path),
          :ok <- Mix.Release.make_cookie(release, cookie_path),
          :ok <- Mix.Release.make_start_erl(release, start_erl_path) do
-      consolidation_path
+      {consolidation_path, release}
     else
       {:error, message} ->
         File.rm_rf!(version_path)
@@ -1212,6 +1220,9 @@ defmodule Mix.Tasks.Release do
 
     {path, reboot?} =
       cond do
+        opts[:runtime_config_path] == false ->
+          {false, false}
+
         path = opts[:runtime_config_path] ->
           {path, false}
 
@@ -1226,7 +1237,10 @@ defmodule Mix.Tasks.Release do
           {default_path, false}
 
         File.exists?(deprecated_path) ->
-          # TODO: Warn from Elixir v1.14 onwards
+          IO.warn(
+            "config/releases.exs is deprecated, use config/runtime.exs or set :runtime_config_path in your release configuration instead"
+          )
+
           {deprecated_path, true}
 
         true ->
@@ -1243,7 +1257,7 @@ defmodule Mix.Tasks.Release do
         release = update_in(release.config_providers, &[{Config.Reader, opts} | &1])
         update_in(release.options, &Keyword.put_new(&1, :reboot_system_after_config, reboot?))
 
-      release.config_providers == [] ->
+      release.config_providers == [] and path != false ->
         skipping("runtime configuration (#{default_path} not found)")
         release
 
@@ -1297,7 +1311,7 @@ defmodule Mix.Tasks.Release do
 
     info(release, """
 
-    Release created at #{path}!
+    Release created at #{path}
 
         # To start your system
         #{cmd} start
@@ -1411,7 +1425,16 @@ defmodule Mix.Tasks.Release do
 
   defp cli_for(:windows, release) do
     {"env.bat", &env_bat_template(release: &1),
-     [{"#{release.name}.bat", cli_bat_template(release: release)}]}
+     [{"#{release.name}.bat", cli_bat_template(release: release) |> maybe_replace_werl()}]}
+  end
+
+  defp maybe_replace_werl(contents) do
+    # TODO: Remove me when we require Erlang/OTP 26+
+    if :erlang.system_info(:otp_release) >= ~c"26" do
+      String.replace(contents, "--werl", "")
+    else
+      contents
+    end
   end
 
   defp elixir_cli_for(:unix, release) do
@@ -1419,7 +1442,10 @@ defmodule Mix.Tasks.Release do
       {"elixir",
        &(&1
          |> File.read!()
-         |> String.replace(~s[ -pa "$SCRIPT_PATH"/../lib/*/ebin], "")
+         |> String.replace(
+           ~s[ -elixir_root "$SCRIPT_PATH"/../lib -pa "$SCRIPT_PATH"/../lib/elixir/ebin ],
+           " "
+         )
          |> replace_erts_bin(release, ~s["$SCRIPT_PATH"/../../erts-#{release.erts_version}/bin/]))},
       {"iex", &File.read!/1}
     ]
@@ -1430,7 +1456,10 @@ defmodule Mix.Tasks.Release do
       {"elixir.bat",
        &(&1
          |> File.read!()
-         |> String.replace(~s[goto expand_erl_libs], ~s[goto run])
+         |> String.replace(
+           ~s[ -elixir_root !SCRIPT_PATH!..\\lib -pa !SCRIPT_PATH!..\\lib\\elixir\\ebin ],
+           " "
+         )
          |> replace_erts_bin(release, ~s[%~dp0\\..\\..\\erts-#{release.erts_version}\\bin\\]))},
       {"iex.bat", &File.read!/1}
     ]
@@ -1451,12 +1480,10 @@ defmodule Mix.Tasks.Release do
   # Helper functions
 
   defp release_mode(release, env_var) do
-    # TODO: Remove otp_release check once we require Erlang/OTP 23+
-    otp_gte_23? = :erlang.system_info(:otp_release) >= '23'
     reboot? = Keyword.get(release.options, :reboot_system_after_config, false)
 
-    if otp_gte_23? and reboot? and release.config_providers != [] do
-      "-elixir -config_provider_reboot_mode #{env_var}"
+    if reboot? and release.config_providers != [] do
+      "-elixir config_provider_reboot_mode #{env_var}"
     else
       "-mode #{env_var}"
     end
