@@ -15,7 +15,7 @@ defmodule Mix.Tasks.ReleaseTest do
           File.mkdir_p!("rel/overlays/empty/directory")
           File.write!("rel/overlays/hello", "world")
 
-          for file <- ~w(rel/vm.args.eex rel/env.sh.eex rel/env.bat.eex) do
+          for file <- ~w(rel/vm.args.eex rel/remote.vm.args.eex rel/env.sh.eex rel/env.bat.eex) do
             File.write!(file, """
             #{file} FOR <%= @release.name %>
             """)
@@ -34,6 +34,9 @@ defmodule Mix.Tasks.ReleaseTest do
           assert root |> Path.join("releases/0.1.0/vm.args") |> File.read!() ==
                    "rel/vm.args.eex FOR release_test\n"
 
+          assert root |> Path.join("releases/0.1.0/remote.vm.args") |> File.read!() ==
+                   "rel/remote.vm.args.eex FOR release_test\n"
+
           assert root |> Path.join("empty/directory") |> File.dir?()
           assert root |> Path.join("hello") |> File.read!() == "world"
         end)
@@ -52,7 +55,8 @@ defmodule Mix.Tasks.ReleaseTest do
           File.mkdir_p!("custom_rel/overlays/empty/directory")
           File.write!("custom_rel/overlays/hello", "world")
 
-          for file <- ~w(custom_rel/vm.args.eex custom_rel/env.sh.eex custom_rel/env.bat.eex) do
+          for file <-
+                ~w(custom_rel/vm.args.eex custom_rel/remote.vm.args.eex custom_rel/env.sh.eex custom_rel/env.bat.eex) do
             File.write!(file, """
             #{file} FOR <%= @release.name %>
             """)
@@ -71,6 +75,9 @@ defmodule Mix.Tasks.ReleaseTest do
           assert root |> Path.join("releases/0.1.0/vm.args") |> File.read!() ==
                    "custom_rel/vm.args.eex FOR release_test\n"
 
+          assert root |> Path.join("releases/0.1.0/remote.vm.args") |> File.read!() ==
+                   "custom_rel/remote.vm.args.eex FOR release_test\n"
+
           assert root |> Path.join("empty/directory") |> File.dir?()
           assert root |> Path.join("hello") |> File.read!() == "world"
         end)
@@ -148,8 +155,8 @@ defmodule Mix.Tasks.ReleaseTest do
           beam =
             File.read!(Path.join(root, "lib/release_test-0.1.0/ebin/Elixir.ReleaseTest.beam"))
 
-          assert {:ok, {_, [{'Docs', _}]}} = :beam_lib.chunks(beam, ['Docs'])
-          assert {:error, _, _} = :beam_lib.chunks(beam, ['Dbgi'])
+          assert {:ok, {_, [{~c"Docs", _}]}} = :beam_lib.chunks(beam, [~c"Docs"])
+          assert {:error, _, _} = :beam_lib.chunks(beam, [~c"Dbgi"])
 
           refute root |> Path.join("bin/start") |> File.exists?()
           refute root |> Path.join("bin/start.bat") |> File.exists?()
@@ -208,6 +215,7 @@ defmodule Mix.Tasks.ReleaseTest do
           assert "bin/demo" in files
           assert "releases/0.1.0/sys.config" in files
           assert "releases/0.1.0/vm.args" in files
+          assert "releases/0.1.0/remote.vm.args" in files
           assert "releases/COOKIE" in files
           assert "releases/start_erl.data" in files
           assert "hello" in files
@@ -288,7 +296,7 @@ defmodule Mix.Tasks.ReleaseTest do
         assert_received {:mix_shell, :info, ["* assembling release_test-0.1.0 on MIX_ENV=dev"]}
 
         assert_received {:mix_shell, :info,
-                         ["\nRelease created at _build/dev/rel/release_test!" <> _]}
+                         ["\nRelease created at _build/dev/rel/release_test\n" <> _]}
 
         assert_received {:mix_shell, :info, ["* skipping runtime configuration" <> _]}
 
@@ -303,6 +311,7 @@ defmodule Mix.Tasks.ReleaseTest do
         assert root |> Path.join("releases/0.1.0/env.sh") |> File.exists?()
         assert root |> Path.join("releases/0.1.0/env.bat") |> File.exists?()
         assert root |> Path.join("releases/0.1.0/vm.args") |> File.exists?()
+        assert root |> Path.join("releases/0.1.0/remote.vm.args") |> File.exists?()
 
         assert root
                |> Path.join("releases/0.1.0/sys.config")
@@ -316,17 +325,18 @@ defmodule Mix.Tasks.ReleaseTest do
         cookie = File.read!(Path.join(root, "releases/COOKIE"))
 
         # Assert runtime
-        open_port(Path.join(root, "bin/release_test"), ['start'])
+        open_port(Path.join(root, "bin/release_test"), [~c"start"])
 
         assert %{
                  app_dir: app_dir,
                  cookie_env: ^cookie,
-                 encoding: {:time_μs, :"£", "£", '£'},
+                 encoding: {:_μ, :"£", "£", ~c"£"},
                  mode: :embedded,
                  node: release_node("release_test"),
                  protocols_consolidated?: true,
                  release_name: "release_test",
                  release_node: "release_test",
+                 release_prog: "release_test" <> ext,
                  release_root: release_root,
                  release_vsn: "0.1.0",
                  root_dir: root_dir,
@@ -343,12 +353,14 @@ defmodule Mix.Tasks.ReleaseTest do
           assert root_dir =~ ~r"_build/dev/rel/(release_test|RELEAS~1)$"
           assert String.ends_with?(sys_config_env, "releases\\0.1.0\\sys")
           assert String.ends_with?(sys_config_init, "releases\\0.1.0\\sys")
+          assert ext == ".bat"
         else
           assert app_dir == Path.join(root, "lib/release_test-0.1.0")
           assert release_root == root
           assert root_dir == root
           assert sys_config_env == Path.join(root, "releases/0.1.0/sys")
           assert sys_config_init == Path.join(root, "releases/0.1.0/sys")
+          assert ext == ""
         end
       end)
     end)
@@ -374,9 +386,10 @@ defmodule Mix.Tasks.ReleaseTest do
         config :release_test, :runtime,
           override: :runtime,
           config_env: config_env(),
-          config_target: config_target()
+          config_target: config_target(),
+          mode: :code.get_mode()
 
-        config :release_test, :encoding, {:runtime, :time_μs, :"£", "£", '£'}
+        config :release_test, :encoding, {:runtime, :_μ, :"£", "£", '£'}
         """)
 
         root = Path.absname("_build/dev/rel/runtime_config")
@@ -399,20 +412,27 @@ defmodule Mix.Tasks.ReleaseTest do
                |> File.chmod(0o555) == :ok
 
         # Assert runtime
-        open_port(Path.join(root, "bin/runtime_config"), ['start'])
+        open_port(Path.join(root, "bin/runtime_config"), [~c"start"])
 
         assert %{
-                 encoding: {:runtime, :time_μs, :"£", "£", '£'},
+                 encoding: {:runtime, :_μ, :"£", "£", ~c"£"},
                  mode: :embedded,
                  node: release_node("runtime_config"),
                  protocols_consolidated?: true,
                  release_name: "runtime_config",
                  release_mode: "embedded",
                  release_node: "runtime_config",
+                 release_prog: "runtime_config" <> ext,
                  release_vsn: "0.1.0",
                  runtime_config:
                    {:ok,
-                    [keep: :static, override: :runtime, config_env: :dev, config_target: :host]},
+                    [
+                      keep: :static,
+                      override: :runtime,
+                      config_env: :dev,
+                      config_target: :host,
+                      mode: :interactive
+                    ]},
                  static_config: {:ok, :was_set},
                  sys_config_env: sys_config_env,
                  sys_config_init: sys_config_init
@@ -421,9 +441,11 @@ defmodule Mix.Tasks.ReleaseTest do
         if match?({:win32, _}, :os.type()) do
           assert sys_config_env =~ "tmp\\runtime_config-0.1.0"
           assert sys_config_init =~ "tmp\\runtime_config-0.1.0"
+          assert ext == ".bat"
         else
           assert sys_config_env =~ "tmp/runtime_config-0.1.0"
           assert sys_config_init =~ "tmp/runtime_config-0.1.0"
+          assert ext == ""
         end
       end)
     end)
@@ -436,7 +458,10 @@ defmodule Mix.Tasks.ReleaseTest do
       Mix.Project.in_project(:release_test, ".", config, fn _ ->
         root = Path.absname("_build/dev/rel/no_dist")
         Mix.Task.run("release", ["no_dist"])
-        open_port(Path.join(root, "bin/no_dist"), ['start'], [{'RELEASE_DISTRIBUTION', 'none'}])
+
+        open_port(Path.join(root, "bin/no_dist"), [~c"start"], [
+          {~c"RELEASE_DISTRIBUTION", ~c"none"}
+        ])
 
         assert %{
                  mode: :embedded,
@@ -471,9 +496,10 @@ defmodule Mix.Tasks.ReleaseTest do
         assert root |> Path.join("releases/0.2.0/demo.rel") |> File.exists?()
         assert root |> Path.join("releases/0.2.0/sys.config") |> File.exists?()
         assert root |> Path.join("releases/0.2.0/vm.args") |> File.exists?()
+        assert root |> Path.join("releases/0.2.0/remote.vm.args") |> File.exists?()
 
         # Assert runtime
-        open_port(Path.join(root, "bin/demo"), ['start'])
+        open_port(Path.join(root, "bin/demo"), [~c"start"])
 
         assert %{
                  app_dir: app_dir,
@@ -483,6 +509,7 @@ defmodule Mix.Tasks.ReleaseTest do
                  protocols_consolidated?: true,
                  release_name: "demo",
                  release_node: "demo",
+                 release_prog: "demo" <> ext,
                  release_root: release_root,
                  release_vsn: "0.2.0",
                  root_dir: root_dir,
@@ -493,9 +520,11 @@ defmodule Mix.Tasks.ReleaseTest do
         if match?({:win32, _}, :os.type()) do
           assert String.ends_with?(app_dir, "demo/lib/release_test-0.1.0")
           assert String.ends_with?(release_root, "demo")
+          assert ext == ".bat"
         else
           assert app_dir == Path.join(root, "lib/release_test-0.1.0")
           assert release_root == root
+          assert ext == ""
         end
 
         assert root_dir == :code.root_dir() |> to_string()
@@ -509,6 +538,7 @@ defmodule Mix.Tasks.ReleaseTest do
       config = [releases: [compile_env_config: []], config_path: "different_config/config.exs"]
 
       File.write!("lib/compile_env.ex", """
+      require Application
       _ = Application.compile_env(:release_test, :static)
       """)
 
@@ -536,8 +566,8 @@ defmodule Mix.Tasks.ReleaseTest do
                  "ERROR! the application :release_test has a different value set for key :static during runtime compared to compile time"
 
         # But now it does
-        env = [{'RELEASE_STATIC', 'was_set'}]
-        open_port(Path.join(root, "bin/compile_env_config"), ['start'], env)
+        env = [{~c"RELEASE_STATIC", ~c"was_set"}]
+        open_port(Path.join(root, "bin/compile_env_config"), [~c"start"], env)
         assert %{} = wait_until_decoded(Path.join(root, "RELEASE_BOOTED"))
       end)
     end)
@@ -548,6 +578,7 @@ defmodule Mix.Tasks.ReleaseTest do
       config = [releases: [no_compile_env_config: [validate_compile_env: false]]]
 
       File.write!("lib/compile_env.ex", """
+      require Application
       _ = Application.compile_env(:release_test, :static)
       """)
 
@@ -557,15 +588,22 @@ defmodule Mix.Tasks.ReleaseTest do
         File.write!("config/runtime.exs", """
         import Config
         config :release_test, :static, String.to_atom(System.get_env("RELEASE_STATIC"))
+        config :release_test, :runtime, mode: :code.get_mode()
         """)
 
         root = Path.absname("_build/dev/rel/no_compile_env_config")
         Mix.Task.run("release", ["no_compile_env_config"])
 
         # It boots with mismatched config
-        env = [{'RELEASE_STATIC', 'runtime'}]
-        open_port(Path.join(root, "bin/no_compile_env_config"), ['start'], env)
-        assert %{} = wait_until_decoded(Path.join(root, "RELEASE_BOOTED"))
+        env = [{~c"RELEASE_STATIC", ~c"runtime"}]
+        open_port(Path.join(root, "bin/no_compile_env_config"), [~c"start"], env)
+
+        assert %{
+                 mode: :embedded,
+                 runtime_config: {:ok, [mode: :embedded]},
+                 static_config: {:ok, :runtime}
+               } =
+                 wait_until_decoded(Path.join(root, "RELEASE_BOOTED"))
       end)
     end)
   end
@@ -577,6 +615,7 @@ defmodule Mix.Tasks.ReleaseTest do
 
       # We write the compile env to guarantee rpc still works
       File.write!("lib/compile_env.ex", """
+      require Application
       _ = Application.compile_env(:release_test, :static)
       """)
 
@@ -592,14 +631,17 @@ defmodule Mix.Tasks.ReleaseTest do
         Mix.Task.run("release")
         script = Path.join(root, "bin/permanent1")
 
-        open_port(script, ['start'])
+        env = [{"RELEASE_DISTRIBUTION", "name"}, {"RELEASE_NODE", "permanent1@127.0.0.1"}]
+        open_port(script, [~c"start"], env)
         wait_until_decoded(Path.join(root, "RELEASE_BOOTED"))
-        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"]) == {"hello world\n", 0}
 
-        assert {pid, 0} = System.cmd(script, ["pid"])
+        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"], env: env) ==
+                 {"hello world\n", 0}
+
+        assert {pid, 0} = System.cmd(script, ["pid"], env: env)
         assert pid != "\n"
 
-        assert System.cmd(script, ["stop"]) == {"", 0}
+        assert System.cmd(script, ["stop"], env: env) == {"", 0}
       end)
     end)
   end
@@ -615,7 +657,7 @@ defmodule Mix.Tasks.ReleaseTest do
       Mix.Project.in_project(:release_test, ".", config, fn _ ->
         File.write!("config/runtime.exs", """
         import Config
-        config :release_test, :runtime, :was_set
+        config :release_test, :runtime, [mode: :code.get_mode()]
         """)
 
         root = Path.absname("_build/dev/rel/eval")
@@ -630,7 +672,13 @@ defmodule Mix.Tasks.ReleaseTest do
         assert String.trim_trailing(hello_world) == "hello_world"
         refute File.exists?(Path.join(root, "RELEASE_BOOTED"))
 
-        open_port(script, ['eval', 'Application.ensure_all_started(:release_test)'])
+        {hello_world, 0} =
+          System.cmd(script, ["eval", "IO.inspect(System.argv( ))", "a", "b", "c"])
+
+        assert String.trim_trailing(hello_world) == ~S(["a", "b", "c"])
+        refute File.exists?(Path.join(root, "RELEASE_BOOTED"))
+
+        open_port(script, [~c"eval", ~c"Application.ensure_all_started(:release_test)"])
 
         assert %{
                  cookie_env: "abcdefghij",
@@ -639,16 +687,19 @@ defmodule Mix.Tasks.ReleaseTest do
                  protocols_consolidated?: true,
                  release_name: "eval",
                  release_node: "eval",
+                 release_prog: "eval" <> ext,
                  release_root: release_root,
                  release_vsn: "0.1.0",
-                 runtime_config: {:ok, :was_set},
+                 runtime_config: {:ok, [mode: :interactive]},
                  static_config: {:ok, :was_set}
                } = wait_until_decoded(Path.join(root, "RELEASE_BOOTED"))
 
         if match?({:win32, _}, :os.type()) do
           assert String.ends_with?(release_root, "eval")
+          assert ext == ".bat"
         else
           assert release_root == root
+          assert ext == ""
         end
       end)
     end)
@@ -664,16 +715,17 @@ defmodule Mix.Tasks.ReleaseTest do
         Mix.Task.run("release")
 
         script = Path.join(root, "bin/permanent2")
-        open_port(script, ['daemon_iex'])
+        env = [{"RELEASE_DISTRIBUTION", "name"}, {"RELEASE_NODE", "permanent2@127.0.0.1"}]
+        open_port(script, [~c"daemon_iex"], env)
 
         assert %{
                  app_dir: app_dir,
                  cookie_env: "abcdefghij",
                  mode: :embedded,
-                 node: release_node("permanent2"),
+                 node: :"permanent2@127.0.0.1",
                  protocols_consolidated?: true,
                  release_name: "permanent2",
-                 release_node: "permanent2",
+                 release_node: "permanent2@127.0.0.1",
                  release_root: ^root,
                  release_vsn: "0.1.0",
                  root_dir: root_dir,
@@ -690,11 +742,20 @@ defmodule Mix.Tasks.ReleaseTest do
 
         assert wait_until(fn ->
                  File.read!(Path.join(root, "tmp/log/erlang.log.1")) =~
-                   "iex(permanent2@#{@hostname})1> "
+                   "iex(permanent2@127.0.0.1)1> "
                end)
 
-        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"]) == {"hello world\n", 0}
-        assert System.cmd(script, ["stop"]) == {"", 0}
+        assert System.cmd(script, ["rpc", "ReleaseTest.hello_world()"], env: env) ==
+                 {"hello world\n", 0}
+
+        assert System.cmd(script, ["restart"], env: env) == {"", 0}
+
+        assert wait_until(fn ->
+                 File.read!(Path.join(root, "tmp/log/erlang.log.1")) =~
+                   "RESTARTED!!!"
+               end)
+
+        assert System.cmd(script, ["stop"], env: env) == {"", 0}
       end)
     end)
   end
@@ -719,6 +780,7 @@ defmodule Mix.Tasks.ReleaseTest do
   end
 
   defp open_port(command, args, env \\ []) do
+    env = for {k, v} <- env, do: {to_charlist(k), to_charlist(v)}
     Port.open({:spawn_executable, to_charlist(command)}, [:hide, args: args, env: env])
   end
 
