@@ -237,6 +237,11 @@ defmodule Mix.Tasks.TestTest do
         # Nothing should get run if we try it again since everything is passing.
         assert mix(["test", "--failed"]) =~ "There are no tests to run"
 
+        # When everything is passing and a file is passed, we return the proper message
+        output = mix(["test", "test/passing_and_failing_test_failed.exs", "--failed"])
+        assert output =~ "There are no tests to run"
+        refute output =~ "does not match"
+
         # `--failed` and `--stale` cannot be combined
         output = mix(["test", "--failed", "--stale"])
         assert output =~ "Combining --failed and --stale is not supported"
@@ -294,7 +299,7 @@ defmodule Mix.Tasks.TestTest do
 
         Port.command(port, "\n")
 
-        message = "undefined function error_not_a_var"
+        message = "undefined variable \"error_not_a_var\""
         assert receive_until_match(port, message, "") =~ "test/b_test_stale.exs"
 
         File.write!("test/b_test_stale.exs", """
@@ -481,6 +486,16 @@ defmodule Mix.Tasks.TestTest do
         refute output =~ "==> foo"
         refute output =~ "Paths given to \"mix test\" did not match any directory/file"
 
+        output = mix(["test", Path.expand("apps/bar/test/bar_tests.exs")])
+
+        assert output =~ """
+               ==> bar
+               ....
+               """
+
+        refute output =~ "==> foo"
+        refute output =~ "Paths given to \"mix test\" did not match any directory/file"
+
         output = mix(["test", "apps/bar/test/bar_tests.exs:10"])
 
         assert output =~ """
