@@ -43,6 +43,7 @@ defmodule MixTest do
       assert_received {:mix_shell, :info, ["Generated install_test app"]}
       refute_received _
 
+      assert Application.app_dir(:crypto) =~ "crypto"
       assert List.keyfind(Application.started_applications(), :install_test, 0)
       assert apply(InstallTest, :hello, []) == :world
     end
@@ -243,7 +244,7 @@ defmodule MixTest do
       Mix.Project.push(GitApp)
       [_latest_rev, rev | _] = get_git_repo_revs("git_repo")
       lockfile = Path.join(tmp_dir, "lock")
-      Mix.Dep.Lock.write(lockfile, %{git_repo: {:git, fixture_path("git_repo"), rev, []}})
+      Mix.Dep.Lock.write(%{git_repo: {:git, fixture_path("git_repo"), rev, []}}, file: lockfile)
       Mix.ProjectStack.pop()
 
       Mix.install(
@@ -277,7 +278,7 @@ defmodule MixTest do
 
       Mix.Project.push(GitApp)
       lockfile = Path.join(tmp_dir, "lock")
-      Mix.Dep.Lock.write(lockfile, %{git_repo: {:git, fixture_path("git_repo"), rev2, []}})
+      Mix.Dep.Lock.write(%{git_repo: {:git, fixture_path("git_repo"), rev2, []}}, file: lockfile)
       Mix.ProjectStack.pop()
 
       Mix.install(
@@ -297,7 +298,7 @@ defmodule MixTest do
       lockfile |> Path.dirname() |> File.mkdir_p!()
       Mix.Project.push(GitApp)
       [_latest_rev, rev | _] = get_git_repo_revs("git_repo")
-      Mix.Dep.Lock.write(lockfile, %{git_repo: {:git, fixture_path("git_repo"), rev, []}})
+      Mix.Dep.Lock.write(%{git_repo: {:git, fixture_path("git_repo"), rev, []}}, file: lockfile)
       Mix.ProjectStack.pop()
 
       Mix.install(
@@ -338,8 +339,14 @@ defmodule MixTest do
       on_exit(fn ->
         :code.set_path(path)
         purge([InstallTest, InstallTest.MixProject, InstallTest.Protocol])
-        Application.stop(:install_test)
-        Application.unload(:install_test)
+
+        ExUnit.CaptureLog.capture_log(fn ->
+          Application.stop(:git_repo)
+          Application.unload(:git_repo)
+
+          Application.stop(:install_test)
+          Application.unload(:install_test)
+        end)
       end)
 
       Mix.State.put(:installed, nil)
