@@ -1,3 +1,16 @@
+home = Path.expand("../tmp/.home", __DIR__)
+File.mkdir_p!(home)
+System.put_env("HOME", home)
+
+mix = Path.expand("../tmp/.mix", __DIR__)
+File.mkdir_p!(mix)
+System.put_env("MIX_HOME", mix)
+
+System.delete_env("XDG_DATA_HOME")
+System.delete_env("XDG_CONFIG_HOME")
+
+## Setup Mix
+
 Mix.start()
 Mix.shell(Mix.Shell.Process)
 Application.put_env(:mix, :colors, enabled: false)
@@ -5,14 +18,20 @@ Application.put_env(:mix, :colors, enabled: false)
 Logger.remove_backend(:console)
 Application.put_env(:logger, :backends, [])
 
+## Setup ExUnit
+
 os_exclude = if match?({:win32, _}, :os.type()), do: [unix: true], else: [windows: true]
 epmd_exclude = if match?({:win32, _}, :os.type()), do: [epmd: true], else: []
-ExUnit.start(trace: "--trace" in System.argv(), exclude: epmd_exclude ++ os_exclude)
+git_exclude = if Mix.SCM.Git.git_version() <= {1, 7, 4}, do: [git_sparse: true], else: []
 
-unless {1, 7, 4} <= Mix.SCM.Git.git_version() do
-  IO.puts(:stderr, "Skipping tests with git sparse checkouts...")
-  ExUnit.configure(exclude: :git_sparse)
-end
+{line_exclude, line_include} =
+  if line = System.get_env("LINE"), do: {[:test], [line: line]}, else: {[], []}
+
+ExUnit.start(
+  trace: !!System.get_env("TRACE"),
+  exclude: epmd_exclude ++ os_exclude ++ git_exclude ++ line_exclude,
+  include: line_include
+)
 
 # Clear environment variables that may affect tests
 System.delete_env("http_proxy")
@@ -136,10 +155,14 @@ defmodule MixTest.Case do
   end
 
   def ensure_touched(file) do
-    ensure_touched(file, File.stat!(file).mtime)
+    ensure_touched(file, file)
   end
 
-  def ensure_touched(file, current) do
+  def ensure_touched(file, current) when is_binary(current) do
+    ensure_touched(file, File.stat!(current).mtime)
+  end
+
+  def ensure_touched(file, current) when is_tuple(current) do
     File.touch!(file)
     mtime = File.stat!(file).mtime
 
@@ -153,11 +176,12 @@ defmodule MixTest.Case do
     end
   end
 
-  def os_newline do
-    case :os.type() do
-      {:win32, _} -> "\r\n"
-      _ -> "\n"
-    end
+  if match?({:win32, _}, :os.type()) do
+    def windows?, do: true
+    def os_newline, do: "\r\n"
+  else
+    def windows?, do: false
+    def os_newline, do: "\n"
   end
 
   def mix(args, envs \\ []) when is_list(args) do
@@ -179,6 +203,10 @@ defmodule MixTest.Case do
     ])
   end
 
+  def force_recompilation(file) do
+    File.write!(file, File.read!(file) <> "\n")
+  end
+
   defp mix_executable do
     Path.expand("../../../bin/mix", __DIR__)
   end
@@ -191,27 +219,21 @@ defmodule MixTest.Case do
     tmp = tmp_path() |> String.to_charlist()
     for path <- :code.get_path(), :string.str(path, tmp) != 0, do: :code.del_path(path)
   end
+
+  def get_git_repo_revs(repo) do
+    File.cd!(fixture_path(repo), fn ->
+      Regex.split(~r/\r?\n/, System.cmd("git", ["log", "--format=%H"]) |> elem(0), trim: true)
+    end)
+  end
 end
 
-## Set up globals
+## Set up Rebar fixtures
 
-home = MixTest.Case.tmp_path(".home")
-File.mkdir_p!(home)
-System.put_env("HOME", home)
-
-mix = MixTest.Case.tmp_path(".mix")
-File.mkdir_p!(mix)
-System.put_env("MIX_HOME", mix)
-
-System.delete_env("XDG_DATA_HOME")
-System.delete_env("XDG_CONFIG_HOME")
-
-rebar = System.get_env("REBAR") || Path.expand("fixtures/rebar", __DIR__)
-File.cp!(rebar, Path.join(mix, "rebar"))
-rebar = System.get_env("REBAR3") || Path.expand("fixtures/rebar3", __DIR__)
-File.cp!(rebar, Path.join(mix, "rebar3"))
-
-## Copy fixtures to tmp
+rebar3_source = System.get_env("REBAR3") || Path.expand("fixtures/rebar3", __DIR__)
+[major, minor | _] = String.split(System.version(), ".")
+rebar3_target = Path.join([mix, "elixir", "#{major}-#{minor}", "rebar3"])
+File.mkdir_p!(Path.dirname(rebar3_target))
+File.cp!(rebar3_source, rebar3_target)
 
 fixtures = ~w(rebar_dep rebar_override)
 
@@ -222,11 +244,13 @@ Enum.each(fixtures, fn fixture ->
   File.cp_r!(source, dest)
 end)
 
-## Generate Git repo fixtures
-System.cmd("git", ~w[config --global user.email "mix@example.com"])
-System.cmd("git", ~w[config --global user.name "mix-repo"])
+## Set up Git fixtures
 
-# Git repo
+System.cmd("git", ~w[config --global user.email mix@example.com])
+System.cmd("git", ~w[config --global user.name mix-repo])
+System.cmd("git", ~w[config --global init.defaultBranch not-main])
+
+### Git repo
 target = Path.expand("fixtures/git_repo", __DIR__)
 
 unless File.dir?(target) do
@@ -308,7 +332,7 @@ unless File.dir?(target) do
   end)
 end
 
-# Deps on Git repo
+### Deps on Git repo
 target = Path.expand("fixtures/deps_on_git_repo", __DIR__)
 
 unless File.dir?(target) do
@@ -368,16 +392,22 @@ end
 target = Path.expand("fixtures/git_rebar", __DIR__)
 
 unless File.dir?(target) do
+  File.mkdir_p!(Path.join(target, "ebin"))
   File.mkdir_p!(Path.join(target, "src"))
 
-  File.write!(Path.join([target, "src", "git_rebar.app.src"]), """
+  # This is used to test that the built-in ebin is ignored
+  # when build_embedded is true.
+  File.write!(Path.join(target, "ebin/.unused"), """
+  """)
+
+  File.write!(Path.join(target, "src/git_rebar.app.src"), """
   {application, git_rebar,
     [
       {vsn, "0.1.0"}
     ]}.
   """)
 
-  File.write!(Path.join([target, "src", "git_rebar.erl"]), """
+  File.write!(Path.join(target, "src/git_rebar.erl"), """
   -module(git_rebar).
   -export([any_function/0]).
   any_function() -> ok.
@@ -396,6 +426,26 @@ Enum.each([:invalidapp, :invalidvsn, :noappfile, :nosemver, :ok], fn dep ->
   File.mkdir_p!(Path.expand("fixtures/deps_status/deps/#{dep}/.git", __DIR__))
 end)
 
+### Archive ebin
+target = Path.expand("fixtures/archive", __DIR__)
+
+unless File.dir?(Path.join(target, "ebin")) do
+  File.mkdir_p!(Path.join(target, "ebin"))
+
+  File.write!(Path.join([target, "ebin", "local_sample.app"]), """
+  {application,local_sample,
+    [
+      {modules,['Elixir.Mix.Tasks.Local.Sample']},
+      {applications,[kernel,stdlib,elixir]}
+    ]
+  }.
+  """)
+
+  [{name, bin}] = Code.compile_file("lib/local.sample.ex", target)
+
+  File.write!(Path.join([target, "ebin", Atom.to_string(name) <> ".beam"]), bin)
+end
+
 ## Generate helper modules
 
 path = MixTest.Case.tmp_path("beams")
